/*!
 * module node-red-contrib-stm32cubemonitor
 * Copyright(c) 2019 STMicroelectronics
 */
module.exports = function(RED) {
  "use strict";
  let settings = RED.settings;
  // safely accessing nested level key from settings
  const logLevel = (((settings || {}).logging || {}).console || {}).level;
  const acquisitionServices = require("@stm32/stm32cubemonitor-acquisitionservices")
    .AcquisitionServices;
  acquisitionServices.initLogger(logLevel);
  const path = require("path");
  const events = require("events");
  const object = require("lodash/fp");

  const logger = RED.log;
  // if stlinkReconnectTime not defined in settings.js file, set default value to 10s
  if (!settings.stlinkReconnectTime) settings.stlinkReconnectTime = 10000;

  // Fallback to "p2p" mode in case the connectType key is not defined in settings.js or key not set to "p2p" or "tcp"
  if (
    !settings.connectionType ||
    (settings.connectionType !== "p2p" && settings.connectionType !== "tcp")
  ) {
    settings.connectionType = "p2p";
  }
  let actualConnectionType = acquisitionServices.initConnectionType(
    settings.connectionType
  );
  if (actualConnectionType !== settings.connectionType) {
    logger.info("tcp server not reachable, fallback to p2p connection type");
    RED.log.warn("------------------------------------------------------");
    RED.log.warn("tcp server not reachable, fallback to p2p connection type");
    RED.log.warn("------------------------------------------------------");
  }

  /**
   * Acquisition out node constructor.
   *
   * @param {*} config - Acquisition out node configuration.
   */
  function AcquisitionOutNode(config) {
    RED.nodes.createNode(this, config);
    this.name = config.name || "";
    this.probeconfig = config.probeconfig || "";
    this.probeconfiguration = RED.nodes.getNode(this.probeconfig);
    var oldmsg = "";

    if (this.probeconfiguration) {
      var node = this;
      node.status({ fill: "grey", shape: "dot", text: "not connected" });
      node.probe = probePool.get(this.probeconfiguration);

      node.probe.on("ready", function(msg) {
        oldmsg = "";
        node.status({
          fill: "green",
          shape: "dot",
          text: `${actualConnectionType} connected ${msg}`
        });
      });
      node.probe.on("closed", function(msg) {
        if (oldmsg !== msg) {
          oldmsg = msg;
          node.status({
            fill: "red",
            shape: "dot",
            text: `not connected ! ${msg}`
          });
        }
      });

      node.probe.on("error", function(err) {
        //only log errors and  ignore warnings
        //status is updated on closed, no need to change it
        if (err.topic === "error") {
          node.error(err.payload);
        }
      });

      node.on("input", function(msg) {
        if (
          !{}.hasOwnProperty.call(msg, "payload") ||
          !{}.hasOwnProperty.call(msg, "topic")
        ) {
          return; // do nothing unless we have a payload
        }

        node.probe.sendCommand(msg.topic, msg.payload, err => {
          if (err) node.error(err + ": " + msg.topic);
        });
      });
      node.probe.connect();
    } else {
      this.error(RED._("Missing probe config"));
    }

    this.on("close", function(done) {
      if (this.probeconfiguration) {
        probePool.close(this.probeconfiguration.probeid, done);
      } else {
        done();
      }
    });
  }
  RED.nodes.registerType("acquisition out", AcquisitionOutNode);

  /**
   * Acquisition in node constructor.
   *
   * @param {*} config - Acquisition in node configuration.
   */
  function AcquisitionInNode(config) {
    RED.nodes.createNode(this, config);
    this.name = config.name || "";
    this.probeconfig = config.probeconfig || "";
    this.probeconfiguration = RED.nodes.getNode(this.probeconfig);
    var oldmsg = "";

    if (this.probeconfiguration) {
      var node = this;
      node.status({ fill: "grey", shape: "dot", text: "not connected" });
      node.probe = probePool.get(this.probeconfiguration);

      this.probe.on("data", function(msgout) {
        const myMsg = object.cloneDeep(msgout);
        node.send([{ payload: myMsg }, null]);
      });
      this.probe.on("error", function(err) {
        node.send([null, err]);
      });
      this.probe.on("ready", function(msg) {
        oldmsg = "";
        node.status({
          fill: "green",
          shape: "dot",
          text: `${actualConnectionType} connected ${msg}`
        });
      });
      this.probe.on("closed", function(msg) {
        if (oldmsg !== msg) {
          oldmsg = msg;
          node.status({
            fill: "red",
            shape: "dot",
            text: `not connected ! ${msg}`
          });
        }
      });
      node.probe.connect();
    } else {
      this.error(RED._("Missing probe config"));
    }

    this.on("close", function(done) {
      if (this.probeconfiguration) {
        probePool.close(this.probeconfiguration.probeid, done);
      } else {
        done();
      }
    });
  }
  RED.nodes.registerType("acquisition in", AcquisitionInNode);

  /**
   * Probe config node constructor.
   *
   * @param {*} config - Probe config node configuration.
   */
  function probe(config) {
    RED.nodes.createNode(this, config);
    this.probeid = config.probeid || "";
    this.protocol = config.protocol || "";
    this.frequency = config.frequency || "";
  }
  RED.nodes.registerType("probe", probe);

  var probePool = (function() {
    var connections = {};
    return {
      /* eslint-disable require-jsdoc */
      get: function(probeConfig) {
        // make local copy of configuration -- perhaps not needed?
        var probeid = probeConfig.probeid,
          frequency = probeConfig.frequency,
          protocol = probeConfig.protocol;

        var id = probeid;
        // just return the connection object if already have one
        // key is the probeid (file path)
        if (connections[id]) {
          return connections[id];
        }

        connections[id] = (function() {
          var obj = {
            _emitter: new events.EventEmitter(),
            acqServ: null,
            on: function(a, b) {
              this._emitter.on(a, b);
            },
            close: function(cb) {
              this.acqServ.close(cb);
            },
            sendCommand: function(topic, data, cb) {
              this.acqServ.sendCommand(topic, data, cb);
            },
            connect: function() {
              this.acqServ.connect();
            }
          };

          var initProbesConfig = function() {
            obj.acqServ = new acquisitionServices(
              probeid,
              settings.stlinkReconnectTime,
              {
                protocol: protocol,
                frequency: frequency
              }
            );
            obj.acqServ.on("close", function(msg) {
              obj._emitter.emit("closed", msg);
            });
            obj.acqServ.on("open", function(msg) {
              obj._emitter.emit("ready", msg);
            });
            obj.acqServ.on("data", function(msg) {
              obj._emitter.emit("data", msg);
            });
            obj.acqServ.on("error", function(msg) {
              obj._emitter.emit("error", msg);
            });
          };
          initProbesConfig();
          return obj;
        })();
        return connections[id];
      },
      close: function(probeid, done) {
        if (connections[probeid]) {
          connections[probeid].close(function() {
            done();
          });
          delete connections[probeid];
        } else {
          done();
        }
      }
      /* eslint-enable require-jsdoc */
    };
  })();

  RED.httpAdmin.get(
    "/blink/*",
    RED.auth.needsPermission("probe.read"),
    function(req, res) {
      let result = acquisitionServices.blinkLed(req.params[0]);
      res.send(result);
    }
  );

  RED.httpAdmin.get(
    "/getprobe",
    RED.auth.needsPermission("probe.read"),
    function(req, res) {
      acquisitionServices.probesDiscovery((receivedList, receivedUpdated) => {
        res.send({ list: receivedList, updated: receivedUpdated });
      });
    }
  );

  const select2Path = require.resolve("select2");
  RED.httpNode.get("/select2/*", function(req, res) {
    var filename = path.join(select2Path, "../..", req.params[0]);
    res.sendFile(filename, function(err) {
      if (err) {
        logger.error(
          "acquisition.js : " +
            filename +
            " not found. Maybe running in dev mode."
        );
      }
    });
  });

  RED.httpNode.get("/images/*", function(req, res) {
    var filename = path.join(__dirname, "images", req.params[0]);
    res.sendFile(filename, function(err) {
      if (err) {
        logger.error(filename + " not found. Maybe running in dev mode.");
      }
    });
  });
};
