/*!
 * module node-red-contrib-stm32cubemonitor
 * Copyright(c) 2019 STMicroelectronics
 *
 * ****************************************************
 * The code to open file is from node-red project, 10-file.js
 * Copyright JS Foundation and other contributors, http://js.foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/

let fs = require("fs-extra");
let os = require("os");
let path = require("path");

let node = this;
node.wstream = null;
node.filename = null;
let lastValues = {};
const FILE_SIZE = 104857600; // 100MB of data

process.on("message", msg => {
  switch (msg.topic) {
    case "initializeLog":
      node.logmode = msg.logmode;
      node.directory = msg.directory;
      node.groupname = msg.groupname;
      createFile();
      break;
    case "log":
      logData(msg.data);
      break;
  }
});

/**
 *
 * Function called to create file of log.
 */
function createFile() {
  node.wstream = null;
  lastValues = {};
  try {
    fs.ensureDirSync(node.directory);
  } catch (err) {
    let error = {};
    error.title = "file.errors.createfail";
    error.type = err.toString();
    error.msg = node.directory;
    process.send(error);
    return;
  }
  var now = new Date();
  var monthValue = now.getMonth() + 1;

  node.filename = path.join(
    node.directory,
    "Log_" +
      node.groupname +
      "_" +
      now.getFullYear() +
      "-" +
      ("0" + monthValue).substr(-2) +
      "-" +
      ("0" + now.getDate()).substr(-2) +
      "_" +
      ("0" + now.getHours()).substr(-2) +
      "h" +
      ("0" + now.getMinutes()).substr(-2) +
      "m" +
      ("0" + now.getSeconds()).substr(-2) +
      "s" +
      ".stcm"
  );
}

/**
 *
 * Function called to log measures in file during acquisition.
 *
 * @param {*} msg - JSON structure of data used as input of the processing node.
 */
function logData(msg) {
  if (node.logmode === "changes") {
    for (let i = 0; i < msg.variabledata.length; i++) {
      if (
        !Object.prototype.hasOwnProperty.call(lastValues, msg.variablename) ||
        msg.variabledata[i].y !== lastValues[msg.variablename]
      ) {
        lastValues[msg.variablename] = msg.variabledata[i].y;
      } else {
        msg.variabledata.splice(i, 1);
        i--;
      }
    }
    if (msg.variabledata.length === 0) {
      return;
    }
  }

  var filename = node.filename;
  var size = null;
  var data = JSON.stringify(msg) + os.EOL;
  var recreateStream = !node.wstream;
  if (node.wstream && node.wstreamIno) {
    // There is already a stream open and we have the inode
    // of the file. Check the file hasn't been deleted
    // or deleted and recreated.
    try {
      var stat = fs.statSync(filename);
      size = stat["size"];
      // File exists - check the inode matches
      if (stat.ino !== node.wstreamIno) {
        // The file has been recreated. Close the current
        // stream and recreate it
        recreateStream = true;
        node.wstream.end();
        delete node.wstream;
        delete node.wstreamIno;
      }
    } catch (err) {
      // File does not exist
      recreateStream = true;
      node.wstream.end();
      delete node.wstream;
      delete node.wstreamIno;
    }
  }
  if (recreateStream) {
    node.wstream = fs.createWriteStream(filename, {
      encoding: "utf8",
      // Append mode
      flags: "a",
      autoClose: true
    });
    node.wstream.on("open", function() {
      try {
        var stat = fs.statSync(filename);
        node.wstreamIno = stat.ino;
      } catch (err) {
        node.warn("file :" + filename + " issue to open");
      }
    });
    node.wstream.on("error", function(err) {
      let error = {};
      error.title = "file.errors.appendfail";
      error.type = err.toString();
      error.msg = msg;
      process.send(error);
    });
  }
  if (filename) {
    // Static filename - write and reuse the stream next time
    node.wstream.write(data, function() {});
  }

  if (size && size > FILE_SIZE) {
    createFile();
  }
}
