/*!
 * module node-red-contrib-stm32cubemonitor
 * Copyright(c) 2019 STMicroelectronics
 */
const path = require("path");
const fs = require("fs");

module.exports = function(RED) {
  "use strict";
  var ElfParser = require("@stm32/stm32cubemonitor-elfparser").ElfParser;
  const logger = RED.log;
  let progressStatus = 0;

  /**
   *
   * Main function to register variables node.
   *
   * @param {*} config - Object holding configuration parameters.
   */
  function initVariables(config) {
    RED.nodes.createNode(this, config);
    let node = this;
    this.groupname = config.groupname;
    this.variablelist = config.variablelist || [];
    this.variablelist.forEach(v => {
      v.type = Number(v.type);
    });
    this.mode = config.mode || "direct";
    this.snapshotheader = config.snapshotheader || "";
    this.triggerstartmode = config.triggerstartmode || "manual";
    this.triggername = config.triggername || "";
    this.triggerthreshold = config.triggerthreshold || "";
    config.frequencyType = config.frequencyType || "0";
    this.frequency = Number.parseFloat(
      config.frequencyType === "custom"
        ? RED.util.evaluateNodeProperty(
            config.frequency,
            config.frequencyType,
            this
          )
        : config.frequencyType
    );
    this.execonfig = config.execonfig;
    let imported = node.execonfig
      ? RED.nodes
          .getNode(node.execonfig)
          .variablelist.filter(elem => elem.checked)
      : [];
    // Remove from manual variablelist variable having the same name than the ones imported
    this.variablelist = this.variablelist.filter(elem => {
      return imported.findIndex(i => i.name === elem.name) === -1;
    });
    this.fullVariableList = this.variablelist.concat(imported);
    this.accesspoint = Number.parseInt(config.accesspoint) || 0;
    if (
      config.frequencyType !== "custom" ||
      (config.frequencyType === "custom" && this.frequency > 0)
    ) {
      let status = this.mode;
      let color = "green";
      if (this.triggerstartmode !== "manual") {
        const found = this.fullVariableList.findIndex(
          elem => elem.name === node.triggername
        );
        if (
          found === -1 ||
          this.triggerthreshold.length === 0 ||
          isNaN(Number(this.triggerthreshold))
        ) {
          this.triggerstartmode = "manual";
          status += " - invalid trigger";
          color = "red";
        } else {
          status += " - trigger enable";
        }
      }
      node.status({ fill: color, shape: "dot", text: status });
      node.on("input", function(msg) {
        var param = {};
        param.groupname = node.groupname;
        param.variablelist = node.fullVariableList;
        param.mode = node.mode;
        param.snapshotheader = node.snapshotheader;
        param.triggerstartmode = node.triggerstartmode;
        param.triggername = node.triggername;
        param.triggerthreshold = node.triggerthreshold;
        param.frequency = node.frequency;
        param.accesspoint = node.accesspoint;
        msg.payload = param;
        node.send(msg);
      });
    } else {
      logger.error(
        "Sampling frequency not correctly set for '" +
          node.groupname +
          "' variable node "
      );
      node.status({
        fill: "red",
        shape: "dot",
        text: "off - invalid frequency"
      });
    }
  }
  RED.nodes.registerType("variables", initVariables);

  /**
   *
   * Main function to register exe-config node.
   *
   * @param {*} n - Object holding configuration parameters.
   */
  function execonfig(n) {
    RED.nodes.createNode(this, n);
    this.exefile = n.exefile;
    this.exefolder = n.exefolder;
    this.fullpathname = n.fullpathname;
    this.variablelist = n.exevariablelist;
    this.configLastUpdate = n.configLastUpdate;
  }
  RED.nodes.registerType("exe-config", execonfig);

  /**
   * Service to return the list of exe files (with extensions .elf, .out or .axf).
   *
   * @param {string} currentDirPath - Current directory.
   * @param {Function} callback - Callback called for each exe file.
   */
  function walkSync(currentDirPath, callback) {
    try {
      var files = fs.readdirSync(currentDirPath, { withFileTypes: true });
      files.forEach(function(file) {
        if (file.isFile()) {
          var name = file.name;
          var filePath = path.join(currentDirPath, name);
          if (
            path.extname(name) === ".elf" ||
            path.extname(name) === ".out" ||
            path.extname(name) === ".axf"
          ) {
            callback(filePath, name);
          }
        }
      });
    } catch (e) {
      // Handle error
      logger.warn(currentDirPath + " path does not exist");
      throw new Error(currentDirPath + " path does not exist");
    }
  }

  RED.httpAdmin.post(
    "/listfiles",
    RED.auth.needsPermission("exe-config.write"),
    function(req, res) {
      var files = [];
      try {
        walkSync(req.body.folderpath, function(filePath, name) {
          var item = { filePath, name };
          files.push(item);
        });
        res.send(files);
      } catch (e) {
        res.status(404).send(e.name + ": " + e.message);
      }
    }
  );

  RED.httpAdmin.post(
    "/getvariables",
    RED.auth.needsPermission("exe-config.write"),
    function(req, res) {
      const exefilename = req.body.id;
      const expand = req.body.exp === "true";
      const exeModifiedTime = fs.statSync(exefilename).mtime.getTime();
      let elfParser = new ElfParser();
      progressStatus = 0;
      elfParser.readElf(
        exefilename,
        list => {
          res.send({ list, exeModifiedTime });
        },
        expand,
        progress => {
          progressStatus = progress;
        }
      );
    }
  );

  RED.httpAdmin.post(
    "/getstat",
    RED.auth.needsPermission("exe-config.write"),
    function(req, res) {
      const exeFilename = req.body.id;
      const exeTime = fs.statSync(exeFilename).mtime.getTime();
      res.send({ exeTime });
    }
  );

  RED.httpNode.get("/getElfParsingProgress", function(req, res) {
    logger.trace("Progress status = " + progressStatus);
    res.send(progressStatus.toString());
  });

  RED.httpNode.get("/images/*", function(req, res) {
    var filename = path.join(__dirname, "images", req.params[0]);
    res.sendFile(filename, function(err) {
      if (err) {
        logger.error(filename + " not found. Maybe running in dev mode.");
      }
    });
  });

  const hyperlistPath = require.resolve("hyperlist");
  RED.httpAdmin.get(
    "/hyperlist.js",
    RED.auth.needsPermission("exe-config.read"),
    function(req, res) {
      res.sendFile(hyperlistPath, function(err) {
        if (err) {
          logger.error(
            "variable.js : " +
              hyperlistPath +
              " not found. Maybe running in dev mode."
          );
        }
      });
    }
  );
  const select2Path = require.resolve("select2");
  RED.httpNode.get("/select2/*", function(req, res) {
    var filename = path.join(select2Path, "../..", req.params[0]);
    res.sendFile(filename, function(err) {
      if (err) {
        logger.error(
          "variable.js : " + filename + " not found. Maybe running in dev mode."
        );
      }
    });
  });
};
