/*!
 * module node-red-contrib-stm32cubemonitor
 * Copyright(c) 2019 STMicroelectronics
 */
/* eslint-disable no-console */
var helper = require("node-red-node-test-helper");
var sinon = require("sinon");
var should = require("should");
const expect = require("chai").expect;
var acquisition = require("../nodes/acquisition");

var acquisitionServices = require("@stm32/stm32cubemonitor-acquisitionservices")
  .AcquisitionServices;

/**
 * Module(s) mocked without sinonjs.
 */
const acqServMock = require("./acquisitionMockFunc.js");
const inAcq = acqServMock.inAcq;

helper.init(require.resolve("node-red"));

describe("acquisition node out", function() {
  let acqServProto;

  var myClock;
  before(function(done) {
    helper.startServer(done);
  });

  after(function(done) {
    helper.stopServer(done);
  });

  beforeEach(function() {
    const myNow = new Date("1");
    myClock = sinon.useFakeTimers(myNow);
    acqServProto = sinon.mock(acquisitionServices.prototype);
    acqServProto.object.connect = acqServMock.connect;
    acqServProto.object.close = acqServMock.close;
    acqServProto.object.sendCommand = acqServMock.sendCommand;
  });

  afterEach(function() {
    helper.unload();
    myClock.restore();
    acqServProto.verify();
  });

  var myProbeConfig = {
    id: "p1",
    type: "probe",
    probeid: "4579",
    frequency: "10",
    protocol: "SWD"
  };

  var groupVar1 = {
    groupname: "Group_Var1",
    variablelist: [
      {
        name: "counterS16",
        address: "0x2000012a",
        type: 4,
        importedfromelf: "true"
      },
      {
        name: "counterS8",
        address: "0x20000091",
        type: 2,
        importedfromelf: "true"
      },
      {
        name: "counterU16",
        address: "0x20000128",
        type: 3,
        importedfromelf: "true"
      },
      {
        name: "counterU8",
        address: "0x20000090",
        type: 1,
        importedfromelf: "true"
      },
      {
        name: "varSLong",
        address: "0x20000180",
        type: 6,
        importedfromelf: "true"
      },
      {
        name: "varSLongLong",
        address: "0x20000158",
        type: 8,
        importedfromelf: "true"
      }
    ],
    mode: "direct",
    frequency: 1,
    accesspoint: 0
  };

  it("should be loaded and not started", function(done) {
    var flow = [
      {
        id: "n1",
        type: "acquisition out",
        name: "myName",
        probeconfig: myProbeConfig
      }
    ];
    helper.load(acquisition, flow, function() {
      var n1 = helper.getNode("n1");
      n1.should.have.property("name", "myName");
      n1.should.have.property("probeconfiguration", null);
      done();
    });
    myClock.tick(50);
  });

  it("should be loaded and started", function(done) {
    acqServMock.config({
      stlinkRes: 0
    }); // set ERROR for STLINK
    var flow = [
      {
        id: "p1",
        z: "f1",
        type: "probe",
        probeid: "4579",
        frequency: "10",
        protocol: "SWD"
      },
      {
        id: "f1",
        type: "tab",
        label: "test flow"
      },
      {
        id: "n1",
        z: "f1",
        type: "acquisition out",
        name: "myName",
        probeconfig: "p1",
        wires: [["probe"]]
      },
      {
        id: "h1",
        z: "f1",
        type: "helper"
      }
    ];
    helper.load(acquisition, flow, function() {
      var n1 = helper.getNode("n1");
      n1.should.have.property("name", "myName");
      n1.probe.on("closed", msg => {
        should(msg).exactly("Error STLINK NOK");
        done();
      });
      n1.on("call:status", msg => {
        if (msg.args[0].text === "not connected ! Device not found (6)") {
          done();
        }
      });

      myClock.tick(1000);
    });
  });

  it("should be loaded, started and received ready event", function(done) {
    acqServMock.config({
      stlinkRes: 1
    }); // set OK for STLINK
    var flow = [
      {
        id: "p1",
        z: "f1",
        type: "probe",
        probeid: "4579",
        frequency: "10",
        protocol: "SWD"
      },
      {
        id: "f1",
        type: "tab",
        label: "test flow"
      },
      {
        id: "n1",
        z: "f1",
        type: "acquisition out",
        name: "myName",
        probeconfig: "p1"
      },
      {
        id: "h1",
        z: "f1",
        type: "helper"
      }
    ];
    helper.load(acquisition, flow, function() {
      var n1 = helper.getNode("n1");
      n1.should.have.property("name", "myName");
      n1.probe.on("ready", msg => {
        should(msg).exactly("(NameOfMCU)");
        done();
      });
      myClock.tick(1000);
    });
  });
  it("should be loaded, started, received ready event and send Command for GroupVar1", function(done) {
    acqServMock.config({
      stlinkRes: 1
    }); // set OK for STLINK
    var flow = [
      {
        id: "p1",
        z: "f1",
        type: "probe",
        probeid: "4579",
        frequency: "10",
        protocol: "SWD"
      },
      {
        id: "f1",
        type: "tab",
        label: "test flow"
      },
      {
        id: "n1",
        z: "f1",
        type: "acquisition out",
        name: "myName",
        probeconfig: "p1"
      },
      {
        id: "h1",
        z: "f1",
        type: "helper"
      }
    ];
    helper.load(acquisition, flow, function() {
      var n1 = helper.getNode("n1");
      n1.should.have.property("name", "myName");
      n1.probe.on("ready", msg => {
        should(msg).exactly("(NameOfMCU)");
        console.log("Ready received then send groupVar1");
        n1.receive({
          topic: "read",
          payload: groupVar1
        });
        myClock.tick(1);
      });
      n1.probe.on("data", function(msgout) {
        console.log("data received from acqServ : ");
        console.log(msgout);
        done();
      });
      myClock.tick(1000);
    });
  });
});

describe("acquisition node in and out", function() {
  let acqServProto;

  var myClock;
  before(function(done) {
    helper.startServer(done);
  });

  after(function(done) {
    helper.stopServer(done);
  });

  beforeEach(function() {
    const myNow = new Date("1");
    myClock = sinon.useFakeTimers(myNow);
    acqServProto = sinon.mock(acquisitionServices.prototype);
    acqServProto.object.open = acqServMock.open;
    acqServProto.object.close = acqServMock.close;
    acqServProto.object.sendCommand = acqServMock.sendCommand;
  });

  afterEach(function() {
    helper.unload();
    myClock.restore();
    acqServProto.verify();
  });

  var groupVar1 = {
    groupname: "Group_Var1",
    variablelist: [
      {
        name: "counterS16",
        address: "0x2000012a",
        type: 4,
        importedfromelf: "true"
      },
      {
        name: "counterS8",
        address: "0x20000091",
        type: 2,
        importedfromelf: "true"
      },
      {
        name: "counterU16",
        address: "0x20000128",
        type: 3,
        importedfromelf: "true"
      },
      {
        name: "counterU8",
        address: "0x20000090",
        type: 1,
        importedfromelf: "true"
      },
      {
        name: "varSLong",
        address: "0x20000180",
        type: 6,
        importedfromelf: "true"
      },
      {
        name: "varSLongLong",
        address: "0x20000158",
        type: 8,
        importedfromelf: "true"
      }
    ],
    mode: "direct",
    frequency: 1,
    accesspoint: 0
  };

  it("should send read on GroupVar1 and receive a read of this Group ", function(done) {
    let nb = 0;
    acqServMock.config({
      stlinkRes: 1
    }); // set OK for STLINK
    var flow = [
      {
        id: "p1",
        z: "f1",
        type: "probe",
        probeid: "4579",
        frequency: "10",
        protocol: "SWD"
      },
      {
        id: "f1",
        type: "tab",
        label: "test flow"
      },
      {
        id: "n1",
        z: "f1",
        type: "acquisition out",
        name: "myNameOut",
        probeconfig: "p1"
      },
      {
        id: "n2",
        z: "f1",
        type: "acquisition in",
        name: "myNameIn",
        probeconfig: "p1"
      },
      {
        id: "h1",
        z: "f1",
        type: "helper"
      }
    ];
    helper.load(acquisition, flow, function() {
      var n1 = helper.getNode("n1");
      n1.should.have.property("name", "myNameOut");
      n1.probe.on("ready", msg => {
        should(msg).exactly("(NameOfMCU)");
        // The ready is received twice because of AcqIn and AcqOut nodes open the Acquisition Services object.
        // then the data is received when the both are ready.
        if (nb === 1) {
          console.log(
            "AckOut : Ready received in a second time then send groupVar1"
          );
          n1.receive({
            topic: "read",
            payload: groupVar1
          });
        } else {
          nb = 1;
        }
        myClock.tick(1);
      });
      n1.probe.on("data", function() {
        console.log("AckOut : data received from acqServ: ");
      });

      var n2 = helper.getNode("n2");
      n2.should.have.property("name", "myNameIn");
      n2.probe.on("ready", msg => {
        should(msg).exactly("(NameOfMCU)");
      });
      n2.probe.on("data", function(msgout) {
        expect(msgout).to.deep.equal(inAcq);
        console.log("AckIn : right data received from acqServ : ");
        done();
      });
      myClock.tick(1000);
    });
  });
});
