/*!
 * module acquisitionservices
 * Copyright(c) 2019 STMicroelectronics
 */
/// <reference types="node" />
import { EventEmitter } from "events";
/**
 * This class aims at offering some generic services to interface with various probe drivers (as stlinkprotocol).
 * It inherits from EventEmitter class.
 */
export declare class AcquisitionServices extends EventEmitter {
  /**
   * This static service aims at discovering the probes connected to the host.
   *
   * @static
   * @param callback this function is called back with the array of discovered probes.
   * A single probe is represented in the array by an object holding a probeName string property,
   * a probeId string property , and an object property property.
   * @memberof AcquisitionServices
   */
  public static probesDiscovery(
    callback: (
      probesList: Array<{
        /** name of the probe */
        probeName: string;
        /** identifier of the probe */
        probeId: string;
        /** object holding a property by type of supported protocol, each of the supported protocol
         * property value is an array of string representing the list of supported frequencies.
         */
        property: {
          [protocol: string]: string[];
        };
      }>,
      updated: boolean
    ) => void
  ): void;
  /**
   * This service aims at identifying a stLink probe by blinking its LED.
   *
   * @static
   * @param {string} probeId the identifier of a probe
   * @returns {{status:{boolean}, err:{string}, errLevel:{string}}} the result of the blink command on the probe
   * @memberof AcquisitionServices
   */
  public static blinkLed(
    probeId: string
  ):
    | {
        status: boolean;
        err: string;
        errLevel?: undefined;
      }
    | {
        status: boolean;
        err: string;
        errLevel: string;
      };
  /**
   * Set the connection type to be used between the driver and the probe.
   *
   * @static
   * @param {string} connectionType The requested connection type, could be either "p2p" or "tcp"
   * @returns {string} The actual connection type that will be used in further commands
   * @memberof AcquisitionServices
   */
  public static initConnectionType(connectionType: string): string;
  /**
   *
   * @static Initialize the logger with the provided level.
   * @param {string} level The level of logger, could be either off, fatal, error, warn, debug, info, trace.
   * @memberof AcquisitionServices
   */
  public static initLogger(level: string): void;
  /**
   * Count of all acquisitions on going from various probes.
   */
  private static AcqCount;
  /**
   * list of connected probes.
   */
  private static listProbe;
  /**
   * Date of the last discovery of connected probes.
   */
  private static lastUpdate;
  /**
   * Type of connection between stLinkDriver and the probe, either "tcp" or "p2p" (default)
   */
  private static connectionType;
  public probeId: string;
  public reconnectTime: number;
  public option: {
    /** protocol to used between probe and MCU target. */
    protocol: string;
    /** frequency to use between probe and MCU target. */
    frequency: string;
    /** connection type either p2p or tcp */
    connectionType: string;
  };
  /**
   * List of groups for which an acquisition is ongoing.
   */
  private groups;
  /**
   * timeout set to relaunch connect() function every "reconnectTime" .
   */
  private tout;
  /**
   * This private service implements the read command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private read;
  /**
   * This private service implements the start command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private start;
  /**
   * This private service implements the stop command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private stop;
  /**
   * This private service implements the write command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private write;
  /**
   * This private service aims at actually connecting the probe to the MCU target.
   * @param probeDiscovery boolean to request for new probe discovery before open of the targeted probe.
   */
  private connectToTarget;
  /**
   * Internal service used to read values in MCU memory target at a given frequency
   * @param context AcquisitionServices object context on which the reading is actually done.
   * @param data parameters of the read command
   * @param callback function called back on error
   * @param oneshot only one single reading of memory target is done if set to true
   */
  private readMemory;
  /**
   * Private method used to stop an ongoing acquisition for a dedicated group
   * @param groupname string representing the name of the group for which the acquisition must be stopped
   * @param callback function called back if no acquisition is ongoing for the provided groupname
   */
  private stopAcquisition;
  /**
   * Private method used to close openned probes
   */
  private closeProbes;
  /**
   * Instantiate a new  Acquisitionservices class object for a dedicated probe and configure the future
   * communication link between the probe and the MCU target
   * @param probeId the identifier of a probe for which a new AcquisitionServices object is instantiated.
   * This probeId is a public member of the object.
   * @param reconnectTime the interval between 2 connections to the probe.
   * @param option object holding 2 properties to configure the communication link between the probe
   * and the MCU target :
   *    protocol : protocol to be used by the probe to communicate with the target MCU
   *    frequency : the frequency to be used by the probe to communicate with the target MCU
   *    connectionType: the type of the connection to be used by the stlink to communicate with the probe,
   *    either p2p or tcp mode.
   */
  constructor(
    probeId: string,
    reconnectTime: number,
    option: {
      /** protocol to used between probe and MCU target. */
      protocol: string;
      /** frequency to use between probe and MCU target. */
      frequency: string;
      /** connection type either p2p or tcp */
      connectionType: string;
    }
  );
  /**
   * This public service aims at actually connecting the probe to the MCU target.
   * @param callback function called back after the attempt to actually connect the probe to the MCU target
   */
  public connect(): void;
  /**
   * This public service aims at closing the connection between the probe and the MCU target
   * @param cb function called back when the connection is effectively closed
   */
  public close(cb: () => void): void;
  /**
   * This service aims at sending a command to the connected probe. The list of possible commands are :
   *  read : to perform a single reading of memory target
   *  write : to write some values in memory target
   *  start: to start an acquisition loop at a configurable frequency
   *  stop : to stop an ongoing acquisition.
   * @param commandType represents the command to send
   * @param data object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  public sendCommand(commandType: string, data: any, callback: (err: any) => void): void;
}
