/*!
 * module acquisitionservices
 * Copyright(c) 2019 STMicroelectronics
 */

import { getLogger, updateLevel } from "@stm32/stm32cubemonitor-logger";
const logger = getLogger("acquisitionservices");

import stlinkprotocol, { resultCode } from "@stm32/stm32cubemonitor-stlink-protocol";
import { EventEmitter } from "events";
import object from "lodash/fp";

/** Duration in ms between consecutive dataSend to NR */
const REFRESH = 50;

/**
 * This class aims at offering some generic services to interface with various probe drivers (as stlinkprotocol).
 * It inherits from EventEmitter class.
 */
export class AcquisitionServices extends EventEmitter {
  /**
   * This static service aims at discovering the probes connected to the host.
   *
   * @static
   * @param callback this function is called back with the array of discovered probes.
   * A single probe is represented in the array by an object holding a probeName string property,
   * a probeId string property , and an object property property.
   * @memberof AcquisitionServices
   */
  public static probesDiscovery(
    callback: (
      probesList: {
        /** name of the probe */
        probeName: string;
        /** identifier of the probe */
        probeId: string;
        /** object holding a property by type of supported protocol, each of the supported protocol
         * property value is an array of string representing the list of supported frequencies.
         */
        property: { [protocol: string]: string[] };
      }[],
      updated: boolean
    ) => void
  ) {
    let updated = false;
    if (AcquisitionServices.AcqCount === 0) {
      AcquisitionServices.listProbe = stlinkprotocol.probeDiscovery(AcquisitionServices.connectionType);
      AcquisitionServices.lastUpdate = Date.now();
      updated = true;
    }
    callback(AcquisitionServices.listProbe, updated);
  }
  /**
   * This service aims at identifying a stLink probe by blinking its LED.
   *
   * @static
   * @param {string} probeId the identifier of a probe
   * @returns {{status:{boolean}, err:{string}, errLevel:{string}}} the result of the blink command on the probe
   * @memberof AcquisitionServices
   */
  public static blinkLed(probeId: string) {
    logger.debug("blinkLed on ", probeId);
    if (AcquisitionServices.AcqCount !== 0) {
      return { status: false, err: "Acquisition ongoing" };
    }

    const res = stlinkprotocol.blinkLed(probeId, AcquisitionServices.connectionType);
    if (res.result === resultCode.STLINK_OK) {
      return { status: true, err: "" };
    }
    const errLvl = res.result === resultCode.STLINK_JTAG_CMD_NOT_SUPPORTED ? "warning" : "error";
    return { status: false, err: stlinkprotocol.errorString(res.result), errLevel: errLvl };
  }
  /**
   * Set the connection type to be used between the driver and the probe.
   *
   * @static
   * @param {string} connectionType The requested connection type, could be either "p2p" or "tcp"
   * @returns {string} The actual connection type that will be used in further commands
   * @memberof AcquisitionServices
   */
  public static initConnectionType(connectionType: string) {
    logger.debug("initConnectionType with ", connectionType);
    if (connectionType === "tcp") {
      AcquisitionServices.connectionType = stlinkprotocol.isTcpAvailable() ? "tcp" : "p2p";
    } else {
      AcquisitionServices.connectionType = connectionType;
    }
    return AcquisitionServices.connectionType;
  }
  /**
   *
   * @static Initialize the logger with the provided level.
   * @param {string} level The level of logger, could be either off, fatal, error, warn, debug, info, trace.
   * @memberof AcquisitionServices
   */
  public static initLogger(level: string) {
    updateLevel(level);
  }
  /**
   * Count of all acquisitions on going from various probes.
   */
  private static AcqCount = 0;
  /**
   * list of connected probes.
   */
  private static listProbe;
  /**
   * Date of the last discovery of connected probes.
   */
  private static lastUpdate = Date.now();

  /**
   * Type of connection between stLinkDriver and the probe, either "tcp" or "p2p" (default)
   */
  private static connectionType: string;
  /**
   * List of groups for which an acquisition is ongoing.
   */
  private groups: any;

  /**
   * timeout set to relaunch connect() function every "reconnectTime" .
   */
  private tout: any;

  /**
   * Instantiate a new  Acquisitionservices class object for a dedicated probe and configure the future
   * communication link between the probe and the MCU target
   * @param probeId the identifier of a probe for which a new AcquisitionServices object is instantiated.
   * This probeId is a public member of the object.
   * @param reconnectTime the interval between 2 connections to the probe.
   * @param option object holding 2 properties to configure the communication link between the probe
   * and the MCU target :
   *    protocol : protocol to be used by the probe to communicate with the target MCU
   *    frequency : the frequency to be used by the probe to communicate with the target MCU
   *    connectionType: the type of the connection to be used by the stlink to communicate with the probe,
   *    either p2p or tcp mode.
   */
  constructor(
    public probeId: string,
    public reconnectTime: number,
    public option: {
      /** protocol to used between probe and MCU target. */
      protocol: string;
      /** frequency to use between probe and MCU target. */
      frequency: string;
      /** connection type either p2p or tcp */
      connectionType: string;
    }
  ) {
    super();
    logger.debug("call to new acquisitionservices ", probeId, reconnectTime, option);
    this.groups = {};
  }

  /**
   * This public service aims at actually connecting the probe to the MCU target.
   * @param callback function called back after the attempt to actually connect the probe to the MCU target
   */
  public connect() {
    let status;
    if (AcquisitionServices.AcqCount === 0) {
      if (Date.now() - AcquisitionServices.lastUpdate > this.reconnectTime) {
        AcquisitionServices.lastUpdate = Date.now();
        status = this.connectToTarget(true);
      } else {
        status = this.connectToTarget(false);
      }
      if (status) {
        this.closeProbes();
      }
    }

    if (this.tout != null) {
      clearTimeout(this.tout);
      this.tout = null;
    }
    this.tout = setTimeout(() => {
      this.connect();
    }, this.reconnectTime);
  }

  /**
   * This public service aims at closing the connection between the probe and the MCU target
   * @param cb function called back when the connection is effectively closed
   */
  public close(cb: () => void) {
    logger.debug("close", this.probeId, this.option.connectionType);
    for (const groupname in this.groups) {
      if (this.groups.hasOwnProperty(groupname)) {
        this.stopAcquisition(groupname, cb);
      }
    }
    if (this.tout != null) {
      clearTimeout(this.tout);
      this.tout = null;
    }
    this.closeProbes();
    cb();
    logger.debug("Emit close" + this.probeId);
    this.emit("close");
  }

  /**
   * This service aims at sending a command to the connected probe. The list of possible commands are :
   *  read : to perform a single reading of memory target
   *  write : to write some values in memory target
   *  start: to start an acquisition loop at a configurable frequency
   *  stop : to stop an ongoing acquisition.
   * @param commandType represents the command to send
   * @param data object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  public sendCommand(commandType: string, data: any, callback: (err) => void) {
    logger.debug("Arguments : commandType = ", commandType, ", data = ", data, ")");
    const myData = object.cloneDeep(data);
    if (commandType) {
      const command = commandType.toLowerCase();
      switch (command) {
        case "read":
          this.read(myData, callback);
          break;
        case "start":
          this.start(myData, callback);
          break;
        case "stop":
          this.stop(myData, callback);
          break;
        case "write":
          this.write(myData, callback);
          break;
        default:
          callback("Unknown command");
          break;
      }
    }
  }

  /**
   * This private service implements the read command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private read(myData: any, callback: (err) => void) {
    if (!this.connectToTarget(false)) {
      callback("Probe " + this.probeId + " not connected");
      return;
    }
    this.readMemory(this, myData, callback, true);
    this.closeProbes();
  }

  /**
   * This private service implements the start command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private start(myData: any, callback: (err) => void) {
    if (!this.connectToTarget(false)) {
      callback("Probe " + this.probeId + " not connected");
      return;
    }

    if (
      myData.hasOwnProperty("groupname") &&
      myData.hasOwnProperty("variablelist") &&
      myData.hasOwnProperty("frequency") &&
      myData.variablelist.length > 0
    ) {
      // if acquisition already started on this group for this probe, don't start a new one
      if (this.groups[myData.groupname]) {
        return;
      }

      const variableslist = myData.variablelist.map((variable) => {
        const myvar = object.cloneDeep(variable);
        myvar.values = [];
        return myvar;
      });
      const myGroup = {
        first: true,
        timerRead: {},
        variables: variableslist
      };
      this.groups[myData.groupname] = myGroup;
      AcquisitionServices.AcqCount++;
      myGroup.timerRead = setImmediate(() => {
        this.readMemory(this, myData, callback);
      });
    } else {
      callback("Parameter missing");
    }
  }

  /**
   * This private service implements the stop command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private stop(myData: any, callback: (err) => void) {
    this.stopAcquisition(myData.groupname, callback);
    this.closeProbes();
  }

  /**
   * This private service implements the write command.
   * @param myData object holding various configuration parameters for the command such as groupname,
   * variablelist, frequency
   * @param callback function called back in case of error (parameter missing, ...)
   */
  private write(myData: any, callback: (err) => void) {
    // if acquisition already on going for this probe, don't re open it
    if (Object.entries(this.groups).length === 0) {
      if (!this.connectToTarget(false)) {
        callback("Probe " + this.probeId + " not connected");
        return;
      }
    }
    logger.debug("call to stlinkprotocol.writeMemoryGroup ", this.probeId, myData, AcquisitionServices.connectionType);
    const { result } = stlinkprotocol.writeMemoryGroup(this.probeId, myData, AcquisitionServices.connectionType);
    if (result === resultCode.STLINK_OK) {
      logger.debug("Successfull write on ", this.probeId, ", data = ", myData, ")");
    } else {
      const errorString = stlinkprotocol.errorString(result);
      const error =
        "Unsuccessful write (" + errorString + ") on " + this.probeId + ", for data = " + JSON.stringify(myData);
      logger.warn(error);
      const respond = {
        payload: errorString,
        topic: "error"
        // TODO change error type according to error
      };
      this.emit("error", respond);
      callback(error);
    }
    // if acquisition already on going for this probe, don't close it
    if (Object.entries(this.groups).length === 0) {
      this.closeProbes();
    }
  }

  /**
   * This private service aims at actually connecting the probe to the MCU target.
   * @param probeDiscovery boolean to request for new probe discovery before open of the targeted probe.
   */
  private connectToTarget(probeDiscovery: boolean) {
    logger.debug(
      "call to stlinkprotocol.connectToTarget ",
      this.probeId,
      this.option.protocol,
      this.option.frequency,
      probeDiscovery,
      AcquisitionServices.connectionType
    );
    const res = stlinkprotocol.connectToTarget(
      this.probeId,
      this.option.protocol,
      this.option.frequency,
      probeDiscovery,
      AcquisitionServices.connectionType
    );
    if (res.result === resultCode.STLINK_OK) {
      logger.debug("Emit open following the success of connect on target ", res.MCUName ? `(${res.MCUName})` : "");
      this.emit("open", res.MCUName ? `(${res.MCUName})` : "");
      return true;
    } else {
      const errorString = stlinkprotocol.errorString(res.result);
      const respond = {
        payload: errorString,
        topic: "warning"
      };
      this.emit("error", respond);
      logger.debug("Emit close because failure of connect to target : ", errorString);
      this.emit("close", errorString);
      return false;
    }
  }

  /**
   * Internal service used to read values in MCU memory target at a given frequency
   * @param context AcquisitionServices object context on which the reading is actually done.
   * @param data parameters of the read command
   * @param callback function called back on error
   * @param oneshot only one single reading of memory target is done if set to true
   */
  private readMemory(context: AcquisitionServices, data: any, callback: (err) => void, oneshot?: boolean) {
    logger.trace("Arguments : context.probeId = ", context.probeId, ", data = ", data, ", oneshot = ", oneshot);
    let res;
    const tab: any[] = [];
    logger.debug("call to stlinkprotocol.readMemoryGroup ", context.probeId, data, AcquisitionServices.connectionType);
    res = stlinkprotocol.readMemoryGroup(context.probeId, data, AcquisitionServices.connectionType);
    if (res.result === resultCode.STLINK_OK && res.data) {
      if (data.mode === "direct") {
        const respond = {
          data: res.data,
          first: oneshot || context.groups[data.groupname].first,
          groupname: data.groupname
        };
        context.emit("data", respond);
      } else {
        const newDataFormat: object[] = [[]];
        for (let j = 0; j < res.data[0].length; j++) {
          for (let i = 0; i < res.data.length; i++) {
            newDataFormat[i] = [];
            newDataFormat[i][0] = res.data[i][j];
          }
          const respond = {
            data: newDataFormat,
            first: oneshot || context.groups[data.groupname].first,
            groupname: data.groupname
          };
          context.emit("data", respond);
          context.groups[data.groupname].first = false;
        }
      }

      if (!oneshot) {
        context.groups[data.groupname].first = false;
        context.groups[data.groupname].timerRead =
          data.frequency === 0
            ? setImmediate(() => {
                context.readMemory(context, data, callback);
              })
            : setTimeout(context.readMemory, 1000 / data.frequency, context, data, callback);
      }
    } else {
      /* res.result != STLINK_OK or no data*/
      context.stopAcquisition(data.groupname, callback);
      let errorMessage;
      if (res.result === resultCode.STLINK_OK) {
        errorMessage = "No data received";
      } else {
        if (typeof res.result === "number") {
          errorMessage = stlinkprotocol.errorString(res.result)
            ? stlinkprotocol.errorString(res.result)
            : "Unknown driver error";
        } else {
          // error comes from snapshot
          errorMessage = res.result;
        }
        const respond = {
          payload: errorMessage,
          topic: "error"
        };
        context.emit("error", respond);
      }
      // in case error is not specific to read operation on erroneous address
      if (res.result !== resultCode.STLINK_JTAG_READ_ERROR) {
        logger.warn("readMemory : emit close because of ", errorMessage);
        context.emit("close", errorMessage);
      }
      callback(errorMessage);
    }
  }

  /**
   * Private method used to stop an ongoing acquisition for a dedicated group
   * @param groupname string representing the name of the group for which the acquisition must be stopped
   * @param callback function called back if no acquisition is ongoing for the provided groupname
   */
  private stopAcquisition(groupname: string, callback: (err) => void) {
    const group = this.groups[groupname];
    if (group) {
      logger.debug("call to stlinkprotocol.stopSnapshot ", this.probeId);
      const res = stlinkprotocol.stopSnapshot(this.probeId);
      if (res !== resultCode.STLINK_OK) {
        const respond = {
          payload: res,
          topic: "warning"
        };
        this.emit("error", respond);
      }
      if (group.timerRead.hasOwnProperty("_onImmediate")) {
        clearImmediate(group.timerRead);
      } else {
        clearTimeout(group.timerRead);
      }
      delete this.groups[groupname];
      AcquisitionServices.AcqCount--;
    } else {
      callback("Unknown group");
    }
  }

  /**
   * Private method used to close openned probes
   */
  private closeProbes() {
    logger.debug("call to stlinkprotocol.closeProbes ", this.probeId, AcquisitionServices.connectionType);
    const res = stlinkprotocol.closeProbes(this.probeId, AcquisitionServices.connectionType);
    if (res.result !== resultCode.STLINK_OK) {
      logger.warn("Could not close probes because of error: " + res.result);
      const respond = {
        payload: stlinkprotocol.errorString(res.result),
        topic: "warning"
      };
      this.emit("error", respond);
    }
  }
}
