/*!
 * module acquisitionservices
 * Copyright(c) 2019 STMicroelectronics
 */

import { expect } from "chai";
import "mocha";
import sinon from "sinon";

/**
 * Tested module
 */
import { AcquisitionServices } from "../src/index";
const REFRESH = 50;
/**
 * Mocked module
 */
import stlinkprotocol, { readMode, readType, resultCode } from "@stm32/stm32cubemonitor-stlink-protocol";

// Generic callback needed for all tests
let errorMessage;
let cbEmitError;
const genCallBackSend = (err) => {
  errorMessage = err;
  cbEmitError++;
};
let cbOpenCalled;
const genCallBackOpen = () => {
  cbOpenCalled = true;
};
let cbCloseCalled;
const genCallBackClose = () => {
  cbCloseCalled = true;
};
let emitError;
let errorResp;

//////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#probesDiscovery()", () => {
  let stlinkprotocolMock;
  let detectedProbesList;
  let emitCallBack;
  const cbProbList = (probeList) => {
    detectedProbesList = probeList;
    emitCallBack = true;
  };
  const expectedProbesList = [
    {
      probeId: "0674FF575250807567012108",
      probeName: "ST-Link v2-1B (0674FF575250807567012108)",
      property: {
        JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz", "560 kHz", "280 kHz", "140 kHz"],
        SWD: ["4.6 MHz", "1.8 MHz", "950 kHz", "400 kHz", "150 kHz"],
        SWIM: ["HIGH", "LOW"]
      }
    },
    {
      probeId: "066DFF485550755187041714",
      probeName: "ST-Link v2-1B (066DFF485550755187041714)",
      property: {
        JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz", "560 kHz", "280 kHz", "140 kHz"],
        SWD: ["4.6 MHz", "1.8 MHz", "950 kHz", "400 kHz", "150 kHz"],
        SWIM: ["HIGH", "LOW"]
      }
    }
  ];

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
    emitCallBack = false;
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
  });

  // ****************************************************************************************************
  it("should discover probe", () => {
    stlinkprotocolMock
      .expects("probeDiscovery")
      .once()
      .returns(expectedProbesList);

    AcquisitionServices.probesDiscovery(cbProbList);

    expect(detectedProbesList).to.deep.equal(expectedProbesList);
    expect(true).equals(emitCallBack);
  });

  // ****************************************************************************************************
  it("should discover no probe", () => {
    stlinkprotocolMock
      .expects("probeDiscovery")
      .once()
      .returns([]);

    AcquisitionServices.probesDiscovery(cbProbList);

    expect(detectedProbesList).to.deep.equal([]);
    expect(true).equals(emitCallBack);
  });
});
//////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#blinkLed()", () => {
  let stlinkprotocolMock;

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
  });

  // ****************************************************************************************************
  it("should blink the led", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("blinkLed")
      .once()
      .returns(res);

    expect(AcquisitionServices.blinkLed("ID1")).to.deep.equal({ status: true, err: "" });
  });

  // ****************************************************************************************************
  it("should fail because of bad parameter ", () => {
    const res = { result: resultCode.STLINK_BAD_PARAMETER };
    stlinkprotocolMock
      .expects("blinkLed")
      .once()
      .returns(res);

    expect(AcquisitionServices.blinkLed("ID1")).to.deep.equal({
      err: "Bad parameter (7)",
      errLevel: "error",
      status: false
    });
  });
});
//////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#connect()", () => {
  let stlinkprotocolMock;
  let resOpen;
  let resClose;
  let emitOpen;
  let emitClose;

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
    cbOpenCalled = false;
    cbCloseCalled = false;
    cbEmitError = 0;
    errorMessage = "";
    emitOpen = false;
    emitClose = false;
    emitError = false;
    resOpen = "";
    resClose = "";
    errorResp = "";
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
  });

  // ****************************************************************************************************
  it("should connect to the device named MCU", () => {
    const res = { result: resultCode.STLINK_OK, MCUName: "MCU" };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400", false)
      .once()
      .returns(res);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("open", (answer) => {
      resOpen = answer;
      emitOpen = true;
    });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });
    acqSer.connect();

    expect(resOpen).to.deep.equal("(" + res.MCUName + ")");
    expect(true).equal(emitOpen);
    acqSer.close(() => {
      resClose = true;
    });
  });
  // ****************************************************************************************************
  it("should connect to the device with no name", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400", false)
      .once()
      .returns(res);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("open", (answer) => {
      resOpen = answer;
      emitOpen = true;
    });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });

    acqSer.connect();
    expect(resOpen).to.deep.equal("");
    expect(true).equal(emitOpen);
    acqSer.close(() => {
      resClose = true;
    });
  });
  // ****************************************************************************************************
  it("should not connect to the device because TIMEOUT error from STlink-protocol", () => {
    const res = { result: resultCode.STLINK_TIMEOUT, MCUName: "MCU" };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("close", (answer) => {
      resClose = answer;
      emitClose = true;
    });
    acqSer.on("open", (answer) => {
      resOpen = answer;
      emitOpen = true;
    });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });
    acqSer.connect();

    expect(resClose).to.deep.equal("Timeout (9)");
    expect(errorResp).to.deep.equal({ payload: "Timeout (9)", topic: "warning" });
    expect(false).equal(emitOpen);
    expect(true).equal(emitError);
    expect(true).equal(emitClose);
    acqSer.close(() => {
      resClose = true;
    });
  });
});
// //////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#sendCommand(READ,data,cb)", () => {
  let stlinkprotocolMock;
  let myClock;
  let answerToCheck;
  let emitData;

  const dataAwaited = {
    data: [[{ x: 145, y: 112 }], [{ x: 145, y: 234 }]],
    first: true,
    groupname: undefined
  };

  const dataToSend = {
    mode: "direct",
    variablelist: [
      { name: "variable1", address: "0x00112233", type: readType.READ_TYPE_U16 },
      { name: "variable2", address: "0x44556677", type: readType.READ_TYPE_S8 }
    ]
  };
  const dataReceivedBySTLink = {
    data: [[{ x: 145, y: 112 }], [{ x: 145, y: 234 }]],
    result: resultCode.STLINK_OK
  };

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
    const myNow = new Date("12367"); // 328097602800000, (x value in the dataStructure)
    myClock = sinon.useFakeTimers(myNow.getTime());
    emitData = false;
    emitError = false;
    cbOpenCalled = false;
    cbCloseCalled = false;
    cbEmitError = 0;
    errorMessage = "";
    errorResp = "";
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
    myClock.restore();
  });

  // ****************************************************************************************************
  it("should send 'READ' command with one shot read on a group of two variables", () => {
    // ************ DECLARATION **************/
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSend)
      .once()
      .returns(dataReceivedBySTLink);

    stlinkprotocolMock
      .expects("closeProbes")
      .withArgs("ID1")
      .once()
      .returns(res);

    // ************ CODE **************/
    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData = true;
      answerToCheck = answer;
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });
    acqSer.sendCommand("READ", dataToSend, genCallBackSend);

    // ************ CHECK RESULT **************/
    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);

    expect(true).equal(emitData);
    // Check the data event information fits to he awaited data structure
    expect(dataAwaited).to.deep.equal(answerToCheck);
  });
  // ****************************************************************************************************
  it("should send 'read' command with one shot read on a group of two variables", () => {
    // ************ DECLARATION **************/
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSend)
      .once()
      .returns(dataReceivedBySTLink);

    stlinkprotocolMock
      .expects("closeProbes")
      .withArgs("ID1")
      .once()
      .returns(res);

    // ************ CODE **************/
    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData = true;
      answerToCheck = answer;
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });
    acqSer.sendCommand("read", dataToSend, genCallBackSend);

    // ************ CHECK RESULT **************/
    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);

    expect(true).equal(emitData);
    // Check the data event information fits to he awaited data structure
    expect(dataAwaited).to.deep.equal(answerToCheck);
  });
});
// //////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#sendCommand(WRITE,data,cb)", () => {
  let stlinkprotocolMock;

  const dataToWrite = {
    address: "0x00112233",
    name: "variable1",
    type: readType.READ_TYPE_U16,
    value: "123456"
  };

  const setOfDataTowrite = [
    { dataToWrite },
    { name: "variable2", address: "0x44556677", type: readType.READ_TYPE_U16, value: "789012" },
    { name: "variable3", address: "0x88990011", type: readType.READ_TYPE_S8, value: "34" }
  ];

  const dataReceivedBySTLink = {
    result: resultCode.STLINK_OK
  };

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
    cbOpenCalled = false;
    cbCloseCalled = false;
    emitError = false;
    cbEmitError = 0;
    errorMessage = "";
    errorResp = "";
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
  });
  // ****************************************************************************************************
  it("should write a unique variable", () => {
    // ************ DECLARATION **************/
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs("ID1", dataToWrite)
      .once()
      .returns(dataReceivedBySTLink);

    // ************ CODE **************/
    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });
    acqSer.sendCommand("WRITE", dataToWrite, genCallBackSend);

    // ************ CHECK RESULT **************/
    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(0).equal(cbEmitError);
  });
  // ****************************************************************************************************
  it("should write a set of variables", () => {
    // ************ DECLARATION **************/
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs("ID1", setOfDataTowrite)
      .once()
      .returns(dataReceivedBySTLink);

    stlinkprotocolMock
      .expects("closeProbes")
      .withArgs("ID1")
      .once()
      .returns(res);

    // ************ CODE **************/
    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });
    acqSer.sendCommand("WRITE", setOfDataTowrite, genCallBackSend);

    // ************ CHECK RESULT **************/
    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);
  });
});
// //////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#sendCommand(START,data,cb)", () => {
  let stlinkprotocolMock;
  let myClock;
  let emitData;
  let nbOfElement;
  let nbOfElementGroup1;
  let nbOfElementGroup2;
  let answerToCheck;

  const dataToSendFhz0 = {
    frequency: 0,
    groupname: "group1",
    variablelist: [
      { address: "0x00112233", name: "variable1", type: readType.READ_TYPE_U16 },
      { address: "0x44556677", name: "variable2", type: readType.READ_TYPE_S8 }
    ]
  };

  const dataToSendFhz100 = {
    frequency: 100,
    groupname: "group2",
    variablelist: [
      { name: "variable3", address: "0x8899AABB", type: readType.READ_TYPE_U16 },
      { name: "variable2", address: "0x66778899", type: readType.READ_TYPE_S8 }
    ]
  };

  const dataReceivedBySTLink = { data: [[{ x: 145, y: 112 }], [{ x: 145, y: 234 }]], result: resultCode.STLINK_OK };

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
    const myNow = new Date("12367"); // 328097602800000, (x value in the dataStructure)
    myClock = sinon.useFakeTimers(myNow.getTime());
    emitError = false;
    nbOfElement = 0;
    nbOfElementGroup1 = 0;
    nbOfElementGroup2 = 0;
    emitData = 0;
    cbOpenCalled = false;
    cbCloseCalled = false;
    cbEmitError = 0;
    errorMessage = "";
    errorResp = "";
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
    myClock.restore();
  });

  // ****************************************************************************************************
  it("blink should fail because an acquisition is ongoing", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendFhz0)
      .atLeast(1)
      .returns(dataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData++;
      answerToCheck = answer.data;
      for (let i = 0; i < answerToCheck.length; i++) {
        if (
          answerToCheck[i][0].x === dataReceivedBySTLink.data[i][0].x &&
          answerToCheck[i][0].y === dataReceivedBySTLink.data[i][0].y
        ) {
          nbOfElement++;
        }
      }
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });

    acqSer.sendCommand("START", dataToSendFhz0, genCallBackSend);

    myClock.tick(REFRESH); // This command allows the execution of the code during 50 ticks

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);
    expect(REFRESH + 1).equal(emitData);
    expect((REFRESH + 1) * 2).equal(nbOfElement);

    expect(AcquisitionServices.blinkLed("ID1")).to.deep.equal({ status: false, err: "Acquisition ongoing" });
  });

  // ****************************************************************************************************
  it("should get 51 measurements on 'data' notification with START command FHZ 0", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendFhz0)
      .atLeast(1)
      .returns(dataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData++;
      answerToCheck = answer.data;
      for (let i = 0; i < answerToCheck.length; i++) {
        if (
          answerToCheck[i][0].x === dataReceivedBySTLink.data[i][0].x &&
          answerToCheck[i][0].y === dataReceivedBySTLink.data[i][0].y
        ) {
          nbOfElement++;
        }
      }
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });

    acqSer.sendCommand("START", dataToSendFhz0, genCallBackSend);

    myClock.tick(REFRESH); // This command allows the execution of the code during 50 ticks

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);
    expect(REFRESH + 1).equal(emitData);
    expect((REFRESH + 1) * 2).equal(nbOfElement);
  });
  // ****************************************************************************************************
  it("should get 6 measurements on 'data' notification with START command FHZ 100", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendFhz100)
      .atLeast(1)
      .returns(dataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData++;
      answerToCheck = answer.data;
      for (let i = 0; i < answerToCheck.length; i++) {
        if (
          answerToCheck[i][0].x === dataReceivedBySTLink.data[i][0].x &&
          answerToCheck[i][0].y === dataReceivedBySTLink.data[i][0].y
        ) {
          nbOfElement++;
        }
      }
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });

    acqSer.sendCommand("START", dataToSendFhz100, genCallBackSend);

    myClock.tick(REFRESH); // This command allows the execution of the code during 50 ticks

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);
    expect(6).equal(emitData);
    expect(6 * 2).equal(nbOfElement);
  });
  // ****************************************************************************************************
  it("should get 51 measurements on 'data' notification on FHZ 0 and 6 measurement on FHz 100", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .twice()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendFhz100)
      .exactly(5 + 1)
      .returns(dataReceivedBySTLink);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendFhz0)
      .exactly(REFRESH + 1)
      .returns(dataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData++;
      answerToCheck = answer.data;
      for (let i = 0; i < answerToCheck.length; i++) {
        if (
          answerToCheck[i][0].x === dataReceivedBySTLink.data[i][0].x &&
          answerToCheck[i][0].y === dataReceivedBySTLink.data[i][0].y
        ) {
          if (answer.groupname === "group1") {
            nbOfElementGroup1++;
          } else if (answer.groupname === "group2") {
            nbOfElementGroup2++;
          }
        }
      }
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });

    acqSer.sendCommand("START", dataToSendFhz100, genCallBackSend);
    acqSer.sendCommand("START", dataToSendFhz0, genCallBackSend);

    myClock.tick(REFRESH); // This command allows the execution of the code during 50 ticks

    // ************ CHECK RESULT **************/
    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);

    // Check the data event on AcquisitionSevices
    expect(51 + 6).equal(emitData);
    // Check the data event information fits to the awaited data structure
    expect(51 * 2).equal(nbOfElementGroup1);
    expect(6 * 2).equal(nbOfElementGroup2);
  });
  // ****************************************************************************************************
  it("should get 51 measurements on 'data' notification with two call of start command on FHz 0", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .twice()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendFhz0)
      .exactly(REFRESH + 1)
      .returns(dataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData++;
      answerToCheck = answer.data;
      for (let i = 0; i < answerToCheck.length; i++) {
        if (
          answerToCheck[i][0].x === dataReceivedBySTLink.data[i][0].x &&
          answerToCheck[i][0].y === dataReceivedBySTLink.data[i][0].y
        ) {
          nbOfElement++;
        }
      }
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });
    acqSer.sendCommand("start", dataToSendFhz0, genCallBackSend);
    myClock.tick(REFRESH / 2);
    acqSer.sendCommand("start", dataToSendFhz0, genCallBackSend);
    myClock.tick(REFRESH / 2);

    // ************ CHECK RESULT **************/
    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);

    // Check the data event on AcquisitionSevices
    expect(51).equal(emitData);
    // Check the data event information fits to he awaited data structure
    expect(51 * 2).equal(nbOfElement);
  });
});
// //////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#sendCommand(START,data,cb) ERROR or INCOMPLETE CASES", () => {
  let stlinkprotocolMock;
  let myClock;
  let emitData;
  let emitClose;
  let closeReceived;
  let dataReceived;
  let nbOfElement;

  const completeDataToSend = {
    frequency: 0,
    groupname: "group1",
    variablelist: [
      { name: "variable1", address: "0x00112233", type: readType.READ_TYPE_U16 },
      { name: "variable2", address: "0x44556677", type: readType.READ_TYPE_S8 }
    ]
  };
  const incompleteDataToSend = {
    // frequency: 0,
    groupname: "group1",
    variablelist: [
      { name: "variable1", address: "0x00112233", type: readType.READ_TYPE_U16 },
      { name: "variable2", address: "0x44556677", type: readType.READ_TYPE_S8 }
    ]
  };
  const dataReceivedBySTLink = { data: [[{ x: 145, y: 112 }], [{ x: 145, y: 234 }]], result: resultCode.STLINK_OK };
  const nodataReceivedBySTLink = { result: resultCode.STLINK_OK };
  const errorReceivedBySTLink = { result: resultCode.STLINK_DEVICE_NOT_OPENED };

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
    const myNow = new Date("12367"); // 328097602800000, (x value in the dataStructure)
    myClock = sinon.useFakeTimers(myNow.getTime());
    emitData = 0;
    closeReceived = "";
    emitError = false;
    nbOfElement = 0;
    dataReceived = [];
    cbOpenCalled = false;
    cbCloseCalled = false;
    cbEmitError = 0;
    errorMessage = "";
    errorResp = "";
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
    myClock.restore();
  });
  // ****************************************************************************************************
  it("should get an error call back on 'data' notification with the unknown 'started' command", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .never()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", completeDataToSend)
      .never()
      .returns(dataReceivedBySTLink);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", incompleteDataToSend)
      .never()
      .returns(dataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData++;
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });
    acqSer.sendCommand("started", incompleteDataToSend, genCallBackSend);

    myClock.tick(REFRESH);

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(1).equal(cbEmitError);

    expect(0).equal(emitData);
    expect("Unknown command").equals(errorMessage);
  });
  // ****************************************************************************************************
  it("should get an error call back on START command because of missing parameters", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", completeDataToSend)
      .never()
      .returns(dataReceivedBySTLink);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", incompleteDataToSend)
      .never()
      .returns(dataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      emitData++;
    });
    acqSer.on("error", (answer) => {
      emitError = true;
      errorResp = answer;
    });
    acqSer.sendCommand("START", incompleteDataToSend, genCallBackSend);

    myClock.tick(REFRESH);

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(1).equal(cbEmitError);

    expect(0).equal(emitData);
    expect("Parameter missing").equals(errorMessage);
  });
  // ****************************************************************************************************
  it("should close the service because of STLINK_OK but no data returned", () => {
    // ************ DECLARATION **************/
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400", false)
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", completeDataToSend)
      .atLeast(1)
      .returns(nodataReceivedBySTLink);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", incompleteDataToSend)
      .never()
      .returns(nodataReceivedBySTLink);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      dataReceived = answer;
      emitData++;
    });
    acqSer.on("close", (answer) => {
      closeReceived = answer;
      emitClose++;
    });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });

    acqSer.sendCommand("START", completeDataToSend, genCallBackSend);

    myClock.tick(REFRESH);

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(1).equal(cbEmitError);

    expect(0).equal(emitData);
    expect(false).equal(emitError);
    expect("No data received").equal(closeReceived);
  });
  // ****************************************************************************************************
  it("should close the service because of STLINK error on readMemory", () => {
    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400", false)
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", completeDataToSend)
      .once()
      .returns(errorReceivedBySTLink);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", incompleteDataToSend)
      .never()
      .returns(errorReceivedBySTLink);

    stlinkprotocolMock
      .expects("errorString")
      .withArgs(5)
      .twice()
      .returns("Device not opened (5)");

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("data", (answer) => {
      dataReceived = answer;
      emitData++;
    });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });
    acqSer.on("close", (answer) => {
      closeReceived = answer;
      emitClose++;
    });

    acqSer.sendCommand("START", completeDataToSend, genCallBackSend);

    myClock.tick(REFRESH);

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(true).equal(emitError);
    expect(1).equal(cbEmitError);
    expect(0).equal(emitData);
    expect("Device not opened (5)").equal(closeReceived);
    expect({ payload: "Device not opened (5)", topic: "error" }).to.deep.equal(errorResp);
  });
  // ****************************************************************************************************
  it("should send 'WRITE' command ,return an error and verify callbak is called", () => {
    // ************ DECLARATION **************/
    const dataToWrite = {
      address: "0x00112233",
      name: "variable1",
      type: readType.READ_TYPE_U16,
      value: "123456"
    };

    const errorValue =
      "Unsuccessful write (" +
      "Device not opened (5)" +
      ") on " +
      "ID1" +
      ", for data = " +
      JSON.stringify(dataToWrite);

    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs("ID1", dataToWrite)
      .once()
      .returns(errorReceivedBySTLink);

    // ************ CODE **************/
    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });
    acqSer.sendCommand("WRITE", dataToWrite, genCallBackSend);

    // ************ CHECK RESULT **************/
    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(true).equal(emitError);
    expect(1).equal(cbEmitError);
    expect(errorValue).contains(cbEmitError);
    expect({ payload: "Device not found (6)", topic: "warning" }).to.deep.equal(errorResp);
  });
});
// //////////////////////////////////////////////////////////////////////////////////////////////////////////
describe("#sendCommand(STOP,data,cb)", () => {
  let stlinkprotocolMock;

  beforeEach(() => {
    stlinkprotocolMock = sinon.mock(stlinkprotocol);
    const myNow = new Date("0");
    cbOpenCalled = false;
    cbCloseCalled = false;
    cbEmitError = 0;
    errorMessage = "";
    emitError = false;
    errorResp = "";
  });

  afterEach(() => {
    stlinkprotocolMock.verify();
  });
  const dataReceivedBySTLink = { data: [[{ x: 145, y: 112 }], [{ x: 145, y: 234 }]], result: resultCode.STLINK_OK };

  // ****************************************************************************************************
  it("should STOP acquisition after 50 ms", async () => {
    let emit = false;
    const dataToSendForStart = {
      frequency: 0,
      groupname: "group1",
      variablelist: [
        { name: "variable1", address: "0x00112233", type: readType.READ_TYPE_U16 },
        { name: "variable2", address: "0x44556677", type: readType.READ_TYPE_S8 }
      ]
    };
    const dataToSendForStop = {
      groupname: "group1"
    };

    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendForStart)
      .atLeast(REFRESH)
      .returns(dataReceivedBySTLink);

    stlinkprotocolMock
      .expects("closeProbes")
      .withArgs("ID1")
      .once()
      .returns(res);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });
    acqSer.sendCommand("START", dataToSendForStart, genCallBackSend);

    await new Promise((resolve, reject) => {
      setTimeout(() => {
        acqSer.sendCommand("STOP", dataToSendForStop, genCallBackSend);
        emit = true;
        resolve("stop");
      }, 50);
    });

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);

    expect(true).equal(emit);
  });
  // ****************************************************************************************************
  it("should stop acquisition after 50 ms", async () => {
    let emit = false;
    const dataToSendForStart = {
      frequency: 0,
      groupname: "group1",
      variablelist: [
        { name: "variable1", address: "0x00112233", type: readType.READ_TYPE_U16 },
        { name: "variable2", address: "0x44556677", type: readType.READ_TYPE_S8 }
      ]
    };
    const dataToSendForStop = { groupname: "group1" };

    const res = { result: resultCode.STLINK_OK };
    stlinkprotocolMock
      .expects("connectToTarget")
      .withArgs("ID1", "SWD", "400")
      .once()
      .returns(res);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs("ID1", dataToSendForStart)
      .atLeast(REFRESH)
      .returns(dataReceivedBySTLink);

    stlinkprotocolMock
      .expects("closeProbes")
      .withArgs("ID1")
      .once()
      .returns(res);

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });
    acqSer.sendCommand("start", dataToSendForStart, genCallBackSend);

    await new Promise((resolve, reject) => {
      setTimeout(() => {
        acqSer.sendCommand("stop", dataToSendForStop, genCallBackSend);
        emit = true;
        resolve("stop'");
      }, 50);
    });

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(false).equal(emitError);
    expect(0).equal(cbEmitError);

    expect(true).equal(emit);
  });
  // ****************************************************************************************************
  it("should get a callback error when stop acquisition on not started group", async () => {
    // ************ DECLARATION **************/
    const dataToSendForStop = { groupname: "group1" };

    const acqSer = new AcquisitionServices("ID1", 15000, { protocol: "SWD", frequency: "400", connectionType: "p2p" });
    acqSer.on("error", (answer) => {
      errorResp = answer;
      emitError = true;
    });

    acqSer.sendCommand("stop", dataToSendForStop, genCallBackSend);

    expect(false).equal(cbCloseCalled);
    expect(false).equal(cbOpenCalled);
    expect(true).equal(emitError);
    expect(1).equal(cbEmitError);
    expect("Unknown group").equal(errorMessage);
    expect({ payload: "Device not found (6)", topic: "warning" }).to.deep.equal(errorResp);
  });
});
