/*!
 * module @stm32/stm32cubemonitor-logger
 * Copyright(c) 2019 STMicroelectronics
 */
const path = require("path");
const log4js = require("log4js");
const os = require("os");

/**
 * Log level for each category.
 * Possible values are: "fatal", "error", "warn", "info", "debug", "trace", "off".
 */
let acquisitionServicesLogLevel = "info";
let elfParserLogLevel = "info";
let stlinkProtocolLogLevel = "info";
let stmStudioLogLevel = "info";
let defaultlogLevel = "info";

/**
 * Log file location.
 * On windows OS, process.env.TMP is : <user directory>/AppData/Local/temp.
 */
if (!process.env.TMP) {
  process.env.TMP = os.homedir;
}
const logDirName = path.join(process.env.TMP, "STM32CubeMonitor", "logs");
const logFileName = path.join(logDirName, "STM32CubeMonitor.log");

/**
 * Log4js configuration:
 *   - log file older than 1 day and newer than daysToKeep is stored in logFileName.yyyy-MM-dd file
 *   - log file older than daysToKeep value is automatically deleted by log4js.
 */
log4js.configure({
  appenders: {
    STM32CubeMonitor: {
      type: "dateFile",
      filename: logFileName,
      pattern: ".yyyy-MM-dd",
      daysToKeep: 10
    }
  },
  categories: {
    acquisitionservices: { appenders: ["STM32CubeMonitor"], level: acquisitionServicesLogLevel },
    elfparser: { appenders: ["STM32CubeMonitor"], level: elfParserLogLevel },
    stlinkprotocol: { appenders: ["STM32CubeMonitor"], level: stlinkProtocolLogLevel },
    stmstudio: { appenders: ["STM32CubeMonitor"], level: stmStudioLogLevel },
    default: { appenders: ["STM32CubeMonitor"], level: defaultlogLevel }
  }
});

/**
 * Service to extract (from stack trace) method name, file name, line and column where log function is called.
 *
 * @param {*} error - Unused parameter.
 * @param {*} structuredStackTrace - Heap stack trace.
 */
function prepareStackTrace(error, structuredStackTrace) {
  var trace = structuredStackTrace[0];
  return {
    // method name
    name: trace.getMethodName() || trace.getFunctionName() || "<anonymous>",
    // file name
    file: trace.getFileName(),
    // line number
    line: trace.getLineNumber(),
    // column number
    column: trace.getColumnNumber()
  };
}

/**
 * Service to capture stack trace for a caller function.
 *
 * @param {Function} caller - Caller function.
 * @returns {object} - Stack trace information.
 */
function getTrace(caller) {
  let original = Error.prepareStackTrace,
    error = {};
  Error.prepareStackTrace = prepareStackTrace;
  Error.captureStackTrace(error, caller || getTrace);
  var stack = error.stack;
  Error.prepareStackTrace = original;
  return stack;
}

/**
 * Service to format information extracted from stack trace.
 * Format is: [filename - function name (line:column).].
 *
 * @param {*} trace - Information extracted from stack trace.
 * @returns {string} - Formatted trace.
 */
function formatter(trace) {
  let fileName = trace.file
    .split("\\")
    .pop()
    .split("/")
    .pop();
  return "[" + fileName + " - " + trace.name + "(" + trace.line + ":" + trace.column + ")] - ";
}

/**
 * Treatment to overload each log function in order to add stack trace information before logging message to log4js.
 * This is done at the mother class level, then applicable to instance created from it, and even after its creation.
 */
let logger = log4js.getLogger();
["trace", "debug", "info", "warn", "error", "fatal"].forEach(function(method) {
  // Save the original log4js function
  var original = logger.constructor.prototype[method];
  logger.constructor.prototype[method] = function log() {
    // Perform treatment only if "method" level is higher than logger category level
    if (this.isLevelEnabled(method)) {
      let args = [];
      // Get information about the stack trace
      let trace = getTrace(log);
      // Format the information got from the stack trace
      args.push(formatter(trace));
      // Concat information of the stack trace with the message(s) to be logged (arguments)
      args = args.concat([].slice.call(arguments));
      // Call the original log4js function with the concatenated information (stack trace + arguments)
      return original.apply(this, args);
    }
  };
});

/**
 *
 * Reconfigured the logger with the provided level level.
 *
 * @param {string} level - The logger level can take following values :
 * "fatal", "error", "warn", "info", "debug", "trace", "off".
 */
function updateLevel(level) {
  if (
    level &&
    (level === "off" ||
      level === "fatal" ||
      level === "error" ||
      level === "warn" ||
      level === "info" ||
      level === "debug" ||
      level === "trace")
  ) {
    acquisitionServicesLogLevel = elfParserLogLevel = stlinkProtocolLogLevel = level;
    stmStudioLogLevel = defaultlogLevel = level;
    log4js.configure({
      appenders: {
        STM32CubeMonitor: {
          type: "dateFile",
          filename: logFileName,
          pattern: ".yyyy-MM-dd",
          daysToKeep: 10
        }
      },
      categories: {
        acquisitionservices: { appenders: ["STM32CubeMonitor"], level: acquisitionServicesLogLevel },
        elfparser: { appenders: ["STM32CubeMonitor"], level: elfParserLogLevel },
        stlinkprotocol: { appenders: ["STM32CubeMonitor"], level: stlinkProtocolLogLevel },
        stmstudio: { appenders: ["STM32CubeMonitor"], level: stmStudioLogLevel },
        default: { appenders: ["STM32CubeMonitor"], level: defaultlogLevel }
      }
    });
  }
}
module.exports.getLogger = log4js.getLogger;
module.exports.updateLevel = updateLevel;
