/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const logger = require("@stm32/stm32cubemonitor-logger").getLogger("stlinkprotocol");
let ffi = require("ffi-napi");
let ref = require("ref-napi");
let StructType = require("ref-struct-di")(ref);
let ArrayType = require("ref-array-di")(ref);
let path = require("path");

// definition of constants used in DLL interface
let defs = {
  DRV_OK: 0x01,
  DRV_ERR: 0x04,
  DRV_NO_DEVICE: 0x82,
  DRV_INVALID_HA: 0x81,
  DRV_INVALID_SRB: 0xe0,
  DRV_FAILED_INIT: 0xe4,
  DRV_SKIP_STEP: 0xf1,

  DRV_MEMORY_PROBLEM: 0x1000,
  DRV_TIMEOUT: 0x1001,
  DRV_BAD_PARAMETER: 0x1002,
  DRV_OPEN_ERR: 0x1003,
  DRV_TRANSFER_ERR: 0x1004,
  DRV_NO_SUCH_INFO: 0x1050,
  DRV_BAD_BUFFER_SIZE: 0x1051,
  DRV_TRUNCATED_DATA: 0x1052,
  DRV_CMD_NOT_AVAILABLE: 0x1053,
  DRV_DEVICE_NOT_SUPPORTED: 0x1054,
  DRV_PERMISSION_ERR: 0x1055,

  DRV_FAILED: 0x2000,
  DRV_TCP_ERROR: 0x2001, // To be refined later
  DRV_TCP_CANT_CONNECT: 0x2002, // Error when TCP client is not able to connect to the server
  DRV_TCP_CLOSE_ERROR: 0x2003, // Error to close the TCP socket (remaining STLink opened ...)

  DRV_SCSI_ERROR: 0x5000,
  DRV_WIN32_ERROR: 0x10000,

  DRV_TCP_ERROR_SIZE: 4,

  DRV_DEFAULT_TIMEOUT: 5000, // ms

  DRV_REQUEST_WRITE: 0, //Through EP2OUT
  DRV_REQUEST_READ_EP3: 3, //Through EP3IN (SWV)
  DRV_REQUEST_READ_EP1: 1, // Through EP1IN

  DRV_SERIAL_NUM_STR_MAX_LEN: 32,

  DRV_DEFAULT_SENSE_LEN: 14,

  DRV_CLOSE_TCP_AUTO: 0, //constant defining close type in STLink_CloseDeviceTcp() closeTcp parameter
  DRV_CLOSE_ONLY_DEVICE: 1, // Close the STLink device but do not close the TCP socket : need to call later STLink_CloseDeviceTcp(CLOSE_ONLY_TCP)
  DRV_CLOSE_ONLY_TCP: 2, // Close the TCP socket (and potentially stlink-server) if no remaining opened STLink device

  DRV_DEFAULT_IF: 0, // Debug Interface
  DRV_STLINK_TCP: 2,

  DRV_OPEN_STLINK_EXCLUSIVE_ACCESS: 1,
  DRV_OPEN_STLINK_SHARED_ACCESS: 0
};

// definition of types used on the DLL interface
let TDeviceInfo2 = StructType(
  {
    StLinkUsbId: ref.types.int32, // DWORD - Tcp server Device cookie in little endian format, to use in STLINK_TCP_CMD_OPEN_DEVICE
    EnumUniqueId: ArrayType(ref.types.char, defs.DRV_SERIAL_NUM_STR_MAX_LEN), // Unique instance ID from system enumeration (equal to device Serial number in some cases, but not always ...)
    VendorId: ref.types.uint16, // Vendor ID from USB device descriptor (system enumeration)
    ProductId: ref.types.uint16, // Product ID from USB device descriptor (system enumeration)
    DeviceUsed: ref.types.byte // On windows, equal to 1 if device interface was already opened from externally when enumerating or trying to open
  },
  { packed: true }
);

let TDeviceRequest = StructType(
  {
    CDBLength: ref.types.byte, // Command header length (CDB on mass storage) - used only on ST-Link/V1
    CDBByte: ArrayType(ref.types.byte, 16), // Command header (CDB on mass storage)
    InputRequest: ref.types.byte, // 0 (output) or input on ST-Link/V1, REQUEST_WRITE, REQUEST_READ_EP1 or REQUEST_READ_EP3 on ST-Link/V2
    Buffer: ref.refType(ref.types.void), // Data for data stage (if any)
    BufferLength: ref.types.int32, // DWORD - Size of data stage (0 if no data stage)
    SenseLength: ref.types.byte, // To be initialized: DEFAULT_SENSE_LEN - used only on ST-Link/V1
    Sense: ArrayType(ref.types.byte, 16) // Used only on ST-Link/V1
  },
  { packed: true }
);

let Handle = ref.types.int64;

// DLL Loading + DLL interface definition
let platform = process.platform;
let stLinkUsbDriver = null;
let stLinkDrv, stLinkDrvFfiMapping;

if (platform === "win32") {
  // Build dll path. In case of asar pack, use the "unpacked" path for the dll.
  stLinkUsbDriver = path.join(__dirname, "STLinkUSBDriver.dll").replace("app.asar", "app.asar.unpacked");
  logger.debug("St Link dll path :", stLinkUsbDriver);
  stLinkDrvFfiMapping = ffi.Library(stLinkUsbDriver, {
    STLink_Reenumerate: ["int", ["int", "bool"]],
    STLink_ReenumerateTcp: ["int", ["int", "bool", "string", "string"]],
    STLink_GetNbDevices: ["int", ["int"]],
    STLink_GetDeviceInfo2: ["int", ["int", "byte", ref.refType(TDeviceInfo2), "long"]],
    STLink_OpenDevice: ["int", ["int", "byte", "byte", ref.refType(Handle)]],
    STLink_OpenDeviceTcp: ["int", ["int", "int32", "byte"]],
    STLink_CloseDevice: ["int", [Handle]],

    STLink_CloseDeviceTcp: ["int", ["int", "byte"]],
    STLink_SendCommand: ["int", [Handle, ref.refType(TDeviceRequest), "long"]],
    STLink_SendCommandTcp: ["int", ["int", ref.refType(TDeviceRequest), "long"]]
  });
} else if (platform === "linux" || platform === "darwin") {
  // Build dll path. In case of asar pack, use the "unpacked" path for the dll.
  stLinkUsbDriver = path.join(__dirname, "libSTLinkUSBDriver").replace("app.asar", "app.asar.unpacked");
  let stLinkDrvSo = ffi.Library(stLinkUsbDriver, {
    _Z18STLink_Reenumerate20TEnumStlinkInterfaceh: ["int", ["int", "bool"]],
    _Z19STLink_GetNbDevices20TEnumStlinkInterface: ["int", ["int"]],
    _Z21STLink_GetDeviceInfo220TEnumStlinkInterfacehP12TDeviceInfo2j: [
      "int",
      ["int", "byte", ref.refType(TDeviceInfo2), "long"]
    ],
    _Z17STLink_OpenDevice20TEnumStlinkInterfacehhPPv: ["int", ["int", "byte", "byte", ref.refType(Handle)]],
    _Z18STLink_CloseDevicePv: ["int", [Handle]],
    _Z18STLink_SendCommandPvP14TDeviceRequestj: ["int", [Handle, ref.refType(TDeviceRequest), "long"]],
    _Z21STLink_ReenumerateTcp20TEnumStlinkInterfacehPcS0_: ["int", ["int", "bool", "string", "string"]],
    _Z20STLink_OpenDeviceTcp20TEnumStlinkInterfacejh: ["int", ["int", "int32", "byte"]],
    _Z21STLink_CloseDeviceTcpjh: ["int", ["int", "byte"]],
    _Z21STLink_SendCommandTcpjP14TDeviceRequestj: ["int", ["int", ref.refType(TDeviceRequest), "long"]]
  });
  stLinkDrvFfiMapping = {
    STLink_Reenumerate: stLinkDrvSo._Z18STLink_Reenumerate20TEnumStlinkInterfaceh,
    STLink_GetNbDevices: stLinkDrvSo._Z19STLink_GetNbDevices20TEnumStlinkInterface,
    STLink_GetDeviceInfo2: stLinkDrvSo._Z21STLink_GetDeviceInfo220TEnumStlinkInterfacehP12TDeviceInfo2j,
    STLink_OpenDevice: stLinkDrvSo._Z17STLink_OpenDevice20TEnumStlinkInterfacehhPPv,
    STLink_CloseDevice: stLinkDrvSo._Z18STLink_CloseDevicePv,
    STLink_SendCommand: stLinkDrvSo._Z18STLink_SendCommandPvP14TDeviceRequestj,
    STLink_ReenumerateTcp: stLinkDrvSo._Z21STLink_ReenumerateTcp20TEnumStlinkInterfacehPcS0_,
    STLink_OpenDeviceTcp: stLinkDrvSo._Z20STLink_OpenDeviceTcp20TEnumStlinkInterfacejh,
    STLink_CloseDeviceTcp: stLinkDrvSo._Z21STLink_CloseDeviceTcpjh,
    STLink_SendCommandTcp: stLinkDrvSo._Z21STLink_SendCommandTcpjP14TDeviceRequestj
  };
} else {
  throw new Error("unsupported plateform :" + platform + " for stLinkUsbDriver");
}

stLinkDrv = (function() {
  let map = {};
  for (let [key, value] of Object.entries(stLinkDrvFfiMapping)) {
    // eslint-disable-next-line require-jsdoc
    let callDriverAPI = (...args) => {
      logger.debug(key, args);
      return value.apply(this, args);
    };
    //property writable set to true for unitary test purpose
    map = Object.defineProperty(map, key, { value: callDriverAPI, writable: true });
  }
  return map;
})();

// module exports
module.exports.stLinkDrv = stLinkDrv;
module.exports.TDeviceInfo2 = TDeviceInfo2;
module.exports.TDeviceRequest = TDeviceRequest;
module.exports.Handle = Handle;
module.exports.defs = defs;
