/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */

var defsDrv = require("./STLinkUSBDriver").defs;

const apidefs = {
  STLINK_SWD_PROTOCOL: "SWD",
  STLINK_JTAG_PROTOCOL: "JTAG",
  // common result codes
  STLINK_OK: 0x01,
  STLINK_ERR: 0x02,
  STLINK_DRIVER_ERR: 0x03,
  STLINK_WRONG_DEVICE_MODE: 0x04,
  STLINK_DEVICE_NOT_OPENED: 0x05,
  STLINK_NO_DEVICE: 0x06,
  STLINK_BAD_PARAMETER: 0x07,
  STLINK_DEVICE_NOT_SUPPORTED: 0x08,
  STLINK_TIMEOUT: 0x09,
  STLINK_DEVICE_NOT_RESPONDING: 0x0a,
  ERROR_BUFFER_READ: "ERROR_BUFFER_READ" // Specific error when extracting data from buffer read in memory target
};

/**
 * Generates error string according to given error code.
 *
 * @param {number} errorCode - Value of the error code.
 * @returns {string} - The error string.
 */
function errorString(errorCode) {
  switch (errorCode) {
    case apidefs.STLINK_BAD_PARAMETER:
      return "Bad parameter";

    case apidefs.STLINK_DEVICE_NOT_OPENED:
      return "Device not opened";

    case apidefs.STLINK_WRONG_DEVICE_MODE:
      return "Device in wrong mode";

    case apidefs.STLINK_DEVICE_NOT_SUPPORTED:
      return "Device not supported";

    case apidefs.STLINK_NO_DEVICE:
      return "Device not found";

    case apidefs.STLINK_ERR:
      return "Error";

    case apidefs.STLINK_DRIVER_ERR:
      return "Driver error";

    case apidefs.STLINK_TIMEOUT:
      return "Timeout";

    case apidefs.STLINK_DEVICE_NOT_RESPONDING:
      return "A device attached to the system is not functioning";

    case apidefs.ERROR_BUFFER_READ:
      return "An error occurred when extracting data from buffer read in memory target ";

    default:
      return "";
  }
}

/**
 * Wrap driver result into a result code.
 *
 * @param {number} drvres - Result of the driver.
 * @returns {number} - Result code.
 */
function drvToAPI(drvres) {
  let apires;
  //system error - Windows error - ERROR_GEN_FAILURE (0x1F)
  const DRV_WIN32_ERROR = defsDrv.DRV_WIN32_ERROR | 0x1f;

  if (drvres == defsDrv.DRV_OK) return apidefs.STLINK_OK;

  switch (drvres) {
    // no need to have a case for defsDrv.DRV_OK, it has been handled before

    case defsDrv.DRV_NO_DEVICE: {
      apires = apidefs.STLINK_NO_DEVICE;
      break;
    }

    case defsDrv.DRV_BAD_PARAMETER: {
      apires = apidefs.STLINK_BAD_PARAMETER;
      break;
    }

    case defsDrv.DRV_DEVICE_NOT_SUPPORTED: {
      apires = apidefs.STLINK_DEVICE_NOT_SUPPORTED;
      break;
    }

    case defsDrv.DRV_TIMEOUT: {
      apires = apidefs.STLINK_TIMEOUT;
      break;
    }

    case defsDrv.DRV_ERR:
    case defsDrv.DRV_TRANSFER_ERR:
    case DRV_WIN32_ERROR: {
      apires = apidefs.STLINK_DEVICE_NOT_RESPONDING;
      break;
    }

    default: {
      // default is generic driver error
      apires = apidefs.STLINK_DRIVER_ERR;
    }
  }

  return apires;
}

module.exports.apidefs = apidefs;
module.exports.drvToAPI = drvToAPI;
module.exports.errorString = errorString;
