/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const readline = require("readline").createInterface({
  input: process.stdin,
  output: process.stdout
});

var stLinkProtRoot = require("./stlinkinterface");
var stLinkProt = require("./stlinkprotocol").cli;
//var clidefs = stLinkProt.clidefs;
var apidefs = require("./stlinkprotocol").apidefs;

/**
 * Add prompt at the beginning of command line.
 *
 */
function prompt() {
  readline.question("stlink>", name => {
    parseInput(name);
  });
}

var nbDevices = 0;
var typeConnect = "p2p";
var accessPointId = 0;

/**
 * Displays device information on console.
 *
 * @param {object} deviceInfo - Device informations to display.
 */
function displayDeviceInfo(deviceInfo) {
  if (deviceInfo == null || deviceInfo === undefined) {
    return;
  }

  console.log("Device Info:");
  console.log("            Device number :" + deviceInfo.DevNumber);
  //console.log('            StLinkUsbId : ' + deviceInfo.StLinkUsbId);
  console.log("            EnumUniqueId : " + deviceInfo.EnumUniqueId);
  console.log("            VendorId : " + deviceInfo.VendorId);
  console.log("            ProductId : " + deviceInfo.ProductId);
  //console.log('            DeviceUsed : ' + deviceInfo.DeviceUsed);
}

/**
 * Parses command line and execute requested command.
 *
 * @param {string} input - Input command line.
 */
function parseInput(input) {
  let args = input.split(" ");

  switch (args[0]) {
    case "q":
    case "quit":
      {
        stLinkProtRoot.closeProbes("", typeConnect);
        readline.close();
        process.exit();
      }
      break;

    case "h":
    case "help": {
      console.log("-------------------------------------------------------------------------------------------");
      console.log("Usage : node stlink ");
      console.log("   Prompt options :");
      console.log("     h/help : display this text");
      console.log("     q/quit : exit this program");
      console.log("     <command> [<args>]:");
      console.log(
        "                 y/settypeconnect <mode>: Set connect type to specified mode either p2p (default value) or tcp " // eslint-disable-line
      );
      console.log("                 e/enum: Enumerate ST-Link devices connected over USB");
      console.log("                 n/nbdev: Number of ST-Link devices connected over USB");
      console.log("                 i/info <nb> : Information for device number <nb> (starting from 0)");
      console.log("                 l/list: List information about enumerated devices");
      console.log("                 o/open <nb>: open device number <nb>");
      console.log("                 c/close <nb>: close device number <nb>");
      console.log("                 b/blink <nb> : Make the led blinks for device number <nb>");
      console.log("                 m/mode <nb> : get current mode of device number <nb>");
      console.log("                 v/version <nb> : get version info of device number <nb>");
      console.log("                 x/exitdfu <nb> : exit DFU mode for device number <nb>");
      console.log("                 j/enterjtag <nb>: enter JTAG mode for device number <nb>");
      console.log("                 s/enterswd <nb> : enter SWD mode for device number <nb>");
      console.log("                 xj/exitjtag <nb> : exit JTAG/SWD mode for device number <nb>");
      console.log("                 initAP <nb> <id>: init Access port <id> for device number <nb>");
      console.log("                 closeAP <nb> <id>: close access port <id> for device number <nb> ");
      console.log("                 id/idcodes <nb> : read ID codes for device number <nb>");
      console.log("                 d/discover : probeDiscovery API");
      console.log("                 t/target <enumid> <protocol> <frequency>: connectToTarget API");
      console.log("                 r32/readmem32 <nb> <addHex> [<size>]: read <size> bytes from device <nb> ");
      console.log("                                                       at address <addHex> in 32-bit mode");
      console.log("                 r16/readmem16 <nb> <addHex> [<size>]: read <size> bytes from device <nb> ");
      console.log("                                                       at address <addHex> in 16-bit mode");
      console.log("                 r8/readmem8 <nb> <addHex> [<size>]: read <size> bytes from device <nb> ");
      console.log("                                                     at address <addHex> in 8-bit mode");
      console.log(
        "                 rm/readmemmode <enumid> <addHex> [<size>] [<mode>] [<accessPoint>] : readMemory API"
      );
      console.log("                 rg/readmemgroup <enumid> <addHex> [<size>] <value> [<accessPoint>]: ");
      console.log("                                        read <size> bytes of value to device <enumid> ");
      console.log("                                        at address <addHex> in 32-bit mode");
      console.log(
        "                 rt/readmemtyped <enumid> <addHex> <type> [<nbelem>] [<accessPoint>]: readMemory API"
      );
      console.log("                 w8 <nb> <addHex> [<size>]: write <size> bytes from device <nb> ");
      console.log("                                                     at address <addHex> in 8-bit mode");
      console.log("                 w32 <nb> <addHex> [<size>]: write <size> bytes from device <nb> ");
      console.log("                                                     at address <addHex> in 32-bit mode");
      console.log("                 wg/writememgroup <enumid> <addHex> [<size>] <value> [<accessPoint>]: ");
      console.log("                                        write <size> bytes of value to device <enumid> ");
      console.log("                                        at address <addHex> in 32-bit mode");
      console.log("-------------------------------------------------------------------------------------------");
      break;
    }

    case "y":
    case "settypeconnect":
      typeConnect = args[1];
      if (typeConnect != "tcp" && typeConnect != "p2p") {
        console.log("unknown connection type");
        typeConnect = "p2p";
      }
      console.log("ST-Link connection type set to " + typeConnect);
      break;

    case "e":
    case "enum": {
      let result = stLinkProt.reenumerate(typeConnect);
      console.log("ST-Link USB enumeration " + (result ? "succeeded" : "failed"));

      // automatically get the number of enumerated devices
      nbDevices = stLinkProt.getNbDevices(typeConnect);

      break;
    }

    case "n":
    case "nbdev": {
      nbDevices = stLinkProt.getNbDevices();
      console.log("Number of enumerated ST-Link devices : " + nbDevices);

      break;
    }

    case "i":
    case "info": {
      if (nbDevices == 0) {
        console.log("No devices available. Try reenumeration ?");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let deviceInfo = stLinkProt.getDeviceInfo(devnb);

      displayDeviceInfo(deviceInfo);

      break;
    }

    case "o":
    case "open": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res;
      if (typeConnect == "tcp") {
        res = stLinkProt.openDeviceTcp(devnb);
      } else {
        res = stLinkProt.openDevice(devnb);
      }

      if (res == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " opened successfully");
      } else {
        console.log("Device " + devnb + " could not be opened : " + stLinkProt.errorString(res));
      }

      break;
    }

    case "c":
    case "close": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res;
      if (typeConnect == "tcp") {
        res = stLinkProt.closeDeviceTcp(devnb);
      } else {
        res = stLinkProt.closeDevice(devnb);
      }

      if (res == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " closed successfully");
      } else {
        console.log("Device " + devnb + " could not be closed : " + stLinkProt.errorString(res));
      }

      break;
    }

    case "m":
    case "mode": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.getCurrentMode(devnb, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : GET CURRENT MODE command sent successfully");
        console.log(
          "        Current Mode : 0x" +
            res.mode.toString(16).padStart(2, "0") +
            " (" +
            stLinkProt.modeString(res.mode) +
            ")"
        );
      }

      if (res.DFUVersion != 0) {
        console.log(
          "Device " + devnb + " : GET CURRENT MODE command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "v":
    case "version": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.getVersion(devnb, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : GET VERSION command sent successfully ");

        console.log("        ST-Link Version : " + res.STLinkVersion);
        console.log("        JTAG Version : " + res.JTAGVersion);
        console.log("        SWIM Version : " + res.SWIMVersion);
      } else {
        console.log(
          "Device " + devnb + " : GET VERSION command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "x":
    case "exitdfu": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.exitDFUMode(devnb);

      if (res == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : EXIT DFU MODE command sent successfully ");
      } else {
        console.log("Device " + devnb + " : EXIT DFU MODE command could not be sent : " + stLinkProt.errorString(res));
      }

      break;
    }

    case "j":
    case "enterjtag": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.enterJTAG(devnb, typeConnect);

      if (res == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : ENTER JTAG command sent successfully ");
      } else {
        console.log("Device " + devnb + " : ENTER JTAG command could not be sent : " + stLinkProt.errorString(res));

        // exit JTAG mode if badly entered
        stLinkProt.exitJTAG(devnb, typeConnect);
      }

      break;
    }

    case "s":
    case "enterswd": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.enterSWD(devnb, typeConnect);

      if (res == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : ENTER SWD command sent successfully ");
      } else {
        console.log("Device " + devnb + " : ENTER SWD command could not be sent : " + stLinkProt.errorString(res));

        // exit JTAG mode if badly entered
        stLinkProt.exitJTAG(devnb, typeConnect);
      }

      break;
    }

    case "xj":
    case "exitjtag": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.exitJTAG(devnb, typeConnect);

      if (res == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : EXIT JTAG command sent successfully ");
      } else {
        console.log("Device " + devnb + " : EXIT JTAG command could not be sent : " + stLinkProt.errorString(res));
      }

      break;
    }

    case "initAP": {
      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (args[2] === undefined) {
        console.log("No Acces port id provided !");
        accessPointId = 0;
      } else {
        accessPointId = Number(args[2]);

        let res = stLinkProt.initAP(devnb, accessPointId, typeConnect);

        if (res == apidefs.STLINK_OK) {
          console.log("Device " + devnb + " : Access point " + accessPointId + " successfully initialized");
        } else {
          console.log(
            "Device " + devnb + " : AP " + accessPointId + " could not be initialized : " + stLinkProt.errorString(res)
          );
          accessPointId = 0;
        }
      }
      break;
    }

    case "closeAP": {
      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }
      let devnb = Number(devnbstr);

      if (args[2] === undefined) {
        console.log("No Acces port id provided !");
      } else {
        let apId = Number(args[2]);
        let res = stLinkProt.closeAP(devnb, apId, typeConnect);

        if (res == apidefs.STLINK_OK) {
          console.log("Device " + devnb + " : Access point " + apId + " successfully closed");
        } else {
          console.log("Device " + devnb + " : AP " + apId + " could not be closed : " + stLinkProt.errorString(res));
        }
      }
      accessPointId = 0;
      break;
    }

    case "id":
    case "idcodes": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.JTAG_ReadIdCodes(devnb, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : JTAG READ ID CODES command sent successfully ");

        console.log("        MCU ID : 0x" + res.MCUID.toString(16).padStart(3, "0"));
        console.log("        MCU Name : " + res.MCUName);
      } else {
        console.log(
          "Device " + devnb + " : JTAG READ ID CODES command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "w8": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnbstr);
        console.log("Device number starts at 0...");
        break;
      }

      let addstr = args[2];

      if (addstr === undefined || addstr == "") {
        console.log("No address provided !");
        break;
      }

      let add = parseInt(addstr, 16);
      if (add === undefined) {
        console.log("Bad address provided : " + addstr);
        break;
      }

      let sizestr = args[3];

      let size;

      if (sizestr === undefined || sizestr == "") {
        console.log("Default size used (1 byte)");
        size = 1;
      } else {
        size = Number(sizestr);
        if (size === undefined) {
          console.log("Bad size provided : " + sizestr);
          break;
        }
      }

      let buffer = Buffer.from([0x02]);
      console.log(devnb, add, size, buffer);
      let res = stLinkProt.JTAG_WriteMemory8Bit(devnb, add, size, buffer, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : JTAG READMEM8 command sent successfully ");
      } else {
        console.log(
          "Device " + devnb + " : JTAG READMEM8 command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "w32": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnbstr);
        console.log("Device number starts at 0...");
        break;
      }

      let addstr = args[2];

      if (addstr === undefined || addstr == "") {
        console.log("No address provided !");
        break;
      }

      let add = parseInt(addstr, 16);
      if (add === undefined) {
        console.log("Bad address provided : " + addstr);
        break;
      }

      let sizestr = args[3];
      let size;

      if (sizestr === undefined || sizestr == "") {
        console.log("Default size used (4 bytes)");
        size = 4;
      } else {
        size = Number(sizestr);
        if (size === undefined) {
          console.log("Bad size provided : " + sizestr);
          break;
        }
      }

      let buffer = Buffer.from([0x01, 0x02, 0x03, 0x04]);

      let res = stLinkProt.JTAG_WriteMemory32Bit(devnb, add, size, buffer, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : JTAG READMEM32 command sent successfully ");
      } else {
        console.log(
          "Device " + devnb + " : JTAG READMEM32 command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "r32":
    case "readmem32": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnbstr);
        console.log("Device number starts at 0...");
        break;
      }

      let addstr = args[2];

      if (addstr === undefined || addstr == "") {
        console.log("No address provided !");
        break;
      }

      let add = parseInt(addstr, 16);
      if (add === undefined) {
        console.log("Bad address provided : " + addstr);
        break;
      }

      let sizestr = args[3];
      let size;

      if (sizestr === undefined || sizestr == "") {
        console.log("Default size used (4 bytes)");
        size = 4;
      } else {
        size = Number(sizestr);
        if (size === undefined) {
          console.log("Bad size provided : " + sizestr);
          break;
        }
      }

      let res = stLinkProt.JTAG_ReadMemory32Bit(devnb, add, size, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : JTAG READMEM32 command sent successfully ");

        for (let i = 0; i < size >> 2; i++) {
          console.log(
            "        Data@ 0x" +
              (add + i * 4).toString(16).padStart(8, "0") +
              " : 0x" +
              res.data[i * 4].toString(16).padStart(2, "0") +
              " 0x" +
              res.data[i * 4 + 1].toString(16).padStart(2, "0") +
              " 0x" +
              res.data[i * 4 + 2].toString(16).padStart(2, "0") +
              " 0x" +
              res.data[i * 4 + 3].toString(16).padStart(2, "0")
          );
        }
      } else {
        console.log(
          "Device " + devnb + " : JTAG READMEM32 command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "r16":
    case "readmem16": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnbstr);
        console.log("Device number starts at 0...");
        break;
      }

      let addstr = args[2];

      if (addstr === undefined || addstr == "") {
        console.log("No address provided !");
        break;
      }

      let add = parseInt(addstr, 16);
      if (add === undefined) {
        console.log("Bad address provided : " + addstr);
        break;
      }

      let sizestr = args[3];
      let size;

      if (sizestr === undefined || sizestr == "") {
        console.log("Default size used (2 bytes)");
        size = 2;
      } else {
        size = Number(sizestr);
        if (size === undefined) {
          console.log("Bad size provided : " + sizestr);
          break;
        }
      }

      let res = stLinkProt.JTAG_ReadMemory16Bit(devnb, add, size, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : JTAG READMEM16 command sent successfully ");

        for (let i = 0; i < size >> 1; i++) {
          console.log(
            "        Data@ 0x" +
              (add + i * 2).toString(16).padStart(8, "0") +
              " : 0x" +
              res.data[i * 2].toString(16).padStart(2, "0") +
              " 0x" +
              res.data[i * 2 + 1].toString(16).padStart(2, "0")
          );
        }
      } else {
        console.log(
          "Device " + devnb + " : JTAG READMEM16 command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "r8":
    case "readmem8": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnbstr);
        console.log("Device number starts at 0...");
        break;
      }

      let addstr = args[2];

      if (addstr === undefined || addstr == "") {
        console.log("No address provided !");
        break;
      }

      let add = parseInt(addstr, 16);
      if (add === undefined) {
        console.log("Bad address provided : " + addstr);
        break;
      }

      let sizestr = args[3];
      let size;

      if (sizestr === undefined || sizestr == "") {
        console.log("Default size used (1 byte)");
        size = 1;
      } else {
        size = Number(sizestr);
        if (size === undefined) {
          console.log("Bad size provided : " + sizestr);
          break;
        }
      }

      let res = stLinkProt.JTAG_ReadMemory8Bit(devnb, add, size, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Device " + devnb + " : JTAG READMEM8 command sent successfully ");

        for (let i = 0; i < size; i++) {
          console.log(
            "        Data@ 0x" +
              (add + i).toString(16).padStart(8, "0") +
              " : 0x" +
              res.data[i].toString(16).padStart(2, "0")
          );
        }
      } else {
        console.log(
          "Device " + devnb + " : JTAG READMEM8 command could not be sent : " + stLinkProt.errorString(res.result)
        );
      }

      break;
    }

    case "rm":
    case "readmemmode": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let probeid = args[1];

      if (probeid === undefined || probeid == "") {
        console.log("No probeid provided !");
        break;
      }

      let addstr = args[2];

      if (addstr === undefined || addstr == "") {
        console.log("No address provided !");
        break;
      }

      if (!addstr.startsWith("0x")) {
        console.log('Address must start with "0x" : ' + addstr);
        break;
      }

      let add = parseInt(addstr, 16);
      if (add === undefined) {
        console.log("Bad address provided : " + addstr);
        break;
      }

      let sizestr = args[3];
      let size;

      if (sizestr === undefined || sizestr == "") {
        console.log("Default size used (4 bytes)");
        size = 4;
      } else {
        size = Number(sizestr);
        if (size === undefined) {
          console.log("Bad size provided : " + sizestr);
          break;
        }
      }

      let modestr = args[4];

      if (args[5] != undefined) accessPointId = Number(args[5]);

      let modenb;
      let mode;

      if (modestr === undefined || modestr == "") {
        console.log("auto read mode used");
        mode = apidefs.READ_MODE_AUTO;
      } else {
        modenb = Number(modestr);
        if (modenb === undefined) {
          console.log("Bad mode provided : " + modestr);
          break;
        }
        switch (modenb) {
          case 32:
            mode = apidefs.READ_MODE_32_BIT;
            break;
          case 16:
            mode = apidefs.READ_MODE_16_BIT;
            break;
          case 8:
            mode = apidefs.READ_MODE_8_BIT;
            break;
          case 0:
            mode = apidefs.READ_MODE_AUTO;
            break;
          default:
            mode = -1;
        }
      }

      if (mode == -1) {
        console.log("Bad mode provided : " + modestr);
        break;
      }

      let res = stLinkProtRoot.readMemoryMode(probeid, addstr, size, mode, accessPointId, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Read memory from target " + probeid + " successful");

        switch (mode) {
          // for 32-bit mode : 32-bit output
          case apidefs.READ_MODE_32_BIT:
            for (let i = 0; i < size >> 2; i++) {
              console.log(
                "        Data@ 0x" +
                  (add + i * 4).toString(16).padStart(8, "0") +
                  " : 0x" +
                  res.data[i * 4].toString(16).padStart(2, "0") +
                  " 0x" +
                  res.data[i * 4 + 1].toString(16).padStart(2, "0") +
                  " 0x" +
                  res.data[i * 4 + 2].toString(16).padStart(2, "0") +
                  " 0x" +
                  res.data[i * 4 + 3].toString(16).padStart(2, "0")
              );
            }

            break;

          // for 16-bit mode : 16-bit output
          case apidefs.READ_MODE_16_BIT:
            for (let i = 0; i < size >> 1; i++) {
              console.log(
                "        Data@ 0x" +
                  (add + i * 2).toString(16).padStart(8, "0") +
                  " : 0x" +
                  res.data[i * 2].toString(16).padStart(2, "0") +
                  " 0x" +
                  res.data[i * 2].toString(16).padStart(2, "0")
              );
            }

            break;

          // for 8-bit and auto mode : 8-bit output
          case apidefs.READ_MODE_8_BIT:
          case apidefs.READ_MODE_AUTO: {
            for (let i = 0; i < size; i++) {
              console.log(
                "        Data@ 0x" +
                  (add + i).toString(16).padStart(8, "0") +
                  " : 0x" +
                  res.data[i].toString(16).padStart(2, "0")
              );
            }

            break;
          }
        }
      } else {
        console.log("Read memory from target " + probeid + " failed : " + stLinkProt.errorString(res.result));
      }

      break;
    }

    case "rt":
    case "readmemtyped": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let probeid = args[1];

      if (probeid === undefined || probeid == "") {
        console.log("No probeid provided !");
        break;
      }

      let addstr = args[2];

      if (addstr === undefined || addstr == "") {
        console.log("No address provided !");
        break;
      }

      if (!addstr.startsWith("0x")) {
        console.log('Address must start with "0x" : ' + addstr);
        break;
      }

      let add = parseInt(addstr, 16);
      if (add === undefined) {
        console.log("Bad address provided : " + addstr);
        break;
      }

      let typestr = args[3];
      let type;

      if (typestr === undefined || typestr == "") {
        type = apidefs.READ_TYPE_U8;
      } else {
        switch (typestr) {
          case "U8":
            type = apidefs.READ_TYPE_U8;
            break;
          case "S8":
            type = apidefs.READ_TYPE_S8;
            break;
          case "U16":
            type = apidefs.READ_TYPE_U16;
            break;
          case "S16":
            type = apidefs.READ_TYPE_S16;
            break;
          case "U32":
            type = apidefs.READ_TYPE_U32;
            break;
          case "S32":
            type = apidefs.READ_TYPE_S32;
            break;
          case "U64":
            type = apidefs.READ_TYPE_U64;
            break;
          case "S64":
            type = apidefs.READ_TYPE_S64;
            break;
          case "F":
            type = apidefs.READ_TYPE_FLOAT;
            break;
          case "D":
            type = apidefs.READ_TYPE_DOUBLE;
            break;

          default:
            type = -1;
        }
      }

      if (type == -1) {
        console.log("Bad type provided : " + typestr);
        break;
      }

      let nbstr = args[4];

      if (args[5] != undefined) accessPointId = Number(args[5]);

      let nbelem;

      if (nbstr === undefined || nbstr == "") {
        console.log("Default number of elements used (1)");
        nbelem = 1;
      } else {
        nbelem = Number(nbstr);
        if (nbelem === undefined) {
          console.log("Bad number of elements provided : " + nbstr);
          break;
        }
      }

      let res = stLinkProtRoot.readMemoryTyped(probeid, addstr, nbelem, type, accessPointId, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Read memory (typed) from target " + probeid + " successful");

        for (let i = 0; i < nbelem; i++) {
          console.log(
            "        Data@ 0x" +
              (add + i * res.elemSize).toString(16).padStart(8, "0") +
              " : " +
              res.data[i].y +
              " (" +
              res.data[i].x +
              ")"
          );
        }
      } else {
        console.log("Read memory from target " + probeid + " failed : " + stLinkProt.errorString(res.result));
      }

      break;
    }

    case "wg":
    case "writememgroup": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let probeid = args[1];

      if (probeid === undefined || probeid == "") {
        console.log("No probeid provided !");
        break;
      }

      let group = [
        { name: "var0", address: "0x20005100", type: apidefs.READ_TYPE_U8, value: "200" },
        { name: "var1", address: "0x20005104", type: apidefs.READ_TYPE_U8, value: "0xC9" },
        { name: "var2", address: "0x20005108", type: apidefs.READ_TYPE_S8, value: "-115" },
        { name: "var3", address: "0x20005000", type: apidefs.READ_TYPE_U32, value: "2147483650" },
        { name: "var4", address: "0x20005020", type: apidefs.READ_TYPE_U32, value: "0x11223344" },
        { name: "var5", address: "0x20005200", type: apidefs.READ_TYPE_S32, value: "2147483622" },
        { name: "var6", address: "0x20005300", type: apidefs.READ_TYPE_S32, value: "-9876" },
        { name: "var7", address: "0x20005400", type: apidefs.READ_TYPE_U16, value: "1025" },
        { name: "var8", address: "0x20005420", type: apidefs.READ_TYPE_S16, value: "-1032" },
        { name: "var9", address: "0x20005430", type: apidefs.READ_TYPE_FLOAT, value: "1000000000000.1" },
        { name: "var10", address: "0x20005450", type: apidefs.READ_TYPE_DOUBLE, value: "14.12" },
        { name: "var11", address: "0x20005460", type: apidefs.READ_TYPE_DOUBLE, value: "-814.12" },
        { name: "var12", address: "0x20005470", type: apidefs.READ_TYPE_DOUBLE, value: "0x1010778810101010" }
      ];

      if (args[4] != undefined) accessPointId = Number(args[4]);

      let res = stLinkProtRoot.writeMemoryGroup(probeid, group, accessPointId, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Write memory group from target " + probeid + " successful");

        console.log(res.data); // the resulting array
      } else {
        console.log("Write memory group from target " + probeid + " failed : " + stLinkProt.errorString(res.result));
      }

      break;
    }
    case "rg":
    case "readmemgroup": {
      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      let probeid = args[1];

      if (probeid === undefined || probeid == "") {
        console.log("No probeid provided !");
        break;
      }

      /*let group = [
        { name: "var1", address: "0x20008000", type: apidefs.READ_TYPE_U8, nb: 4 },
        { name: "var6", address: "0x20008810", type: apidefs.READ_TYPE_S32, nb: 2 },
        { name: "var2", address: "0x20008010", type: apidefs.READ_TYPE_S8, nb: 3 },
        { name: "var7", address: "0x20009000", type: apidefs.READ_TYPE_DOUBLE },
        { name: "var10", address: "0x2000A010", type: apidefs.READ_TYPE_FLOAT, nb: 3 },
        { name: "var3", address: "0x20008400", type: apidefs.READ_TYPE_U16 },
        { name: "var4", address: "0x20008410", type: apidefs.READ_TYPE_S16, nb: 1 },
        { name: "var5", address: "0x20008800", type: apidefs.READ_TYPE_U32 },
        { name: "var8", address: "0x20009010", type: apidefs.READ_TYPE_U64 },
        { name: "var9", address: "0x2000A000", type: apidefs.READ_TYPE_S64, nb: 2 }
      ];*/

      let group = [
        { name: "var0", address: "0x20008000", type: apidefs.READ_TYPE_U8, nb: 4 },
        { name: "var6", address: "0x20008810", type: apidefs.READ_TYPE_S32, nb: 2 },
        { name: "var2", address: "0x20008010", type: apidefs.READ_TYPE_S8, nb: 3 },
        { name: "var7", address: "0x20009000", type: apidefs.READ_TYPE_DOUBLE },
        { name: "var1", address: "0x2000A010", type: apidefs.READ_TYPE_FLOAT, nb: 3 },
        { name: "var3", address: "0x20008400", type: apidefs.READ_TYPE_U16 },
        { name: "var4", address: "0x20008410", type: apidefs.READ_TYPE_S16, nb: 1 },
        { name: "var5", address: "0x20008800", type: apidefs.READ_TYPE_U32 },
        { name: "var8", address: "0x20009010", type: apidefs.READ_TYPE_U64 },
        { name: "var9", address: "0x2000A000", type: apidefs.READ_TYPE_S64, nb: 2 }
      ];

      if (args[4] != undefined) accessPointId = Number(args[4]);

      let res = stLinkProtRoot.readMemoryGroup(probeid, group, accessPointId, typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Read memory group from target " + probeid + " successful");

        console.log(res.data); // the resulting array
      } else {
        console.log("Read memory group from target " + probeid + " failed : " + stLinkProt.errorString(res.result));
      }

      break;
    }

    case "l":
    case "list": {
      stLinkProt.reenumerate(typeConnect);

      nbDevices = stLinkProt.getNbDevices(typeConnect);

      if (nbDevices == 0) {
        console.log("No devices found !");
        break;
      }

      for (let i = 0; i < nbDevices; i++) {
        let devInfo = stLinkProt.getDeviceInfo(i, typeConnect);

        displayDeviceInfo(devInfo);
      }

      break;
    }

    case "discover":
    case "d": {
      let listprobe = [];
      listprobe = stLinkProtRoot.probeDiscovery(typeConnect);
      nbDevices = stLinkProt.getNbDevices(typeConnect);

      if (nbDevices > 0) {
        listprobe.forEach(probe => {
          console.log("Device Info:");
          console.log(`            Device Name : ${probe.probeName}`);
          console.log(`            Device Id : ${probe.probeId}`);
          for (const [protocol, frequencies] of Object.entries(probe.property)) {
            console.log(`            Protocol : ${protocol} at following frequencies : ${frequencies.join()}.`);
          }
        });
      } else {
        console.log("No devices found !");
      }
      break;
    }

    case "target":
    case "t": {
      // if (nbDevices == 0) {
      //   console.log("No devices found !");
      //   break;
      // }

      let probeid = args[1];

      if (probeid === undefined || probeid == "") {
        console.log("No probeid provided !");
        break;
      }

      let protocol = args[2];

      if (protocol === undefined || protocol == "") {
        console.log("No protocol provided !");
        break;
      }

      let frequency = args[3];

      // do not ask for re-discovery by default, to avoid side effects when multiple probes are connected
      let res = stLinkProtRoot.connectToTarget(probeid, protocol, frequency, true, typeConnect);

      // update number of found devices, in case it has been updated by connectToTarget call
      nbDevices = stLinkProt.getNbDevices(typeConnect);

      if (res.result == apidefs.STLINK_OK) {
        console.log("Connection to target " + probeid + " successful");
        console.log("        MCU ID : 0x" + res.MCUID.toString(16).padStart(3, "0"));
        console.log("        MCU Name : " + res.MCUName);
      } else {
        console.log("Connection to target " + probeid + " failed : " + stLinkProt.errorString(res.result));
      }

      break;
    }
    case "b":
    case "blink": {
      let devnbstr = args[1];

      if (devnbstr === undefined || devnbstr == "") {
        console.log("No device number provided !");
        break;
      }

      let devnb = Number(devnbstr);
      if (devnb === undefined || devnb > nbDevices - 1) {
        console.log("Bad device number provided : " + devnb);
        console.log("Device number starts at 0...");
        break;
      }

      let res = stLinkProt.JTAG_BlinkLed(devnb, typeConnect);

      if (res == apidefs.STLINK_OK) {
        console.log("The Led of device " + devnb + " is blinking");
      } else {
        console.log("The Led of device " + devnb + " could not be blinked : " + stLinkProt.errorString(res));
      }

      break;
    }
    default: {
      console.log("unknown command : " + input);
    }
  }

  prompt();
}

prompt();
