/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const STLogger = require("@stm32/stm32cubemonitor-logger");
const logger = STLogger.getLogger("stlinkprotocol");
let stLinkDrv = require("./STLinkUSBDriver").stLinkDrv;
let TDeviceRequest = require("./STLinkUSBDriver").TDeviceRequest;
let defsDrv = require("./STLinkUSBDriver").defs;

let apicom = require("./apicommon");
let apidefs = apicom.apidefs;
let drvToAPI = apicom.drvToAPI;

let JTAGV3Frequencies = {};
let SWDV3Frequencies = {};
let JTAGV3CurrentFreq = {};
let SWDV3CurrentFreq = {};

const defs = {
  // ST-Link JTAG command
  STLINK_CMD_JTAG: 0xf2,

  // JTAG sub-commands + parameters
  STLINK_JTAG_ENTER2: 0x30, // to enter JTAG/SWD with new API
  STLINK_JTAG_ENTER_SWD: 0xa3,
  STLINK_JTAG_ENTER_JTAG_NO_CORE_RESET: 0xa4,
  STLINK_JTAG_AP_NO_CORE: 0,
  STLINK_JTAG_AP_CORTEXM_CORE: 1,
  STLINK_JTAG_EXIT: 0x21,

  STLINK_JTAG_READ_IDCODES: 0x31, // New in API2

  STLINK_JTAG_READMEM_32BIT: 0x07,
  STLINK_JTAG_READMEM_16BIT: 0x47, // New in ST-Link/V2 from version J26
  STLINK_JTAG_READMEM_8BIT: 0x0c,
  STLINK_JTAG_WRITEMEM_8BIT: 0x0d,
  STLINK_JTAG_GETLASTRWSTATUS2: 0x3e, // Added in V2J15
  STLINK_JTAG_WRITEMEM_32BIT: 0x08,

  STLINK_SWD_SET_FREQ: 0x43, // New in ST-Link/V2 from version J20
  STLINK_JTAG_SET_FREQ: 0x44, // New in ST-Link/V2 from version J24
  STLINK_JTAG_BLINK_LED: 0x49, // New in ST-Link/V2 from version J28
  STLINK_JTAG_INIT_AP: 0x4b, // New in ST-Link/V2 from version J28
  STLINK_JTAG_CLOSE_AP_DBG: 0x4c, // New in ST-Link/V2 from version J28
  STLINK_JTAG_SET_COM_FREQ: 0x61, // New in ST-Link/V3
  STLINK_JTAG_GET_COM_FREQ: 0x62, // New in ST-Link/V3

  // JTAG status codes
  STLINK_JTAG_OK: 0x80,
  STLINK_JTAG_UNKNOWN_ERROR: 0x01,
  STLINK_JTAG_SPI_ERROR: 0x02, //internal error in JTAG protocol implementation entering SWD/JTAG mode
  STLINK_JTAG_DMA_ERROR: 0x03,
  STLINK_JTAG_UNKNOWN_JTAG_CHAIN: 0x04, //unexpected IRSCAN chain width
  STLINK_JTAG_NO_DEVICE_CONNECTED: 0x05, //no target detected on the JTAG chain
  STLINK_JTAG_INTERNAL_ERROR: 0x06, //command ignored (unknown parameter or wrong ST-Link mode when called)
  STLINK_JTAG_CMD_WAIT: 0x07,
  STLINK_JTAG_CMD_ERROR: 0x08, //Bad parameter when setting JTAG frequency
  STLINK_JTAG_GET_IDCODE_ERROR: 0x09,
  STLINK_JTAG_ALIGNMENT_ERROR: 0x0a,
  STLINK_JTAG_DBG_POWER_ERROR: 0x0b,
  STLINK_JTAG_WRITE_ERROR: 0x0c,
  STLINK_JTAG_WRITE_VERIF_ERROR: 0x0d,
  STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE: 0x0e, // New from V2J24

  STLINK_SWD_AP_WAIT: 0x10, //SWD protocol implementation error
  STLINK_SWD_AP_FAULT: 0x11, //SWD protocol implementation error
  STLINK_SWD_AP_ERROR: 0x12, //SWD protocol implementation error
  STLINK_SWD_AP_PARITY_ERROR: 0x13, //SWD protocol implementation error
  STLINK_SWD_DP_WAIT: 0x14, //SWD protocol implementation error
  STLINK_SWD_DP_FAULT: 0x15, //SWD protocol implementation error
  STLINK_SWD_DP_ERROR: 0x16, //SWD protocol implementation error
  STLINK_SWD_DP_PARITY_ERROR: 0x17, //SWD protocol implementation error

  STLINK_SWD_AP_WDATA_ERROR: 0x18,
  STLINK_SWD_AP_STICKY_ERROR: 0x19, // No more specific to SWD from V2J24
  STLINK_SWD_AP_STICKYORUN_ERROR: 0x1a, // No more specific to SWD from V2J24
  STLINK_AP_ALREADY_USED: 0x1b,
  STLINK_TRACE_AP_TURNAROUND: 0x1c,
  STLINK_BAD_AP: 0x1d, //Incorrect Access Port index

  STLINK_SWV_NOT_AVAILABLE: 0x20, // New in ST-Link/V2
  STLINK_NO_JUMP_TO_USB_LOADER: 0x21, // New in ST-Link/V2, from V2J18S4 - The jump is not supported by the hardware

  STLINK_JTAG_TCPID_NOT_FOUND: 0x30, // New in ST-Link/V2 from version J28
  STLINK_JTAG_TCPID_MAX_REACHED: 0x31, // New in ST-Link/V2 from version J28

  STLINK_JTAG_CONF_CHANGED: 0x40, //configuration has been changed regarding previous settings due to new clock.
  STLINK_JTAG_FREQ_NOT_SUPPORTED: 0x41,
  STLINK_JTAG_UNKNOWN_CMD: 0x42,

  STLINK_JTAG_RUNNING: 0x80,
  STLINK_JTAG_HALT: 0x81,

  STLINK_JTAG_OLD_ERROR: 0x81 //0x81 is the general error code of first API
};

const jtagapidefs = {
  STLINK_JTAG_PROTOCOL_NOT_SUPPORTED: 0xf0,
  STLINK_SWD_PROTOCOL_NOT_SUPPORTED: 0xf1,
  STLINK_JTAG_FREQUENCY_NOT_SUPPORTED: 0xf2,
  STLINK_SWD_FREQUENCY_NOT_SUPPORTED: 0xf3,
  STLINK_JTAG_READ_ERROR: 0xf4,
  STLINK_JTAG_ERROR: 0xf5,
  STLINK_JTAG_UNKNOWN_JTAG_CHAIN: 0xf6,
  STLINK_JTAG_NO_DEVICE_CONNECTED: 0xf7,
  STLINK_JTAG_WRITE_ERROR: 0xf8,
  STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE: 0xf9,
  STLINK_SWD_ERROR: 0xfa,
  STLINK_AP_ERROR: 0xfb,
  STLINK_SWV_NOT_AVAILABLE: 0xfc,
  STLINK_JTAG_TCP_ERROR: 0xfd,
  STLINK_JTAG_16BIT_ACCESS_NOT_SUPPORTED: 0xfe,
  STLINK_JTAG_CMD_NOT_SUPPORTED: 0xff
};

/**
 * Wrap JTAG command status to known result code.
 *
 * @param {number} JTAGstatus - JTAG command status.
 * @returns {number} - Result code.
 */
function JTAGtoAPI(JTAGstatus) {
  switch (JTAGstatus) {
    case defs.STLINK_JTAG_OK:
      return apidefs.STLINK_OK;

    case defs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN:
      return jtagapidefs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN;
    case defs.STLINK_JTAG_NO_DEVICE_CONNECTED:
      return jtagapidefs.STLINK_JTAG_NO_DEVICE_CONNECTED;
    case defs.STLINK_JTAG_WRITE_ERROR:
    case defs.STLINK_JTAG_WRITE_VERIF_ERROR:
      return jtagapidefs.STLINK_JTAG_WRITE_ERROR;
    case defs.STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE:
      return jtagapidefs.STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE;

    case defs.STLINK_SWD_AP_WAIT:
    case defs.STLINK_SWD_AP_FAULT:
    case defs.STLINK_SWD_AP_ERROR:
    case defs.STLINK_SWD_AP_PARITY_ERROR:
    case defs.STLINK_SWD_DP_WAIT:
    case defs.STLINK_SWD_DP_FAULT:
    case defs.STLINK_SWD_DP_ERROR:
    case defs.STLINK_SWD_DP_PARITY_ERROR:
      return jtagapidefs.STLINK_SWD_ERROR;

    case defs.STLINK_SWD_AP_WDATA_ERROR:
      return jtagapidefs.STLINK_JTAG_READ_ERROR;

    case defs.STLINK_SWD_AP_STICKY_ERROR:
    case defs.STLINK_SWD_AP_STICKYORUN_ERROR:
    case defs.STLINK_AP_ALREADY_USED:
    case defs.STLINK_TRACE_AP_TURNAROUND:
    case defs.STLINK_BAD_AP:
      return jtagapidefs.STLINK_AP_ERROR;

    case defs.STLINK_SWV_NOT_AVAILABLE:
      return jtagapidefs.STLINK_SWV_NOT_AVAILABLE;

    case defs.STLINK_JTAG_TCPID_NOT_FOUND:
    case defs.STLINK_JTAG_TCPID_MAX_REACHED:
      return jtagapidefs.STLINK_JTAG_TCP_ERROR;

    case defs.STLINK_JTAG_FREQ_NOT_SUPPORTED:
      return jtagapidefs.STLINK_JTAG_FREQUENCY_NOT_SUPPORTED;

    case defs.STLINK_JTAG_UNKNOWN_CMD:
      return jtagapidefs.STLINK_JTAG_CMD_NOT_SUPPORTED;

    default:
      return jtagapidefs.STLINK_JTAG_ERROR;
  }
}

const STM32Add = {
  STM32_DBGMCU_IDCODE_ADDR: 0xe0042000,
  STM32_DBGMCU_ADDR: 0x40015800,
  STM32H_IDCODE_ADDR: 0x5c001000,
  STM32W_IDCODE_ADDR: 0x40004000,
  STM32L5_IDCODE_ADDR: 0xe0044000,
  STM32MP1_IDCODE_ADDR: 0x50081000
};

const JTAGFrequencies = {
  "18 MHz": 2,
  "9 MHz": 4,
  "4.5 MHz": 8,
  "2.25 MHz": 16,
  "1.12 MHz - Default": 32,
  "560 kHz": 64,
  "280 kHz": 128,
  "140 kHz": 256
};

const SWDFrequencies = {
  "4.6 MHz": 0,
  "1.8 MHz - Default": 1,
  "950 kHz": 3,
  //"650 kHz": 5,
  //"480 kHz": 7,
  "400 kHz": 9,
  //"360 kHz": 10,
  //"240 kHz": 15,
  "150 kHz": 25
};

const STM32MCU = {
  0x000: "STM32",
  0x410: "STM32F10xxx",
  0x411: "STM32F2xxxx",
  0x412: "STM32F10xxx",
  0x413: "STM32F40xxx/41xxx",
  0x414: "STM32F10xxx",
  0x415: "STM32L47xxx/48xxx",
  0x416: "STM32L1xxx6(8/B)",
  0x417: "STM32L05xxx/06xxx",
  0x418: "STM32F105xx/107xx",
  0x419: "STM32F42xxx/43xxx",
  0x420: "STM32F10xxx",
  0x421: "STM32F446xx",
  0x422: "STM32F302xB(C)/303xB(C)/STM32F358xx",
  0x423: "STM32F401xB(C)",
  0x424: "STM32",
  0x425: "STM32L031xx/041xx",
  0x427: "STM32L1xxxC",
  0x428: "STM32F10xxx",
  0x429: "STM32L1xxx6(8/B)A",
  0x430: "STM32F10xxx",
  0x431: "STM32F411xx",
  0x432: "STM32F373xx/STM32F378xx",
  0x433: "STM32F401xD(E)",
  0x434: "STM32F469xx/479xx",
  0x435: "STM32L43xxx/44xxx",
  0x436: "STM32L1xxxD",
  0x437: "STM32L1xxxE",
  0x438: "STM32F303x4(6/8)/334xx/328xx",
  0x439: "STM32F301xx/302x4(6/8)/STM32F318xx",
  0x440: "STM32F05xxx/STM32F030x8",
  0x441: "STM32F412xx",
  0x442: "STM32F030xC/STM32F09xxx",
  0x444: "STM32F03xx4/6",
  0x445: "STM32F04xxx/STM32F070x6",
  0x446: "STM32F302xD(E)/303xD(E)/STM32F398xx",
  0x447: "STM32L07xxx/08xxx",
  0x448: "STM32F070xB/STM32F071xx/072xx",
  0x449: "STM32F74xxx/75xxx",
  0x450: "STM32H74xxx/75xxx",
  0x451: "STM32F76xxx/77xxx",
  0x452: "STM32F72xxx/73xxx",
  0x456: "STM32G05xxx/06xxx",
  0x457: "STM32L01xxx/02xxx",
  0x458: "STM32F410xx",
  0x460: "STM32G07xxx/08xxx",
  0x461: "STM32L496xx/4A6xx",
  0x462: "STM32L45xxx/46xxx",
  0x463: "STM32F413xx/423xx",
  0x464: "STM32L412xx/422xx",
  0x466: "STM32G03xxx/04xxx",
  0x467: "STM32G0Bxxx/0Cxxx",
  0x468: "STM32G431xx/441xx",
  0x469: "STM32G47xxx/48xxx",
  0x470: "STM32L4Rxx/4Sxx",
  0x471: "STM32L4P5xx/Q5xx",
  0x472: "STM32L552xx/562xx",
  0x479: "STM32G491xx/4A1xx",
  0x480: "STM32H7A3xx/B3xx",
  0x481: "STM32U59xxx/5Axxx",
  0x482: "STM32U57xxx/585xx",
  0x483: "STM32H72xxx/73xxx",
  0x494: "STM32WB10xx/WB15xx",
  0x495: "STM32WB5xxx/WB3xxx",
  0x496: "STM32WB5xxx/WB3xxx",
  0x497: "STM32WLExxx/WL5xxx",
  0x500: "STM32MP15xx",
  0x9a8: "STM32",
  0x9b0: "STM32"
};

/**
 * Maps error code to a string describing error.
 *
 * @param {number} errorCode - The error code value.
 * @returns {string} - The error string.
 */
function errorString(errorCode) {
  switch (errorCode) {
    case jtagapidefs.STLINK_JTAG_PROTOCOL_NOT_SUPPORTED:
      return "JTAG protocol not supported";

    case jtagapidefs.STLINK_SWD_PROTOCOL_NOT_SUPPORTED:
      return "SWD protocol not supported";

    case jtagapidefs.STLINK_JTAG_FREQUENCY_NOT_SUPPORTED:
      return "JTAG frequency not supported";

    case jtagapidefs.STLINK_SWD_FREQUENCY_NOT_SUPPORTED:
      return "SWD frequency not supported";

    case jtagapidefs.STLINK_JTAG_READ_ERROR:
      return "JTAG/SWD read error";

    case jtagapidefs.STLINK_JTAG_ERROR:
      return "JTAG/SWD error";

    case jtagapidefs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN:
      return "Unknown JTAG chain";

    case jtagapidefs.STLINK_JTAG_NO_DEVICE_CONNECTED:
      return "JTAG/SWD no device connected";

    case jtagapidefs.STLINK_JTAG_WRITE_ERROR:
      return "JTAG/SWD write error";

    case jtagapidefs.STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE:
      return "JTAG/SWD already opened in other mode";

    case jtagapidefs.STLINK_SWD_ERROR:
      return "SWD protocol implementation error";

    case jtagapidefs.STLINK_AP_ERROR:
      return "Access Point error";

    case jtagapidefs.STLINK_SWV_NOT_AVAILABLE:
      return "SWV not available";

    case jtagapidefs.STLINK_JTAG_TCP_ERROR:
      return "TCP error";

    case jtagapidefs.STLINK_JTAG_16BIT_ACCESS_NOT_SUPPORTED:
      return "16-bit access mode not supported";

    case jtagapidefs.STLINK_JTAG_CMD_NOT_SUPPORTED:
      return "JTAG commmand not supported";

    default:
      return "";
  }
}

/**
 * Returns the list of possible JTAG frequencies.
 */
function getJTAGFrequencies() {
  return Object.keys(JTAGFrequencies);
}

/**
 * Returns the list of possible SWD frequencies.
 */
function getSWDFrequencies() {
  return Object.keys(SWDFrequencies);
}

/**
 * Returns the list of supported SWD frequencies for a ST-LINK V3.
 */
function v3getSWDFrequencies() {
  return Object.keys(SWDV3Frequencies);
}

/**
 * Returns the list of supported JTAG frequencies for a ST-LINK V3.
 */
function v3getJTAGFrequencies() {
  return Object.keys(JTAGV3Frequencies);
}

/**
 * Gets the MCU name according to it Id.
 *
 * @param {number} mcuid - Id of the MCU.
 * @returns {string} - MCU String.
 */
function getSTM32MCUString(mcuid) {
  let MCUString = STM32MCU[mcuid];

  if (MCUString != undefined) {
    return MCUString;
  } else {
    return "Unknown MCU";
  }
}

/**
 * Checks given device is openned.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @returns {number} - OK if openned, error code otherwise.
 */
function checkOpenedDevice(deviceDesc) {
  if (deviceDesc == undefined) {
    logger.error("Device not defined");
    return apidefs.STLINK_NO_DEVICE;
  }

  if (deviceDesc.state != "opened") {
    logger.warn("Device not opened : ", deviceDesc);
    return apidefs.STLINK_DEVICE_NOT_OPENED;
  }

  return apidefs.STLINK_OK;
}

/**
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the commands.
 */
function enterJTAG(deviceDesc, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc);

  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.warn("Device not opened : ", deviceDesc);
    return check;
  }

  // check if protocol is supported
  if (deviceDesc.JTAG == undefined) {
    logger.warn("JTAG protocol not supported : ", deviceDesc);
    return jtagapidefs.STLINK_JTAG_PROTOCOL_NOT_SUPPORTED;
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 4;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_ENTER2;
  req.CDBByte[2] = defs.STLINK_JTAG_ENTER_JTAG_NO_CORE_RESET;
  req.CDBByte[3] = deviceDesc.accessPointId;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("JTAG read error : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 * Enter SWD mode for given device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the commands.
 */
function enterSWD(deviceDesc, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc);

  let check = checkOpenedDevice(deviceDesc);

  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return check;
  }

  // check if protocol is supported
  if (deviceDesc.SWD == undefined) {
    logger.warn("SWD protocol not supported : ", deviceDesc);
    return jtagapidefs.STLINK_SWD_PROTOCOL_NOT_SUPPORTED;
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 4;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_ENTER2;
  req.CDBByte[2] = defs.STLINK_JTAG_ENTER_SWD;
  req.CDBByte[3] = deviceDesc.accessPointId;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;

  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);

    logger.trace("SWD read status : ", JTAGstatus);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("SWD read error : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 * Exit JTAG mode for given device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the commands.
 */
function exitJTAG(deviceDesc, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc);

  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return check;
  }

  let req = new TDeviceRequest();

  req.CDBLength = 2;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_EXIT;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = null;
  req.BufferLength = 0;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  logger.trace("drvResult = ", drvResult);
  return drvToAPI(drvResult);
}

/**
 * Fill supported frequencies for given protocol.
 *
 * @param {*} buffer - Buffer where to read frequencies.
 * @param {string} protocol - The connection mode either JTAG or SWD.
 */
function fillSupportedFreq(buffer, protocol) {
  let nbfreq = buffer.readUInt32LE(8);
  let freqValue;
  if (protocol == apidefs.STLINK_JTAG_PROTOCOL) {
    JTAGV3CurrentFreq = buffer.readUInt32LE(4);
  } else {
    SWDV3CurrentFreq = buffer.readUInt32LE(4);
  }
  for (let i = 0; i < nbfreq; i++) {
    freqValue = buffer.readUInt32LE(12 + 4 * i);
    if (protocol == apidefs.STLINK_JTAG_PROTOCOL) {
      JTAGV3Frequencies[freqValue + " kHz"] = freqValue;
    } else {
      SWDV3Frequencies[freqValue + " kHz"] = freqValue;
    }
  }
}

/**
 * Reads code Id for a given device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command and Id of the MCU.
 */
function JTAG_ReadIdCodes(deviceDesc, typeConnect) {
  logger.trace("Argument : deviceDesc = ", deviceDesc);
  const CORTEX_DESIGNER_ARM_ID = 0x477;

  let check = checkOpenedDevice(deviceDesc);

  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return { result: check };
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(12).fill(0);

  req.CDBLength = 2;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_READ_IDCODES;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  let mcuid = 0;

  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    let JTAGstatus = resp.readUInt16LE(0);

    if (JTAGstatus == defs.STLINK_JTAG_OK) {
      let cortexID = resp.readUInt32LE(4);

      // First check the DESIGNER ID from cortexIdcode
      if ((cortexID & 0x00000fff) != CORTEX_DESIGNER_ARM_ID) {
        logger.error("Device not supported : ", deviceDesc);
        return { result: apidefs.STLINK_DEVICE_NOT_SUPPORTED };
      }

      // Then find STM32 MCU code from specific target memory regions
      let tempAPId = deviceDesc.accessPointId;

      //Try to find mcuid value with AP0
      if (deviceDesc.accessPointId == undefined) {
        deviceDesc.accessPointId = 0; //Set the access port to 0
      }

      mcuid = getMCUId(deviceDesc, deviceDesc.accessPointId, typeConnect);

      //Try to find the mcuID when AP0 didn't work (with AP1)
      if (mcuid == 0) {
        deviceDesc.accessPointId = 1; //Set the access port to 1
        mcuid = getMCUId(deviceDesc, deviceDesc.accessPointId);
      }
      deviceDesc.accessPointId = tempAPId;

      // return found MCUID, or 0 if nothing found after all trials
      logger.trace("mcuid = ", mcuid);
      return {
        result: apidefs.STLINK_OK,
        MCUID: mcuid
      };
    } else {
      logger.error("JTAG read error : ", JTAGstatus);
      return { result: JTAGtoAPI(JTAGstatus) };
    }
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return { result: drvToAPI(drvResult) };
  }
}

/**
 *
 * @param {number} deviceDesc - Device descriptor.
 * @param {number} accessPort - Access Port number.
 * @param {number} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - The mcuid found.
 */
function getMCUId(deviceDesc, accessPort, typeConnect) {
  let read;
  let mcuid = 0;
  let cmd_result = initAP(deviceDesc, accessPort, typeConnect); //Access port initialization

  if (cmd_result != apidefs.STLINK_OK) {
    logger.error("initAP error = ", cmd_result);
    return 0;
  }

  for (const address in STM32Add) {
    if (mcuid == 0) {
      read = JTAG_ReadMemory32Bit(deviceDesc, STM32Add[address], 4, typeConnect);
      if (read.result == apidefs.STLINK_OK) {
        mcuid =
          STM32Add[address] != STM32Add.STM32W_IDCODE_ADDR
            ? read.data.readUInt32LE(0) & 0x00000fff
            : (read.data.readUInt32LE(0) >> 12) & 0x00000fff;
      }
    }
  }
  cmd_result = closeAP(deviceDesc, accessPort, typeConnect); //Close access port
  if (cmd_result != apidefs.STLINK_OK) {
    logger.error("closeAP error = ", cmd_result);
  }

  return mcuid;
}

/**
 * Set the JTAG frequency for a given device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} frequency - The frequency to set.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - The result of the command.
 */
function JTAG_SetJTAGFrequency(deviceDesc, frequency, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc, ", frequency arg = ", frequency);

  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return check;
  }

  // check if protocol is supported
  if (deviceDesc.JTAG == undefined) {
    logger.error("JTAG protocol not supported : ", deviceDesc);
    return jtagapidefs.STLINK_JTAG_PROTOCOL_NOT_SUPPORTED;
  }

  // check if frequency is valid
  let newFreq;
  if (frequency == "default") {
    newFreq = JTAGFrequencies["1.12 MHz - Default"];
  } else {
    newFreq = JTAGFrequencies[frequency];
    if (newFreq == undefined) {
      logger.error("JTAG frequency not supported : ", deviceDesc, " FHz = ", frequency);
      return jtagapidefs.STLINK_JTAG_FREQUENCY_NOT_SUPPORTED;
    }
  }

  // send command
  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 4;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_SET_FREQ;
  req.CDBByte[2] = newFreq & 0xff;
  req.CDBByte[3] = (newFreq >> 8) & 0xff;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);
    logger.trace("JTAG status = ", JTAGstatus);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 * Set the SWD frequency for a given device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} frequency - The frequency to set.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - The result of the command.
 */
function JTAG_SetSWDFrequency(deviceDesc, frequency, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc, ", frequency arg = ", frequency);
  let check = checkOpenedDevice(deviceDesc);

  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return check;
  }

  // check if protocol is supported
  if (deviceDesc.SWD == undefined) {
    logger.error("SWD protocol not supported : ", deviceDesc);

    return jtagapidefs.STLINK_SWD_PROTOCOL_NOT_SUPPORTED;
  }

  // check if frequency is valid
  let newFreq;
  if (frequency == "default") {
    newFreq = SWDFrequencies["1.8 MHz - Default"];
  } else {
    newFreq = SWDFrequencies[frequency];
    if (newFreq == undefined) {
      logger.error("Frequency requested not supported : ", deviceDesc, " FHz = ", frequency);
      return jtagapidefs.STLINK_SWD_FREQUENCY_NOT_SUPPORTED;
    }
  }

  // send command
  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 4;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_SWD_SET_FREQ;
  req.CDBByte[2] = newFreq & 0xff;
  req.CDBByte[3] = (newFreq >> 8) & 0xff;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);
    logger.trace("JTAG status ) ", JTAGstatus);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 * Sets the STLINK V3 frequency for a given device on specified mode.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} protocol - The connection mode either JTAG or SWD.
 * @param {string} frequency - The frequency to set.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_SetComFrequency(deviceDesc, protocol, frequency, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc, ", frequency arg = ", frequency);
  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return check;
  }

  // check if protocol is supported
  if (protocol == apidefs.STLINK_SWD_PROTOCOL && deviceDesc.SWD == undefined) {
    logger.error("SWD protocol not supported : ", deviceDesc);
    return jtagapidefs.STLINK_SWD_PROTOCOL_NOT_SUPPORTED;
  } else if (protocol == apidefs.STLINK_JTAG_PROTOCOL && deviceDesc.JTAG == undefined) {
    logger.error("JTAG protocol not supported : ", deviceDesc);
    return jtagapidefs.STLINK_JTAG_PROTOCOL_NOT_SUPPORTED;
  }

  // check if frequency is valid
  let newFreq;
  if (protocol == apidefs.STLINK_JTAG_PROTOCOL) {
    newFreq = frequency == "default" ? JTAGV3CurrentFreq : JTAGV3Frequencies[frequency];
  } else {
    newFreq = frequency == "default" ? SWDV3CurrentFreq : SWDV3Frequencies[frequency];
  }

  if (newFreq == undefined) {
    logger.error("Frequency requested not supported : ", deviceDesc, " FHz = ", frequency);
    return jtagapidefs.STLINK_SWD_FREQUENCY_NOT_SUPPORTED;
  }

  // send command
  let req = new TDeviceRequest();

  let resp = Buffer.alloc(8).fill(0);

  req.CDBLength = 10;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_SET_COM_FREQ;
  req.CDBByte[2] = protocol == apidefs.STLINK_JTAG_PROTOCOL ? 1 : 0;
  req.CDBByte[3] = 0;
  req.CDBByte[4] = newFreq & 0xff;
  req.CDBByte[5] = (newFreq >> 8) & 0xff;
  req.CDBByte[6] = (newFreq >> 16) & 0xff;
  req.CDBByte[7] = (newFreq >> 24) & 0xff;

  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);
    logger.trace("JTAG status ) ", JTAGstatus);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 * Gets supported target frequencies for a given ST-LINK V3 device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} protocol - The connection mode either JTAG or SWD.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_GetTargetFrequencies(deviceDesc, protocol, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc);
  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return check;
  }

  let resp = Buffer.alloc(52).fill(0);

  let req = new TDeviceRequest();

  req.CDBLength = 3;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_GET_COM_FREQ;
  req.CDBByte[2] = protocol == apidefs.STLINK_JTAG_PROTOCOL ? 1 : 0;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);
    logger.trace("JTAG status ) ", JTAGstatus);
    fillSupportedFreq(resp, protocol);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 *  Get status of the last read or write command on given device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command.
 */
function GetLastReadWriteStatus(deviceDesc, typeConnect) {
  // deviceDesc is assumed to be existing and opened
  logger.trace("Arguments : deviceDesc = ", deviceDesc);
  let req = new TDeviceRequest();

  // allocate 12 bytes for answer
  let resp = Buffer.alloc(12).fill(0);

  req.CDBLength = 16;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_GETLASTRWSTATUS2;
  req.CDBByte[2] = deviceDesc.accessPointId;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    let JTAGStatus = resp.readUInt16LE(0);

    if (JTAGStatus == defs.STLINK_JTAG_OK) {
      logger.trace("JTAG read = ", apidefs.STLINK_OK);
      return { result: apidefs.STLINK_OK };
    } else {
      // erroneous address is in the second 32-bit answer
      let badaddress = resp.readUInt32LE(4);
      logger.error("Bad address : ", badaddress);
      return {
        result: JTAGtoAPI(JTAGStatus),
        bad_addr: badaddress
      };
    }
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return { result: drvToAPI(drvResult) };
  }
}

/**
 * Writes 32 bit aligned buffer of specified size at provided address in device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {number} address - Address where to write.
 * @param {number} size - Size of the buffer.
 * @param {object} buffer - Data to write.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command, size and data written.
 */
function JTAG_WriteMemory32Bit(deviceDesc, address, size, buffer, typeConnect) {
  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    return { result: check };
  }

  // size must not be <= 0 or > 6144 for 32-bit accesses
  if (size <= 0 || size > 6144) {
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  // size and address must be 4-bytes aligned (32-bit access)
  if (size % 4 != 0 || address % 4 != 0) {
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  let req = new TDeviceRequest();

  //   let resp = new Buffer(buffer);
  let resp = buffer;

  req.CDBLength = 10;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_WRITEMEM_32BIT;
  req.CDBByte[2] = address & 0xff;
  req.CDBByte[3] = (address >> 8) & 0xff;
  req.CDBByte[4] = (address >> 16) & 0xff;
  req.CDBByte[5] = (address >> 24) & 0xff;
  req.CDBByte[6] = size & 0xff;
  req.CDBByte[7] = (size >> 8) & 0xff;
  req.CDBByte[8] = deviceDesc.accessPointId;

  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.InputRequest = defsDrv.DRV_REQUEST_WRITE;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    let readResult = GetLastReadWriteStatus(deviceDesc, typeConnect);

    if (readResult.result == apidefs.STLINK_OK) {
      // return the buffer read out from target
      return {
        result: apidefs.STLINK_OK,
        data: resp,
        size: size
      };
    } else {
      return { result: readResult.result };
    }
  } else {
    return { result: drvToAPI(drvResult) };
  }
}

/**
 * Writes 8 bit aligned buffer of specified size at provided address in device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {number} address - Address where to write.
 * @param {number} size - Size of the buffer.
 * @param {object} buffer - Data to write.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command, size and data written.
 */
function JTAG_WriteMemory8Bit(deviceDesc, address, size, buffer, typeConnect) {
  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    return { result: check };
  }

  // size must not be <= 0 or > 6144 for 32-bit accesses
  if (size <= 0 || size > 6144) {
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  let req = new TDeviceRequest();

  //   let resp = new Buffer(buffer);
  let resp = buffer;

  req.CDBLength = 10;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_WRITEMEM_8BIT;
  req.CDBByte[2] = address & 0xff;
  req.CDBByte[3] = (address >> 8) & 0xff;
  req.CDBByte[4] = (address >> 16) & 0xff;
  req.CDBByte[5] = (address >> 24) & 0xff;
  req.CDBByte[6] = size & 0xff;
  req.CDBByte[7] = (size >> 8) & 0xff;
  req.CDBByte[8] = deviceDesc.accessPointId;

  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.InputRequest = defsDrv.DRV_REQUEST_WRITE;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    let readResult = GetLastReadWriteStatus(deviceDesc, typeConnect);

    if (readResult.result == apidefs.STLINK_OK) {
      // return the buffer read out from target
      return {
        result: apidefs.STLINK_OK,
        data: resp,
        size: size
      };
    } else {
      return { result: readResult.result };
    }
  } else {
    return { result: drvToAPI(drvResult) };
  }
}

/**
 * Read 32 bit aligned buffer of specified size at provided address in device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {number} address - Address where to write.
 * @param {number} size - Size to read.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command, size and data read.
 */
function JTAG_ReadMemory32Bit(deviceDesc, address, size, typeConnect) {
  logger.trace("Arguments : deviceDesc = ", deviceDesc, ", address = ", address, ", size = ", size);
  let check = checkOpenedDevice(deviceDesc);

  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return { result: check };
  }

  // size must not be <= 0 or > 6144 for 32-bit accesses
  if (size <= 0 || size > 6144) {
    logger.error("size n not in the range : ", size);
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  // size and address must be 4-bytes aligned (32-bit access)
  if (size % 4 != 0 || address % 4 != 0) {
    logger.error("size and address must be 4-bytes aligned (32-bit access): ", size, " : ", address);
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(size).fill(0);

  req.CDBLength = 10;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_READMEM_32BIT;
  req.CDBByte[2] = address & 0xff;
  req.CDBByte[3] = (address >> 8) & 0xff;
  req.CDBByte[4] = (address >> 16) & 0xff;
  req.CDBByte[5] = (address >> 24) & 0xff;
  req.CDBByte[6] = size & 0xff;
  req.CDBByte[7] = (size >> 8) & 0xff;
  req.CDBByte[8] = deviceDesc.accessPointId;

  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    let readResult = GetLastReadWriteStatus(deviceDesc, typeConnect);

    if (readResult.result == apidefs.STLINK_OK) {
      // return the buffer read out from target
      logger.trace("Read memory = ", resp, " with size ) ", size);
      return {
        result: apidefs.STLINK_OK,
        data: resp,
        size: size
      };
    } else {
      logger.error("Read error : ", readResult.result);
      return { result: readResult.result };
    }
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return { result: drvToAPI(drvResult) };
  }
}

/**
 * Read 16 bit aligned buffer of specified size at provided address in device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {number} address - Address where to write.
 * @param {number} size - Size to read.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command, size and data read.
 */
function JTAG_ReadMemory16Bit(deviceDesc, address, size, typeConnect) {
  logger.trace("Arguments : deviceDesc = ", deviceDesc, ", address = ", address, ", size = ", size);
  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return { result: check };
  }

  // size must not be <= 0 or > 6144 for 16-bit accesses
  if (size <= 0 || size > 6144) {
    logger.error("size n not in the range : ", size);
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  // size and address must be 2-bytes aligned (16-bit access)
  if (size % 2 != 0 || address % 2 != 0) {
    logger.error("size and address must be 2-bytes aligned (16-bit access) : ", size, " ", address);
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(size).fill(0);

  req.CDBLength = 10;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_READMEM_16BIT;
  req.CDBByte[2] = address & 0xff;
  req.CDBByte[3] = (address >> 8) & 0xff;
  req.CDBByte[4] = (address >> 16) & 0xff;
  req.CDBByte[5] = (address >> 24) & 0xff;
  req.CDBByte[6] = size & 0xff;
  req.CDBByte[7] = (size >> 8) & 0xff;
  req.CDBByte[8] = deviceDesc.accessPointId;

  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    let readResult = GetLastReadWriteStatus(deviceDesc, typeConnect);

    if (readResult.result == apidefs.STLINK_OK) {
      // return the buffer read out from target
      logger.trace("Read memory = ", resp, " with size = ", size);
      return {
        result: apidefs.STLINK_OK,
        data: resp,
        size: size
      };
    } else {
      logger.error("Read error : ", readResult.result);
      return { result: readResult.result };
    }
  } else {
    logger.error("Driver not ready : ", drvToAPI(drvResult));
    return { result: drvToAPI(drvResult) };
  }
}

/**
 * Read 8 bit aligned buffer of specified size at provided address in device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {number} address - Address where to write.
 * @param {number} size - Size to read.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command, size and data read.
 */
function JTAG_ReadMemory8Bit(deviceDesc, address, size, typeConnect) {
  logger.trace("Arguments : deviceDesc = ", deviceDesc, ", address = ", address, ", size = ", size);
  let check = checkOpenedDevice(deviceDesc);

  if (check != apidefs.STLINK_OK) {
    logger.error("Device not opened : ", deviceDesc);
    return { result: check };
  }

  // size must not be <= 0 or > 64 for 8-bit accesses
  if (size <= 0 || size > 64) {
    logger.error("size n not in the range : ", size);
    return { result: apidefs.STLINK_BAD_PARAMETER };
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(size).fill(0);

  req.CDBLength = 10;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_READMEM_8BIT;
  req.CDBByte[2] = address & 0xff;
  req.CDBByte[3] = (address >> 8) & 0xff;
  req.CDBByte[4] = (address >> 16) & 0xff;
  req.CDBByte[5] = (address >> 24) & 0xff;
  req.CDBByte[6] = size & 0xff;
  req.CDBByte[7] = (size >> 8) & 0xff;
  req.CDBByte[8] = deviceDesc.accessPointId;

  req.Buffer = resp;

  // minimum two bytes are read
  if (size == 1) {
    req.BufferLength = 2;
  } else {
    req.BufferLength = resp.length;
  }

  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    let readResult = GetLastReadWriteStatus(deviceDesc, typeConnect);

    if (readResult.result == apidefs.STLINK_OK) {
      // return the buffer read out from target
      logger.trace("data read = ", resp, " with size ", size);
      return {
        result: apidefs.STLINK_OK,
        data: resp,
        size: size
      };
    } else {
      logger.error("Read error : ", readResult.result);
      return { result: readResult.result };
    }
  } else {
    logger.error(" JTAG_ReadMemory8Bit drvResult error = ", drvResult);
    return { result: drvToAPI(drvResult) };
  }
}

/**
 * Initializes access point for specified device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {number} accessPointId - Access point index.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result od the command.
 */
function initAP(deviceDesc, accessPointId, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc);

  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.warn("Device not opened : ", deviceDesc);
    return check;
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 4;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_INIT_AP;
  req.CDBByte[2] = accessPointId;
  req.CDBByte[3] = defs.STLINK_JTAG_AP_CORTEXM_CORE;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("JTAG init AP error : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 * Closes Access point on specified device.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {number} accessPointId - Access point index.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result od the command.
 */
function closeAP(deviceDesc, accessPointId, typeConnect) {
  logger.trace("deviceDesc arg = ", deviceDesc);

  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.warn("Device not opened : ", deviceDesc);
    return check;
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 3;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_CLOSE_AP_DBG;
  req.CDBByte[2] = accessPointId;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (drvResult == defsDrv.DRV_OK) {
    // JTAG response is an u16 in Little Endian
    let JTAGstatus = resp.readUInt16LE(0);
    return JTAGtoAPI(JTAGstatus);
  } else {
    logger.error("JTAG close AP error : ", drvToAPI(drvResult));
    return drvToAPI(drvResult);
  }
}

/**
 * Blink led.
 *
 * @param {object} deviceDesc - Device descriptor.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} Result of the command.
 */
function JTAG_BlinkLed(deviceDesc, typeConnect) {
  logger.trace("JTAG_BlinkLed arguments : deviceDesc = ", deviceDesc, "typeConnect = ", typeConnect);

  let check = checkOpenedDevice(deviceDesc);
  if (check != apidefs.STLINK_OK) {
    logger.warn("Device not opened : ", deviceDesc);
    return check;
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 2;
  req.CDBByte[0] = defs.STLINK_CMD_JTAG;
  req.CDBByte[1] = defs.STLINK_JTAG_BLINK_LED;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let drvResult;
  if (typeConnect == "tcp") {
    drvResult = stLinkDrv.STLink_SendCommandTcp(deviceDesc.Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    drvResult = stLinkDrv.STLink_SendCommand(deviceDesc.Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }
  if (drvResult == defsDrv.DRV_OK) {
    logger.trace("Led blink ok");
    return apidefs.STLINK_OK;
  } else {
    logger.error("Led blink error : ", drvResult);
    return drvToAPI(drvResult);
  }
}

module.exports.apidefs = jtagapidefs;
module.exports.errorString = errorString;
module.exports.enterJTAG = enterJTAG;
module.exports.initAP = initAP;
module.exports.closeAP = closeAP;
module.exports.enterSWD = enterSWD;
module.exports.exitJTAG = exitJTAG;
module.exports.JTAG_ReadIdCodes = JTAG_ReadIdCodes;
module.exports.getJTAGFrequencies = getJTAGFrequencies;
module.exports.getSWDFrequencies = getSWDFrequencies;
module.exports.JTAG_SetJTAGFrequency = JTAG_SetJTAGFrequency;
module.exports.v3getSWDFrequencies = v3getSWDFrequencies;
module.exports.v3getJTAGFrequencies = v3getJTAGFrequencies;
module.exports.JTAG_SetSWDFrequency = JTAG_SetSWDFrequency;
module.exports.JTAG_SetComFrequency = JTAG_SetComFrequency;
module.exports.JTAG_GetTargetFrequencies = JTAG_GetTargetFrequencies;
module.exports.JTAG_WriteMemory32Bit = JTAG_WriteMemory32Bit;
module.exports.JTAG_WriteMemory8Bit = JTAG_WriteMemory8Bit;
module.exports.JTAG_ReadMemory32Bit = JTAG_ReadMemory32Bit;
module.exports.JTAG_ReadMemory16Bit = JTAG_ReadMemory16Bit;
module.exports.JTAG_ReadMemory8Bit = JTAG_ReadMemory8Bit;
module.exports.getSTM32MCUString = getSTM32MCUString;
module.exports.JTAG_BlinkLed = JTAG_BlinkLed;
