"use strict";
/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const logger = require("@stm32/stm32cubemonitor-logger").getLogger("stlinkprotocol");
const stlinkprotocol = require("./stlinkprotocol");
const apidefs = require("./stlinkprotocol").apidefs;

//------------------------------------------------
// Snapshot module variable definition
// Const used to identify the header.
const EXPECTED_START_MARK = "8f25a7"; // id of startMark header.

// Flag definitions for DataTraceHeaderT.flags
const flag = {
  SNP_TRC_FLAG_OVERFLOW: 1,
  SNP_TRC_START_STOP_ACK: 2,
  SNP_TRC_RECORD_SKIPPED: 4
};

// Type of variable definition of the target
const targetType = {
  SNP_TRC_DATA_ACCESS_8BIT: 0, // signed or unsigned for acquisition; signed only for trigger
  SNP_TRC_DATA_ACCESS_16BIT: 1, // signed or unsigned for acquisition; signed only for trigger
  SNP_TRC_DATA_ACCESS_32BIT: 2, // signed or unsigned for acquisition; signed only for trigger
  SNP_TRC_DATA_ACCESS_64BIT: 3, // For management of double
  SNP_TRC_DATA_ACCESS_TRIG_U8: 4, // For trigger only: unsigned; note that bit2 discriminates
  SNP_TRC_DATA_ACCESS_TRIG_U16: 5, // For trigger only: unsigned; note that bit2 discriminates
  SNP_TRC_DATA_ACCESS_TRIG_U32: 6, // For trigger only: unsigned; note that bit2 discriminates
  SNP_TRC_DATA_ACCESS_TRIG_FLOAT: 7, // For trigger only
  SNP_TRC_DATA_ACCESS_TRIG_DOUBLE: 8 // For trigger only
};

// Unit of acquisition
const timeUnit = {
  second: 5,
  millisecond: 4,
  microsecond: 3,
  nanosecond: 2
};

// Mapping of type variable bettween target and application.
const targetTypeMap = {
  [apidefs.READ_TYPE_U8]: targetType.SNP_TRC_DATA_ACCESS_8BIT,
  [apidefs.READ_TYPE_S8]: targetType.SNP_TRC_DATA_ACCESS_8BIT,
  [apidefs.READ_TYPE_U16]: targetType.SNP_TRC_DATA_ACCESS_16BIT,
  [apidefs.READ_TYPE_S16]: targetType.SNP_TRC_DATA_ACCESS_16BIT,
  [apidefs.READ_TYPE_U32]: targetType.SNP_TRC_DATA_ACCESS_32BIT,
  [apidefs.READ_TYPE_S32]: targetType.SNP_TRC_DATA_ACCESS_32BIT,
  [apidefs.READ_TYPE_FLOAT]: targetType.SNP_TRC_DATA_ACCESS_32BIT,
  [apidefs.READ_TYPE_U64]: targetType.SNP_TRC_DATA_ACCESS_64BIT,
  [apidefs.READ_TYPE_S64]: targetType.SNP_TRC_DATA_ACCESS_64BIT,
  [apidefs.READ_TYPE_DOUBLE]: targetType.SNP_TRC_DATA_ACCESS_64BIT
};

// Mapping of type variable bettween target and application for trigger variable.
const triggerTypeMap = {
  [apidefs.READ_TYPE_U8]: targetType.SNP_TRC_DATA_ACCESS_TRIG_U8,
  [apidefs.READ_TYPE_S8]: targetType.SNP_TRC_DATA_ACCESS_8BIT,
  [apidefs.READ_TYPE_U16]: targetType.SNP_TRC_DATA_ACCESS_TRIG_U16,
  [apidefs.READ_TYPE_S16]: targetType.SNP_TRC_DATA_ACCESS_16BIT,
  [apidefs.READ_TYPE_U32]: targetType.SNP_TRC_DATA_ACCESS_TRIG_U32,
  [apidefs.READ_TYPE_S32]: targetType.SNP_TRC_DATA_ACCESS_32BIT,
  [apidefs.READ_TYPE_FLOAT]: targetType.SNP_TRC_DATA_ACCESS_TRIG_FLOAT,
  [apidefs.READ_TYPE_U64]: targetType.SNP_TRC_DATA_ACCESS_64BIT,
  [apidefs.READ_TYPE_S64]: targetType.SNP_TRC_DATA_ACCESS_64BIT,
  [apidefs.READ_TYPE_DOUBLE]: targetType.SNP_TRC_DATA_ACCESS_TRIG_DOUBLE
};

// Error message.
const SNAP_ERR = {
  INIT_SNP: "Snapshot mode - Error occured during snapshot initialization.",
  RESET_FLAG: "Snapshot mode - Error occured while the flag reset.",
  READ_HEADER: "Snapshot mode - Error occured while the header reading.",
  STOP_SNP: "Snapshot mode - Error occured while the interruption of the acquisition.",
  START_TRACE: "Snapshot mode - Error occured while the activation of trace on the target.",
  SNAPSHOT_HEADER_NOT_FOUND: "Snapshot mode - Wrong 'Header address' from acquisition parameters.",
  VERSION_NOT_SUPPORTED:
    "Snapshot mode - This version is not supported, snapshot_embedded_SW V8 and after are supported",
  TRIGGER_NOT_SUPPORTED: "Snapshot mode - The trigger mode is not supported.",
  GROUP_VAR_CORRUPT: "Snapshot mode - Invalid variable group configuration for snapshot.",
  MAX_VAR_NUM_EXCEED: "Snapshot mode - The maximum number of variable is exceed."
};

// Trigger mode.
const TRIGGER_MODE = {
  SNP_TRC_TRIGGER_START_IMMEDIATE: "0x00",
  SNP_TRC_TRIGGER_RISING: "0x01",
  SNP_TRC_TRIGGER_FALLING: "0x02"
};

// Mapping of trigger mode message from UI to target variable.
const TRIGGER_MODE_MAP = {
  ["manual"]: TRIGGER_MODE.SNP_TRC_TRIGGER_START_IMMEDIATE,
  ["rising-edge"]: TRIGGER_MODE.SNP_TRC_TRIGGER_RISING,
  ["falling-edge"]: TRIGGER_MODE.SNP_TRC_TRIGGER_FALLING
};

/**
 * This class intanciates a new snapshot object or return the object already created according to the ProbId.
 */
class Snapshot {
  /**
   * @param {*} probeId - The probe to be used to communicate with the target MCU.
   */
  constructor(probeId) {
    if (Snapshot.ongoing[probeId]) {
      return Snapshot.ongoing[probeId];
    }
    this.probeId = probeId; // The probe to be used to communicate with the target MCU.

    // Variable from the header
    this.h_readPointer; // read pointer.
    this.h_readPointer; // read pointer.
    this.h_writePointer; // Write pointer.
    this.h_writePointerC; // Write pointer copy.
    this.h_flags; // flags informs if trace is in overflow, started or skipped.
    this.h_timestampValue; // use for incrementing timestamp value.
    this.h_readPointerEnd; // Value computed according to write pointer.
    this.h_bufferSize = 0; // This size of the buffer used to save the trace (variable to be trace).
    this.h_nbVar; // Number of var to be read.
    this.h_nbMaxVar; // Maximum number of variables (32bits words) in one record.
    this.h_headerVersion; // Identifier of the header format.
    this.h_triggerAddress; // // Address of variable for trigger.
    this.h_triggerAccessType; // Type of variable for trigger.
    this.h_triggerThreshold; // Threshold of trigger.
    this.h_triggerMode; // Mode of trigger.
    this.h_triggerCmd; // Trigger command.
    this.h_triggerModeStrHex; // header target address of the trigger mode.
    this.h_triggerAddressStrHex; // header target address of trigger address.
    this.h_triggerAccessTypeStrHex; // header target address of access type of trigger.
    this.h_triggerThresholdStrHex; // header target address of the trigger threshold.
    this.h_triggerThresholdStrHexDouble; // header target address of the trigger thresholdDouble.
    this.h_triggerThresholdStrHexDouble2; // header target address of the trigger thresholdDouble2.
    this.h_triggerCmdStrHex; // header target address of Trigger command.
    this.h_bufferStartAddrStrHex; // satrt address of the buffer.
    this.h_listOfVarStrHex; // Address of the list of the variable to be trace (variable address).
    this.h_listOfTypeStrHex; // Address of the list of the access type of variable to be trace.
    this.h_nbVarAddrStrHex; // Number of variable to be trace.
    this.h_readPointerAddStrHex; // Address of the read point header.
    this.h_readPointerCopyAddStrHex; // Address of the read point header.
    this.h_writePointerAddStrHex; // Address of the read point header.
    this.h_writePointerCopyAddStrHex; // Address of the read point header.
    this.h_flagAddStrHex; // Address of the flags.

    // Variable use to manage the acquisition
    this.expectedInitialReadPointer; // This value is computed and informs about the position of read pointer for starting the trace .
    this.recordSize; // It is the size of one record, for exemple if 5 variable is traced the size is 6, 1 time stamp + 5 variables.
    this.VarsToBeRead; // Variable to be trace.
    this.headerSize; // The size of each element of header (2 or 4 bytes).
    this.traceIsStarted = false; // Inform that the trace is started
    this.snaptimestamp; // Timestamp variable
    this.unitOfTimestamp;
    this.valueOfTimestamp;
    this.myGroupToBeRead = undefined; // used to save the previous group of variable to be read.
    this.timeSynchroStart;
    this.nbOfDouble = 0;
    this.synchroFinished = true;
    this.synchroStopDone = true;
    this.snapResult;
    this.previousMeas = [];
    this.accessPointId = 0;
    this.typeConnect = apidefs.DEFAULT_CONNECT_TYPE;
    this.headerOfTriggerBuilt = false;
    this.readPointerEndReached = false;

    // Trace header feild is shared between the application and the target.
    // This variable structure com from the target.
    this.header = [
      { name: "startMark[0]", address: "", type: apidefs.READ_TYPE_S8, nb: 1 },
      { name: "headerFieldSize", address: "", type: apidefs.READ_TYPE_S8, nb: 1 },
      { name: "headerVersion", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "bufferStartAddr", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "bufferEndAddr", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "nbVarMax", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "bufferFormat", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "timestamp_base_unit", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "writePointer", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "writePointerCopy", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "readPointer", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "readPointerCopy", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "flags", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "subSamplingRate", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "nbVar", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "g_varList[0].accessType", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "g_varList[0].address", address: "", type: apidefs.READ_TYPE_U32, nb: 1 }
    ];

    // This variable structure com from the target.
    this.headerOfTrigger = [
      { name: "triggerAddress", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerAccessType", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerThreshold", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerMode", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerCmd", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerStopParam", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerState", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerNPretrig", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerBIgnoreTrig", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerThresholdDouble", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
      { name: "triggerThresholdDouble2", address: "", type: apidefs.READ_TYPE_U32, nb: 1 }
    ];

    Snapshot.ongoing[probeId] = this;
    return this;
  }

  /**
   * This function return the instance according to the probe ID.
   *
   * @param {*} probeId - The probe to be used to communicate with the target MCU.
   */
  static getInstance(probeId) {
    return Snapshot.ongoing[probeId];
  }

  /**
   * Main service called to activate the snapshot mode and read the buffer.
   *
   * @param {*} groupToBeRead - Group of variables to be trace.
   * @param {*} acqParam - Acquisition parameters.
   * @param {*} aPId - Id of the access point.
   * @param {*} tConnect - The connection protocol to use either p2p or tcp.
   * @returns {*} - The snapResult object containing variables values.
   */
  getTrace(groupToBeRead, acqParam, aPId = 0, tConnect = apidefs.DEFAULT_CONNECT_TYPE) {
    if (this.snaptimestamp === undefined) {
      this.snaptimestamp = +new Date();
    }

    if (this.myGroupToBeRead === undefined) {
      let initSNP = this.initSnapshot(acqParam, groupToBeRead, aPId, tConnect); // First off all we need to check if target is compatible with snapshot mode and read there content.
      if (initSNP !== apidefs.STLINK_OK) {
        logger.debug("Error: " + stlinkprotocol.errorString(initSNP));
        const ERROR_RESP = { result: initSNP };
        return ERROR_RESP;
      }
      // case first time acquisition or server reset
      if (this.traceIsStarted) {
        // Case the target acquisition is ongoing
        this.resetFlag();
        const err = this.stopSnapshot(this.probeId); // We have to stop the acquisition
        if (err !== apidefs.STLINK_OK) {
          logger.debug("Snapshot Interruption - error = ", err);
        }
        this.synchroStop(); // check that the target is stopped
        this.snapResult = this.respEmpty();
      } else {
        this.initTarget();
        this.startTrace(); // Start acquisition in the target
        this.timeSynchroStart = setTimeout(() => {
          this.synchroStart;
        }, 0); // Check that the target acquisition is started
        this.snapResult = this.respEmpty();
      }
    } else if (this.traceIsStarted) {
      // The server has already done acquisition and one acquisition is ongoing!!!
      if (JSON.stringify(this.myGroupToBeRead) !== JSON.stringify(groupToBeRead)) {
        // The variables to be trace have changed
        this.resetFlag();
        const err = this.stopSnapshot(this.probeId); // Stop the trace.
        if (err !== apidefs.STLINK_OK) {
          logger.debug("Snapshot Interruption - error = ", err);
        }
        this.synchroStop(); // check that the target is stopped
        return { result: SNAP_ERR.GROUP_VAR_CORRUPT };
      } else if (this.synchroStopDone && this.synchroFinished) {
        // Case the target trace is ongoing the buffer can be read
        this.snapResult = this.readBuffer();
        if (this.readPointerEndReached) {
          let res = this.readBuffer();
          if (res.result === apidefs.STLINK_OK) {
            for (let i = 0; i < res.data.length; i++) {
              this.snapResult.data[i] = this.snapResult.data[i].concat(res.data[i]);
            }
          }
        }
      } else if (!this.synchroStopDone) {
        // The target is in the stop proccess
        this.snapResult = this.respEmpty();
      }
    } else {
      if (!this.synchroStopDone) {
        // The target is in the stop proccess
        this.snapResult = this.respEmpty();
      } else {
        // The target has restarted, the trace is stopped
        this.initTarget();
        this.startTrace(); // Start the target trace.
        this.timeSynchroStart = setTimeout(() => {
          this.synchroStart;
        }, 0); // Check that the trace is started.
        this.snapResult = this.respEmpty();
      }
    }
    this.myGroupToBeRead = groupToBeRead; // Save the previous group of variable for next acquisition.
    return this.snapResult;
  }

  /**
   * This function interrupts the snapshot acquisition which is runnning.
   *
   * @returns {*} 1 if ok, error string otherwise.
   */
  snapshotInterrupt() {
    if (this.myGroupToBeRead !== undefined && this.synchroFinished === true && this.synchroStopDone === true) {
      const err = this.stopSnapshot(this.probeId); //Acquisition stopped
      if (err !== apidefs.STLINK_OK) {
        logger.debug("Snapshot Interruption - error = ", err);
      }
      this.synchroStop();
      delete Snapshot.ongoing[this.probeId];
      return err;
    }
    //Nothing to do, no ongoing snapshot
    return apidefs.STLINK_OK;
  }

  /**
   * This private service stops the acquisition in the target.
   * The nbVar variable is change to 0.
   *
   * @returns {*} 1 if ok, error string otherwise.
   */
  stopSnapshot() {
    this.myGroupToBeRead = undefined;
    let myGroup = [];

    let elt = new Object();
    elt.name = "nbVar";
    elt.address = this.h_nbVarAddrStrHex;
    elt.type = apidefs.READ_TYPE_U32;
    elt.value = "0x00"; // The nbVar value stops the target acquisition.
    myGroup.push(elt);

    if (this.h_triggerCmd === 1) {
      let elt = new Object();
      elt.name = "mode";
      elt.address = this.h_triggerCmdStrHex;
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = "0";
      myGroup.push(elt);
    }

    const writeMem = stlinkprotocol.writeMemoryGroup(this.probeId, myGroup, this.accessPointId, this.typeConnect); // Write address of variable in the list
    if (writeMem.result !== apidefs.STLINK_OK) {
      logger.debug("error: " + stlinkprotocol.errorString(writeMem.result));
      return SNAP_ERR.STOP_SNP;
    }
    this.synchroFinished = false;
    this.synchroStopDone = false; // The target acquisition is stopped we have to re-synchronize the target to application
    logger.debug("Write pointer and write copy pointer updated");
    return apidefs.STLINK_OK;
  }

  /**
   * This private service wait the flag change to 0x02.
   * This means that the acquisition is stopped.
   */
  synchroStop() {
    //Todo use the  "async function synchroStop() .. await "
    let i = 0;
    this.readHeader();

    while ((this.h_flags & 0x02) !== flag.SNP_TRC_START_STOP_ACK && i < 50) {
      /*await*/ this.sleep(100);
      this.headerOfTriggerBuilt = false;
      this.readHeader();
      i++;
    }

    // Reset flags
    const myGroup = [
      {
        address: this.h_flagAddStrHex,
        type: apidefs.READ_TYPE_U32,
        value: "0x00" // Reset flag
      }
    ];
    const flagUpdate = stlinkprotocol.writeMemoryGroup(this.probeId, myGroup, this.accessPointId, this.typeConnect);
    if (flagUpdate.result !== apidefs.STLINK_OK) {
      logger.debug("Error: " + stlinkprotocol.errorString(flagUpdate.result));
    }
    this.synchroStopDone = true;
    return this.readHeader();
  }

  /**
   * This private service reset the flag value of the header.
   */
  resetFlag() {
    // Reset flags
    const myGroup = [
      {
        address: this.h_flagAddStrHex,

        type: apidefs.READ_TYPE_U32,
        value: "0x00" // Reset flag
      }
    ];
    const flagUpdate = stlinkprotocol.writeMemoryGroup(this.probeId, myGroup);
    if (flagUpdate.result !== apidefs.STLINK_OK) {
      logger.debug("Error: " + stlinkprotocol.errorString(flagUpdate.result));
      return SNAP_ERR.RESET_FLAG;
    }
    return apidefs.STLINK_OK;
  }

  /**
   * This private service wait the flag change to 0x02.
   * This means that the acquisition is started.
   *
   */
  synchroStart() {
    this.readHeader();

    if ((this.synchroStopDone && (this.h_flags & 0x02) == flag.SNP_TRC_START_STOP_ACK) || this.synchroFinished) {
      const myGroup = [
        {
          address: this.h_flagAddStrHex,
          type: apidefs.READ_TYPE_U32,
          value: "0x00" // Reset flag
        }
      ];
      const flagUpdate = stlinkprotocol.writeMemoryGroup(this.probeId, myGroup, this.accessPointId, this.typeConnect);
      if (flagUpdate.result !== apidefs.STLINK_OK) {
        logger.debug("Error: " + stlinkprotocol.errorString(flagUpdate.result));
      }
      this.abortTime(this.timeSynchroStart);
      this.synchroFinished = true;
      return this.readHeader();
    }

    this.timeSynchroStart = setTimeout(() => {
      this.synchroStart;
    }, 100);
  }

  /**
   * Stop the setTimeout.
   *
   * @param {*} timeSynchro - Timer to be abort.
   */
  abortTime(timeSynchro) {
    clearTimeout(timeSynchro);
  }

  /**
   * This private service makes an empty response.
   * It is used during the synchronization with the target.
   * During synchronization the target is not in acquiisitiopn mode.
   *
   */
  respEmpty() {
    const res = { result: apidefs.STLINK_OK };
    const val = [];
    if (this.VarsToBeRead !== undefined) {
      for (let i = 0; i < this.VarsToBeRead.length; i++) {
        val[i] = [];
        val[i].push({
          //y: undefined,
          x: this.snaptimestamp
        });
      }
    } else {
      val[0] = [];
      val[0].push({
        //y: undefined,
        x: this.snaptimestamp
      });
    }
    res.data = val;
    logger.debug("No measurement - result = ", res);
    return res;
  }

  /**
   * Private service to init snapshot mode.
   *
   * @param {*} acqParam - Acquisition parameters.
   * @param {*} groupToBeRead - Group of variables to be trace.
   * @param {*} APId - Id of the access point.
   * @param {*} ConnectionType - Type of connection.
   */
  initSnapshot(acqParam, groupToBeRead, APId, ConnectionType) {
    logger.debug("probe = ", this.probeId, ", acqParam = ", acqParam, ", groupToBeRead = ", groupToBeRead);

    this.accessPointId = APId;
    this.typeConnect = ConnectionType;
    this.VarsToBeRead = groupToBeRead;
    //Trigger parameter init
    this.h_triggerMode = acqParam.triggerstartmode;
    this.getAddressOfTrig(acqParam.triggername);
    this.getTypeOfTrig(acqParam.triggername);
    this.h_triggerThreshold = Number(acqParam.triggerthreshold);
    // read startMark value
    const initGroup = [
      {
        address: acqParam.snapshotheader,
        type: apidefs.READ_TYPE_U32
      }
    ];

    const startHeader = stlinkprotocol.readMemoryGroup(this.probeId, initGroup, this.accessPointId, this.typeConnect);
    if (startHeader.result !== apidefs.STLINK_OK) {
      logger.debug("Read group memory error: " + stlinkprotocol.errorString(startHeader.result));
      return SNAP_ERR.INIT_SNP;
    }

    logger.debug("startHeader = " + JSON.stringify(startHeader));
    const val = startHeader.data[0][0].y.toString(16).padStart(8, "0");
    const startMarkVal = val.substr(2, 8);
    this.headerSize = Number(val.substr(0, 2));

    if (startMarkVal !== EXPECTED_START_MARK) {
      logger.debug("StartMark value not expected", startMarkVal);
      return SNAP_ERR.SNAPSHOT_HEADER_NOT_FOUND;
    }

    if (this.headerSize !== 2 && this.headerSize !== 4) {
      logger.debug("Header size not supported by the target, error " + stlinkprotocol.errorString(startHeader.result));
      return SNAP_ERR.INIT_SNP;
    }

    this.buildHeader(acqParam.snapshotheader, this.headerSize);
    this.readHeader(); // reading of the header
    // Check version and trigger capability
    const VERSION = 0x00ff & ("0x" + this.h_headerVersion);
    if (VERSION < 8) {
      return SNAP_ERR.VERSION_NOT_SUPPORTED;
    }

    // Check the max number of variable is exceed
    if (this.VarsToBeRead.length > this.h_nbMaxVar) {
      return SNAP_ERR.MAX_VAR_NUM_EXCEED;
    }

    const TRIGGER_SUPPORTED = (0xff00 & ("0x" + this.h_headerVersion)) >> 8;

    if (TRIGGER_SUPPORTED !== 7 && this.h_triggerMode !== "manual") {
      return SNAP_ERR.TRIGGER_NOT_SUPPORTED;
    } else {
      this.buildTriggerHeader(acqParam.snapshotheader, this.headerSize, this.h_nbMaxVar);
      this.readHeader();
    }

    if (this.h_nbVar > 0) {
      this.traceIsStarted = true;
    } else {
      this.traceIsStarted = false;
    }

    return apidefs.STLINK_OK;
  }

  /**
   * This private service return the address of the trigger.
   *
   * @param {*} name - Name of the trigger.
   */
  getAddressOfTrig(name) {
    this.VarsToBeRead.forEach(element => {
      if (name === element.name) {
        this.h_triggerAddress = element.address;
      }
    });
  }

  /**
   * This private service return the type of the trigger.
   *
   * @param {*} name - Name of the trigger.
   */
  getTypeOfTrig(name) {
    this.VarsToBeRead.forEach(element => {
      if (name === element.name) {
        this.h_triggerAccessType = triggerTypeMap[element.type];
      }
    });
  }

  /**
   * Private service to build the header according to the startMark address.
   *
   * @param {*} startMark - Header start address.
   * @param {*} headerSize - Size of each element of header (2 or 4).
   */
  buildHeader(startMark, headerSize) {
    const startMarkVal = parseInt(startMark, 16);

    for (let i = 0; i < this.header.length; i++) {
      switch (this.header[i].name) {
        case "startMark[0]":
          this.header[i].address = startMark;
          break;
        case "headerFieldSize":
          this.header[i].address = "0x" + (startMarkVal + 3).toString(16);
          break;
        default:
          this.header[i].address = "0x" + (startMarkVal + (i - 1) * headerSize).toString(16); // all other fields
          break;
      }
      logger.debug(this.header[i].name + ": @" + this.header[i].address);
    }
  }

  /**
   * Private service to build the header according to the startMark address.
   *
   * @param {*} startMark - Header start address.
   * @param {*} headerSize - Size of each element of header (2 or 4).
   * @param {*} nbMaxVar - Max number of variable reserved in target.
   */
  buildTriggerHeader(startMark, headerSize, nbMaxVar) {
    const startMarkVal = parseInt(startMark, 16);

    for (let i = 0; i < this.headerOfTrigger.length; i++) {
      this.headerOfTrigger[i].address =
        "0x" +
        (
          startMarkVal +
          this.header.length * headerSize +
          (nbMaxVar - 1) * 2 * headerSize +
          (i - 1) * headerSize
        ).toString(16); // the address of trigger part is located after header structure

      logger.debug(this.headerOfTrigger[i].name + ": @" + this.headerOfTrigger[i].address);
    }

    this.headerOfTriggerBuilt = true;
  }

  /**
   * This private service read the header field from the target.
   * In case nbVar is to 0 some variables are updated.
   */
  readHeader() {
    const resOfGroup = stlinkprotocol.readMemoryGroup(this.probeId, this.header, this.accessPointId, this.typeConnect);
    if (resOfGroup.result !== apidefs.STLINK_OK) {
      logger.debug("Error: " + stlinkprotocol.errorString(resOfGroup.result));
      return SNAP_ERR.READ_HEADER;
    }
    for (let i = 0; i < this.header.length; i++) {
      const val = resOfGroup.data[i][0];
      logger.debug(this.header[i].name + ": " + val.y.toString(16));
      switch (this.header[i].name) {
        case "headerVersion":
          this.h_headerVersion = val.y.toString(16);
          break;
        case "bufferStartAddr":
          this.h_bufferStartAddrStrHex = val.y.toString(16);
          break;
        case "bufferEndAddr":
          this.h_bufferEndAddrStrHex = val.y.toString(16);
          this.h_bufferSize =
            (parseInt(this.h_bufferEndAddrStrHex, 16) - parseInt(this.h_bufferStartAddrStrHex, 16) + 1) / 4;
          break;
        case "nbVar":
          this.h_nbVar = val.y;
          this.h_nbVarAddrStrHex = this.header[i].address;
          break;
        case "flags":
          this.h_flags = val.y;
          this.h_flagAddStrHex = this.header[i].address;
          break;
        case "timestamp_base_unit":
          this.h_timestampValue = val.y;
          break;
        case "readPointer":
          this.h_readPointer = val.y;
          this.h_readPointerAddStrHex = this.header[i].address;
          break;
        case "readPointerCopy":
          this.h_readPointerCopyAddStrHex = this.header[i].address;
          break;
        case "writePointer":
          this.h_writePointer = val.y;
          this.h_writePointerAddStrHex = this.header[i].address;
          break;
        case "writePointerCopy":
          this.h_writePointerC = val.y;
          this.h_writePointerCopyAddStrHex = this.header[i].address;
          break;
        case "nbVarMax":
          this.h_nbMaxVar = Number(val.y);
          break;
        case "g_varList[0].accessType":
          this.h_listOfTypeStrHex = this.header[i].address;
          break;
        case "g_varList[0].address":
          this.h_listOfVarStrHex = this.header[i].address;
          break;
      }
    }

    if (this.headerOfTriggerBuilt) {
      const trigerHeaderVal = stlinkprotocol.readMemoryGroup(
        this.probeId,
        this.headerOfTrigger,
        this.accessPointId,
        this.typeConnect
      );

      for (let i = 0; i < this.headerOfTrigger.length; i++) {
        const val = trigerHeaderVal.data[i][0];
        logger.debug(this.headerOfTrigger[i].name + ": " + val.y.toString(16));
        switch (this.headerOfTrigger[i].name) {
          case "triggerAddress":
            this.h_triggerAddressStrHex = this.headerOfTrigger[i].address;
            break;
          case "triggerAccessType":
            this.h_triggerAccessTypeStrHex = this.headerOfTrigger[i].address;
            break;
          case "triggerThreshold":
            this.h_triggerThresholdStrHex = this.headerOfTrigger[i].address;
            break;
          case "triggerMode":
            this.h_triggerModeStrHex = this.headerOfTrigger[i].address;
            break;
          case "triggerCmd":
            this.h_triggerCmdStrHex = this.headerOfTrigger[i].address;
            this.h_triggerCmd = val.y;
            break;
          case "triggerThresholdDouble":
            this.h_triggerThresholdStrHexDouble = this.headerOfTrigger[i].address;
            break;
          case "triggerThresholdDouble2":
            this.h_triggerThresholdStrHexDouble2 = this.headerOfTrigger[i].address;
            break;
        }
      }
    }

    if (this.h_nbVar == 0) {
      // in this case the trace is not started or has been interrupted (by reset of taregt for example).
      this.h_readPointerEnd = undefined;
      this.previousMeas[0] = undefined;
    }
  }

  /**
   * Service to initialize the target.
   * Target header is changed with all vairables to be trace.
   * The read pointer is updated with the expected value.
   * Update the trigger parameters.
   * At the end of this procedure the trace is started in target side.
   */
  initTarget() {
    // Add one or more variables in the list to trace
    let myGroup = [];
    this.nbOfDouble = 0;
    for (let i = 0; i < this.VarsToBeRead.length; i++) {
      let elt = new Object();
      // Type of variable to be trace
      elt.address = "0x" + (parseInt(this.h_listOfTypeStrHex, 16) + i * this.headerSize * 2).toString(16);
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = targetTypeMap[this.VarsToBeRead[i].type].toString(16) || targetType.SNP_TRC_DATA_ACCESS_32BIT;
      myGroup.push(elt);
      // Address of variable
      elt = new Object();
      elt.address = "0x" + (parseInt(this.h_listOfVarStrHex, 16) + i * this.headerSize * 2).toString(16);
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = this.VarsToBeRead[i].address.toString(16);
      myGroup.push(elt);

      // compute the number of octect according type of variable
      if (
        this.VarsToBeRead[i].type == apidefs.READ_TYPE_S64 ||
        this.VarsToBeRead[i].type == apidefs.READ_TYPE_U64 ||
        this.VarsToBeRead[i].type == apidefs.READ_TYPE_DOUBLE
      ) {
        this.nbOfDouble += 1; // We have to inform the target about the number of double to be trace, the double is coded in 2 32 bits words
      }
    }

    // update Readpointer according to expected value.
    Math.trunc((this.recordSize = 4 * (1 + Number(this.VarsToBeRead.length) + this.nbOfDouble))); // it is the nb of octets by record
    const nbRecords = Math.trunc((this.h_bufferSize * 4) / this.recordSize); // nb of records that buffer can save, records is one acquisition, this acquisition is composed of 1 timestamp + varaibles that user want to trace.
    this.expectedInitialReadPointer = parseInt(this.h_bufferStartAddrStrHex, 16) + (nbRecords - 1) * this.recordSize;
    if (this.expectedInitialReadPointer !== this.h_readPointer) {
      let elt = new Object();
      // Update readpointer
      elt.address = this.h_readPointerAddStrHex;
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = "0x" + this.expectedInitialReadPointer.toString(16);
      myGroup.push(elt);
      // Update readpointerCopy
      elt = new Object();
      elt.address = this.h_readPointerCopyAddStrHex;
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = "0x" + this.expectedInitialReadPointer.toString(16);
      myGroup.push(elt);
    }

    let elt = new Object();
    // update write pointer in start buffer position
    elt = new Object();
    elt.name = "writePointer";
    elt.address = this.h_writePointerAddStrHex;
    elt.type = apidefs.READ_TYPE_U32;
    elt.value = "0x" + this.h_bufferStartAddrStrHex;
    myGroup.push(elt);
    elt = new Object();
    elt.name = "writePointerCopy";
    elt.address = this.h_writePointerCopyAddStrHex;
    elt.type = apidefs.READ_TYPE_U32;
    elt.value = "0x" + this.h_bufferStartAddrStrHex;
    myGroup.push(elt);

    if (this.headerOfTriggerBuilt) {
      // Trigger address
      elt = new Object();
      elt.name = "address";
      elt.address = this.h_triggerAddressStrHex;
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = this.h_triggerAddress;
      myGroup.push(elt);
      // Trigger type address
      elt = new Object();
      elt.name = "accessType";
      elt.address = this.h_triggerAccessTypeStrHex;
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = this.h_triggerAccessType.toString(16);
      myGroup.push(elt);
      // Trigger threshold
      if (this.h_triggerAccessType === targetType.SNP_TRC_DATA_ACCESS_TRIG_FLOAT) {
        const buf = Buffer.allocUnsafe(4);
        buf.writeFloatBE(this.h_triggerThreshold, 0);
        this.h_triggerThreshold = buf.toString("hex");
        elt = new Object();
        elt.name = "threshold";
        elt.address = this.h_triggerThresholdStrHex;
        elt.type = apidefs.READ_TYPE_U32;
        elt.value = "0x" + this.h_triggerThreshold;
        myGroup.push(elt);
      } else if (this.h_triggerAccessType === targetType.SNP_TRC_DATA_ACCESS_TRIG_DOUBLE) {
        // The target have 2 specifical 32 bits words to manange Double variable
        const buf = Buffer.allocUnsafe(8);
        buf.writeDoubleBE(this.h_triggerThreshold, 0);
        this.h_triggerThreshold = buf.toString("hex");
        let thresholdDouble = this.h_triggerThreshold.substr(8, 16);
        let thresholdDouble2 = this.h_triggerThreshold.substr(0, 8);
        elt = new Object();
        elt.name = "thresholdDouble";
        elt.address = this.h_triggerThresholdStrHexDouble;
        elt.type = apidefs.READ_TYPE_U32;
        elt.value = "0x" + thresholdDouble;
        myGroup.push(elt);
        elt = new Object();
        elt.name = "thresholdDouble2";
        elt.address = this.h_triggerThresholdStrHexDouble2;
        elt.type = apidefs.READ_TYPE_U32;
        elt.value = "0x" + thresholdDouble2;
        myGroup.push(elt);
      } else if (this.h_triggerThreshold < 0) {
        this.h_triggerThreshold = (this.h_triggerThreshold >>> 0).toString(16);
        elt = new Object();
        elt.name = "threshold";
        elt.address = this.h_triggerThresholdStrHex;
        elt.type = apidefs.READ_TYPE_U32;
        elt.value = "0x" + this.h_triggerThreshold;
        myGroup.push(elt);
      } else {
        elt = new Object();
        elt.name = "threshold";
        elt.address = this.h_triggerThresholdStrHex;
        elt.type = apidefs.READ_TYPE_U32;
        elt.value = "0x" + this.h_triggerThreshold.toString(16);
        myGroup.push(elt);
      }

      // Trigger mode
      elt = new Object();
      elt.name = "mode";
      elt.address = this.h_triggerModeStrHex;
      elt.type = apidefs.READ_TYPE_U32;
      elt.value = TRIGGER_MODE_MAP[this.h_triggerMode];
      myGroup.push(elt);

      // Trigger command
      elt = new Object();
      elt.name = "cmd";
      elt.address = this.h_triggerCmdStrHex;
      elt.type = apidefs.READ_TYPE_U32;
      if (this.h_triggerMode !== "manual") {
        elt.value = "0x01";
      } else {
        elt.value = "0x00";
      }
      myGroup.push(elt);
    }
    // Write all parameters to probeId, the target should start to write in the buffer
    const writeMem = stlinkprotocol.writeMemoryGroup(this.probeId, myGroup, this.accessPointId, this.typeConnect); // Write address of variable in the list
    if (writeMem.result === apidefs.STLINK_OK) {
      logger.debug("The target is ready to dump trace");
    } else {
      logger.debug("error: " + stlinkprotocol.errorString(writeMem.result));
    }

    this.readHeader(); // update the information from the target
    const unitTimestamp = (this.h_timestampValue & 0xf000) >> 12;
    switch (unitTimestamp) {
      case timeUnit.millisecond:
        this.unitOfTimestamp = 1;
        break;
      case timeUnit.second:
        this.unitOfTimestamp = 10;
        break;
      case timeUnit.microsecond:
        this.unitOfTimestamp = 0.1;
        break;
      case timeUnit.nanosecond:
        this.unitOfTimestamp = 0.01;
        break;
      default:
        this.unitOfTimestamp = 1;
        break;
    }
    this.valueOfTimestamp = this.h_timestampValue & 0x0fff;
  }

  /**
   * This service update the nbVar variable in the target header with the number of variable we have to trace.
   * When this variable change from 0 to another value the trace starts.
   */
  startTrace() {
    let myGroup = [];
    let elt = new Object();
    elt.address = this.h_nbVarAddrStrHex;
    elt.type = apidefs.READ_TYPE_U32;
    elt.value = "0x" + (this.VarsToBeRead.length + (this.nbOfDouble << 16)).toString(16).padStart(8, "0"); // due to target implementation the number of double is save in MSB of nbVar variriable
    myGroup.push(elt);
    const writeMem = stlinkprotocol.writeMemoryGroup(this.probeId, myGroup, this.accessPointId, this.typeConnect); // Write address of variable in the list
    if (writeMem.result !== apidefs.STLINK_OK) {
      logger.debug("error: " + stlinkprotocol.errorString(writeMem.result));
      return SNAP_ERR.START_TRACE;
    }

    logger.debug("The target start to dump trace");
    this.traceIsStarted = true; // the trace is started
    return this.readHeader(); // update the information from the target
  }

  /**
   * Service to read the traces in the buffer of target.
   * According the position of writePointer and readPointer the area for reading is computed.
   * The target is informed that we are reading the buffer by change the readPointer before reading and
   * by change readPointerCopy after reading.
   */
  readBuffer() {
    this.readPointerEndReached = false;
    this.readHeader(); // Update the information from the target
    let add;
    let result = undefined;
    var nbRecords = 0;
    loop: while (result === undefined) {
      if (this.h_readPointerEnd === undefined) {
        logger.debug("readPointerEnd === undefined");
        //First reading pointer set at startAddress
        add = parseInt(this.h_bufferStartAddrStrHex, 16);
        nbRecords = Math.trunc((this.h_writePointerC - add) / this.recordSize);
        if (nbRecords > 0) {
          this.h_readPointerEnd = add + nbRecords * this.recordSize;
        }
        return this.getValue(nbRecords, add);
      } else if (this.h_readPointerEnd < this.h_writePointerC) {
        logger.debug("readPointerEnd < writePointerC");
        // Cas the lower part of the buffer is reading during the target is Writing the upper part of the buffer
        add = this.h_readPointerEnd;
        nbRecords = Math.trunc((this.h_writePointerC - this.h_readPointerEnd) / this.recordSize);
        this.h_readPointerEnd = add + nbRecords * this.recordSize;
        return this.getValue(nbRecords, add);
      } else if (this.h_readPointerEnd === this.expectedInitialReadPointer + this.recordSize) {
        logger.debug("readPointerEnd === expectedInitialReadPointer + recordSize");
        // it is the last recording of the buffer.
        this.h_readPointerEnd = parseInt(this.h_bufferStartAddrStrHex, 16); // Gome back to the begining og buffer.
        continue loop;
      } else if (this.h_readPointerEnd > this.expectedInitialReadPointer + this.recordSize) {
        logger.debug("readPointerEnd > expectedInitialReadPointer + recordSize");
        // Cas the reading buffer is bigger than expectedInitialReadPointer --> the last record.
        this.h_readPointerEnd = undefined;
        //this.readBuffer();
        continue loop;
      } else if (this.h_readPointerEnd > this.h_writePointerC || (this.h_flags & 0x01) === flag.SNP_TRC_FLAG_OVERFLOW) {
        logger.debug("readPointerEnd > writePointerC || (flags & 0x01) === flag.SNP_TRC_FLAG_OVERFLOW");
        // Cas we are reading the greater part of the buffer during the target is Writing the lower part of the buffer
        // The reading part can be go to the end of buffer
        add = this.h_readPointerEnd;
        nbRecords = 1 + Math.trunc((this.expectedInitialReadPointer - this.h_readPointerEnd) / this.recordSize);
        this.h_readPointerEnd = add + nbRecords * this.recordSize;
        this.readPointerEndReached = true;
        return this.getValue(nbRecords, add);
      } else {
        logger.debug(
          "Not new value from buffer: WritePointer @ 0x",
          this.h_writePointer.toString(16),
          " - ReadPointer @ 0x",
          this.h_writePointer.toString(16)
        );
        return this.getValue(nbRecords, add);
      }
    }
  }

  /**
   * This function read the buffer.
   *
   * @param {*} nbRecords - Nunber of records.
   * @param {*} add - Startting address of the buffer recording.
   */
  getValue(nbRecords, add) {
    let res = { result: apidefs.STLINK_BAD_PARAMETER };
    if (nbRecords >= 1 && this.h_readPointerEnd !== undefined) {
      // a new value has been saved in the buffer
      // Update the read pointer for reading
      const readpointerLastRec = this.h_readPointerEnd - this.recordSize;
      const pointer = [
        {
          address: this.h_readPointerAddStrHex,
          type: apidefs.READ_TYPE_U32,
          value: "0x" + readpointerLastRec.toString(16)
        }
      ];

      let readValue = stlinkprotocol.writeMemoryGroup(this.probeId, pointer, this.accessPointId, this.typeConnect);
      if ((readValue.result = apidefs.STLINK_OK)) {
        logger.debug("Read pointer changed: @ 0x", readpointerLastRec.toString(16));
      } else {
        logger.debug("Error: " + stlinkprotocol.errorString(readValue.result));
      }

      const buffer = [
        {
          address: "0x" + add.toString(16),
          type: apidefs.READ_TYPE_U32,
          nb: (nbRecords * this.recordSize) / 4
        }
      ];

      const varRecorded = stlinkprotocol.readMemoryGroup(this.probeId, buffer, this.accessPointId, this.typeConnect);
      if ((this.h_flags & 0x01) === flag.SNP_TRC_FLAG_OVERFLOW) {
        logger.debug("Reset falgs OVERFLOW: ");
        const myGroup = [
          {
            address: this.h_flagAddStrHex,
            type: apidefs.READ_TYPE_FLOAT,
            value: "0x00" // Reset flag
          }
        ];
        stlinkprotocol.writeMemoryGroup(this.probeId, myGroup, this.accessPointId, this.typeConnect);
      }

      // Update readPointerCopy: readPointerCopy = readPointer is used by the target
      // to determinate that the reading is terminated.
      const pointerCopy = [
        {
          address: this.h_readPointerCopyAddStrHex,
          type: apidefs.READ_TYPE_U32,
          value: "0x" + readpointerLastRec.toString(16)
        }
      ];

      readValue = stlinkprotocol.writeMemoryGroup(this.probeId, pointerCopy, this.accessPointId, this.typeConnect);
      if ((readValue.result = apidefs.STLINK_OK)) {
        logger.debug("Read pointerCopy changed: @", readpointerLastRec.toString(16));
      } else {
        logger.debug("Error: " + stlinkprotocol.errorString(readValue.result));
      }
      let buffVal = [];

      // Format result to be compatible with node acquisition service.
      const size = nbRecords * this.recordSize;
      let j = 0;
      let timestamp;
      // Create table used for response.
      for (let k = 0; k < Number(this.VarsToBeRead.length); k++) {
        buffVal[k] = [];
      }
      if (varRecorded.result === apidefs.STLINK_OK) {
        for (let i = 0; i < size >> 2; i++) {
          logger.debug(
            "Data for buffer @ 0x" + (add + i * 4).toString(16).padStart(8, "0") + " : 0x" + varRecorded.data[0][i].y
          );

          if (j === 0) {
            timestamp =
              this.snaptimestamp + (1 + varRecorded.data[0][i].y) * this.valueOfTimestamp * this.unitOfTimestamp; //unit value: 10 = Second; 1 = millisecond; 0.1 = microsecond; 0.01 = nanosecond
          } else {
            const type = this.VarsToBeRead[j - 1].type;
            let val;
            if (type === apidefs.READ_TYPE_S8 || type === apidefs.READ_TYPE_S16 || type === apidefs.READ_TYPE_S32) {
              val = varRecorded.data[0][i].y >> 0;
            } else if (type === apidefs.READ_TYPE_FLOAT) {
              var reExp = /[0-9A-Fa-f]{8}/g;

              if (reExp.test(varRecorded.data[0][i].y.toString(16))) {
                const buf = Buffer.alloc(8, varRecorded.data[0][i].y.toString(16), "hex"); //new Buffer(varRecorded.data[0][i].y.toString(16), "hex");
                val = buf.readFloatBE(0);
              } else {
                logger.error("Error - Value received is not valid: ", varRecorded.data[0][i].y.toString(16));

                // todo inform acquisitionServices that the vqlue is not a valid value.
              }
            } else if (
              type === apidefs.READ_TYPE_DOUBLE ||
              type === apidefs.READ_TYPE_S64 ||
              type === apidefs.READ_TYPE_U64
            ) {
              let str64Bits = varRecorded.data[0][i + 1].y
                .toString(16)
                .padStart(8, "0")
                .concat(varRecorded.data[0][i].y.toString(16).padStart(8, "0"));
              let val64Bits = Buffer.alloc(8, str64Bits, "hex");

              if (type == apidefs.READ_TYPE_DOUBLE) {
                val = val64Bits.readDoubleBE(0);
              } else if (type == apidefs.READ_TYPE_S64) {
                // TODO use BigInt for accurate 64-bit value !!
                val = val64Bits.readInt32BE(0) * Math.pow(2, 32) + val64Bits.readInt32BE(4);
              } else if (type == apidefs.READ_TYPE_U64) {
                // TODO use BigInt for accurate 64-bit value !!
                val = val64Bits.readUInt32BE(0) * Math.pow(2, 32) + val64Bits.readUInt32BE(4);
              }

              i++;
            } else {
              val = varRecorded.data[0][i].y >> 0;
            }

            buffVal[j - 1].push({
              // -1 due to timestamp value
              y: val, //varRecorded.data[0][i].y,
              x: timestamp
            });

            // save value in case the next trace is empty
            this.previousMeas[j - 1] = {
              // -1 due to timestamp value
              y: val, //varRecorded.data[0][i].y,
              x: timestamp
            };
          }

          j++;
          if (j === 1 + Number(this.VarsToBeRead.length)) {
            j = 0;
          }
        }

        // Result format
        res.result = apidefs.STLINK_OK;
        res.data = buffVal;
        logger.debug("Result = ", res.result);
        logger.debug("Result = ", res.data);
        return res;
      } else {
        logger.debug("Error Value: " + stlinkprotocol.errorString(varRecorded.result));
      }
    } else {
      if (this.previousMeas[0] !== undefined) {
        // no measurement done, the previous measurment is pushed
        res.data = this.previousMeas;
      } else {
        let val = [];
        if (this.VarsToBeRead !== undefined) {
          for (let i = 0; i < this.VarsToBeRead.length; i++) {
            val[i] = [];
            val[i].push({
              //y: undefined,
              x: this.snaptimestamp + this.valueOfTimestamp * this.unitOfTimestamp
            });
          }
        } else {
          val[0] = [];
          val[0].push({
            //y: undefined,
            x: this.snaptimestamp + this.valueOfTimestamp * this.unitOfTimestamp
          });
        }
        res.data = val;
      }
      res.result = apidefs.STLINK_OK;
      logger.debug("No measurement - result = ", res);
      return res;
    }
  }

  /**
   * This function wait the delay defined in milliseconds.
   *
   * @param {*} delay - Delay for waiting.
   */
  sleep(delay) {
    var start = new Date().getTime();
    for (var i = 0; i < 1e7; i++) {
      if (new Date().getTime() - start > delay) {
        break;
      }
    }
  }
}

Snapshot.ongoing = [];

//export constructor
exports.Snapshot = Snapshot;

// export used for test only:
exports.SNAP_ERR = SNAP_ERR;
exports.targetType = targetType;
exports.targetTypeMap = targetTypeMap;
