/*!
 * module @stm32/stm32cubemonitor-stlink-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
declare module "@stm32/stm32cubemonitor-stlink-protocol" {
  /**
   * Returned result codes
   */
  const enum resultCode {
    // common result codes
    STLINK_OK = 0x01,
    STLINK_ERR = 0x02,
    STLINK_DRIVER_ERR = 0x03,
    STLINK_WRONG_DEVICE_MODE = 0x04,
    STLINK_DEVICE_NOT_OPENED = 0x05,
    STLINK_NO_DEVICE = 0x06,
    STLINK_BAD_PARAMETER = 0x07,
    STLINK_DEVICE_NOT_SUPPORTED = 0x08,
    STLINK_TIMEOUT = 0x09,

    // JTAG/SWD result codes
    STLINK_JTAG_PROTOCOL_NOT_SUPPORTED = 0xf0,
    STLINK_SWD_PROTOCOL_NOT_SUPPORTED = 0xf1,
    STLINK_JTAG_FREQUENCY_NOT_SUPPORTED = 0xf2,
    STLINK_SWD_FREQUENCY_NOT_SUPPORTED = 0xf3,
    STLINK_JTAG_READ_ERROR = 0xf4,
    STLINK_JTAG_ERROR = 0xf5,
    STLINK_JTAG_UNKNOWN_JTAG_CHAIN = 0xf6,
    STLINK_JTAG_NO_DEVICE_CONNECTED = 0xf7,
    STLINK_JTAG_WRITE_ERROR = 0xf8,
    STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE = 0xf9,
    STLINK_SWD_ERROR = 0xfa,
    STLINK_AP_ERROR = 0xfb,
    STLINK_SWV_NOT_AVAILABLE = 0xfc,
    STLINK_JTAG_TCP_ERROR = 0xfd,
    STLINK_JTAG_16BIT_ACCESS_NOT_SUPPORTED = 0xfe,
    STLINK_JTAG_CMD_NOT_SUPPORTED = 0xff
  }

  /**
   * List the connected probes
   * @param typeConnect the connection protocol to use either p2p or tcp
   * @returns an array of connected probes' descriptors
   */
  export function probeDiscovery(
    typeConnect: string
  ): Array<{
    probeName: string;
    probeId: string;
    property: { [protocol: string]: string[] };
  }>;

  /**
   * Service to connect to a target MCU through a probe
   * @param probeId the probe to be used to connect to the target MCU
   * @param protocol the protocol to be used by the probe to communicate with the target MCU
   * @param frequency the frequency to be used by the probe to communicate with the target MCU
   * @param rediscovery request or not a new discovery procedure
   * @param typeConnect the connection protocol to use either p2p or tcp
   * @returns operation result code, and if result is STLINK_OK, returns also MCU ID and MCU Name string
   */
  export function connectToTarget(
    probeId: string,
    protocol: string,
    frequency: string,
    rediscovery: boolean,
    typeConnect: string
  ): { result: resultCode; MCUID?: number; MCUName?: string };

  /**
   * readMode parameter
   */
  const enum readMode {
    READ_MODE_8_BIT = 0x01,
    READ_MODE_16_BIT = 0x02,
    READ_MODE_32_BIT = 0x03,
    READ_MODE_AUTO = 0x04 // default
  }

  /**
   * readType parameter
   */
  const enum readType {
    READ_TYPE_U8 = 0x01,
    READ_TYPE_S8 = 0x02,
    READ_TYPE_U16 = 0x03,
    READ_TYPE_S16 = 0x04,
    READ_TYPE_U32 = 0x05,
    READ_TYPE_S32 = 0x06,
    READ_TYPE_U64 = 0x07,
    READ_TYPE_S64 = 0x08,
    READ_TYPE_FLOAT = 0x09,
    READ_TYPE_DOUBLE = 0x0a
  }

  /**
   * Service to read data from a target memory area, specifying the type of data to be read
   *
   * @param probeId the probe to be used to communicate with the target MCU
   * @param baseAddress the base address to read from in the target memory space
   *                    (as a 0x-started string in hexadecimal representation)
   * @param nbElem the number of elements of the specified type to be read
   * @param type the read type : u8-s8 - address and size may have any value
   *                             u16-s16 - address and size must be multiples of 2
   *                             u32-s32-float - address and size must be multiples of 4
   *                             u64-s64-double - address and size must be multiples of 8
   * @param accessPointId id of the access point
   * @param typeConnect the connection protocol to use either p2p or tcp
   * @returns operation result code, and if result is STLINK_OK, returns also an array of {x,y} duplets
   *          objects with x for timestamp and y for value according to requested type
   */

  export function readMemoryTyped(
    probeId: string,
    baseAddress: string,
    nbElem: number,
    type: readType,
    accessPointId: number,
    typeConnect: string
  ): { result: resultCode; data?: [{ x: number; y: number }] };

  /**
   * Service to read data from a target memory area, specifying a memory access mode
   *
   * @param probeId the probe to be used to communicate with the target MCU
   * @param baseAddress the base address to read from in the target memory space
   *                    (as a 0x-started string in hexadecimal representation)
   * @param sizeInBytes the size in bytes of the memory area to read
   * @param mode the read mode : 8-bit - address and size may have any value, but read operation is slower
   *                             16-bit - address and size must be multiples of 2, but read operation is faster
   *                             32-bit - address and size must be multiples of 4, but read operation is fastest
   *                             automatic (default) - address and size may have any value, and an optimized
   *                                                   mix of 8-bit and 32-bit operations is performed
   * @param accessPointId id of the access point
   * @param typeConnect the connection protocol to use either p2p or tcp
   * @returns operation result code, and if result is STLINK_OK, returns also an array of data bytes and a timestamp
   */

  export function readMemoryMode(
    probeId: string,
    baseAddress: string,
    sizeInBytes: number,
    mode: readMode,
    accessPointId: number,
    typeConnect: string
  ): { result: resultCode; data?: number[]; time?: number };

  /**
   * Service to a group of data from a target memory area
   *
   * @param probeId the probe to be used to communicate with the target MCU
   * @param group structure containing:
   *              - variablelist the group of data to be read from the target memory
   *              this is an array, containing objects each describing one variable with
   *              its name, its address (as a 0x-started string in hexadecimal representation),
   *              its type (from READ TYPE numeric constants), and the number of items to be read
   *              - a list of parameters for trigger and snapshot mode
   *              - The access point Id
   * @param typeConnect the connection protocol to use either p2p or tcp
   * @returns operation result code, and if result is STLINK_OK, returns also an array
   *          containing one array of {x,y} duplets objects for each variable from the group parameter;
   *          x contains the timestamp of read operation, y contains the read value according to the variable type
   *          variable indexes in this array are the same as in the group parameter
   */

  export function readMemoryGroup(
    probeId: string,
    group: {
      variablelist: [{ name: string; address: string; type: number }];
      mode?: string;
      snapshotheader?: string;
      triggerstart?: string;
      triggerThreshold?: number;
      triggerstop?: string;
      triggeraddress?: string;
      triggertype?: string;
      accesspoint?: number;
    },
    typeConnect: string
  ): { result: resultCode; data?: [[{ x: number; y: number }]] };

  /**
   * Service to write a group of data from a target memory area
   *
   * @param probeId the probe to be used to communicate with the target MCU
   * @param group the group of data to be read from the target memory
   *              this is an array, containing objects each describing one variable with
   *              its name, its address (as a 0x-started string in hexadecimal representation),
   *              its type (from READ TYPE numeric constants), and its value to write
   *              And The access point Id.
   * @param typeConnect the connection protocol to use either "p2p" or "tcp".
   * @returns operation result code
   */

  export function writeMemoryGroup(
    probeId: string,
    group: {
      variablelist: [{ name: string; address: string; type: number; value: string }];
      accesspoint?: number;
    },
    typeConnect: string
  ): { result: resultCode };

  /**
   * Service to properly disconnect each probe from its target
   * and close each probe connection with host
   *
   * @param probeId the probe to be used to communicate with the target MCU
   *        use "" to disconnect all connected probes
   * @param typeConnect the connection protocol to use either p2p or tcp
   * @returns operation result code
   */

  export function closeProbes(probeId: string, typeConnect: string): { result: resultCode };

  /**
   * Service to get error String matching error code
   *
   *  @param errorCode the error code value
   *  @returns error string
   */
  export function errorString(errorCode: resultCode): string;

  /**
   * Service to stop the snapshot process running in target.
   *
   * @param probeId the probe identifier in which the snapshot is activated.
   * @returns operation result, 1 if ok, error string otherwise
   */
  export function stopSnapshot(probeId: string): any;

  /**
   * Service to make the probe led blinked.
   *
   * @param {string} probeId - The probe to be used to communicate with the target MCU.
   * @param {string} connectType - The connection protocol to use, either "p2p" or "tcp".
   */
  export function blinkLed(probeId: string, connectType: string);

  /**
   * Check TCP Server availability.
   *
   * @returns {boolean} True if TCP server is available, false otherwise.
   */
  export function isTcpAvailable();
}
