/*!
 * module cubemonitor-stlink-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const logger = require("@stm32/stm32cubemonitor-logger").getLogger("stlinkinterface");
var stlinkprotocol = require("./stlinkprotocol.js");
var Snapshot = require("./snapshot.js").Snapshot;
var apidefs = require("./stlinkprotocol").apidefs;

/**
 * Service to read a group of variables to a target memory area.
 *
 * @param {string} probeId - The probe to be used to communicate with the target MCU.
 * @param {object} group - Containing
 *              -variablelist {Array<Object>}- The list of variables to be read
 *               from the target memory. This is an array, containing objects each describing one
 *               variable with its address (as a 0x-started string in hexadecimal representation),
 *               withits type (from READ TYPE numeric constants), and the number of items to be read.
 *              - a list of parameters for acquisition, either for direct or snapshot mode.
 *               In snapshot case the header parameter from target is defined: snapshotheader address,
 *               trigger parameters.
 *              - The access point identifier.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {object} Operation result code.
 */
function readMemoryGroup(probeId, group, typeConnect) {
  logger.debug("Arguments : probeId = ", probeId, ", variablelist = ", group.variablelist, ", mode = ", group.mode);
  //Case direct mode
  if (group.mode === undefined || group.mode == "direct") {
    let res = stlinkprotocol.readMemoryGroup(probeId, group.variablelist, group.accesspoint, typeConnect);
    logger.trace("Direct acquisition: ", res.result);
    logger.trace("Direct acquisition: ", res.data);
    return res;
  } else {
    //Snapshot mode
    let snap = new Snapshot(probeId);
    let res = snap.getTrace(group.variablelist, group, group.accesspoint, typeConnect);
    if (res == undefined) {
      // may be the target is crash or reset during acquisition
      let response = new Object();
      response.result = apidefs.STLINK_TIMEOUT;
      logger.error("Snapshot acquisition: ", stlinkprotocol.errorString(response.result));
      return response;
    } else {
      logger.trace("Snapshot acquisition: ", res.result);
      logger.trace("Snapshot acquisition: ", res.data);
      return res;
    }
  }
}

/**
 * Service to write a group of data to a target memory area.
 *
 * @param {string} probeId - The probe to be used to communicate with the target MCU.
 * @param {Array<object>} group - The list of data to write to the target memory
 *              this is an array, containing objects each describing one variable with
 *              its name, its address (as a 0x-started string in hexadecimal representation),
 *              its type (from READ TYPE numeric constants), the number of items to write,
 *              and the value to write.
 *                         - plus access point identifier.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {object} Operation result code.
 */
function writeMemoryGroup(probeId, group, typeConnect) {
  return stlinkprotocol.writeMemoryGroup(probeId, group.variablelist, group.accesspoint, typeConnect);
}

/**
 * Service return the error code as words.
 *
 * @param {*} errorCode - Error code returned.
 */
function errorString(errorCode) {
  return stlinkprotocol.errorString(errorCode);
}

/**
 * Service to connect to a target MCU through a probe.
 *
 * @param {*} probeId - The probe to be used to connect to the target MCU.
 * @param {*} protocol - The protocol to be used by the probe to communicate with the target MCU.
 * @param {*} frequency - The frequency to be used by the probe to communicate with the target MCU.
 * @param {*} rediscovery - Discovery perform a second time if no device has been ever found.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {*} Operation result code, and if result is STLINK_OK, returns also MCU ID and MCU Name string.
 */
function connectToTarget(probeId, protocol, frequency, rediscovery, typeConnect) {
  return stlinkprotocol.connectToTarget(probeId, protocol, frequency, rediscovery, typeConnect);
}

/**
 * Service to properly disconnect each probe from its target
 * and close each probe connection.
 *
 * @param {string} probeId - The probe to be used to communicate with the target MCU.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 */
function closeProbes(probeId, typeConnect) {
  return stlinkprotocol.closeProbes(probeId, typeConnect);
}

/**
 * Service to read data from a target memory area, specifying a memory access mode.
 *
 * @param {*} probeId  - The probe to be used to communicate with the target MCU.
 * @param {*} baseAddress - The base address to read from in the target memory space
 *                    (as a 0x-started string in hexadecimal representation).
 * @param {*} sizeInBytes - The size in bytes of the memory area to read.
 * @param {*} mode - The read mode :
 *                             8-bit - address and size may have any value, but read operation is 8-bit only (slower)
 *                             16-bit - address and size must be multiples of 2, but read operation is faster
 *                             32-bit - address and size must be multiples of 4, but read operation is fastest
 *                             automatic (default) - address and size may have any value, and an optimized
 *                                                   mix of 8-bit and 32-bit operations is performed.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {*} Operation result code, and if result is STLINK_OK, returns also an array of data bytes and a timestamp.
 */
function readMemoryMode(probeId, baseAddress, sizeInBytes, mode, accessPointId, typeConnect) {
  return stlinkprotocol.readMemoryMode(probeId, baseAddress, sizeInBytes, mode, accessPointId, typeConnect);
}

/**
 * Service to read data from a target memory area, specifying the type of data to be read.
 *
 * @param {*} probeId - The probe to be used to communicate with the target MCU.
 * @param {*} baseAddress - The base address to read from in the target memory space
 *                    (as a 0x-started string in hexadecimal representation).
 * @param {*} nbElem - The number of elements of the specified type to be read.
 * @param {*} type - The read type : u8-s8 - address and size may have any value
 *                             u16-s16 - address and size must be multiples of 2
 *                             u32-s32-float - address and size must be multiples of 4
 *                             u64-s64-double - address and size must be multiples of 8.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {object} Operation result code, and if result is STLINK_OK, returns also an array of duplets objects
 * with x for timestamp and y for value according to requested type.
 */
function readMemoryTyped(probeId, baseAddress, nbElem, type, accessPointId, typeConnect) {
  return stlinkprotocol.readMemoryTyped(probeId, baseAddress, nbElem, type, accessPointId, typeConnect);
}

/**
 * Service to stop the acquisition in the target.
 *
 * @param {string} probeId  - The probe to be used to communicate with the target MCU.
 * @returns {*} Operation result, 1 if ok, error string otherwise.
 */
function stopSnapshot(probeId) {
  let snap = Snapshot.getInstance(probeId);
  if (snap) {
    return snap.snapshotInterrupt();
  }
  return apidefs.STLINK_OK;
}

/**
 * List the connected probes.
 *
 * @typedef {object} probe The protocol
 * @property {string} probeName The protocol name.
 * @property {string} probeId The list of supported frequency for this protocol.
 * @property {object.<string, string[]>} property The list of supported frequency for this protocol.
 *
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {Array.<probe>}
 * An array of connected probes' descriptors.
 */
function probeDiscovery(typeConnect) {
  return stlinkprotocol.probeDiscovery(typeConnect);
}

/**
 * Check TCP Server availability.
 *
 * @returns {boolean} True if TCP server is available, false otherwise.
 */
function isTcpAvailable() {
  return stlinkprotocol.isTcpAvailable();
}

/**
 * Service to make the probe led blink.
 *
 * @param {string} probeId - The probe identifier on which the led should blink.
 * @param {string} typeConnect - The connection protocol to use, either "p2p" or "tcp".
 */
function blinkLed(probeId, typeConnect) {
  return stlinkprotocol.blinkLed(probeId, typeConnect);
}

// Exports for Acquisition Services module
module.exports.probeDiscovery = probeDiscovery;
module.exports.connectToTarget = connectToTarget;
module.exports.closeProbes = closeProbes;
module.exports.readMemoryGroup = readMemoryGroup;
module.exports.writeMemoryGroup = writeMemoryGroup;
module.exports.readMemoryMode = readMemoryMode;
module.exports.readMemoryTyped = readMemoryTyped;
module.exports.errorString = errorString;
module.exports.stopSnapshot = stopSnapshot;
module.exports.blinkLed = blinkLed;
module.exports.isTcpAvailable = isTcpAvailable;
