/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const logger = require("@stm32/stm32cubemonitor-logger").getLogger("stlinkprotocol");
const { stLinkDrv, TDeviceInfo2, TDeviceRequest, Handle, defs: defsDrv } = require("./STLinkUSBDriver");
var apicom = require("./apicommon");
var drvToAPI = apicom.drvToAPI;
var JTAG = require("./jtagcommands");
//var SWIM = require("./swimcommands");

var ref = require("ref-napi");

const defs = {
  // ST-Link GET VERSION command
  STLINK_CMD_GET_VERSION: 0xf1,
  STLINK_CMD_GET_VERSION_EXT: 0xfb,

  // ST-Link GET CURRENT MODE command
  STLINK_CMD_GET_CURRENT_MODE: 0xf5,
  // GET CURRENT MODE returned values
  STLINK_DEV_DFU_MODE: 0x00,
  STLINK_DEV_MASS_MODE: 0x01,
  STLINK_DEV_JTAG_MODE: 0x02,
  STLINK_DEV_SWIM_MODE: 0x03,
  STLINK_DEV_BTLD_MODE: 0x04,

  // ST-Link DFU command
  STLINK_CMD_DFU: 0xf3,
  // DFU sub-command
  STLINK_DFU_EXIT: 0x07,

  // ST-Link SWIM command
  STLINK_CMD_SWIM: 0xf4,
  // SWIM sub-commands
  STLINK_SWIM_MODE_ENTER: 0x00,
  STLINK_SWIM_MODE_EXIT: 0x01,
  STLINK_SWIM_GET_VERSION: 0x02,
  STLINK_SWIM_SET_HOST_SPEED_MODE: 0x03,
  STLINK_SWIM_RESET_ENTRY_SEQUENCE: 0x04,
  STLINK_SWIM_RESET_SRST: 0x05,
  STLINK_SWIM_RESET_COMM_RESET: 0x06,
  STLINK_SWIM_RESET_CHIP_RESET_ON: 0x07,
  STLINK_SWIM_RESET_CHIP_RESET_OFF: 0x08,
  STLINK_SWIM_GET_STATUS: 0x09,
  STLINK_SWIM_WOTF_START: 0x0a,
  STLINK_SWIM_ROTF_START: 0x0b,
  STLINK_SWIM_ROTF_GET_DATA: 0x0c,
  STLINK_SWIM_GET_BUFFER_LENGTH: 0x0d,

  // minimum FW version supported
  STLINK_JTAG_MIN_FW_VERSION: 25,
  STLINK_SWIM_MIN_FW_VERSION: 5,
  STLINK_JTAG_MIN_FW_VERSION_FOR_16BIT_ACCESSES: 26,
  STLINK_JTAG_MIN_FW_VERSION_FOR_AP_AND_TCP: 28,

  // TCP Constants
  STLINK_TCP_CMD_REFRESH_DEVICE_LIST: 0x00,
  STLINK_TCP_CMD_GET_NB_DEV: 0x01,
  STLINK_TCP_CMD_GET_DEV_INFO: 0x02,
  STLINK_TCP_CMD_OPEN_DEV: 0x03,
  STLINK_TCP_CMD_CLOSE_DEV: 0x04,
  STLINK_TCP_CMD_SEND_USB_CMD: 0x05,
  STLINK_TCP_GET_SERVER_API_VERSION: 0x06,
  STLINK_TCP_CMD_GET_NB_OF_DEV_CLIENTS: 0x07
};

// definitions for interface towards CLI only
const clidefs = {
  // GET CURRENT MODE returned values
  STLINK_DEV_DFU_MODE: defs.STLINK_DEV_DFU_MODE,
  STLINK_DEV_MASS_MODE: defs.STLINK_DEV_MASS_MODE,
  STLINK_DEV_JTAG_MODE: defs.STLINK_DEV_JTAG_MODE,
  STLINK_DEV_SWIM_MODE: defs.STLINK_DEV_SWIM_MODE,
  STLINK_DEV_BTLD_MODE: defs.STLINK_DEV_BTLD_MODE
};

// definitions for interfaces towards CLI and Acquisition Services
const apidefs = {
  // DEBUG PROTOCOL to use in connectToTarget API
  STLINK_SWIM_PROTOCOL: "SWIM",

  // READ MODE to use in readMemory API
  READ_MODE_8_BIT: 0x01,
  READ_MODE_16_BIT: 0x02,
  READ_MODE_32_BIT: 0x03,
  READ_MODE_AUTO: 0x04,

  // READ TYPE to use in ReadMemoryType API
  READ_TYPE_U8: 0x01,
  READ_TYPE_S8: 0x02,
  READ_TYPE_U16: 0x03,
  READ_TYPE_S16: 0x04,
  READ_TYPE_U32: 0x05,
  READ_TYPE_S32: 0x06,
  READ_TYPE_U64: 0x07,
  READ_TYPE_S64: 0x08,
  READ_TYPE_FLOAT: 0x09,
  READ_TYPE_DOUBLE: 0x0a
};

const TCP_3 = "/0";
const DEFAULT_TCP_SERVER_OPTION = "-a";

// merge the other API definitions with this file's
// definitions to provide them all at once to Acquisition Services
Object.assign(apidefs, apicom.apidefs);
Object.assign(apidefs, JTAG.apidefs);

var nbDevices = 0;
var Devices = [];
var listProbe = [];

const MAX_FRAME_SIZE = 200;

/**
 * Reenumerate the connected devices on specified prootcol.
 *
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 */
function reenumerate(typeConnect) {
  logger.debug("STLINKProtocol- reenumerate", typeConnect);

  let result;
  switch (typeConnect) {
    case "p2p":
      result = stLinkDrv.STLink_Reenumerate(defsDrv.DRV_DEFAULT_IF, false);
      if (result == defsDrv.DRV_OK) {
        // automatically get the number of enumerated devices
        nbDevices = stLinkDrv.STLink_GetNbDevices(defsDrv.DRV_DEFAULT_IF);
      } else {
        logger.debug("result = ", drvToAPI(result));
        nbDevices = 0;
      }
      break;

    case "tcp":
      result = stLinkDrv.STLink_ReenumerateTcp(defsDrv.DRV_STLINK_TCP, false, TCP_3, DEFAULT_TCP_SERVER_OPTION);
      if (result == defsDrv.DRV_OK) {
        // automatically get the number of enumerated devices
        nbDevices = stLinkDrv.STLink_GetNbDevices(defsDrv.DRV_STLINK_TCP);
      } else {
        logger.debug("result = ", drvToAPI(result));
        nbDevices = 0;
      }
      break;
    default:
      logger.error("typeConnect unknown");
      break;
  }

  logger.trace("nbDevices found = ", nbDevices);
  // reset array of devices
  Devices = [];

  return drvToAPI(result);
}

/**
 * Gets the number of connected devices.
 *
 * @returns {number} - The number of connected devices.
 */
function getNbDevices() {
  return nbDevices;
}

/**
 * Gets the information on the specified device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - The list of informations on the device or null.
 */
function getDeviceInfo(deviceIndex, typeConnect) {
  logger.debug("STLinkProtocol getDeviceInfo", deviceIndex, typeConnect);
  if (deviceIndex > nbDevices - 1) {
    logger.warn("No device with index = ", deviceIndex);
    return null;
  }

  let deviceInfo = new TDeviceInfo2();

  deviceInfo.ref().fill(0);

  let param_1 = defsDrv.DRV_DEFAULT_IF;

  if (typeConnect == "tcp") {
    param_1 = defsDrv.DRV_STLINK_TCP;
  }

  let result = stLinkDrv.STLink_GetDeviceInfo2(param_1, deviceIndex, deviceInfo.ref(), deviceInfo.ref().length);

  if (result != defsDrv.DRV_OK) {
    logger.error("No device with index = ", deviceIndex);
    return null;
  }

  Devices[deviceIndex] = {
    DevNumber: deviceIndex,
    StLinkUsbId: "0x" + deviceInfo.StLinkUsbId.toString(16).padStart(8, "0"),
    //At that stage, cookie contains cookie 1 value
    Cookie: deviceInfo.StLinkUsbId,
    EnumUniqueId: getUniqueIDString(deviceInfo.EnumUniqueId),
    VendorId: getVendorString(deviceInfo.VendorId),
    ProductId: getProductString(deviceInfo.ProductId),
    DeviceUsed: deviceInfo.DeviceUsed ? true : false
  };
  return Devices[deviceIndex];
}

/**
 * Opens the specified device on p2p connection protocol.
 *
 * @param {number} deviceIndex - The index of the device.
 * @returns {number} - The status of the open operation.
 */
function openDevice(deviceIndex) {
  if (deviceIndex > nbDevices - 1) {
    logger.error("No device with index = ", deviceIndex);
    return apidefs.STLINK_BAD_PARAMETER;
  }

  let handlePtr = ref.alloc(Handle);

  let result = stLinkDrv.STLink_OpenDevice(
    defsDrv.DRV_DEFAULT_IF,
    deviceIndex,
    defsDrv.DRV_OPEN_STLINK_SHARED_ACCESS,
    handlePtr
  );

  if (result != defsDrv.DRV_OK) {
    Devices[deviceIndex].state = "closed";
    logger.error("deviceIndex(", deviceIndex, ") is not opened. Result = ", result);
    Devices[deviceIndex].Handle = undefined;
  } else {
    Devices[deviceIndex].Handle = handlePtr.deref();
    Devices[deviceIndex].state = "opened";
  }

  return drvToAPI(result);
}

/**
 * Opnes the specified device on tcp protocol.
 *
 * @param {number} deviceIndex - The index of the device.
 * @returns {number} - The status of the open operation.
 */
function openDeviceTcp(deviceIndex) {
  if (deviceIndex > nbDevices - 1) {
    logger.error("No device with index = ", deviceIndex);
    return apidefs.STLINK_BAD_PARAMETER;
  }

  let result;
  logger.debug("Cookie 1: " + Devices[deviceIndex].Cookie);
  result = stLinkDrv.STLink_OpenDeviceTcp(
    defsDrv.DRV_STLINK_TCP,
    Devices[deviceIndex].Cookie,
    defsDrv.DRV_OPEN_STLINK_SHARED_ACCESS
  );

  if (result != defsDrv.DRV_OK) {
    Devices[deviceIndex].state = "closed";
    logger.error("deviceIndex(", deviceIndex, ") is not opened. Result = ", result);
    Devices[deviceIndex].Cookie = undefined;
  } else {
    Devices[deviceIndex].state = "opened";
    logger.debug("Cookie 2: " + Devices[deviceIndex].Cookie);
  }

  return drvToAPI(result);
}

/**
 * Close the specified device on p2p connection protocol.
 *
 * @param {number} deviceIndex - The index of the device.
 * @returns {number} - The status of the close operation.
 */
function closeDevice(deviceIndex) {
  let check = checkOpenedDevice(deviceIndex);
  if (check != apidefs.STLINK_OK) {
    logger.warn("Device (", deviceIndex, ") not ready to be closed = ", check);
    return check;
  }

  if (Devices[deviceIndex].accessPointId != undefined) {
    closeAccessPoint(deviceIndex, Devices[deviceIndex].accessPointId, "p2p");
  }

  let result = stLinkDrv.STLink_CloseDevice(Devices[deviceIndex].Handle);

  if (result == defsDrv.DRV_OK) {
    Devices[deviceIndex].state = "closed";
    logger.trace("deviceIndex (", deviceIndex, ") is closed.");
  }
  return drvToAPI(result);
}

/**
 * Close the specified device on tcp protocol.
 *
 * @param {number} deviceIndex - The index of the device.
 * @returns {number} - The status of the close operation.
 */
function closeDeviceTcp(deviceIndex) {
  let check = checkOpenedDevice(deviceIndex);
  if (check != apidefs.STLINK_OK) {
    logger.warn("Device (", deviceIndex, ") not ready to be closed = ", check);
    return check;
  }

  if (Devices[deviceIndex].accessPointId != undefined) {
    closeAccessPoint(deviceIndex, Devices[deviceIndex].accessPointId, "tcp");
  }
  const result = stLinkDrv.STLink_CloseDeviceTcp(Devices[deviceIndex].Cookie, 1);

  if (result == defsDrv.DRV_OK) {
    Devices[deviceIndex].state = "closed";

    logger.trace("deviceIndex (", deviceIndex, ") is closed.");
  }

  return drvToAPI(result);
}

/**
 * Gets the mode on which the link is configured can be:
 * JTAG(0x02), SWIM(0x03), MASS STORAGE(0x01) or DFU(0x00).
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the commands and if OK the mode and if DFU, it version.
 */
function getCurrentMode(deviceIndex, typeConnect) {
  logger.trace("deviceIndex = ", deviceIndex);
  let check = checkOpenedDevice(deviceIndex);

  if (check != apidefs.STLINK_OK) {
    logger.error("Device (", deviceIndex, ") not possible to get current mode = ", check);
    return { result: check };
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(2).fill(0);

  req.CDBLength = 2;
  req.CDBByte[0] = defs.STLINK_CMD_GET_CURRENT_MODE;
  req.CDBByte[1] = 0x00;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let result;
  if (typeConnect == "tcp") {
    result = stLinkDrv.STLink_SendCommandTcp(Devices[deviceIndex].Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    result = stLinkDrv.STLink_SendCommand(Devices[deviceIndex].Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (result != defsDrv.DRV_OK) {
    logger.error("Result = ", drvToAPI(result));
    return { result: drvToAPI(result) };
  }

  // mode response is an u16 in Little Endian
  let tmpmode = resp.readUInt16LE(0);

  // actual mode byte is the LSB of this LE-u16
  Devices[deviceIndex].mode = tmpmode & 0x00ff;

  // if mode is DFU, DFU version is in MSB of this LE-u16
  if (Devices[deviceIndex].mode == defs.STLINK_DEV_DFU_MODE) {
    Devices[deviceIndex].DFUVersion = (tmpmode & 0xff00) >> 8;
  } else {
    Devices[deviceIndex].DFUVersion = 0;
  }

  logger.trace("Devices[", deviceIndex, "] = ", Devices[deviceIndex]);
  return {
    result: apidefs.STLINK_OK,
    mode: Devices[deviceIndex].mode,
    DFUVersion: Devices[deviceIndex].DFUVersion
  };
}

/**
 * Gets the versions of the specified decice.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command and versions of ST-LINK, JTAG and SWIM.
 */
function getVersion(deviceIndex, typeConnect) {
  logger.trace("deviceIndex = ", deviceIndex);

  let check = checkOpenedDevice(deviceIndex);

  if (check != apidefs.STLINK_OK) {
    logger.warn("Device (", deviceIndex, ") not possible to get version = ", check);
    return { result: check };
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(6).fill(0);

  req.CDBLength = 2;
  req.CDBByte[0] = defs.STLINK_CMD_GET_VERSION;
  req.CDBByte[1] = 0x00;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let result;
  if (typeConnect == "tcp") {
    result = stLinkDrv.STLink_SendCommandTcp(Devices[deviceIndex].Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    result = stLinkDrv.STLink_SendCommand(Devices[deviceIndex].Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (result != defsDrv.DRV_OK) {
    logger.error("Result = ", drvToAPI(result));
    return { result: drvToAPI(result) };
  }

  Devices[deviceIndex].STLinkVersion = (resp[0] >> 4) & 0x0f; // 4 MSB of first byte
  Devices[deviceIndex].JTAGVersion = ((resp[0] << 2) & 0x3c) | ((resp[1] >> 6) & 0x03); // 4 LSB of first byte + 2 MSB of second byte
  Devices[deviceIndex].SWIMVersion = resp[1] & 0x3f; // 6 LSB of second byte

  //ST-Link/V3 versions must be retrieved using ST_GetVersionExt()
  if (Devices[deviceIndex].STLinkVersion >= 3) {
    return getVersionExt(deviceIndex, typeConnect);
  }

  return {
    result: apidefs.STLINK_OK,
    STLinkVersion: Devices[deviceIndex].STLinkVersion,
    JTAGVersion: Devices[deviceIndex].JTAGVersion,
    SWIMVersion: Devices[deviceIndex].SWIMVersion
  };
}

/**
 * Exit the DFU mode for hte specified device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function exitDFUMode(deviceIndex, typeConnect) {
  logger.trace("deviceIndex = ", deviceIndex);
  let check = checkOpenedDevice(deviceIndex, defs.STLINK_DEV_DFU_MODE);
  if (check != apidefs.STLINK_OK) {
    logger.warn("Device (", deviceIndex, ") not possible to exit DFU mode = ", check);
    return check;
  }

  let req = new TDeviceRequest();

  req.CDBLength = 2;
  req.CDBByte[0] = defs.STLINK_CMD_DFU;
  req.CDBByte[1] = defs.STLINK_DFU_EXIT;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = null; // no answer expected
  req.BufferLength = 0;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let result;
  if (typeConnect == "tcp") {
    result = stLinkDrv.STLink_SendCommandTcp(Devices[deviceIndex].Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    result = stLinkDrv.STLink_SendCommand(Devices[deviceIndex].Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  return drvToAPI(result);
}

/**
 * Internal - Checks specified device is opened in the given mode.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} expectedMode - The mode the device is supposed to be.
 * @returns {number} - Value apidefs.STLINK_OK if openned, result of the command otherwise.
 */
function checkOpenedDevice(deviceIndex, expectedMode) {
  if (deviceIndex > nbDevices - 1) {
    return apidefs.STLINK_BAD_PARAMETER;
  }
  if (Devices[deviceIndex] == undefined) {
    return apidefs.STLINK_DEVICE_NOT_OPENED;
  }
  if (Devices[deviceIndex].state != "opened") {
    return apidefs.STLINK_DEVICE_NOT_OPENED;
  }

  if (expectedMode != undefined && Devices[deviceIndex].mode != expectedMode) {
    return apidefs.STLINK_WRONG_DEVICE_MODE;
  }
  return apidefs.STLINK_OK;
}

/**
 * Internal - Generates error string according to given error code.
 *
 * @param {number} errorCode - Value of the error code.
 * @returns {string} - The error string.
 */
function errorString(errorCode) {
  let errCodeStr = " (" + errorCode.toString(16) + ")";
  let errorStr = apicom.errorString(errorCode);

  if (errorStr == "") {
    errorStr = JTAG.errorString(errorCode);
  }

  return errorStr + errCodeStr;
}

/**
 * Internal - Gets vendor string according to given vendor Id.
 *
 * @param {number} vendorID - Id of vendor.
 * @returns {string} - Vendor string.
 */
function getVendorString(vendorID) {
  const VENDORS = {
    "0483": "ST-Microelectronics"
  };

  let rawStr = vendorID.toString(16).padStart(4, "0");

  if (VENDORS[rawStr] !== undefined) {
    return VENDORS[rawStr];
  }

  return rawStr;
}

/**
 * Internal - Gets product Id string according to given product ID.
 *
 * @param {number} productID - Id of the product.
 * @returns {string} - Product string.
 */
function getProductString(productID) {
  const PRODUCTS = {
    "3744": "ST-Link v1",
    "3748": "ST-Link v2",
    "374A": "ST-Link v2-1A",
    "374B": "ST-Link v2-1B",
    "374E": "ST-Link v3-E",
    "374F": "ST-Link v3"
  };

  let rawStr = productID
    .toString(16)
    .padStart(4, "0")
    .toUpperCase();

  if (PRODUCTS[rawStr] !== undefined) {
    return PRODUCTS[rawStr];
  }

  return rawStr;
}

/**
 * Internal - Gets Unique Id string according to given tab.
 *
 * @param {Array} UniqueIDTab - Array of char code composing Id.
 * @returns {string} - Unique Id string.
 */
function getUniqueIDString(UniqueIDTab) {
  let uniqueID = "";

  for (var i = 0; i < defsDrv.DRV_SERIAL_NUM_STR_MAX_LEN; i++) {
    if (UniqueIDTab[i] == 0) {
      break;
    }
    uniqueID = uniqueID + String.fromCharCode(UniqueIDTab[i]);
  }

  return uniqueID;
}

/**
 * Internal - Retrieves meode string from given mode.
 *
 * @param {number} mode - Mode Id.
 * @returns {string} - String mode or empty.
 */
function modeString(mode) {
  const map = {
    [defs.STLINK_DEV_DFU_MODE]: "DFU",
    [defs.STLINK_DEV_MASS_MODE]: "MASS/IDLE",
    [defs.STLINK_DEV_JTAG_MODE]: "JTAG",
    [defs.STLINK_DEV_SWIM_MODE]: "SWIM",
    [defs.STLINK_DEV_BTLD_MODE]: "BTLD"
  };
  return map[mode] || "";
}

/**
 * Gets the versions of specified device for ST-Link v3.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the command and versions of ST-LINK, JTAG and SWIM.
 */
function getVersionExt(deviceIndex, typeConnect) {
  logger.trace("deviceIndex = ", deviceIndex);
  let check = checkOpenedDevice(deviceIndex);

  if (check != apidefs.STLINK_OK) {
    logger.warn("Device (", deviceIndex, ") not possible to get extended version = ", check);
    return { result: check };
  }

  let req = new TDeviceRequest();

  let resp = Buffer.alloc(12).fill(0);

  req.CDBLength = 2;
  req.CDBByte[0] = defs.STLINK_CMD_GET_VERSION_EXT;
  req.CDBByte[1] = 0x00;
  req.InputRequest = defsDrv.DRV_REQUEST_READ_EP1;
  req.Buffer = resp;
  req.BufferLength = resp.length;
  req.SenseLength = defsDrv.DRV_DEFAULT_SENSE_LEN;

  let result;
  if (typeConnect == "tcp") {
    result = stLinkDrv.STLink_SendCommandTcp(Devices[deviceIndex].Cookie, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  } else {
    result = stLinkDrv.STLink_SendCommand(Devices[deviceIndex].Handle, req.ref(), defsDrv.DRV_DEFAULT_TIMEOUT);
  }

  if (result != defsDrv.DRV_OK) {
    logger.error("Result = ", drvToAPI(result));
    return { result: drvToAPI(result) };
  }

  Devices[deviceIndex].STLinkVersion = resp[0];
  Devices[deviceIndex].SWIMVersion = resp[1];
  Devices[deviceIndex].JTAGVersion = resp[2];

  return {
    result: apidefs.STLINK_OK,
    STLinkVersion: Devices[deviceIndex].STLinkVersion,
    JTAGVersion: Devices[deviceIndex].JTAGVersion,
    SWIMVersion: Devices[deviceIndex].SWIMVersion
  };
}

// JTAG low-level interface routing from "jtagcommands.js"
// added device index verification + passing device descriptor as parameter

/**
 * Route enter JTAG command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function enterJTAG(deviceIndex, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    logger.warn("Device index (", deviceIndex, ") does not exist");
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.enterJTAG(Devices[deviceIndex], typeConnect);
}

/**
 * Route enter SWD command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function enterSWD(deviceIndex, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    logger.warn("Device index (", deviceIndex, ") does not exist");
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.enterSWD(Devices[deviceIndex], typeConnect);
}

/**
 * Route exit JTAG command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function exitJTAG(deviceIndex, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    logger.warn("Device index (", deviceIndex, ") does not exist");
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.exitJTAG(Devices[deviceIndex], typeConnect);
}

/**
 * Route Read ID codes command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_ReadIdCodes(deviceIndex, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    logger.warn("Device index (", deviceIndex, ") does not exist");
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.JTAG_ReadIdCodes(Devices[deviceIndex], typeConnect);
}

/**
 * Route read memory 32 bits command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {number} address - Address of data to read.
 * @param {number} size - Size to read.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_ReadMemory32Bit(deviceIndex, address, size, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    logger.warn("Device index (", deviceIndex, ") does not exist");
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.JTAG_ReadMemory32Bit(Devices[deviceIndex], address, size, typeConnect);
}

/**
 * Route write memory 32 bits command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {number} address - Address of data to write.
 * @param {number} size - Size to write.
 * @param {*} buffer - Data to be written.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_WriteMemory32Bit(deviceIndex, address, size, buffer, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.JTAG_WriteMemory32Bit(Devices[deviceIndex], address, size, buffer, typeConnect);
}

/**
 * Route write memory 8 bits command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {number} address - Address of data to write.
 * @param {number} size - Size to write.
 * @param {*} buffer - Data to be written.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_WriteMemory8Bit(deviceIndex, address, size, buffer, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.JTAG_WriteMemory8Bit(Devices[deviceIndex], address, size, buffer, typeConnect);
}

/**
 * Route read memory 16 bits command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {number} address - Address of data to read.
 * @param {number} size - Size to read.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_ReadMemory16Bit(deviceIndex, address, size, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    logger.warn("Device index (", deviceIndex, ") does not exist");
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.JTAG_ReadMemory16Bit(Devices[deviceIndex], address, size, typeConnect);
}

/**
 * Route read memory 8 bits command to selected device.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {number} address - Address of data to read.
 * @param {number} size - Size to read.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_ReadMemory8Bit(deviceIndex, address, size, typeConnect) {
  if (deviceIndex > nbDevices - 1) {
    logger.warn("Device index (", deviceIndex, ") does not exist");
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.JTAG_ReadMemory8Bit(Devices[deviceIndex], address, size, typeConnect);
}

// Functions exported for Acquisition Services module
// ------------------------------------------------------------------------------------------

/**
 * Get the list of connected devices.
 *
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {Array<object>} Array of devices.
 */
function probeDiscovery(typeConnect) {
  logger.debug("Launch enumerate");
  let enum_res = reenumerate(typeConnect); // nbDevices updated here

  if (enum_res != apidefs.STLINK_OK) {
    logger.info("No board discovered, result = ", enum_res);

    return [];
  }

  // reset list of probes communicated to upper layers
  listProbe = [];

  for (let i = 0; i < nbDevices; i++) {
    // Get probe general information
    logger.debug("i value, typeConnect value", i, typeConnect);
    getDeviceInfo(i, typeConnect);
  }

  for (let i = 0; i < nbDevices; i++) {
    if (Devices[i] == null) {
      logger.warn("Device index ", i, " no information");
      break;
    }
    // Open probe
    let open_res;
    if (typeConnect == "tcp") {
      open_res = openDeviceTcp(i);
    } else {
      open_res = openDevice(i);
    }

    if (open_res != apidefs.STLINK_OK) {
      logger.warn("Open device failed for device index ", i, " result = ", open_res);
      break;
    }
    // Check current probe mode, if DFU then use exitDFU command to have this probe go back in Idle mode
    let mode_res = getCurrentMode(i, typeConnect);

    if (mode_res.result != apidefs.STLINK_OK) {
      logger.warn("Get current mode failed for device index ", i, " result = ", mode_res.result);
      break;
    }

    // On V3 Should never return this valus
    if (mode_res.mode == defs.STLINK_DEV_DFU_MODE) {
      let dfu_res = exitDFUMode(i, typeConnect);

      if (dfu_res != apidefs.STLINK_OK) {
        logger.warn("Exit DFU mode failed for device index ", i, " result = ", dfu_res);
        break;
      }
    }

    // Get detailed probe FW versions and update probe protocol support information
    let vers_res = getVersion(i, typeConnect);

    if (vers_res.result == apidefs.STLINK_OK) {
      let shortEnumUniqueId = Devices[i].EnumUniqueId;
      if (Devices[i].EnumUniqueId.length > 5) {
        shortEnumUniqueId = Devices[i].EnumUniqueId.substring(Devices[i].EnumUniqueId.length - 5);
      }
      const probe = {
        probeName: Devices[i].ProductId + " (..." + shortEnumUniqueId + ")",
        probeId: Devices[i].EnumUniqueId,
        property: {},
        version: { STLinkVersion: Devices[i].STLinkVersion, JTAGVersion: Devices[i].JTAGVersion }
      };

      let supported_protocols = 0;

      if (vers_res.JTAGVersion >= 1 && vers_res.STLinkVersion >= 3) {
        getTargetFrequencies(Devices[i], typeConnect);

        Devices[i].SWD = JTAG.v3getSWDFrequencies();
        Devices[i].JTAGVersion = vers_res.JTAGVersion;
        probe.property.SWD = Devices[i].SWD;

        Devices[i].JTAG = JTAG.v3getJTAGFrequencies();
        probe.property.JTAG = Devices[i].JTAG;
        supported_protocols += 2;
      } else if (vers_res.JTAGVersion >= defs.STLINK_JTAG_MIN_FW_VERSION) {
        Devices[i].SWD = JTAG.getSWDFrequencies();
        Devices[i].JTAGVersion = vers_res.JTAGVersion;
        probe.property.SWD = Devices[i].SWD;

        Devices[i].JTAG = JTAG.getJTAGFrequencies();
        probe.property.JTAG = Devices[i].JTAG;

        supported_protocols += 2;
      }

      // add this probe in the list only if at least one protocol is supported
      if (supported_protocols > 0) {
        listProbe.push(probe);
      }
    } else {
      logger.warn("Get version failed for device index ", i, " result = ", vers_res.result);
    }
    if (typeConnect == "tcp") {
      closeDeviceTcp(i);
    } else {
      closeDevice(i);
    }
  }

  logger.debug("List of Probe : ");
  listProbe.forEach(function(elt) {
    logger.debug("List of probe element : ", elt);
  });
  return listProbe;
}

/**
 * Gets supported frequencies of selected device.
 *
 * @param {object} device - The device descriptor.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 */
function getTargetFrequencies(device, typeConnect) {
  let cmd_result = JTAG.JTAG_GetTargetFrequencies(device, apidefs.STLINK_JTAG_PROTOCOL, typeConnect);
  if (cmd_result != apidefs.STLINK_OK) {
    logger.error("JTAG_GetComFrequency error = ", cmd_result);
  }
  cmd_result = JTAG.JTAG_GetTargetFrequencies(device, apidefs.STLINK_SWD_PROTOCOL, typeConnect);
  if (cmd_result != apidefs.STLINK_OK) {
    logger.error("JTAG_GetComFrequency error = ", cmd_result);
  }
}

/**
 * Connects to selected device.
 *
 * @param {number} probeId - The device unique Id.
 * @param {string} protocol - Protocol to use iather SWIM, JTAG or SWD.
 * @param {string} frequency - The frequency to set.
 * @param {boolean} rediscovery - Either or not a discovery procedure is requested.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} - Result of the commands.
 */
function connectToTarget(probeId, protocol, frequency, rediscovery, typeConnect) {
  logger.debug(
    "Arguments : probeId = ",
    probeId,
    ", protocol = ",
    protocol,
    ", frequency = ",
    frequency,
    ", rediscovery = ",
    rediscovery,
    ",typeConnect = ",
    typeConnect
  );
  let res = {
    result: apidefs.STLINK_BAD_PARAMETER
  };

  // probeId is the probe enum unique ID
  if (probeId == undefined || probeId == "") {
    logger.error("Undefined probeId");
    return res;
  }

  if (
    protocol != apidefs.STLINK_SWD_PROTOCOL &&
    protocol != apidefs.STLINK_JTAG_PROTOCOL &&
    protocol != apidefs.STLINK_SWIM_PROTOCOL
  ) {
    logger.error("Unknown protocol : ", protocol);
    return res;
  }

  let freq;
  if (frequency == undefined || frequency == "") {
    freq = "default";
  } else {
    freq = frequency;
  }

  // if requested explicitly,
  // or if no device has been ever found, perform discovery
  if (rediscovery === true || Devices.length == 0) {
    probeDiscovery(typeConnect);
  }

  let probeIndex = Devices.findIndex(function(elem) {
    return elem.EnumUniqueId == probeId;
  });

  if (probeIndex == -1) {
    logger.error("No device for the probeId ", probeId);
    res.result = apidefs.STLINK_NO_DEVICE;
    return res;
  }

  let open_res;

  if (typeConnect == "tcp") {
    open_res = openDeviceTcp(probeIndex);
  } else {
    open_res = openDevice(probeIndex);
  }

  if (open_res != apidefs.STLINK_OK) {
    logger.error("Device not opened on probeIndex ", probeIndex);
    res.result = apidefs.STLINK_DEVICE_NOT_OPENED;
    return res;
  }

  //set access point id to default value
  Devices[probeIndex].accessPointId = undefined;

  // change frequency according to protocol
  let cmd_result;
  let idcode;

  switch (protocol) {
    case apidefs.STLINK_SWD_PROTOCOL: {
      if (Devices[probeIndex].STLinkVersion >= 3) {
        cmd_result = JTAG.JTAG_SetComFrequency(Devices[probeIndex], apidefs.STLINK_SWD_PROTOCOL, freq, typeConnect);
        if (cmd_result != apidefs.STLINK_OK) {
          logger.error("JTAG_SetComFrequency error = ", cmd_result);
          break;
        }
      } else {
        cmd_result = JTAG.JTAG_SetSWDFrequency(Devices[probeIndex], freq, typeConnect);
        if (cmd_result != apidefs.STLINK_OK) {
          logger.error("JTAG_SetSWDFrequency error = ", cmd_result);
          break;
        }
      }

      // change mode into SWD
      cmd_result = JTAG.enterSWD(Devices[probeIndex], typeConnect);
      if (cmd_result != apidefs.STLINK_OK) {
        logger.error("enterSWD error = ", cmd_result);
        break;
      }

      // check mode is SWD/JTAG
      let getmode = getCurrentMode(probeIndex, typeConnect);

      if (getmode.result != apidefs.STLINK_OK) {
        cmd_result = getmode.result;

        logger.error("getCurrentMode result error = ", cmd_result);
        break;
      }

      if (getmode.mode != defs.STLINK_DEV_JTAG_MODE) {
        cmd_result = apidefs.STLINK_WRONG_DEVICE_MODE;
        logger.error("getCurrentMode mode error = ", cmd_result);
        break;
      }

      // read ID codes and return
      let ids = JTAG.JTAG_ReadIdCodes(Devices[probeIndex], typeConnect);
      cmd_result = ids.result;
      if (cmd_result == apidefs.STLINK_OK) {
        idcode = ids.MCUID;
        break;
      }

      break;
    }

    case apidefs.STLINK_JTAG_PROTOCOL: {
      if (Devices[probeIndex].STLinkVersion >= 3) {
        cmd_result = JTAG.JTAG_SetComFrequency(Devices[probeIndex], apidefs.STLINK_JTAG_PROTOCOL, freq, typeConnect);
        if (cmd_result != apidefs.STLINK_OK) {
          logger.error("JTAG_SetComFrequency error = ", cmd_result);
          break;
        }
      } else {
        cmd_result = JTAG.JTAG_SetJTAGFrequency(Devices[probeIndex], freq, typeConnect);
        if (cmd_result != apidefs.STLINK_OK) {
          logger.error("JTAG_SetJTAGFrequency error = ", cmd_result);
          break;
        }
      }

      // change mode into JTAG
      cmd_result = JTAG.enterJTAG(Devices[probeIndex], typeConnect);
      if (cmd_result != apidefs.STLINK_OK) {
        logger.error("enterJTAG error = ", cmd_result);
        break;
      }

      // check mode is SWD/JTAG
      let getmode = getCurrentMode(probeIndex, typeConnect);
      if (getmode.result != apidefs.STLINK_OK) {
        cmd_result = getmode.result;
        logger.error("getCurrentMode result error = ", cmd_result);
        break;
      }

      if (getmode.mode != defs.STLINK_DEV_JTAG_MODE) {
        cmd_result = apidefs.STLINK_WRONG_DEVICE_MODE;
        logger.error("getCurrentMode mode error = ", cmd_result);
        break;
      }

      // read ID codes and return
      let ids = JTAG.JTAG_ReadIdCodes(Devices[probeIndex], typeConnect);
      cmd_result = ids.result;
      if (cmd_result == apidefs.STLINK_OK) {
        idcode = ids.MCUID;
        break;
      }

      break;
    }

    case apidefs.STLINK_SWIM_PROTOCOL: {
      // TODO
      cmd_result = apidefs.STLINK_ERR;
      logger.error("SWIM error = ", cmd_result);
      break;
    }

    default:
      cmd_result = apidefs.STLINK_ERR;
  }

  res.result = cmd_result;

  if (cmd_result != apidefs.STLINK_OK) {
    logger.error("Error on connect device then close probe ", probeId);

    if (typeConnect == "tcp") {
      closeDeviceTcp(probeIndex);
    } else {
      closeDevice(probeIndex);
    }
  } else {
    res.MCUID = idcode;
    res.MCUName = JTAG.getSTM32MCUString(idcode);
  }
  logger.debug("Result = ", res);
  return res;
}

/**
 * Service to properly disconnect each probe from its target,
 * and close each probe connection.
 *
 * @param {*} probeId - The device unique Id or "" to disconnect all.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {object} Operation result code.
 */
function closeProbes(probeId, typeConnect) {
  logger.debug("Argument : probeId = ", probeId);
  let start, stop;
  let res = { result: apidefs.STLINK_OK };
  // probeId is the probe enum unique ID
  if (probeId == undefined || probeId == "") {
    start = 0;
    stop = nbDevices;
  } else {
    let probeIndex = Devices.findIndex(function(elem) {
      return elem.EnumUniqueId == probeId;
    });

    if (probeIndex != -1) {
      start = probeIndex;
      stop = start + 1;
    } else {
      res.result = apidefs.STLINK_NO_DEVICE;
      logger.error("probe does not exist : ", probeId);
      return res;
    }
  }
  for (let i = start; i < stop; i++) {
    let mode = getCurrentMode(i, typeConnect).mode;

    // exit JTAG/SWD mode
    if (mode == defs.STLINK_DEV_JTAG_MODE) {
      exitJTAG(i, typeConnect);
    }

    // exit SWIM mode
    if (mode == defs.STLINK_DEV_SWIM_MODE) {
      //TODO exitSWIM(i);
    }

    if (checkOpenedDevice(i) == apidefs.STLINK_OK) {
      if (typeConnect == "tcp") {
        res.result = closeDeviceTcp(i);
      } else {
        res.result = closeDevice(i);
      }
    }
  }
  logger.debug("End of closing");
  return res;
}

/**
 * Service to read data from a target memory area, specifying the type of data to be read.
 *
 * @param {number} probeId - The unique If of th eprobe to be used to communicate with the target MCU.
 * @param {number} baseAddress - The base address to read from in the target memory space
 *                    (as a 0x-started string in hexadecimal representation).
 * @param {number} nbElem - The number of elements of the specified type to be read.
 * @param {number} type - The read type : u8-s8 - address and size may have any value
 *                             u16-s16 - address and size must be multiples of 2
 *                             u32-s32-float - address and size must be multiples of 4
 *                             u64-s64-double - address and size must be multiples of 8.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {object} Operation result code, and if result is STLINK_OK, returns also an array of duplets
 *          objects with x for timestamp and y for value according to requested type.
 */
function readMemoryTyped(probeId, baseAddress, nbElem, type, accessPointId, typeConnect) {
  logger.debug(
    "Arguments  : probeId = ",
    probeId,
    ", baseAddress = ",
    baseAddress,
    ", nbElem = ",
    nbElem,
    ", type = ",
    type
  );
  let res = {
    result: apidefs.STLINK_BAD_PARAMETER
  };

  let readType;

  // probeId is the probe enum unique ID, it is mandatory
  if (probeId == undefined || probeId == "") {
    logger.error("Probe undefined");
    return res;
  }

  // check that this probe is known
  let probeIndex = Devices.findIndex(function(elem) {
    return elem.EnumUniqueId == probeId;
  });

  if (probeIndex == -1) {
    res.result = apidefs.STLINK_NO_DEVICE;
    logger.error("No index for probeId ", probeId);
    return res;
  }

  //check that this probe is opened
  if (Devices[probeIndex].state != "opened") {
    res.result = apidefs.STLINK_DEVICE_NOT_OPENED;
    logger.error("Device with probeId ", probeId, " is not opened");
    return res;
  }

  // check that the mode is correct if it is provided
  if (type === undefined) {
    logger.error("type of read not defined");
    return res;
  }

  if (
    type == apidefs.READ_TYPE_U8 ||
    type == apidefs.READ_TYPE_S8 ||
    type == apidefs.READ_TYPE_U16 ||
    type == apidefs.READ_TYPE_S16 ||
    type == apidefs.READ_TYPE_U32 ||
    type == apidefs.READ_TYPE_S32 ||
    type == apidefs.READ_TYPE_U64 ||
    type == apidefs.READ_TYPE_S64 ||
    type == apidefs.READ_TYPE_FLOAT ||
    type == apidefs.READ_TYPE_DOUBLE
  ) {
    readType = type;
  } else {
    logger.error("type of read not allowed : ", type);
    return res;
  }

  // check that base address is a valid hexadecimal string
  let add = parseInt(baseAddress, 16);
  if (add === undefined) {
    logger.error("Read address not defined");
    return res;
  }

  let remainingBytes;
  let elemSize;
  let data = [];

  // get the timestamp for the next memory read operation
  let timestamp = +new Date();

  let initAPRes = initAccessPoint(probeIndex, accessPointId, typeConnect);
  if (initAPRes != apidefs.STLINK_OK) {
    res.result = initAPRes;
    logger.error("Access point", accessPointId, " could not be initialized");
    return res;
  }

  switch (readType) {
    // read from target an array of bytes
    case apidefs.READ_TYPE_U8:
    case apidefs.READ_TYPE_S8: {
      elemSize = 1;
      remainingBytes = nbElem * elemSize;

      let read8 = JTAG.JTAG_ReadMemory8Bit(Devices[probeIndex], add, remainingBytes, typeConnect);

      if (read8.result != apidefs.STLINK_OK) {
        logger.error("JTAG_ReadMemory8Bit failed, error is ", read8.result);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: read8.result };
      }

      if (readType == apidefs.READ_TYPE_U8) {
        for (let i = 0; i < remainingBytes; i++) {
          data.push({
            y: read8.data.readUInt8(i),
            x: timestamp
          });
        }
      }

      if (readType == apidefs.READ_TYPE_S8) {
        for (let i = 0; i < remainingBytes; i++) {
          data.push({
            y: read8.data.readInt8(i),
            x: timestamp
          });
        }
      }

      break;
    }

    // read from target an array of 16-bit values
    case apidefs.READ_TYPE_U16:
    case apidefs.READ_TYPE_S16: {
      elemSize = 2;
      remainingBytes = nbElem * elemSize;

      let read16;

      // check if 16-bit access are supported by this version of ST-Link; if not, do 8-bit access
      if (
        Devices[probeIndex].JTAGVersion < defs.STLINK_JTAG_MIN_FW_VERSION_FOR_16BIT_ACCESSES ||
        (Devices[probeIndex].JTAGVersion == 0 && Devices[probeIndex].STLinkVersion >= 3)
      ) {
        // size and address must be 2-bytes aligned (16-bit access)
        // check this in case of 8-bit access if 16-bit access not available
        if (remainingBytes % 2 != 0 || add % 2 != 0) {
          logger.error("size and address is not 2-bytes aligned (16-bit access). Address ", add);
          closeAccessPoint(probeIndex, accessPointId, typeConnect);
          return res;
        }

        read16 = JTAG.JTAG_ReadMemory8Bit(Devices[probeIndex], add, remainingBytes, typeConnect);
      } else {
        read16 = JTAG.JTAG_ReadMemory16Bit(Devices[probeIndex], add, remainingBytes, typeConnect);
      }

      if (read16.result != apidefs.STLINK_OK) {
        logger.error("JTAG_ReadMemory16Bit failed, error is ", read16.result);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: read16.result };
      }

      if (readType == apidefs.READ_TYPE_U16) {
        for (let i = 0; i < remainingBytes; i += 2) {
          data.push({
            y: read16.data.readUInt16LE(i),
            x: timestamp
          });
        }
      }

      if (readType == apidefs.READ_TYPE_S16) {
        for (let i = 0; i < remainingBytes; i += 2) {
          data.push({
            y: read16.data.readInt16LE(i),
            x: timestamp
          });
        }
      }

      break;
    }

    // read from target an array of 32-bit values
    case apidefs.READ_TYPE_U32:
    case apidefs.READ_TYPE_S32:
    case apidefs.READ_TYPE_FLOAT: {
      elemSize = 4;
      remainingBytes = nbElem * elemSize;

      let read32 = JTAG.JTAG_ReadMemory32Bit(Devices[probeIndex], add, remainingBytes, typeConnect);

      if (read32.result != apidefs.STLINK_OK) {
        logger.error("JTAG_ReadMemory32Bit failed, error is ", read32.result);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: read32.result };
      }

      if (readType == apidefs.READ_TYPE_U32) {
        for (let i = 0; i < remainingBytes; i += 4) {
          data.push({
            y: read32.data.readUInt32LE(i),
            x: timestamp
          });
        }
      }

      if (readType == apidefs.READ_TYPE_S32) {
        for (let i = 0; i < remainingBytes; i += 4) {
          data.push({
            y: read32.data.readInt32LE(i),
            x: timestamp
          });
        }
      }

      if (readType == apidefs.READ_TYPE_FLOAT) {
        for (let i = 0; i < remainingBytes; i += 4) {
          data.push({
            y: read32.data.readFloatLE(i),
            x: timestamp
          });
        }
      }

      break;
    }

    case apidefs.READ_TYPE_U64:
    case apidefs.READ_TYPE_S64:
    case apidefs.READ_TYPE_DOUBLE: {
      elemSize = 8;
      remainingBytes = nbElem * elemSize;

      // address must be 8-bytes aligned (64-bit data)
      if (add % 8 != 0) {
        logger.error("address is not 8-bytes aligned (64-bit access). Address ", add);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return res;
      }

      let read64 = JTAG.JTAG_ReadMemory32Bit(Devices[probeIndex], add, remainingBytes, typeConnect);

      if (read64.result != apidefs.STLINK_OK) {
        logger.error("Read 64 bits failed, error is ", read64.result);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: read64.result };
      }

      if (readType == apidefs.READ_TYPE_U64) {
        for (let i = 0; i < remainingBytes; i += 8) {
          // TODO use BigInt for accurate 64-bit value !!

          data.push({
            y: read64.data.readUInt32LE(i + 4) * Math.pow(2, 32) + read64.data.readUInt32LE(i),
            x: timestamp
          });
        }
      }

      if (readType == apidefs.READ_TYPE_S64) {
        for (let i = 0; i < remainingBytes; i += 8) {
          // TODO use BigInt for accurate 64-bit value !!
          data.push({
            y: read64.data.readInt32LE(i + 4) * Math.pow(2, 32) + read64.data.readInt32LE(i),
            x: timestamp
          });
        }
      }

      if (readType == apidefs.READ_TYPE_DOUBLE) {
        for (let i = 0; i < remainingBytes; i += 8) {
          data.push({
            y: read64.data.readDoubleLE(i),
            x: timestamp
          });
        }
      }

      break;
    }

    default: {
      logger.error("readType unknown : ", readType);
      closeAccessPoint(probeIndex, accessPointId, typeConnect);
      return res;
    }
  }
  res.result = apidefs.STLINK_OK;
  res.data = data;
  res.elemSize = elemSize;
  logger.debug("Result = ", res);
  return res;
}

/**
 * Service to read data from a target memory area, specifying a memory access mode.
 *
 * @param {number} probeId - The unique If of the probe to be used to communicate with the target MCU.
 * @param {number} baseAddress - The base address to read from in the target memory space
 *                    (as a 0x-started string in hexadecimal representation).
 * @param {number} sizeInBytes - The size in bytes of the memory area to read.
 * @param {number} mode - The read mode : 8-bit address and size may have any value, but read operation
 *                                      is 8-bit only (slower)
 *                             16-bit - address and size must be multiples of 2, but read operation is faster
 *                             32-bit - address and size must be multiples of 4, but read operation is fastest
 *                             automatic (default) - address and size may have any value, and an optimized
 *                                                   mix of 8-bit and 32-bit operations is performed.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {object} - Operation result code, and if result is STLINK_OK, returns also an array of data bytes
 *                     and a timestamp.
 */
function readMemoryMode(probeId, baseAddress, sizeInBytes, mode = apidefs.READ_MODE_AUTO, accessPointId, typeConnect) {
  logger.debug(
    "Arguments : probeId = ",
    probeId,
    ", baseAddress = ",
    baseAddress,
    ", sizeInBytes = ",
    sizeInBytes,
    ", mode = ",
    mode
  );
  let res = {
    result: apidefs.STLINK_BAD_PARAMETER
  };

  let readMode;

  // probeId is the probe enum unique ID, it is mandatory
  if (probeId == undefined || probeId == "") {
    logger.error("probeId undefined");
    return res;
  }

  // check that this probe is known
  let probeIndex = Devices.findIndex(function(elem) {
    return elem.EnumUniqueId == probeId;
  });

  if (probeIndex == -1) {
    res.result = apidefs.STLINK_NO_DEVICE;
    logger.error("no device for probeId ", probeId);
    return res;
  }

  //check that this probe is opened
  if (Devices[probeIndex].state != "opened") {
    res.result = apidefs.STLINK_DEVICE_NOT_OPENED;
    logger.error("device not opened for probeId ", probeId, " wih index ", probeIndex);
    return res;
  }

  // check that the mode is correct if it is provided
  if (
    mode == apidefs.READ_MODE_8_BIT ||
    mode == apidefs.READ_MODE_16_BIT ||
    mode == apidefs.READ_MODE_32_BIT ||
    mode == apidefs.READ_MODE_AUTO
  ) {
    readMode = mode;
  } else {
    logger.error("Unknown readMode ", readMode);
    return res;
  }

  // check that base address is a valid hexadecimal string
  let add = parseInt(baseAddress, 16);
  if (add === undefined) {
    logger.error("Address undefined");
    return res;
  }

  let initAPRes = initAccessPoint(probeIndex, accessPointId, typeConnect);
  if (initAPRes != apidefs.STLINK_OK) {
    res.result = initAPRes;
    logger.error("Access point", accessPointId, " could not be initialized");
    return res;
  }

  let remainingBytes = sizeInBytes;
  let bytesToRead;
  let data = [];

  // get the timestamp for the next memory read operation
  let timestamp = +new Date();

  switch (readMode) {
    // read from target an array of bytes
    case apidefs.READ_MODE_8_BIT: {
      let read8 = JTAG.JTAG_ReadMemory8Bit(Devices[probeIndex], add, remainingBytes, typeConnect);

      if (read8.result != apidefs.STLINK_OK) {
        logger.error("JTAG_ReadMemory8Bit read error ", read8.result);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: read8.result };
      }

      // copy read8.data into data array
      data.push(...read8.data);

      break;
    }

    // read from target an array of 16-bit values
    case apidefs.READ_MODE_16_BIT: {
      // if 16-bit access mode is not supported by the ST-Link FW, return an error
      if (
        Devices[probeIndex].JTAGVersion < defs.STLINK_JTAG_MIN_FW_VERSION_FOR_16BIT_ACCESSES ||
        (Devices[probeIndex].JTAGVersion == 0 && Devices[probeIndex].STLinkVersion >= 3)
      ) {
        logger.error("JTAG Read Memory 16Bit not supported ");
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: apidefs.STLINK_JTAG_16BIT_ACCESS_NOT_SUPPORTED };
      }
      let read16 = JTAG.JTAG_ReadMemory16Bit(Devices[probeIndex], add, remainingBytes, typeConnect);

      if (read16.result != apidefs.STLINK_OK) {
        logger.error("JTAG_ReadMemory16Bit read error ", read16.result);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: read16.result };
      }

      // copy read16.data into data array
      data.push(...read16.data);
      break;
    }

    // read from target an array of 32-bit values
    case apidefs.READ_MODE_32_BIT: {
      let read32 = JTAG.JTAG_ReadMemory32Bit(Devices[probeIndex], add, remainingBytes, typeConnect);

      if (read32.result != apidefs.STLINK_OK) {
        logger.error("JTAG_ReadMemory32Bit read error ", read32.result);
        closeAccessPoint(probeIndex, accessPointId, typeConnect);
        return { result: read32.result };
      }

      // copy read32.data into data array
      data.push(...read32.data);

      break;
    }

    // read from target an array of bytes, automatically optimizing read operations
    case apidefs.READ_MODE_AUTO: {
      let addressToRead = add;

      // if address is not 32 bits aligned
      if (addressToRead % 4 != 0) {
        // if start address is odd, 16 bit access not supported or only one byte to read
        // one 8-bit read operation to read out max 3 bytes to align the next bytes on 32-bit boundary
        if (
          addressToRead & (0x01 == 1) ||
          (addressToRead % 4 == 2 &&
            (Devices[probeIndex].JTAGVersion < defs.STLINK_JTAG_MIN_FW_VERSION_FOR_16BIT_ACCESSES ||
              (Devices[probeIndex].JTAGVersion == 0 && Devices[probeIndex].STLinkVersion >= 3) ||
              remainingBytes == 1))
        ) {
          if (remainingBytes < 4 - (addressToRead % 4)) {
            bytesToRead = remainingBytes;
          } else {
            bytesToRead = 4 - (addressToRead % 4);
          }

          let read8 = JTAG.JTAG_ReadMemory8Bit(Devices[probeIndex], addressToRead, bytesToRead, typeConnect);

          if (read8.result != apidefs.STLINK_OK) {
            logger.error("JTAG_ReadMemory8Bit read error ", read8.result);
            closeAccessPoint(probeIndex, accessPointId, typeConnect);
            return { result: read8.result };
          }

          // copy read8.data into data array
          data.push(...read8.data);
        } else if (addressToRead % 4 == 2 && remainingBytes > 1) {
          //one 16-bit read operation to align the next bytes on 32-bit boundary
          bytesToRead = 2;
          let read16 = JTAG.JTAG_ReadMemory16Bit(Devices[probeIndex], addressToRead, bytesToRead, typeConnect);

          if (read16.result != apidefs.STLINK_OK) {
            logger.error("JTAG_ReadMemory16Bit read error ", read16.result);
            closeAccessPoint(probeIndex, accessPointId, typeConnect);
            return { result: read16.result };
          }

          // copy read16.data into data array
          data.push(...read16.data);
        }

        // update remaining bytes counter and address to read from
        remainingBytes = remainingBytes - bytesToRead;
        addressToRead = addressToRead + bytesToRead;
      }
      // then read as many 32-bit words as possible
      bytesToRead = Math.floor(remainingBytes / 4) * 4;

      // TODO : implement the JTAG FW auto-increment functionality which works on 1K pages ?? see DbgDev

      if (bytesToRead > 0) {
        // read requested bytes in 32-bit mode
        let read32 = JTAG.JTAG_ReadMemory32Bit(Devices[probeIndex], addressToRead, bytesToRead, typeConnect);

        if (read32.result != apidefs.STLINK_OK) {
          logger.error("JTAG_ReadMemory32Bit read error ", read32.result);
          closeAccessPoint(probeIndex, accessPointId, typeConnect);
          return { result: read32.result };
        }

        // copy read32.data into data array
        data.push(...read32.data);

        // update remaining bytes counter
        remainingBytes = remainingBytes - bytesToRead;
        addressToRead = addressToRead + bytesToRead;
      }

      // Now read remaining bytes (if any)
      // first if 16bit access is supported and it remains at least 2 bytes, read them
      if (
        remainingBytes >= 2 &&
        !(
          Devices[probeIndex].JTAGVersion < defs.STLINK_JTAG_MIN_FW_VERSION_FOR_16BIT_ACCESSES ||
          (Devices[probeIndex].JTAGVersion == 0 && Devices[probeIndex].STLinkVersion >= 3)
        )
      ) {
        bytesToRead = 2;
        let read16 = JTAG.JTAG_ReadMemory16Bit(Devices[probeIndex], addressToRead, bytesToRead, typeConnect);

        if (read16.result != apidefs.STLINK_OK) {
          logger.error("JTAG_ReadMemory16Bit read error ", read16.result);
          closeAccessPoint(probeIndex, accessPointId, typeConnect);
          return { result: read16.result };
        }

        // copy read16.data into data array
        data.push(...read16.data);

        // update remaining bytes counter
        remainingBytes = remainingBytes - bytesToRead;
        addressToRead = addressToRead + bytesToRead;
      }

      // at last, use one 8-bit read operation to read out remaining bytes that are not 32-bit aligned (max 3)
      if (remainingBytes > 0) {
        let read8 = JTAG.JTAG_ReadMemory8Bit(Devices[probeIndex], addressToRead, remainingBytes, typeConnect);

        if (read8.result != apidefs.STLINK_OK) {
          logger.error("JTAG_ReadMemory8Bit read error ", read8.result);
          closeAccessPoint(probeIndex, accessPointId, typeConnect);
          return { result: read8.result };
        }

        // copy read8.data into data array
        data.push(...read8.data);
      }

      break;
    }

    default: {
      logger.error("Read mode unknoown : ", readMode);
      closeAccessPoint(probeIndex, accessPointId, typeConnect);
      return res;
    }
  }

  res.result = apidefs.STLINK_OK;
  res.data = data;
  res.time = timestamp;

  logger.debug("res = ", res);
  return res;
}

/**
 * This function is used to compute the size of a variable item inside a group of variables.
 *
 * @param {number} type - Type of variable.
 * @param {number} nb - Number of items to be considered.
 * @returns {number} Result of the computation.
 */
function computeVarSize(type, nb) {
  let res;
  switch (type) {
    case apidefs.READ_TYPE_U8:
    case apidefs.READ_TYPE_S8:
      res = 1 * nb;
      return res;

    case apidefs.READ_TYPE_U16:
    case apidefs.READ_TYPE_S16:
      res = 2 * nb;
      return res;
    case apidefs.READ_TYPE_U32:
    case apidefs.READ_TYPE_S32:
    case apidefs.READ_TYPE_FLOAT:
      res = 4 * nb;
      return res;
    case apidefs.READ_TYPE_U64:
    case apidefs.READ_TYPE_S64:
    case apidefs.READ_TYPE_DOUBLE:
      return (res = 8 * nb);
    default:
      res = undefined;
      return res;
  }
}

/**
 *  Service to read a group of data from a target memory area.
 *  This is optimized so as to use the minimum number of read operations,
 *  grouping if possible several variables into a single "frame" reading,
 *  based on the max frame size (depending on HW constraints).
 *
 *  @param {string} probeId - The probe to be used to communicate with the target MCU.
 *  @param {Array<object>} group - The group of variables to be read to the target memory
 *              this is an array, containing objects each describing one variable with
 *              its name, its address (as a 0x-started string in hexadecimal representation),
 *              its type (from READ TYPE numeric constants), and the number of items to be read.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {object} Operation result code, and if result is STLINK_OK, returns also an array
 *          containing one array of {x,y} duplets objects for each variable from the group parameter;
 *          x contains timestamp of read operation, y contains the value according to the variable type
 *          variable indexes in this array are the same as in the group parameter.
 */
function readMemoryGroup(probeId, group, accessPointId = 0, typeConnect) {
  let res = {
    result: apidefs.STLINK_BAD_PARAMETER
  };

  // probeId is the probe enum unique ID, it is mandatory
  if (probeId == undefined || probeId == "") {
    logger.error("probeId undefined");
    return res;
  }

  // check that this probe is known
  let probeIndex = Devices.findIndex(function(elem) {
    return elem.EnumUniqueId == probeId;
  });

  if (probeIndex == -1) {
    res.result = apidefs.STLINK_NO_DEVICE;
    logger.error("no device for probeId ", probeId);
    return res;
  }

  //check that this probe is opened
  if (Devices[probeIndex].state != "opened") {
    res.result = apidefs.STLINK_DEVICE_NOT_OPENED;
    logger.error("device not opened for probeId ", probeId, " with index ", probeIndex);
    return res;
  }

  // check that the group is correct if it is provided
  if (group === undefined) {
    logger.error("group undefined");
    return res;
  }
  let arrFrame = [];

  for (let i = 0; i < group.length; i++) {
    // extract info from current property value and check it
    let varToRead = group[i];
    let buildNewFrame = true;

    if (varToRead.address == undefined || varToRead.type == undefined) {
      logger.error("address or type undefined");
      return res;
    }

    // test that address is a string
    if (typeof varToRead.address !== "string") {
      logger.error("address is not a string");
      return res;
    }

    // Test that address starts with "0x"
    if (varToRead.address.substring(0, 2) != "0x") {
      logger.error("address is not started with a 0x : ", varToRead.address);
      return res;
    }

    if (varToRead.nb != undefined && typeof varToRead.nb !== "number") {
      logger.error("The number is not correctly defined");
      return res;
    }

    // default number of elements is 1
    if (varToRead.nb == undefined) {
      varToRead.nb = 1;
    }

    let varSize = computeVarSize(varToRead.type, varToRead.nb);

    let varStartAddress = parseInt(varToRead.address, 16);
    let varEndAddress = varStartAddress + varSize;

    for (let j = 0; j < arrFrame.length; j++) {
      let frameStartAddress = arrFrame[j].start;
      let frameEndAddress = arrFrame[j].end;

      //variable included in frame
      if (varStartAddress > frameStartAddress && varEndAddress < frameEndAddress) {
        // add the indice of current variable into the array of indices of current frame
        arrFrame[j].indices.push(i);
        buildNewFrame = false;
        break;
      } else if (varStartAddress <= frameStartAddress && frameEndAddress - varStartAddress < MAX_FRAME_SIZE) {
        // add the indice of current variable into the array of indices of current frame
        arrFrame[j].indices.push(i);

        // update frame start address
        arrFrame[j].start = varStartAddress;
        buildNewFrame = false;
        break;
      } else if (varEndAddress >= frameEndAddress && varEndAddress - frameStartAddress < MAX_FRAME_SIZE) {
        // Variable could be added to the end of the frame

        // add the indice of current variable into the array of indices of current frame
        arrFrame[j].indices.push(i);

        // update frame end address
        arrFrame[j].end = varEndAddress;
        buildNewFrame = false;
        break;
      }
    }

    if (buildNewFrame) {
      var newframe = {};

      // the new frame contains only the current variable
      newframe.start = varStartAddress;
      newframe.end = varEndAddress;

      // init the array of indices with the indice of the current variable
      newframe.indices = [];
      newframe.indices.push(i);

      // add this new frame to the frame array
      arrFrame.push(newframe);
    }
  }

  for (let j = 0; j < arrFrame.length; j++) {
    logger.trace(
      " liste arrFrame[",
      j,
      "] : 0x",
      arrFrame[j].start.toString(16).padStart(8, "0"),
      "- 0x",
      arrFrame[j].end.toString(16).padStart(8, "0"),
      arrFrame[j].indices
    );
  }

  // result array
  // each array item contains itself an array of values corresponding to the indice of the variable into
  // the original group array
  let result = [];

  for (let i = 0; i < arrFrame.length; i++) {
    // get the timestamp for the next memory read operation
    let timestamp = +new Date();

    // read the current frame, result is a byte array

    let read = readMemoryMode(
      probeId,
      "0x" + arrFrame[i].start.toString(16).padStart(8, "0"),
      arrFrame[i].end - arrFrame[i].start,
      apidefs.READ_MODE_AUTO,
      accessPointId,
      typeConnect
    );

    if (read.result != apidefs.STLINK_OK) {
      res.result = read.result;
      logger.error("ReadMemoryMode failed ", res);
      return res;
    }

    // initialise a buffer from the returned bytes to be able to use the number extraction function (ex readUInt16LE)
    let buffer = Buffer.from(read.data);

    // here the variables contained in the frame have to be extracted by using their indices contained
    // in arrFrame[i].indices
    for (let j = 0; j < arrFrame[i].indices.length; j++) {
      let indiceInGroup = arrFrame[i].indices[j];

      let varInGroup = group[indiceInGroup];

      // offset of variable v into the frame is (group[v].address - arrFrame[i].start)
      let offsetInFrame = varInGroup.address - arrFrame[i].start;

      // abnormal case ==> error
      if (offsetInFrame < 0 || offsetInFrame > MAX_FRAME_SIZE) {
        logger.error("Offset in frame is out of range ", offsetInFrame);
        return res;
      }

      // each item of result is itself an array containing one or more values
      result[indiceInGroup] = [];

      // extract variable values from data read from target, according to variable type
      try {
        switch (varInGroup.type) {
          case apidefs.READ_TYPE_U8: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readUInt8(offsetInFrame + k),
                x: timestamp
              });
            }
            break;
          }

          case apidefs.READ_TYPE_S8: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readInt8(offsetInFrame + k),
                x: timestamp
              });
            }
            break;
          }

          case apidefs.READ_TYPE_U16: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readUInt16LE(offsetInFrame + k * 2),
                x: timestamp
              });
            }
            break;
          }

          case apidefs.READ_TYPE_S16: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readInt16LE(offsetInFrame + k * 2),
                x: timestamp
              });
            }
            break;
          }

          case apidefs.READ_TYPE_U32: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readUInt32LE(offsetInFrame + k * 4),
                x: timestamp
              });
            }
            break;
          }

          case apidefs.READ_TYPE_S32: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readInt32LE(offsetInFrame + k * 4),
                x: timestamp
              });
            }
            break;
          }
          case apidefs.READ_TYPE_U64: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y:
                  buffer.readUInt32LE(offsetInFrame + k * 8 + 4) * Math.pow(2, 32) +
                  buffer.readUInt32LE(offsetInFrame + k * 8),
                x: timestamp
              });
            }
            break;
          }

          case apidefs.READ_TYPE_S64: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y:
                  buffer.readInt32LE(offsetInFrame + k * 8 + 4) * Math.pow(2, 32) +
                  buffer.readInt32LE(offsetInFrame + k * 8),
                x: timestamp
              });
            }
            break;
          }
          case apidefs.READ_TYPE_FLOAT: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readFloatLE(offsetInFrame + k * 4),
                x: timestamp
              });
            }
            break;
          }

          case apidefs.READ_TYPE_DOUBLE: {
            for (let k = 0; k < varInGroup.nb; k++) {
              result[indiceInGroup].push({
                y: buffer.readDoubleLE(offsetInFrame + k * 8),
                x: timestamp
              });
            }
            break;
          }

          default: {
            logger.error("type is not allowed : ", varInGroup.type);
            return res;
          }
        }
      } catch (err) {
        logger.error("An error occurred when extracting data from buffer read in memory target : ", err);
        return {
          result: apidefs.ERROR_BUFFER_READ
        };
      }
    }
  }

  res.result = apidefs.STLINK_OK;
  res.data = result;
  logger.debug("res = ", res.result);
  logger.debug("res = ", res.data);
  return res;
}

/**
 * Service to write a group of data to a target memory area.
 *
 * @param {string} probeId - The probe to be used to communicate with the target MCU.
 * @param {Array<object>} group - The group of data to be write to the target memory
 *              this is an array, containing objects each describing one variable with
 *              its name, its address (as a 0x-started string in hexadecimal representation),
 *              its type (from READ TYPE numeric constants), and the number of items to be write,
 *              and the bale to be write.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol to use either p2p or tcp.
 * @returns {object} Operation result code.
 */
function writeMemoryGroup(probeId, group, accessPointId = 0, typeConnect) {
  logger.debug("probeId = ", probeId, ", group = ", group);
  let res = { result: apidefs.STLINK_BAD_PARAMETER };

  // probeId is the probe enum unique ID, it is mandatory
  if (probeId == undefined || probeId == "") {
    logger.error("probeId undefined");
    return res;
  }

  // check that this probe is known
  let probeIndex = Devices.findIndex(function(elem) {
    return elem.EnumUniqueId == probeId;
  });

  if (probeIndex == -1) {
    res.result = apidefs.STLINK_NO_DEVICE;
    logger.error("no device for probeId ", probeId);
    return res;
  }

  //check that this probe is opened
  if (Devices[probeIndex].state != "opened") {
    res.result = apidefs.STLINK_DEVICE_NOT_OPENED;
    logger.error("device not opened for probeId ", probeId, " with index ", probeIndex);
    return res;
  }

  // check that the group is correct if it is provided
  if (group === undefined) {
    logger.error("group undefined");
    return res;
  }

  res.result = apidefs.STLINK_OK;
  for (let i = 0; i < group.length; i++) {
    let varToWrite = group[i];
    logger.trace("group[i] = " + JSON.stringify(varToWrite));

    let valueString, valueDec, mode;
    let valueHex = [];

    let address = Number(varToWrite.address);

    if (varToWrite.value.toLowerCase().startsWith("0x")) {
      logger.trace("varToWrite is in Hexadecimal form");
      // Remove the prefix "0x"
      valueString = varToWrite.value.substring(2);
      // Add padding if the user wrote an odd number of digits (e.g. 0x3)
      if ((valueString.length - 2) % 2 == 1) {
        valueString = "0" + valueString;
      }
      // Add padding (worst case is the "double" type which is a 64 bits)
      while (valueString.length < 16) {
        valueString = "00" + valueString;
      }
      logger.trace("valueString = " + valueString);

      // Convert string in an array of hexadecimal values (in big endian)
      let buffLength = valueString.length / 2;
      valueHex = Buffer.allocUnsafe(buffLength);
      for (let j = 0; j < buffLength; j++) {
        valueHex[j] = parseInt(valueString.substring(j * 2, j * 2 + 2), 16);
        logger.trace("valueHex[" + j + "] = " + valueHex[j]);
      }
      // Set the mode in "Hexadecimal"
      mode = "Hex";
    } else {
      logger.trace("varToWrite is in Decimal form");
      if (varToWrite.value.length <= 10) {
        valueDec = Number(varToWrite.value);
      } else {
        valueDec = varToWrite.value;
      }
      // Set the mode in "Decimal"
      mode = "Dec";
      logger.trace("value = " + valueDec);
    }

    let initAPRes = initAccessPoint(probeIndex, accessPointId, typeConnect);
    if (initAPRes != apidefs.STLINK_OK) {
      res.result = initAPRes;
      logger.error("Access point", accessPointId, " could not be initialized");
      return res;
    }

    switch (varToWrite.type) {
      case apidefs.READ_TYPE_U8: {
        let buffer = Buffer.allocUnsafe(1);

        if (mode == "Dec") {
          buffer.writeUInt8(valueDec, 0);
        } else {
          // Read the last value of valueHex (in big endian) and write it in buffer (in little endian)
          buffer.writeUInt8(valueHex.slice(valueHex.length - 1).readUInt8(0), 0);
        }
        res = JTAG_WriteMemory8Bit(Devices[probeIndex].DevNumber, address, buffer.length, buffer, typeConnect);

        break;
      }
      case apidefs.READ_TYPE_S8: {
        let buffer = Buffer.allocUnsafe(1);

        if (mode == "Dec") {
          buffer.writeInt8(valueDec, 0);
        } else {
          // Read the last value of valueHex (in big endian) and write it in buufer (in little endian)
          buffer.writeInt8(valueHex.slice(valueHex.length - 1).readInt8(0), 0);
        }

        res = JTAG_WriteMemory8Bit(Devices[probeIndex].DevNumber, address, buffer.length, buffer, typeConnect);

        break;
      }

      case apidefs.READ_TYPE_U16: {
        let buffer = Buffer.allocUnsafe(2);

        if (mode == "Dec") {
          buffer.writeUInt16LE(valueDec, 0);
        } else {
          // Read the last two value of valueHex (in big endian) and write them in buffer (in little endian)
          buffer.writeUInt16LE(valueHex.slice(valueHex.length - 2).readUInt16BE(0), 0);
        }

        res = JTAG_WriteMemory8Bit(Devices[probeIndex].DevNumber, address, buffer.length, buffer, typeConnect);

        break;
      }

      case apidefs.READ_TYPE_S16: {
        let buffer = Buffer.allocUnsafe(2);

        if (mode == "Dec") {
          buffer.writeInt16LE(valueDec, 0);
        } else {
          // Read the last two values of valueHex (in big endian) and write them in buffer (in little endian)
          buffer.writeInt16LE(valueHex.slice(valueHex.length - 2).readInt16BE(0), 0);
        }

        res = JTAG_WriteMemory8Bit(Devices[probeIndex].DevNumber, address, buffer.length, buffer, typeConnect);

        break;
      }

      case apidefs.READ_TYPE_U32: {
        let buffer = Buffer.allocUnsafe(4);

        if (mode == "Dec") {
          buffer.writeUInt32LE(valueDec, 0);
        } else {
          // Read the last four values of valueHex (in big endian) and write them in buffer (in little endian)
          buffer.writeUInt32LE(valueHex.slice(valueHex.length - 4).readUInt32BE(0), 0);
        }

        res = JTAG_WriteMemory32Bit(Devices[probeIndex].DevNumber, address, buffer.length, buffer, typeConnect);

        break;
      }

      case apidefs.READ_TYPE_S32: {
        let buffer = Buffer.allocUnsafe(4);

        if (mode == "Dec") {
          buffer.writeInt32LE(valueDec, 0);
        } else {
          // Read the last four values of valueHex (in big endian) and write them in buffer (in little endian)
          buffer.writeInt32LE(valueHex.slice(valueHex.length - 4).readInt32BE(0), 0);
        }
        res = JTAG_WriteMemory32Bit(Devices[probeIndex].DevNumber, address, buffer.length, buffer, typeConnect);
        break;
      }

      case apidefs.READ_TYPE_FLOAT: {
        let buffer = Buffer.allocUnsafe(4);
        if (mode == "Dec") {
          buffer.writeFloatLE(valueDec, 0);
        } else {
          // Read the last four values of valueHex (in big endian) and write them in buffer (in little endian)
          buffer.writeFloatLE(valueHex.slice(valueHex.length - 4).readFloatBE(0), 0);
        }
        res = JTAG_WriteMemory32Bit(Devices[probeIndex].DevNumber, address, buffer.length, buffer, typeConnect);
        break;
      }

      case apidefs.READ_TYPE_DOUBLE: {
        let buffer = Buffer.allocUnsafe(8);
        if (mode == "Dec") {
          buffer.writeDoubleLE(valueDec, 0);
        } else {
          // Read the last eight values of valueHex (in big endian) and write them in buffer (in little endian)
          buffer.writeDoubleLE(valueHex.slice(valueHex.length - 8).readDoubleBE(0), 0);
        }

        res = JTAG_WriteMemory32Bit(
          Devices[probeIndex].DevNumber,
          varToWrite.address,
          buffer.length,
          buffer,
          typeConnect
        );
        break;
      }

      default:
        logger.error("type : " + varToWrite.type);
        res = { result: apidefs.STLINK_BAD_PARAMETER };
        break;
    }

    if (res.result !== apidefs.STLINK_OK) {
      // One issue occurs when writing current value (or if the variable type is unknown)
      // Stop writing the others values and returns
      logger.error(
        "Unable to write value ",
        varToWrite.value,
        " for ",
        varToWrite.name,
        " of type ",
        varToWrite.type,
        " in address ",
        varToWrite.address
      );
      break;
    }
  }

  return res;
}

/**
 * Initializes the access point to use.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} Result of the command.
 */
function initAccessPoint(deviceIndex, accessPointId, typeConnect) {
  logger.trace("Initializing AP nb: ", accessPointId);
  //if access point already initialized
  if (Devices[deviceIndex].accessPointId == accessPointId) {
    return apidefs.STLINK_OK;
  }
  // if another access point is initialized, first close it
  if (Devices[deviceIndex].accessPointId != undefined) {
    let cmd_result = closeAccessPoint(deviceIndex, Devices[deviceIndex].accessPointId, typeConnect);
    if (cmd_result != apidefs.STLINK_OK) {
      logger.error("initAP error, could not close previously initialized AP= ", cmd_result);
      return cmd_result;
    }
  }

  let cmd_result = apidefs.STLINK_OK;
  // Init Access Point up to version V2J28
  if (
    Devices[deviceIndex].STLinkVersion >= 3 ||
    (Devices[deviceIndex].STLinkVersion == 2 &&
      Devices[deviceIndex].JTAGVersion >= defs.STLINK_JTAG_MIN_FW_VERSION_FOR_AP_AND_TCP)
  ) {
    cmd_result = JTAG.initAP(Devices[deviceIndex], accessPointId, typeConnect);
    if (cmd_result != apidefs.STLINK_OK) {
      logger.error("initAP error = ", cmd_result);
    } else {
      Devices[deviceIndex].accessPointId = accessPointId;
    }
  }
  return cmd_result;
}

/**
 * Closes given access point.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {number} accessPointId - Id of the access point.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} Result of the command.
 */
function closeAccessPoint(deviceIndex, accessPointId, typeConnect) {
  logger.trace("Closing AP nb: ", accessPointId);
  if (
    Devices[deviceIndex].STLinkVersion >= 3 ||
    (Devices[deviceIndex].STLinkVersion == 2 &&
      Devices[deviceIndex].JTAGVersion >= defs.STLINK_JTAG_MIN_FW_VERSION_FOR_AP_AND_TCP)
  ) {
    Devices[deviceIndex].accessPointId = undefined;
    return JTAG.closeAP(Devices[deviceIndex], accessPointId, typeConnect);
  }
}

/**
 * Check TCP Server availability.
 *
 * @returns {boolean} True if TCP server is available, false otherwise.
 */
function isTcpAvailable() {
  logger.debug("isTcpAvailable");
  var result = stLinkDrv.STLink_ReenumerateTcp(defsDrv.DRV_STLINK_TCP, false, TCP_3, DEFAULT_TCP_SERVER_OPTION);
  if (result === defsDrv.DRV_TCP_CANT_CONNECT) {
    logger.info("isTcpAvailable : can't connect to Tcp");
  }
  return result !== defsDrv.DRV_TCP_CANT_CONNECT;
}

/**
 * Blink Led on the given probeId device api for cli.
 *
 * @param {number} deviceIndex - The index of the device.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function JTAG_BlinkLed(deviceIndex, typeConnect) {
  logger.debug("JTAG_BlinkLed arguments deviceIndex = ", deviceIndex, ", typeConnect = ", typeConnect);
  if (deviceIndex > nbDevices - 1) {
    return apidefs.STLINK_BAD_PARAMETER;
  }

  return JTAG.JTAG_BlinkLed(Devices[deviceIndex], typeConnect);
}
/**
 * Blink Led on the given probeId device.
 *
 * @param {string} probeId - The probeId on which the led should blink.
 * @param {string} typeConnect - The connection protocol either p2p or tcp.
 * @returns {number} - Result of the command.
 */
function blinkLed(probeId, typeConnect) {
  logger.debug("blinkLed arguments probeId = ", probeId, ", typeConnect = ", typeConnect);
  let res = { result: apidefs.STLINK_BAD_PARAMETER };
  // probeId is the probe enum unique ID, it is mandatory
  if (probeId === undefined || probeId === "") {
    logger.error("probeId undefined");
    return res;
  }

  probeDiscovery(typeConnect);
  // check that this probe is known
  let probeIndex = Devices.findIndex(function(elem) {
    return elem.EnumUniqueId == probeId;
  });

  if (probeIndex === -1) {
    res.result = apidefs.STLINK_NO_DEVICE;
    logger.error("no device for probeId ", probeId);
    return res;
  }

  // Open probe
  let open_res;
  if (typeConnect === "tcp") {
    open_res = openDeviceTcp(probeIndex);
  } else {
    open_res = openDevice(probeIndex);
  }

  if (open_res != apidefs.STLINK_OK) {
    logger.error("Open device failed for device index ", probeIndex, " result = ", open_res);
    res.result = open_res;
    return res;
  }

  if (
    Devices[probeIndex].STLinkVersion >= 3 ||
    (Devices[probeIndex].STLinkVersion === 2 &&
      Devices[probeIndex].JTAGVersion >= defs.STLINK_JTAG_MIN_FW_VERSION_FOR_AP_AND_TCP)
  ) {
    res.result = JTAG.JTAG_BlinkLed(Devices[probeIndex], typeConnect);
  } else {
    res.result = apidefs.STLINK_JTAG_CMD_NOT_SUPPORTED;
  }

  if (typeConnect === "tcp") {
    closeDeviceTcp(probeIndex);
  } else {
    closeDevice(probeIndex);
  }

  return res;
}
// Common constants definitions
module.exports.apidefs = apidefs;

// ----------------------------------------------
// Exports for Command Line Interface (CLI)
module.exports.cli = {};

// low-level definitions
module.exports.cli.clidefs = clidefs;

// display strings
module.exports.cli.errorString = errorString;
module.exports.cli.modeString = modeString;

// generic interface
module.exports.cli.reenumerate = reenumerate;
module.exports.cli.getNbDevices = getNbDevices;
module.exports.cli.openDevice = openDevice;
module.exports.cli.openDeviceTcp = openDeviceTcp;
module.exports.cli.closeDevice = closeDevice;
module.exports.cli.closeDeviceTcp = closeDeviceTcp;
module.exports.cli.getDeviceInfo = getDeviceInfo;
module.exports.cli.getCurrentMode = getCurrentMode;
module.exports.cli.getVersion = getVersion;
module.exports.cli.exitDFUMode = exitDFUMode;
module.exports.cli.initAP = initAccessPoint;
module.exports.cli.closeAP = closeAccessPoint;

// JTAG interface
module.exports.cli.enterJTAG = enterJTAG;
module.exports.cli.enterSWD = enterSWD;
module.exports.cli.exitJTAG = exitJTAG;
module.exports.cli.JTAG_ReadIdCodes = JTAG_ReadIdCodes;
module.exports.cli.JTAG_ReadMemory32Bit = JTAG_ReadMemory32Bit;
module.exports.cli.JTAG_ReadMemory16Bit = JTAG_ReadMemory16Bit;
module.exports.cli.JTAG_ReadMemory8Bit = JTAG_ReadMemory8Bit;
module.exports.cli.JTAG_WriteMemory32Bit = JTAG_WriteMemory32Bit;
module.exports.cli.JTAG_WriteMemory8Bit = JTAG_WriteMemory8Bit;
module.exports.cli.JTAG_BlinkLed = JTAG_BlinkLed;

// SWIM interface

// ----------------------------------------------
// Exports for Acquisition Services module
module.exports.probeDiscovery = probeDiscovery;
module.exports.connectToTarget = connectToTarget;
module.exports.closeProbes = closeProbes;
module.exports.readMemoryMode = readMemoryMode;
module.exports.readMemoryTyped = readMemoryTyped;
module.exports.readMemoryGroup = readMemoryGroup;
module.exports.writeMemoryGroup = writeMemoryGroup;
module.exports.errorString = errorString;
module.exports.isTcpAvailable = isTcpAvailable;
module.exports.blinkLed = blinkLed;
