/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
/**
 * Provided definitions required by mocked functions to return status (for nominal and error cases).
 */
const stLinkDefs = require("../src/STLinkUSBDriver").defs;

/**
 * Function required by mocked functions to write response in buffer memory.
 */
const writeValueInRespBuffer = require("./testsUtilities").writeValueInRespBuffer;

/**
 * Options used by STLink_SendCommandMocked.
 * Value set by STLink_SendCommandSetOptions.
 * Valut returned by STLink_SendCommandGetOptions.
 */
let sendCommandOptions = "";

/**
 * Mocked STLink_GetDeviceInfo2 (returning default device information for nominal use case).
 *
 * @param {number} interface - Device interface (constant value stLinkDefs.DRV_DEFAULT_IF).
 * @param {number} index - Device index.
 * @param {number[]} deviceInfo - Device information.
 * @param {number} deviceInfoLength - Device information length.
 */
function STLink_GetDeviceInfo2Mocked(interface, index, deviceInfo, deviceInfoLength) {
  // console.log("STLink_GetDeviceInfo2Mocked called !");
  for (let i = 0; i < deviceInfoLength; i++) {
    deviceInfo[i] = 0x30 + i;
  }
  // Set the end of the EnumUniqueId
  deviceInfo[8] = 0x00;
  // set the Vendor Id
  deviceInfo[36] = 0x83;
  deviceInfo[37] = 0x4;
  // set the Product Id
  deviceInfo[38] = 0x4a;
  deviceInfo[39] = 0x37;
  return stLinkDefs.DRV_OK;
}

/**
 * Set the options of STLink_SendCommandMocked.
 *
 * @param {string} options - Options.
 */
function STLink_SendCommandSetOptions(options) {
  sendCommandOptions = options;
}

/**
 * Get the options of STLink_SendCommandMocked.
 *
 * @returns {string} Options - Options.
 */
function STLink_SendCommandGetOptions() {
  return sendCommandOptions;
}

/**
 * Mocked STLink_SendCommand.
 * Performs different treatments following "sendCommandOptions" value.
 *
 * @param {any} handle - Command handle.
 * @param {*} reqRef - Request reference.
 */
function STLink_SendCommandMocked(handle, reqRef) {
  // console.log("STLink_SendCommandMocked called !");

  let options = STLink_SendCommandGetOptions();
  // console.log("options = " + options);

  let unknownSendCommand = false;
  let responseSize;

  // Read the "resp" buffer address from reqRef
  // let addr = reqRef.readUInt64LE(18);
  // console.log("addr = " + addr.toString(16).padStart(16, "0"));

  // Decode the SendCommand request
  let sendCommandMeta = reqRef[1];
  let sendCommandSub = reqRef[2];
  // console.log("sendCommandMeta = " + sendCommandMeta);
  // console.log("sendCommandSub = " + sendCommandSub);

  // Compute the expected response corresponding to the SendCommand request
  let computedResponse = [];
  let memoryAddress = 0;
  switch (sendCommandMeta) {
    case 0xf2: // STLINK_CMD_JTAG
      switch (sendCommandSub) {
        case 0x07: // STLINK_JTAG_READMEM_32BIT
        case 0x0c: // STLINK_JTAG_READMEM_8BIT
        case 0x47: // STLINK_JTAG_READMEM_16BIT
          memoryAddress = reqRef.readUInt32LE(3);
          // console.log("ReadIdCodes - Address = " + memoryAddress);

          switch (options) {
            case "READ_MEMORY_RETURNS_ERROR":
              return stLinkDefs.DRV_TIMEOUT;

            case "SECOND_READ_MEMORY_RETURNS_ERROR":
              responseSize = (reqRef[8] << 8) | reqRef[7];
              // console.log("responseSize = " + responseSize);
              for (let i = 1; i < responseSize + 1; i++) {
                computedResponse.push(i);
              }
              STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32_DBGMCU_IDCODE_ADDR":
              if (memoryAddress === 0xe0042000) {
                // STM32Add.STM32_DBGMCU_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32_DBGMCU_ADDR":
              if (memoryAddress === 0x40015800) {
                // STM32Add.STM32_DBGMCU_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32H_IDCODE_ADDR":
              if (memoryAddress === 0x5c001000) {
                // STM32Add.STM32H_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32W_IDCODE_ADDR":
              if (memoryAddress === 0x40004000) {
                // STM32Add.STM32H_IDCODE_ADDR
                computedResponse = [0x22, 0x10, 0x46, 0x44]; // Random value (>> 12) - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "READ_STM32_MCU_CODE_AT_STM32_DBGMCU_IDCODE_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "READ_STM32_MCU_CODE_AT_STM32_DBGMCU_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "READ_STM32_MCU_CODE_AT_STM32H_IDCODE_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "READ_STM32_MCU_CODE_AT_STM32W_IDCODE_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "STM32_MCU_CODE_IS_NOT_SET_IN_TARGET":
              computedResponse = [0x00, 0x00, 0x00, 0x11]; // MCU ID = 0
              break;

            case "UNKNOWN_MCU_ID":
              computedResponse = [0x61, 0x05, 0x03, 0x04]; // Random value - 0x561 = Unknown MCU ID
              break;

            default:
              // nominal case
              if (memoryAddress === 0xe0042000) {
                // STM32Add.STM32_DBGMCU_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                responseSize = (reqRef[8] << 8) | reqRef[7];
                // console.log("responseSize = " + responseSize);
                for (let i = 1; i < responseSize + 1; i++) {
                  computedResponse.push(i);
                }
              }
              break;
          }
          break;

        case 0x30: // STLINK_JTAG_ENTER2 (for enterSWD & enterJTAG)
          switch (options) {
            case "ENTER_SWD_OR_JTAG_RETURNS_ERROR":
              return stLinkDefs.STLINK_JTAG_SPI_ERROR;

            case "ENTER_SWD_OR_JTAG_FAILS":
              computedResponse = [0x06]; // random value (different of STLINK_JTAG_OK)
              break;

            default:
              computedResponse = [0x80]; // STLINK_JTAG_OK (0x80)
              break;
          }
          break;

        case 0x21: // STLINK_JTAG_EXIT
          if (options === "EXIT_JTAG_RETURNS_ERROR") {
            return stLinkDefs.DRV_ERR;
          } else {
            return stLinkDefs.DRV_OK;
          }

        case 0x31: // STLINK_JTAG_READ_IDCODES
          switch (options) {
            case "READ_ID_CODES_SETS_UNKNOWN_JTAG_CHAIN":
              computedResponse = [0x04, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN
              break;
            case "READ_ID_CODES_SETS_NO_DEVICE_CONNECTED":
              computedResponse = [0x05, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_NO_DEVICE_CONNECTED
              break;
            case "READ_ID_CODES_SETS_WRITE_ERROR":
              computedResponse = [0x0c, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_WRITE_ERROR
              break;
            case "READ_ID_CODES_SETS_ALREADY_OPENED_IN_OTHER_MODE":
              computedResponse = [0x0e, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE
              break;
            case "READ_ID_CODES_SETS_SWD_AP_WAIT":
              computedResponse = [0x10, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_SWD_DP_WAIT
              break;
            case "READ_ID_CODES_SETS_SWD_AP_STICKY_ERROR":
              computedResponse = [0x19, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_SWD_AP_STICKY_ERROR
              break;
            case "READ_ID_CODES_SETS_SWV_NOT_AVAILABLE":
              computedResponse = [0x20, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_SWV_NOT_AVAILABLE
              break;
            case "READ_ID_CODES_SETS_TCPID_NOT_FOUND":
              computedResponse = [0x30, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_TCPID_NOT_FOUND
              break;
            case "READ_ID_CODES_SETS_UNKNOWN_JTAG_ERROR":
              computedResponse = [0x00, 0, 0, 0, 0x77, 0x04]; // first value is an unknown JTAG error
              break;
            case "ID_CODES_ARE_ERRONEOUS":
              computedResponse = [0x80, 0, 0, 0, 0x06, 0x04]; // STLINK_JTAG_OK (0x80) but erroneous ID Codes
              break;
            default:
              // No error case
              computedResponse = [0x80, 0, 0, 0, 0x77, 0x04]; // STLINK_JTAG_OK (0x80) + CORTEX_DESIGNER_ARM_ID (0x77) + CORTEX_DESIGNER_ARM_ID (0x04)
              break;
          }
          break;

        case 0x3e: // STLINK_JTAG_GETLASTRWSTATUS2
          switch (options) {
            case "READ_LAST_READ_WRITE_STATUS_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "LAST_READ_WRITE_STATUS_IS_ERRONEOUS":
              computedResponse = [0, 0, 0, 0, 0x0a, 0x0b, 0x0c, 0x0d]; // 0x0a, 0x0b, 0x0c, 0x0d represents the bad address
              break;

            default:
              computedResponse = [0x80, 0];
              break;
          }
          break;

        case 0x43: // STLINK_SWD_SET_FREQ
        case 0x44: // STLINK_JTAG_SET_FREQ
        case 0x61: // STLINK_JTAG_SET_COM_FREQ
          switch (options) {
            case "SET_FREQUENCY_RETURNS_ERROR":
              return stLinkDefs.DRV_NO_SUCH_INFO;

            case "SET_FREQUENCY_FAILS":
              computedResponse = [0x04]; // random value (different of STLINK_JTAG_OK)
              break;

            default:
              computedResponse = [0x80]; // STLINK_JTAG_OK (0x80)
              break;
          }
          break;

        case 0x0d: // STLINK_JTAG_WRITEMEM_8BIT
        case 0x08: // STLINK_JTAG_WRITEMEM_32BIT
          if (options === "") {
            return stLinkDefs.DRV_OK;
          } else {
            return stLinkDefs.DRV_PERMISSION_ERR;
          }

        case 0x62: // STLINK_JTAG_GET_COM_FREQ
          if (reqRef[3] == 1) {
            //JTAG
            computedResponse = [0x80, 0, 0, 0, 0x40, 0x1f, 0, 0, 2, 0, 0, 0, 0xc0, 0x5d, 0, 0, 0x40, 0x1f, 0, 0]; //returning 2 frequencies
          } else {
            // SWD
            computedResponse = [
              0x80,
              0,
              0,
              0,
              0x40,
              0x1f,
              0,
              0,
              3,
              0,
              0,
              0,
              0xe4,
              0x0c,
              0,
              0,
              0xe8,
              3,
              0,
              0,
              0xc8,
              0,
              0,
              0
            ]; //returning 3 frequencies
          }
          break;

        case 0x4b: // STLINK_INIT_AP
          computedResponse = [0x80];
          break;

        case 0x4c: // STLINK_CLOSE_AP
          computedResponse = [0x80];
          break;

        case 0x49: // STLINK_JTAG_BLINK_LED
          if (options === "") {
            return stLinkDefs.DRV_OK;
          } else {
            return stLinkDefs.DRV_CMD_NOT_AVAILABLE;
          }

        default:
          unknownSendCommand = true;
          break;
      }
      break;

    case 0xf1: // STLINK_CMD_GET_VERSION
      switch (options) {
        case "GET_VERSION_RETURNS_ERROR":
          return stLinkDefs.DRV_ERR;

        case "UNSUPPORTED_FIRMWARE_VERSION":
          computedResponse = [0x0, 0x22]; // // This value is lower than allowed JTAG version (and then the probe does not support the JTAG protocol)
          break;

        case "STLINK_V3_VERSION":
        case "STLINK_V3_NO_JTAG":
        case "GET_EXT_VERSION_RETURNS_ERROR":
          computedResponse = [0x30, 0x00];
          break;

        default:
          computedResponse = [0x0c, 0x22]; // Random value (can be another value)
          break;
      }
      break;

    case 0xfb: //STLINK_CMD_GET_VERSION_EXT
      switch (options) {
        case "GET_EXT_VERSION_RETURNS_ERROR":
          return stLinkDefs.DRV_ERR;
        case "STLINK_V3_NO_JTAG":
          computedResponse = [0x03, 0x00, 0x00, 0x00, 0x03];
          break;
        default:
          computedResponse = [0x03, 0x01, 0x04, 0x02, 0x03];
          break;
      }
      break;

    case 0xf5: // STLINK_CMD_GET_CURRENT_MODE
      switch (options) {
        case "GET_CURRENT_MODE_RETURNS_ERROR":
          return stLinkDefs.DRV_FAILED;
        case "GET_CURRENT_MODE_FAILS":
          computedResponse = [0x10, 0x10]; // First byte is a random value (different of STLINK_DEV_JTAG_MODE)
          break;
        case "SET_DFU_MODE":
        case "EXIT_DFU_MODE_RETURNS_ERROR":
          computedResponse = [0x00, 0x10]; // First byte is STLINK_DEV_DFU_MODE
          break;
        case "SET_MASS_MODE":
          computedResponse = [0x01, 0x10]; // First byte is STLINK_DEV_MASS_MODE
          break;
        case "SET_JTAG_MODE":
          computedResponse = [0x02, 0x10]; // First byte is STLINK_DEV_JTAG_MODE
          break;
        case "SET_SWIM_MODE":
          computedResponse = [0x03, 0x10]; // First byte is STLINK_DEV_SWIM_MODE
          break;
        case "SET_BTLD_MODE":
          computedResponse = [0x04, 0x10]; // First byte is STLINK_DEV_BTLD_MODE
          break;
        default:
          computedResponse = [0x02, 0x10]; // Default mode is STLINK_DEV_JTAG_MODE
          break;
      }
      break;

    case 0xf3: // STLINK_CMD_DFU
      switch (sendCommandSub) {
        case 0x07: // STLINK_DFU_EXIT
          if (options === "EXIT_DFU_MODE_RETURNS_ERROR") {
            return stLinkDefs.DRV_ERR;
          } else {
            return stLinkDefs.DRV_OK;
          }

        default:
          unknownSendCommand = true;
          break;
      }
      break;

    default:
      unknownSendCommand = true;
      break;
  }

  if (unknownSendCommand === true) {
    console.error(
      "Unknown SendCommand P2P - sendCommandMeta = " +
        sendCommandMeta.toString(16).padStart(16, "0") +
        " - sendCommandSub = " +
        sendCommandSub.toString(16).padStart(16, "0")
    );
    return stLinkDefs.DRV_FAILED;
  }

  // Write the expected response in response buffer
  let addrInBytesArray = [
    reqRef[18],
    reqRef[19],
    reqRef[20],
    reqRef[21],
    reqRef[22],
    reqRef[23],
    reqRef[24],
    reqRef[25]
  ];
  writeValueInRespBuffer(addrInBytesArray, computedResponse);
  return stLinkDefs.DRV_OK;
}

/**
 * Mocked STLink_SendCommand.
 * Performs different treatments following "sendCommandOptions" value.
 *
 * @param {any} handle - Command handle.
 * @param {*} reqRef - Request reference.
 */
function STLink_SendCommandTcpMocked(handle, reqRef) {
  let options = STLink_SendCommandGetOptions();
  // console.log("options = " + options);

  let unknownSendCommand = false;
  let responseSize;

  // Read the "resp" buffer address from reqRef
  // let addr = reqRef.readUInt64LE(18);
  // console.log("addr = " + addr.toString(16).padStart(16, "0"));

  // Decode the SendCommand request
  let sendCommandMeta = reqRef[1];
  let sendCommandSub = reqRef[2];
  // console.log("sendCommandMeta = " + sendCommandMeta);
  // console.log("sendCommandSub = " + sendCommandSub);

  // Compute the expected response corresponding to the SendCommand request
  let computedResponse = [];
  let memoryAddress = 0;
  switch (sendCommandMeta) {
    case 0xf2: // STLINK_CMD_JTAG
      switch (sendCommandSub) {
        case 0x07: // STLINK_JTAG_READMEM_32BIT
        case 0x0c: // STLINK_JTAG_READMEM_8BIT
        case 0x47: // STLINK_JTAG_READMEM_16BIT
          memoryAddress = reqRef.readUInt32LE(3);
          // console.log("ReadIdCodes - Address = " + memoryAddress);

          switch (options) {
            case "READ_MEMORY_RETURNS_ERROR":
              return stLinkDefs.DRV_TIMEOUT;

            case "SECOND_READ_MEMORY_RETURNS_ERROR":
              responseSize = (reqRef[8] << 8) | reqRef[7];
              // console.log("responseSize = " + responseSize);
              for (let i = 1; i < responseSize + 1; i++) {
                computedResponse.push(i);
              }
              STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32_DBGMCU_IDCODE_ADDR":
              if (memoryAddress === 0xe0042000) {
                // STM32Add.STM32_DBGMCU_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32_DBGMCU_ADDR":
              if (memoryAddress === 0x40015800) {
                // STM32Add.STM32_DBGMCU_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32H_IDCODE_ADDR":
              if (memoryAddress === 0x5c001000) {
                // STM32Add.STM32H_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "STM32_MCU_CODE_IS_SET_IN_STM32W_IDCODE_ADDR":
              if (memoryAddress === 0x40004000) {
                // STM32Add.STM32H_IDCODE_ADDR
                computedResponse = [0x22, 0x10, 0x46, 0x44]; // Random value (>> 12) - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                computedResponse = [0x00, 0x00, 0x03, 0x04]; // MCU ID = 0
              }
              break;

            case "READ_STM32_MCU_CODE_AT_STM32_DBGMCU_IDCODE_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "READ_STM32_MCU_CODE_AT_STM32_DBGMCU_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "READ_STM32_MCU_CODE_AT_STM32H_IDCODE_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "READ_STM32_MCU_CODE_AT_STM32W_IDCODE_ADDR_RETURNS_ERROR":
              return stLinkDefs.DRV_MEMORY_PROBLEM;

            case "STM32_MCU_CODE_IS_NOT_SET_IN_TARGET":
              computedResponse = [0x00, 0x00, 0x00, 0x11]; // MCU ID = 0
              break;

            case "UNKNOWN_MCU_ID":
              computedResponse = [0x61, 0x05, 0x03, 0x04]; // Random value - 0x561 = Unknown MCU ID
              break;

            default:
              // nominal case
              if (memoryAddress === 0xe0042000) {
                // STM32Add.STM32_DBGMCU_IDCODE_ADDR
                computedResponse = [0x61, 0x04, 0x03, 0x04]; // Random value - 0x461 = MCU ID STM32L496xx/4A6xx
              } else {
                responseSize = (reqRef[8] << 8) | reqRef[7];
                // console.log("responseSize = " + responseSize);
                for (let i = 1; i < responseSize + 1; i++) {
                  computedResponse.push(i);
                }
              }
              break;
          }
          break;

        case 0x30: // STLINK_JTAG_ENTER2 (for enterSWD & enterJTAG)
          switch (options) {
            case "ENTER_SWD_OR_JTAG_RETURNS_ERROR":
              return stLinkDefs.DRV_ERR;

            case "ENTER_SWD_OR_JTAG_FAILS":
              computedResponse = [0x06]; // random value (different of STLINK_JTAG_OK)
              break;

            default:
              computedResponse = [0x80]; // STLINK_JTAG_OK (0x80)
              break;
          }
          break;

        case 0x21: // STLINK_JTAG_EXIT
          if (options === "EXIT_JTAG_RETURNS_ERROR") {
            return stLinkDefs.DRV_ERR;
          } else {
            return stLinkDefs.DRV_OK;
          }

        case 0x31: // STLINK_JTAG_READ_IDCODES
          switch (options) {
            case "READ_ID_CODES_SETS_UNKNOWN_JTAG_CHAIN":
              computedResponse = [0x04, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN
              break;
            case "READ_ID_CODES_SETS_NO_DEVICE_CONNECTED":
              computedResponse = [0x05, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_NO_DEVICE_CONNECTED
              break;
            case "READ_ID_CODES_SETS_WRITE_ERROR":
              computedResponse = [0x0c, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_WRITE_ERROR
              break;
            case "READ_ID_CODES_SETS_ALREADY_OPENED_IN_OTHER_MODE":
              computedResponse = [0x0e, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE
              break;
            case "READ_ID_CODES_SETS_SWD_AP_WAIT":
              computedResponse = [0x10, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_SWD_DP_WAIT
              break;
            case "READ_ID_CODES_SETS_SWD_AP_STICKY_ERROR":
              computedResponse = [0x19, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_SWD_AP_STICKY_ERROR
              break;
            case "READ_ID_CODES_SETS_SWV_NOT_AVAILABLE":
              computedResponse = [0x20, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_SWV_NOT_AVAILABLE
              break;
            case "READ_ID_CODES_SETS_TCPID_NOT_FOUND":
              computedResponse = [0x30, 0, 0, 0, 0x77, 0x04]; // first value is defs.STLINK_JTAG_TCPID_NOT_FOUND
              break;
            case "READ_ID_CODES_SETS_UNKNOWN_JTAG_ERROR":
              computedResponse = [0x00, 0, 0, 0, 0x77, 0x04]; // first value is an unknown JTAG error
              break;
            case "ID_CODES_ARE_ERRONEOUS":
              computedResponse = [0x80, 0, 0, 0, 0x06, 0x04]; // STLINK_JTAG_OK (0x80) but erroneous ID Codes
              break;
            default:
              // No error case
              computedResponse = [0x80, 0, 0, 0, 0x77, 0x04]; // STLINK_JTAG_OK (0x80) + CORTEX_DESIGNER_ARM_ID (0x77) + CORTEX_DESIGNER_ARM_ID (0x04)
              break;
          }
          break;

        case 0x3e: // STLINK_JTAG_GETLASTRWSTATUS2
          switch (options) {
            case "READ_LAST_READ_WRITE_STATUS_RETURNS_ERROR":
              return stLinkDefs.DRV_ERR;

            case "LAST_READ_WRITE_STATUS_IS_ERRONEOUS":
              computedResponse = [0, 0, 0, 0, 0x0a, 0x0b, 0x0c, 0x0d]; // 0x0a, 0x0b, 0x0c, 0x0d represents the bad address
              break;

            default:
              computedResponse = [0x80, 0];
              break;
          }
          break;

        case 0x43: // STLINK_SWD_SET_FREQ
        case 0x44: // STLINK_JTAG_SET_FREQ
        case 0x61: // STLINK_JTAG_SET_COM_FREQ
          switch (options) {
            case "SET_FREQUENCY_RETURNS_ERROR":
              return stLinkDefs.DRV_ERR;

            case "SET_FREQUENCY_FAILS":
              computedResponse = [0x04]; // random value (different of STLINK_JTAG_OK)
              break;

            default:
              computedResponse = [0x80]; // STLINK_JTAG_OK (0x80)
              break;
          }
          break;

        case 0x0d: // STLINK_JTAG_WRITEMEM_8BIT
        case 0x08: // STLINK_JTAG_WRITEMEM_32BIT
          if (options === "") {
            return stLinkDefs.DRV_OK;
          } else {
            return stLinkDefs.DRV_ERR;
          }

        case 0x62: // STLINK_JTAG_GET_COM_FREQ
          if (reqRef[3] == 1) {
            //JTAG
            computedResponse = [0x80, 0, 0, 0, 0x40, 0x1f, 0, 0, 2, 0, 0, 0, 0xc0, 0x5d, 0, 0, 0x40, 0x1f, 0, 0]; //returning 2 frequencies
          } else {
            // SWD
            computedResponse = [
              0x80,
              0,
              0,
              0,
              0x40,
              0x1f,
              0,
              0,
              3,
              0,
              0,
              0,
              0xe4,
              0x0c,
              0,
              0,
              0xe8,
              3,
              0,
              0,
              0xc8,
              0,
              0,
              0
            ]; //returning 3 frequencies
          }
          break;
        case 0x4b: // STLINK_INIT_AP
          computedResponse = [0x80];
          break;

        case 0x4c: // STLINK_CLOSE_AP
          computedResponse = [0x80];
          break;

        case 0x49: // STLINK_JTAG_BLINK_LED
          if (options === "") {
            return stLinkDefs.DRV_OK;
          } else {
            return stLinkDefs.DRV_CMD_NOT_AVAILABLE;
          }

        default:
          unknownSendCommand = true;
          break;
      }
      break;

    case 0xf1: // STLINK_CMD_GET_VERSION
      switch (options) {
        case "GET_VERSION_RETURNS_ERROR":
          return stLinkDefs.DRV_ERR;

        case "UNSUPPORTED_FIRMWARE_VERSION":
          computedResponse = [0x0, 0x22]; // // This value is lower than allowed JTAG version (and then the probe does not support the JTAG protocol)
          break;

        case "STLINK_V3_VERSION":
        case "STLINK_V3_NO_JTAG":
        case "GET_EXT_VERSION_RETURNS_ERROR":
          computedResponse = [0x30, 0x00];
          break;

        default:
          computedResponse = [0x0c, 0x22]; // Random value (can be another value)
          break;
      }
      break;

    case 0xfb: //STLINK_CMD_GET_VERSION_EXT
      switch (options) {
        case "GET_EXT_VERSION_RETURNS_ERROR":
          return stLinkDefs.DRV_ERR;
        case "STLINK_V3_NO_JTAG":
          computedResponse = [0x03, 0x00, 0x00, 0x00, 0x03];
          break;
        default:
          computedResponse = [0x03, 0x01, 0x04, 0x02, 0x03];
          break;
      }
      break;

    case 0xf5: // STLINK_CMD_GET_CURRENT_MODE
      switch (options) {
        case "GET_CURRENT_MODE_RETURNS_ERROR":
          return stLinkDefs.DRV_FAILED;
        case "GET_CURRENT_MODE_FAILS":
          computedResponse = [0x10, 0x10]; // First byte is a random value (different of STLINK_DEV_JTAG_MODE)
          break;
        case "SET_DFU_MODE":
        case "EXIT_DFU_MODE_RETURNS_ERROR":
          computedResponse = [0x00, 0x10]; // First byte is STLINK_DEV_DFU_MODE
          break;
        case "SET_MASS_MODE":
          computedResponse = [0x01, 0x10]; // First byte is STLINK_DEV_MASS_MODE
          break;
        case "SET_JTAG_MODE":
          computedResponse = [0x02, 0x10]; // First byte is STLINK_DEV_JTAG_MODE
          break;
        case "SET_SWIM_MODE":
          computedResponse = [0x03, 0x10]; // First byte is STLINK_DEV_SWIM_MODE
          break;
        case "SET_BTLD_MODE":
          computedResponse = [0x04, 0x10]; // First byte is STLINK_DEV_BTLD_MODE
          break;
        default:
          computedResponse = [0x02, 0x10]; // Default mode is STLINK_DEV_JTAG_MODE
          break;
      }
      break;

    case 0xf3: // STLINK_CMD_DFU
      switch (sendCommandSub) {
        case 0x07: // STLINK_DFU_EXIT
          if (options === "EXIT_DFU_MODE_RETURNS_ERROR") {
            return stLinkDefs.DRV_ERR;
          } else {
            return stLinkDefs.DRV_OK;
          }

        default:
          unknownSendCommand = true;
          break;
      }
      break;

    default:
      unknownSendCommand = true;
      break;
  }

  if (unknownSendCommand === true) {
    console.error(
      "Unknown SendCommand TCP - sendCommandMeta = " +
        sendCommandMeta.toString(16).padStart(16, "0") +
        " - sendCommandSub = " +
        sendCommandSub.toString(16).padStart(16, "0")
    );
    return stLinkDefs.DRV_FAILED;
  }

  // Write the expected response in response buffer
  let addrInBytesArray = [
    reqRef[18],
    reqRef[19],
    reqRef[20],
    reqRef[21],
    reqRef[22],
    reqRef[23],
    reqRef[24],
    reqRef[25]
  ];
  writeValueInRespBuffer(addrInBytesArray, computedResponse);
  return stLinkDefs.DRV_OK;
}

module.exports.STLink_GetDeviceInfo2Mocked = STLink_GetDeviceInfo2Mocked;
module.exports.STLink_SendCommandSetOptions = STLink_SendCommandSetOptions;
module.exports.STLink_SendCommandMocked = STLink_SendCommandMocked;
module.exports.STLink_SendCommandTcpMocked = STLink_SendCommandTcpMocked;
