/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */

const expect = require("chai").expect;
const sinon = require("sinon");
const STLogger = require("@stm32/stm32cubemonitor-logger");
const logger = STLogger.getLogger("snapshot");

/**
 * Tested module.
 */
//stnapshot.js

/**
 * Module(s) mocked with sinonjs.
 */
const stLinkProto = require("../src/stlinkprotocol");

/**
 * Provided definitions required for the assert.
 */
const apidefs = require("../src/stlinkprotocol").apidefs;

/**
 * Constructor of snapshot.
 */
const Snapshot = require("../src/snapshot").Snapshot;
const targetType = require("../src/snapshot").targetType;
const SNAP_ERR = require("../src/snapshot").SNAP_ERR;
const targetTypeMap = require("../src/snapshot").targetTypeMap;

/**
 * Tests Scenarios.
 */

describe("#buildHeader()", function() {
  let header = [
    { name: "startMark[0]", address: "0x200008c4", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerFieldSize", address: "0x200008c7", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerVersion", address: "0x200008c8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferStartAddr", address: "0x200008cc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferEndAddr", address: "0x200008d0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVarMax", address: "0x200008d4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferFormat", address: "0x200008d8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "timestamp_base_unit", address: "0x200008dc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointer", address: "0x200008e0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointerCopy", address: "0x200008e4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointer", address: "0x200008e8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointerCopy", address: "0x200008ec", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "flags", address: "0x200008f0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "subSamplingRate", address: "0x200008f4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVar", address: "0x200008f8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].accessType", address: "0x200008fc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].address", address: "0x20000900", type: apidefs.READ_TYPE_U32, nb: 1 }
  ];

  let myAcqParam = {
    snapshotheader: "0x200008c4",
    triggeraddress: "0x20000000",
    triggerthreshold: "10",
    triggerstop: "manual"
  };

  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  beforeEach(function() {
    logger.info("#buildHeader()");
  });

  it("should work if the buildHeader works", function() {
    logger.info("it should work if the buildHeader works()");
    const snap = new Snapshot(probe.probeId); //rewire("../src/snapshot");
    snap.buildHeader(myAcqParam.snapshotheader, 4); //get builHeader private function

    let myHeader = snap.header;

    expect(myHeader).to.deep.equal(header);
  });
});

describe("#getTargetType()", function() {
  beforeEach(function() {
    logger.info("#getTargetType()");
  });

  it("should work if the getTargetType works", function() {
    logger.info("it should work if the getTargetType works()");

    // U8
    let myTargetType = targetType[apidefs.READ_TYPE_U8];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_8BIT);

    // S8
    myTargetType = targetType[apidefs.READ_TYPE_S8];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_8BIT);

    // U16
    myTargetType = targetType[apidefs.READ_TYPE_U16];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_16BIT);

    // S16
    myTargetType = targetType[apidefs.READ_TYPE_S16];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_16BIT);

    // U32
    myTargetType = targetType[apidefs.READ_TYPE_U32];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_32BIT);

    // S32
    myTargetType = targetType[apidefs.READ_TYPE_S32];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_32BIT);

    // S32
    myTargetType = targetType[apidefs.READ_TYPE_FLOAT];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_32BIT);

    // U64
    myTargetType = targetType[apidefs.READ_TYPE_U64];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_64BIT);

    // S64
    myTargetType = targetType[apidefs.READ_TYPE_S64];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_64BIT);

    // DOUBLE
    myTargetType = targetType[apidefs.READ_TYPE_DOUBLE];
    expect(myTargetType).to.deep.equal(targetTypeMap.SNP_TRC_DATA_ACCESS_64BIT);
  });
});

describe("#readHeader()", function() {
  let stlinkprotocolMock;
  let header = [
    { name: "startMark[0]", address: "0x200008c4", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerFieldSize", address: "0x200008c7", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerVersion", address: "0x200008c8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferStartAddr", address: "0x200008cc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferEndAddr", address: "0x200008d0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVarMax", address: "0x200008d4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferFormat", address: "0x200008d8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "timestamp_base_unit", address: "0x200008dc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointer", address: "0x200008e0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointerCopy", address: "0x200008e4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointer", address: "0x200008e8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointerCopy", address: "0x200008ec", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "flags", address: "0x200008f0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "subSamplingRate", address: "0x200008f4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVar", address: "0x200008f8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].accessType", address: "0x200008fc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].address", address: "0x20000900", type: apidefs.READ_TYPE_U32, nb: 1 }
  ];

  let headerResponse = [];
  headerResponse.push(
    [{ x: "startMark[0]", y: "0x008f25a7" }],
    [{ x: "headerFieldSize", y: "0x00000004" }],
    [{ x: "headerVersion", y: "0x00000608" }],
    [{ x: "bufferStartAddr", y: "0x20000058" }],
    [{ x: "bufferEndAddr", y: "0x200008ef" }],
    [{ x: "nbVarMax", y: "0x0000000a" }],
    [{ x: "bufferFormat", y: "0x00000001" }],
    [{ x: "timestamp_base_unit", y: "0x000040a0" }],
    [{ x: "writePointer", y: "0x20000058" }],
    [{ x: "writePointerCopy", y: "0x20000058" }],
    [{ x: "readPointer", y: "0x200008b8" }],
    [{ x: "readPointerCopy", y: "0x200008b8" }],
    [{ x: "flags", y: "0x00000000" }],
    [{ x: "subSamplingRate", y: "0x00000001" }],
    [{ x: "nbVar", y: "0x00001006" }],
    [{ x: "g_varList[0].accessType", y: "" }],
    [{ x: "g_varList[0].address", y: "" }]
  );

  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let expectedReadObject = { result: 1, data: headerResponse };

  beforeEach(function() {
    logger.info("#readHeader()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.restore();
  });

  it("should work if the readHeader works", function() {
    logger.info("it should work if the readHeader works()");

    const snap = new Snapshot(probe.probeId);
    snap.header = header;

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .returns(expectedReadObject);

    //myReadHeader();
    snap.readHeader();

    //bufferStartAddrStrHexa
    let bufferStartAddrStrHexa = snap.h_bufferStartAddrStrHex;
    expect(bufferStartAddrStrHexa).to.deep.equal(headerResponse[3][0].y);

    //bufferEndAddrStrHexa
    let bufferEndAddrStrHexa = snap.h_bufferEndAddrStrHex;
    expect(bufferEndAddrStrHexa).to.deep.equal(headerResponse[4][0].y);

    //nbVar
    let nbVar = snap.h_nbVar;
    expect(nbVar).to.deep.equal(headerResponse[14][0].y);

    //flags
    let flags = snap.h_flags;
    expect(flags).to.deep.equal(headerResponse[12][0].y);

    //timestampValue
    let timestampValue = snap.h_timestampValue;
    expect(timestampValue).to.deep.equal(headerResponse[7][0].y);

    //readPointer
    let readPointer = snap.h_readPointer;
    expect(readPointer).to.deep.equal(headerResponse[10][0].y);

    //readPointerCopyAddStrHex
    let readPointerCopyAddStrHex = snap.h_readPointerCopyAddStrHex;
    expect(readPointerCopyAddStrHex).to.deep.equal(header[11].address);

    //writePointer
    let writePointer = snap.h_writePointer;
    expect(writePointer).to.deep.equal(headerResponse[8][0].y);

    //writePointerC
    let writePointerC = snap.h_writePointerC;
    expect(writePointerC).to.deep.equal(headerResponse[9][0].y);

    //bufferSize
    let bufferSize = snap.h_bufferSize;
    expect(bufferSize).to.deep.equal(
      (parseInt(bufferEndAddrStrHexa, 16) - parseInt(bufferStartAddrStrHexa, 16) + 1) / 4
    );

    //listOfTypeStrHex
    let listOfTypeStrHex = snap.h_listOfTypeStrHex;
    expect(listOfTypeStrHex).to.deep.equal(header[15].address);

    //listOfVarStrHex
    let listOfVarStrHex = snap.h_listOfVarStrHex;
    expect(listOfVarStrHex).to.deep.equal(header[16].address);
  });
});

describe("#initTarget()", function() {
  let stlinkprotocolMock;
  let expectedReadObject = { result: 1 };
  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let myGroup = [
    {
      address: "0x12345678",
      type: apidefs.READ_TYPE_U8,
      nb: 4
    },
    {
      address: "0x12340000",
      type: apidefs.READ_TYPE_S64
    },
    {
      address: "0x2345000",
      type: apidefs.READ_TYPE_S32,
      nb: 3
    },
    {
      address: "0x12345000",
      type: apidefs.READ_TYPE_S8
    },
    {
      address: "0x123400FF",
      type: apidefs.READ_TYPE_U64
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U32
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U16
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_S16
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_DOUBLE
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_FLOAT
    }
  ];

  beforeEach(function() {
    logger.info("#initTarget()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.verify();
    stlinkprotocolMock.restore();
  });

  it("should work if the initTarget works", function() {
    logger.info("it should work if the initTarget works()");

    const snap = new Snapshot(probe.probeId);

    snap.headerSize = 4;
    snap.VarsToBeRead = myGroup;
    snap.h_listOfTypeStrHex = "0x200008fc";
    snap.h_listOfVarStrHex = "0x20000900";
    snap.h_writePointer = parseInt("0x20000058", 16);
    snap.h_readPointer = parseInt("0x20000058", 16);
    snap.h_bufferStartAddrStrHex = "0x20000058";
    snap.h_triggerAccessType = "0x05";
    snap.h_triggerThreshold = "0x00";

    //Triger variable
    snap.valOfTriggerMode = "manual";

    let bufferSize = 50 * (myGroup.length + 1);
    snap.h_bufferSize = bufferSize;
    let recordSize;

    // update Readpointer according to expected value.
    Math.trunc((recordSize = 4 * (1 + Number(myGroup.length) + 3))); // it is the nb of octets by record

    let nbRecords = Math.trunc((bufferSize * 4) / recordSize); // nb of records that buffer can save, records is one acquisition, this acquisition is composed of 1 timestamp + varaibles that user want to trace.
    let myExpectedInitialReadPointer = snap.h_writePointer + (nbRecords - 1) * recordSize;

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedReadObject);

    snap.initTarget();
    expect(snap.expectedInitialReadPointer).to.deep.equal(myExpectedInitialReadPointer);
  });

  it("should work if the initTarget works with tile unit = millisecond", function() {
    logger.info("it should work if the initTarget works() with tile unit = millisecond");

    const snap = new Snapshot(probe.probeId);

    snap.headerSize = 4;
    snap.VarsToBeRead = myGroup;
    snap.h_listOfTypeStrHex = "0x200008fc";
    snap.h_listOfVarStrHex = "0x20000900";
    snap.h_writePointer = parseInt("0x20000058", 16);
    snap.h_readPointer = parseInt("0x20000058", 16);
    snap.h_bufferStartAddrStrHex = "0x20000058";
    snap.h_triggerAccessType = "0x05";
    snap.h_triggerThreshold = "0x00";

    //Triger variable
    snap.valOfTriggerMode = "manual";
    snap.h_timestampValue = 0x4000;
    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedReadObject);

    snap.initTarget();
    expect(snap.unitOfTimestamp).to.deep.equal(1);
  });

  it("should work if the initTarget works with tile unit = second", function() {
    logger.info("it should work if the initTarget works() with tile unit = second");

    const snap = new Snapshot(probe.probeId);

    snap.headerSize = 4;
    snap.VarsToBeRead = myGroup;
    snap.h_listOfTypeStrHex = "0x200008fc";
    snap.h_listOfVarStrHex = "0x20000900";
    snap.h_writePointer = parseInt("0x20000058", 16);
    snap.h_readPointer = parseInt("0x20000058", 16);
    snap.h_bufferStartAddrStrHex = "0x20000058";
    snap.h_triggerAccessType = "0x05";
    snap.h_triggerThreshold = "0x00";

    //Triger variable
    snap.valOfTriggerMode = "manual";
    snap.h_timestampValue = 0x5000;
    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedReadObject);

    snap.initTarget();
    expect(snap.unitOfTimestamp).to.deep.equal(10);
  });

  it("should work if the initTarget works with tile unit = microsecond", function() {
    logger.info("it should work if the initTarget works() with tile unit = microsecond");

    const snap = new Snapshot(probe.probeId);

    snap.headerSize = 4;
    snap.VarsToBeRead = myGroup;
    snap.h_listOfTypeStrHex = "0x200008fc";
    snap.h_listOfVarStrHex = "0x20000900";
    snap.h_writePointer = parseInt("0x20000058", 16);
    snap.h_readPointer = parseInt("0x20000058", 16);
    snap.h_bufferStartAddrStrHex = "0x20000058";
    snap.h_triggerAccessType = "0x05";
    snap.h_triggerThreshold = "0x00";

    //Triger variable
    snap.valOfTriggerMode = "manual";
    snap.h_timestampValue = 0x3000;
    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedReadObject);

    snap.initTarget();
    expect(snap.unitOfTimestamp).to.deep.equal(0.1);
  });

  it("should work if the initTarget works with tile unit = nanosecond", function() {
    logger.info("it should work if the initTarget works() with tile unit = nanosecond");

    const snap = new Snapshot(probe.probeId);

    snap.headerSize = 4;
    snap.VarsToBeRead = myGroup;
    snap.h_listOfTypeStrHex = "0x200008fc";
    snap.h_listOfVarStrHex = "0x20000900";
    snap.h_writePointer = parseInt("0x20000058", 16);
    snap.h_readPointer = parseInt("0x20000058", 16);
    snap.h_bufferStartAddrStrHex = "0x20000058";
    snap.h_triggerAccessType = "0x05";
    snap.h_triggerThreshold = "0x00";

    //Triger variable
    snap.valOfTriggerMode = "manual";
    snap.h_timestampValue = 0x2000;
    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedReadObject);

    snap.initTarget();
    expect(snap.unitOfTimestamp).to.deep.equal(0.01);
  });
});

describe("#readBuffer()", function() {
  let stlinkprotocolMock;
  let header = [
    { name: "startMark[0]", address: "0x200008c4", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerFieldSize", address: "0x200008c7", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerVersion", address: "0x200008c8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferStartAddr", address: "0x200008cc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferEndAddr", address: "0x200008d0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVarMax", address: "0x200008d4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferFormat", address: "0x200008d8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "timestamp_base_unit", address: "0x200008dc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointer", address: "0x200008e0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointerCopy", address: "0x200008e4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointer", address: "0x200008e8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointerCopy", address: "0x200008ec", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "flags", address: "0x200008f0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "subSamplingRate", address: "0x200008f4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVar", address: "0x200008f8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].accessType", address: "0x200008fc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].address", address: "0x20000900", type: apidefs.READ_TYPE_U32, nb: 1 }
  ];

  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let myGroup = [
    {
      address: "0x12345678",
      type: apidefs.READ_TYPE_U8,
      nb: 4
    },
    {
      address: "0x12340000",
      type: apidefs.READ_TYPE_S64
    },
    {
      address: "0x2345000",
      type: apidefs.READ_TYPE_S32,
      nb: 3
    },
    {
      address: "0x12345000",
      type: apidefs.READ_TYPE_S8
    },
    {
      address: "0x123400FF",
      type: apidefs.READ_TYPE_U64
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U32
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U16
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_S16
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_DOUBLE
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_FLOAT
    }
  ];

  beforeEach(function() {
    logger.info("#initTarget()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.restore();
  });

  it("should work if the readBuffer with readPointerEnd undefined works", function() {
    logger.info("it should work if the readBuffer with readPointerEnd undefined works()");

    let expectedWriteResObject = { result: 1 };
    let expectedReadData = [];
    expectedReadData.push(
      [{ x: "0", y: parseInt("0x008f25a7", 16) }],
      [{ x: "0", y: parseInt("0x00000004", 16) }],
      [{ x: "0", y: parseInt("0x00000608", 16) }],
      [{ x: "0", y: parseInt("0x20000058", 16) }],
      [{ x: "0", y: parseInt("0x200008ef", 16) }],
      [{ x: "0", y: parseInt("0x0000000a", 16) }],
      [{ x: "0", y: parseInt("0x00000001", 16) }],
      [{ x: "0", y: parseInt("0x000040a0", 16) }],
      [{ x: "0", y: parseInt("0x20000058", 16) }],
      [{ x: "0", y: parseInt("0x20000058", 16) }],
      [{ x: "0", y: parseInt("0x200008b8", 16) }],
      [{ x: "0", y: parseInt("0x200008b8", 16) }],
      [{ x: "0", y: parseInt("0x00000000", 16) }],
      [{ x: "0", y: parseInt("0x00000001", 16) }],
      [{ x: "0", y: parseInt("0x00001006", 16) }],
      [{ x: "0", y: parseInt("") }],
      [{ x: "0", y: parseInt("") }]
    );
    let expectedReadObject = { result: 1, data: expectedReadData };

    const snap = new Snapshot(probe.probeId);
    snap.header = header;

    snap.h_readPointerEnd = undefined;

    // Buffersize
    let bufferSize = 50 * (myGroup.length + 1);
    snap.h_bufferSize = bufferSize;

    // Nbr of var for readding
    snap.VarsToBeRead = myGroup;

    // update Readpointer according to expected value.
    let recordSize;
    Math.trunc((recordSize = 4 * (1 + Number(myGroup.length) + 3))); // it is the nb of octets by record
    snap.recordSize = recordSize;

    //Triger variable
    snap.valOfTriggerMode = "manual";

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedWriteResObject);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .returns(expectedReadObject);

    //launch function with writePointerC == bufferStartAddrStrHexa there is no recording
    snap.readBuffer();

    let myReadPointerEnd;
    myReadPointerEnd = undefined;
    expect(snap.h_readPointerEnd).to.deep.equal(myReadPointerEnd);
  });

  it("should work if the readBuffer with readPointerEnd undefined and nbRecords > 0  works", function() {
    logger.info("it should work if the readBuffer with readPointerEnd undefined and nbRecords > 0 works()");

    let expectedWriteResObject = { result: 1 };
    let expectedReadData = [];
    expectedReadData.push(
      [{ x: "0", y: parseInt("0x008f25a7") }],
      [{ x: "0", y: parseInt("0x00000004") }],
      [{ x: "0", y: parseInt("0x00000608") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200008ef") }],
      [{ x: "0", y: parseInt("0x0000000a") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x000040a0") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x200008b8") }],
      [{ x: "0", y: parseInt("0x200008b8") }],
      [{ x: "0", y: parseInt("0x00000000") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x00001006") }],
      [{ x: "0", y: parseInt("") }],
      [{ x: "0", y: parseInt("") }]
    );
    let expectedReadObject = { result: 1, data: expectedReadData };

    let readPointer = [
      {
        address: "0x200008e8",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readPointerC = [
      {
        address: "0x200008ec",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readBuffer = [
      {
        address: "0x20000058",
        type: apidefs.READ_TYPE_U32,
        nb: 28
      }
    ];

    let expectedBuffervValuedData = [];
    expectedBuffervValuedData.push([
      { x: "0", y: "10" },
      { x: "1", y: "20" },
      { x: "2", y: "30" },
      { x: "3", y: "16" },
      { x: "4", y: "16" },
      { x: "5", y: "16" },
      { x: "6", y: "16" },
      { x: "7", y: "10" },
      { x: "8", y: "10" },
      { x: "9", y: "10" },
      { x: "10", y: "10" },
      { x: "11", y: "10" },
      { x: "12", y: "10" },
      { x: "13", y: "10" },
      { x: "14", y: "10" },
      { x: "15", y: "10" },
      { x: "16", y: "10" },
      { x: "17", y: "10" },
      { x: "18", y: "10" },
      { x: "19", y: "10" },
      { x: "20", y: "10" },
      { x: "21", y: "10" },
      { x: "22", y: "10" },
      { x: "23", y: "10" },
      { x: "24", y: "10" },
      { x: "25", y: "10" },
      { x: "26", y: "10" },
      { x: "27", y: "10" }
    ]);
    let expectedBufferValueObject = { result: 1, data: expectedBuffervValuedData };

    const snap = new Snapshot(probe.probeId);
    snap.header = header;

    snap.h_readPointerEnd = undefined;

    // Buffersize
    let bufferSize = 50 * (myGroup.length + 1);
    snap.h_bufferSize = bufferSize;

    // Nbr of var for readding
    snap.VarsToBeRead = myGroup;

    //Triger variable
    snap.valOfTriggerMode = "manual";

    // update Readpointer according to expected value.
    let recordSize;
    Math.trunc((recordSize = 4 * (1 + Number(myGroup.length) + 3))); // it is the nb of octets by record
    snap.recordSize = recordSize;

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedWriteResObject);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointer)
      .returns(expectedWriteResObject);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, readBuffer)
      .returns(expectedBufferValueObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointerC)
      .returns(expectedWriteResObject);

    snap.readBuffer();

    let myReadPointerEnd;
    let add = parseInt(snap.h_bufferStartAddrStrHex, 16);
    let nbRecords = Math.trunc((snap.h_writePointerC - add) / recordSize);
    myReadPointerEnd = add + nbRecords * recordSize;
    expect(snap.h_readPointerEnd).to.deep.equal(myReadPointerEnd);
  });

  it("should work if the readBuffer with readPointerEnd <  writePointerC works", function() {
    logger.info("it should work if the readBuffer with readPointerEnd <  writePointerC works()");

    let expectedWriteResObject = { result: 1 };
    let expectedReadData = [];
    expectedReadData.push(
      [{ x: "0", y: parseInt("0x008f25a7") }],
      [{ x: "0", y: parseInt("0x00000004") }],
      [{ x: "0", y: parseInt("0x00000608") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200008ef") }],
      [{ x: "0", y: parseInt("0x0000000a") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x000040a0") }],
      [{ x: "0", y: parseInt("0x200000e4") }],
      [{ x: "0", y: parseInt("0x200000e4") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x00000000") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x00001006") }],
      [{ x: "0", y: parseInt("") }],
      [{ x: "0", y: parseInt("") }]
    );
    let expectedReadObject = { result: 1, data: expectedReadData };

    let readPointer = [
      {
        address: "0x200008e8",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readPointerC = [
      {
        address: "0x200008ec",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readBuffer = [
      {
        address: "0x20000058",
        type: apidefs.READ_TYPE_U32,
        nb: 28
      }
    ];

    let expectedBuffervValuedData = [];
    expectedBuffervValuedData.push([
      { x: "0", y: "10" },
      { x: "1", y: "20" },
      { x: "2", y: "30" },
      { x: "3", y: "16" },
      { x: "4", y: "16" },
      { x: "5", y: "16" },
      { x: "6", y: "16" },
      { x: "7", y: "10" },
      { x: "8", y: "10" },
      { x: "9", y: "10" },
      { x: "10", y: "10" },
      { x: "11", y: "10" },
      { x: "12", y: "10" },
      { x: "13", y: "10" },
      { x: "14", y: "10" },
      { x: "15", y: "10" },
      { x: "16", y: "10" },
      { x: "17", y: "10" },
      { x: "18", y: "10" },
      { x: "19", y: "10" },
      { x: "20", y: "10" },
      { x: "21", y: "10" },
      { x: "22", y: "10" },
      { x: "23", y: "10" },
      { x: "24", y: "10" },
      { x: "25", y: "10" },
      { x: "26", y: "10" },
      { x: "27", y: "10" }
    ]);
    let expectedBufferValueObject = { result: 1, data: expectedBuffervValuedData };

    const snap = new Snapshot(probe.probeId);
    snap.header = header;

    snap.readPointerEnd = undefined;

    // Buffersize
    let bufferSize = 50 * (myGroup.length + 1);
    snap.bufferSize = bufferSize;

    // Nbr of var for readding
    snap.VarsToBeRead = myGroup;

    //Triger variable
    snap.valOfTriggerMode = "manual";

    // update Readpointer according to expected value.
    let recordSize;
    Math.trunc((recordSize = 4 * (1 + Number(myGroup.length) + 3))); // it is the nb of octets by record
    snap.recordSize = recordSize;

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedWriteResObject);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointer)
      .returns(expectedWriteResObject);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, readBuffer)
      .returns(expectedBufferValueObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointerC)
      .returns(expectedWriteResObject);

    let myReadPointerEnd = 536871112;
    snap.readPointerEnd = myReadPointerEnd;

    snap.readBuffer();

    let add = myReadPointerEnd;
    let nbRecords = Math.trunc((snap.h_writePointerC - add) / recordSize);
    myReadPointerEnd = add + nbRecords * recordSize;
    expect(snap.h_readPointerEnd).to.deep.equal(myReadPointerEnd);
  });

  it("should work if the readBuffer with readPointerEnd == expectedInitialReadPointer + recordSize works", function() {
    logger.info(
      "it should work if the readBuffer with readPointerEnd == expectedInitialReadPointer + recordSize works()"
    );

    let expectedWriteResObject = { result: 1 };
    let expectedReadData = [];
    expectedReadData.push(
      [{ x: "0", y: parseInt("0x008f25a7") }],
      [{ x: "0", y: parseInt("0x00000004") }],
      [{ x: "0", y: parseInt("0x00000608") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200008ef") }],
      [{ x: "0", y: parseInt("0x0000000a") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x000040a0") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x00000000") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x00001006") }],
      [{ x: "0", y: parseInt("") }],
      [{ x: "0", y: parseInt("") }]
    );
    let expectedReadObject = { result: 1, data: expectedReadData };

    let readPointer = [
      {
        address: "0x200008e8",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readPointerC = [
      {
        address: "0x200008ec",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readBuffer = [
      {
        address: "0x20000058",
        type: apidefs.READ_TYPE_U32,
        nb: 28
      }
    ];

    let expectedBuffervValuedData = [];
    expectedBuffervValuedData.push([
      { x: "0", y: "10" },
      { x: "1", y: "20" },
      { x: "2", y: "30" },
      { x: "3", y: "16" },
      { x: "4", y: "16" },
      { x: "5", y: "16" },
      { x: "6", y: "16" },
      { x: "7", y: "10" },
      { x: "8", y: "10" },
      { x: "9", y: "10" },
      { x: "10", y: "10" },
      { x: "11", y: "10" },
      { x: "12", y: "10" },
      { x: "13", y: "10" },
      { x: "14", y: "10" },
      { x: "15", y: "10" },
      { x: "16", y: "10" },
      { x: "17", y: "10" },
      { x: "18", y: "10" },
      { x: "19", y: "10" },
      { x: "20", y: "10" },
      { x: "21", y: "10" },
      { x: "22", y: "10" },
      { x: "23", y: "10" },
      { x: "24", y: "10" },
      { x: "25", y: "10" },
      { x: "26", y: "10" },
      { x: "27", y: "10" }
    ]);
    let expectedBufferValueObject = { result: 1, data: expectedBuffervValuedData };

    const snap = new Snapshot(probe.probeId);
    snap.header = header;

    // Buffersize
    let bufferSize = 50 * (myGroup.length + 1);
    snap.h_bufferSize = bufferSize;

    // Nbr of var for readding
    snap.VarsToBeRead = myGroup;

    //Triger variable
    snap.valOfTriggerMode = "manual";

    // update Readpointer according to expected value.
    let recordSize;
    Math.trunc((recordSize = 4 * (1 + Number(myGroup.length) + 3))); // it is the nb of octets by record
    snap.recordSize = recordSize;

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedWriteResObject);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointer)
      .returns(expectedWriteResObject);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, readBuffer)
      .returns(expectedBufferValueObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointerC)
      .returns(expectedWriteResObject);

    let myReadPointerEnd = parseInt("0x200008ef", 16); //536873160‬

    snap.h_readPointerEnd = myReadPointerEnd + recordSize;
    snap.expectedInitialReadPointer = myReadPointerEnd;

    snap.readBuffer();

    myReadPointerEnd = parseInt(snap.h_bufferStartAddrStrHex, 16);

    expect(snap.h_readPointerEnd).to.deep.equal(myReadPointerEnd);
  });

  it("should work if the readBuffer with readPointerEnd > expectedInitialReadPointer + recordSize works", function() {
    logger.info(
      "it should work if the readBuffer with readPointerEnd > expectedInitialReadPointer + recordSize works()"
    );

    let expectedWriteResObject = { result: 1 };
    let expectedReadData = [];
    expectedReadData.push(
      [{ x: "0", y: parseInt("0x008f25a7") }],
      [{ x: "0", y: parseInt("0x00000004") }],
      [{ x: "0", y: parseInt("0x00000608") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200008ef") }],
      [{ x: "0", y: parseInt("0x0000000a") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x000040a0") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x00000000") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x00001006") }],
      [{ x: "0", y: parseInt("") }],
      [{ x: "0", y: parseInt("") }]
    );
    let expectedReadObject = { result: 1, data: expectedReadData };

    let readPointer = [
      {
        address: "0x200008e8",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readPointerC = [
      {
        address: "0x200008ec",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readBuffer = [
      {
        address: "0x20000058",
        type: apidefs.READ_TYPE_U32,
        nb: 28
      }
    ];

    let expectedBuffervValuedData = [];
    expectedBuffervValuedData.push([
      { x: "0", y: "10" },
      { x: "1", y: "20" },
      { x: "2", y: "30" },
      { x: "3", y: "16" },
      { x: "4", y: "16" },
      { x: "5", y: "16" },
      { x: "6", y: "16" },
      { x: "7", y: "10" },
      { x: "8", y: "10" },
      { x: "9", y: "10" },
      { x: "10", y: "10" },
      { x: "11", y: "10" },
      { x: "12", y: "10" },
      { x: "13", y: "10" },
      { x: "14", y: "10" },
      { x: "15", y: "10" },
      { x: "16", y: "10" },
      { x: "17", y: "10" },
      { x: "18", y: "10" },
      { x: "19", y: "10" },
      { x: "20", y: "10" },
      { x: "21", y: "10" },
      { x: "22", y: "10" },
      { x: "23", y: "10" },
      { x: "24", y: "10" },
      { x: "25", y: "10" },
      { x: "26", y: "10" },
      { x: "27", y: "10" }
    ]);
    let expectedBufferValueObject = { result: 1, data: expectedBuffervValuedData };

    const snap = new Snapshot(probe.probeId);
    snap.header = header;

    // Buffersize
    let bufferSize = 50 * (myGroup.length + 1);
    snap.bufferSize = bufferSize;

    // Nbr of var for readding
    snap.VarsToBeRead = myGroup;

    //Triger variable
    snap.valOfTriggerMode = "manual";

    // update Readpointer according to expected value.
    let recordSize;
    Math.trunc((recordSize = 4 * (1 + Number(myGroup.length) + 3))); // it is the nb of octets by record
    snap.recordSize = recordSize;

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedWriteResObject);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointer)
      .returns(expectedWriteResObject);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, readBuffer)
      .returns(expectedBufferValueObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointerC)
      .returns(expectedWriteResObject);

    stlinkprotocolMock.expects("readMemoryGroup").returns(expectedReadObject);

    let myReadPointerEnd = parseInt("0x200008ef", 16) + 1;

    snap.h_readPointerEnd = myReadPointerEnd + recordSize + 1;
    snap.expectedInitialReadPointer = myReadPointerEnd;

    myReadPointerEnd = undefined;

    snap.readBuffer();

    expect(snap.h_readPointerEnd).to.deep.equal(myReadPointerEnd);
  });

  it("should work if the readBuffer with readPointerEnd > writePointerC or flags == OVERFLOW works", function() {
    logger.info("it should work if the readBuffer with readPointerEnd > writePointerC or flags == OVERFLOW works()");

    let expectedWriteResObject = { result: 1 };
    let expectedReadData = [];
    expectedReadData.push(
      [{ x: "0", y: parseInt("0x008f25a7") }],
      [{ x: "0", y: parseInt("0x00000004") }],
      [{ x: "0", y: parseInt("0x00000608") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200008ef") }],
      [{ x: "0", y: parseInt("0x0000000a") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x000040a0") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x20000058") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x200000c8") }],
      [{ x: "0", y: parseInt("0x00000000") }],
      [{ x: "0", y: parseInt("0x00000001") }],
      [{ x: "0", y: parseInt("0x00001006") }],
      [{ x: "0", y: parseInt("0x00001006") }],
      [{ x: "0", y: parseInt("0x00001006") }]
    );
    let expectedReadObject = { result: 1, data: expectedReadData };

    let readPointer = [
      {
        address: "0x200008e8",
        type: apidefs.READ_TYPE_U32,
        value: "0x20000090"
      }
    ];

    let readPointerC = [
      {
        address: "0x200008ec",
        type: apidefs.READ_TYPE_U32,
        value: "0x200008e1"
      }
    ];

    let readBuffer = [
      {
        address: "0x200008a9",
        type: apidefs.READ_TYPE_U32,
        nb: 28
      }
    ];

    let expectedBuffervValuedData = [];
    expectedBuffervValuedData.push([
      { x: "0", y: "10" },
      { x: "1", y: "20" },
      { x: "2", y: "30" },
      { x: "3", y: "16" },
      { x: "4", y: "16" },
      { x: "5", y: "16" },
      { x: "6", y: "16" },
      { x: "7", y: "10" },
      { x: "8", y: "10" },
      { x: "9", y: "10" },
      { x: "10", y: "10" },
      { x: "11", y: "10" },
      { x: "12", y: "10" },
      { x: "13", y: "10" },
      { x: "14", y: "10" },
      { x: "15", y: "10" },
      { x: "16", y: "10" },
      { x: "17", y: "10" },
      { x: "18", y: "10" },
      { x: "19", y: "10" },
      { x: "20", y: "10" },
      { x: "21", y: "10" },
      { x: "22", y: "10" },
      { x: "23", y: "10" },
      { x: "24", y: "10" },
      { x: "25", y: "10" },
      { x: "26", y: "10" },
      { x: "27", y: "10" }
    ]);
    let expectedBufferValueObject = { result: 1, data: expectedBuffervValuedData };

    const snap = new Snapshot(probe.probeId);

    snap.header = header;

    // Buffersize
    let bufferSize = 50 * (myGroup.length + 1);
    snap.h_bufferSize = bufferSize;

    // Nbr of var for readding
    snap.VarsToBeRead = myGroup;

    //Triger variable
    snap.valOfTriggerMode = "manual";

    // update Readpointer according to expected value.
    let recordSize;
    Math.trunc((recordSize = 4 * (1 + Number(myGroup.length) + 3))); // it is the nb of octets by record
    snap.recordSize = recordSize;

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedWriteResObject);
    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointer)
      .returns(expectedWriteResObject);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, readBuffer)
      .returns(expectedBufferValueObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, readPointerC)
      .returns(expectedWriteResObject);

    stlinkprotocolMock.expects("readMemoryGroup").returns(expectedReadObject);

    let myReadPointerEnd = parseInt("0x200008e1", 16) - recordSize; // it is 2 records on the end od buffer
    let expectedInitialReadPointer = parseInt("0x200008e1", 16);

    snap.h_readPointerEnd = myReadPointerEnd;
    snap.readBuffer();

    let add = myReadPointerEnd;
    let nbRecords = 1 + Math.trunc((expectedInitialReadPointer - myReadPointerEnd) / recordSize);
    myReadPointerEnd = add + nbRecords * recordSize;
    expect(snap.h_readPointerEnd).to.deep.equal(myReadPointerEnd);
  });
});

describe("#initSnapshot()", function() {
  let stlinkprotocolMock;

  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let myAcqParam = {
    snapshotheader: "0x200008c4",
    triggerstartmode: "manual",
    triggername: "Double1",
    triggerthreshold: "10"
  };

  let myGroup = [
    {
      address: "0x12345678",
      type: apidefs.READ_TYPE_U8,
      nb: 4
    },
    {
      address: "0x12340000",
      type: apidefs.READ_TYPE_S64
    },
    {
      address: "0x2345000",
      type: apidefs.READ_TYPE_S32,
      nb: 3
    },
    {
      address: "0x12345000",
      type: apidefs.READ_TYPE_S8
    },
    {
      address: "0x123400FF",
      type: apidefs.READ_TYPE_U64
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U32
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U16
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_S16
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_DOUBLE
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_FLOAT
    }
  ];

  let initGroup = [
    {
      address: myAcqParam.snapshotheader,
      type: apidefs.READ_TYPE_U32
    }
  ];

  let header = [
    { name: "startMark[0]", address: "0x200008c4", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerFieldSize", address: "0x200008c7", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerVersion", address: "0x200008c8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferStartAddr", address: "0x200008cc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferEndAddr", address: "0x200008d0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVarMax", address: "0x200008d4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferFormat", address: "0x200008d8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "timestamp_base_unit", address: "0x200008dc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointer", address: "0x200008e0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointerCopy", address: "0x200008e4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointer", address: "0x200008e8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointerCopy", address: "0x200008ec", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "flags", address: "0x200008f0", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "subSamplingRate", address: "0x200008f4", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVar", address: "0x200008f8", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].accessType", address: "0x200008fc", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].address", address: "0x20000900", type: apidefs.READ_TYPE_U32, nb: 1 }
  ];

  let headerResponse = [];
  headerResponse.push(
    [{ x: "startMark[0]", y: "0x008f25a7" }],
    [{ x: "headerFieldSize", y: "0x00000004" }],
    [{ x: "headerVersion", y: "608" }],
    [{ x: "bufferStartAddr", y: "0x20000058" }],
    [{ x: "bufferEndAddr", y: "0x200008ef" }],
    [{ x: "nbVarMax", y: "0x0000000a" }],
    [{ x: "bufferFormat", y: "0x00000001" }],
    [{ x: "timestamp_base_unit", y: "0x000040a0" }],
    [{ x: "writePointer", y: "0x20000058" }],
    [{ x: "writePointerCopy", y: "0x20000058" }],
    [{ x: "readPointer", y: "0x200008b8" }],
    [{ x: "readPointerCopy", y: "0x200008b8" }],
    [{ x: "flags", y: "0x00000000" }],
    [{ x: "subSamplingRate", y: "0x00000001" }],
    [{ x: "nbVar", y: "0x00001006" }],
    [{ x: "g_varList[0].accessType", y: "" }],
    [{ x: "g_varList[0].address", y: "" }]
  );
  let expectedReadObject = { result: 1, data: headerResponse };

  let headerResponseNbVar0 = [];
  headerResponseNbVar0.push(
    [{ x: "startMark[0]", y: "0x008f25a7" }],
    [{ x: "headerFieldSize", y: "0x00000004" }],
    [{ x: "headerVersion", y: "608" }],
    [{ x: "bufferStartAddr", y: "0x20000058" }],
    [{ x: "bufferEndAddr", y: "0x200008ef" }],
    [{ x: "nbVarMax", y: "0x0000000a" }],
    [{ x: "bufferFormat", y: "0x00000001" }],
    [{ x: "timestamp_base_unit", y: "0x000040a0" }],
    [{ x: "writePointer", y: "0x20000058" }],
    [{ x: "writePointerCopy", y: "0x20000058" }],
    [{ x: "readPointer", y: "0x200008b8" }],
    [{ x: "readPointerCopy", y: "0x200008b8" }],
    [{ x: "flags", y: "0x00000000" }],
    [{ x: "subSamplingRate", y: "0x00000001" }],
    [{ x: "nbVar", y: "0x00000000" }],
    [{ x: "g_varList[0].accessType", y: "" }],
    [{ x: "g_varList[0].address", y: "" }]
  );

  let headerOfTrigger = [
    { name: "triggerAddress", address: "0x2000094c", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerAccessType", address: "0x20000950", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerThreshold", address: "0x20000954", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerMode", address: "0x20000958", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerCmd", address: "0x2000095c", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerStopParam", address: "0x20000960", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerState", address: "0x20000964", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerNPretrig", address: "0x20000968", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerBIgnoreTrig", address: "0x2000096c", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerThresholdDouble", address: "0x20000970", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "triggerThresholdDouble2", address: "0x20000974", type: apidefs.READ_TYPE_U32, nb: 1 }
  ];

  let expectedReadObjectNbVar0 = { result: 1, data: headerResponseNbVar0 };

  beforeEach(function() {
    logger.info("#initSnapshot()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.restore();
  });

  it("should work if the initSnapshot with headersize not defined works", function() {
    logger.info("it should work if the initSnapshot with headersize not defined works()");
    const snap = new Snapshot(probe.probeId);

    let initRes = [[{ x: "0", y: "048f25a7" }]];
    let initResponse = { result: 1, data: initRes };

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, initGroup)
      .returns(initResponse);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .twice()
      .returns(expectedReadObject);

    stlinkprotocolMock.expects("readMemoryGroup").returns(expectedReadObject);

    snap.initSnapshot(myAcqParam, myGroup);

    expect(true).to.deep.equal(snap.traceIsStarted);
  });

  it("should work if the initSnapshot with headersize not defined and wrong startMarkVal works", function() {
    logger.info("it should work if the initSnapshot with headersize not defined and wrong startMarkVal works()");
    const snap = new Snapshot(probe.probeId);

    let initRes = [[{ x: "0", y: "048f25a7" }]];
    let initResponse = { result: 1, data: initRes };

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, initGroup)
      .returns(initResponse);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .twice()
      .returns(expectedReadObjectNbVar0);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, headerOfTrigger)
      .twice()
      .returns(expectedReadObjectNbVar0);

    snap.h_nbVar = 0;
    snap.headerSize = undefined;

    snap.initSnapshot(myAcqParam, myGroup);

    expect(false).to.deep.equal(snap.traceIsStarted);
  });

  it("should work if the initSnapshot with wrong response works", function() {
    logger.info("it should work if the initSnapshot with wrong response works()");
    const snap = new Snapshot(probe.probeId);

    let initRes = [[{ x: "0", y: "048f25a7" }]];
    let initResponse = { result: 3, data: initRes };

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, initGroup)
      .returns(initResponse);

    snap.h_timestampValue = 1562054794097;

    snap.initSnapshot(myAcqParam, myGroup);

    expect(true).to.deep.equal(true);
  });
});

describe("#getTrace()", function() {
  // Vérifier le comportement des tests
  let stlinkprotocolMock;
  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let myAcqParam = {
    snapshotheader: "0x200008c4",
    triggerstartmode: "manual",
    triggername: "Double1",
    triggerthreshold: "10"
  };

  let myGroup = [
    {
      address: "0x12345678",
      type: apidefs.READ_TYPE_U8,
      nb: 4
    },
    {
      address: "0x12340000",
      type: apidefs.READ_TYPE_S64
    },
    {
      address: "0x2345000",
      type: apidefs.READ_TYPE_S32,
      nb: 3
    },
    {
      address: "0x12345000",
      type: apidefs.READ_TYPE_S8
    },
    {
      address: "0x123400FF",
      type: apidefs.READ_TYPE_U64
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U32
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_U16
    },
    {
      address: "0x123400AA",
      type: apidefs.READ_TYPE_S16
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_DOUBLE
    },
    {
      address: "0x12340088",
      type: apidefs.READ_TYPE_FLOAT
    }
  ];

  beforeEach(function() {
    logger.info("#initSnapshot()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.restore();
  });

  let headerResponse = [];
  headerResponse.push(
    [{ x: "startMark[0]", y: "8f25a7" }],
    [{ x: "headerFieldSize", y: "0x00000004" }],
    [{ x: "headerVersion", y: "608" }],
    [{ x: "bufferStartAddr", y: "0x20000058" }],
    [{ x: "bufferEndAddr", y: "0x200008ef" }],
    [{ x: "nbVarMax", y: "0x0000000a" }],
    [{ x: "bufferFormat", y: "0x00000001" }],
    [{ x: "timestamp_base_unit", y: "0x000040a0" }],
    [{ x: "writePointer", y: "0x20000058" }],
    [{ x: "writePointerCopy", y: "0x20000058" }],
    [{ x: "readPointer", y: "0x200008b8" }],
    [{ x: "readPointerCopy", y: "0x200008b8" }],
    [{ x: "flags", y: "0x00000000" }],
    [{ x: "subSamplingRate", y: "0x00000001" }],
    [{ x: "nbVar", y: "0x00001006" }],
    [{ x: "g_varList[0].accessType", y: "" }],
    [{ x: "g_varList[0].address", y: "" }]
  );
  let expectedReadObject = { result: 1, data: headerResponse };

  const initGroup = [
    {
      address: myAcqParam.snapshotheader,
      type: apidefs.READ_TYPE_U32
    }
  ];

  it("should work if the getTrace with the GroupToBeRead is undefined and Initsnapshot return error works", function() {
    logger.info("it should work if the getTrace with myGroupToBeRead is undefined and with Initsnapshot return error");
    const snap = new Snapshot(probe.probeId);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, initGroup)
      .returns(expectedReadObject);

    snap.traceIsStarted = false;

    let res = snap.getTrace(myGroup, myAcqParam);

    expect(res.result).to.deep.equal(SNAP_ERR.INIT_SNP);
  });
});

describe("#stopSnapshot()", function() {
  let stlinkprotocolMock;
  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let expectedReadObject = { result: 1 };

  beforeEach(function() {
    logger.info("#initSnapshot()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.restore();
  });

  it("should work if the stopSnapshot works", function() {
    logger.info("it should work if the stopSnapshot works()");
    const snap = new Snapshot(probe.probeId);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .atLeast(1)
      .returns(expectedReadObject);

    snap.stopSnapshot(probe.probeId);

    expect(snap.synchroFinished).to.deep.equal(false);
    expect(snap.synchroStopDone).to.deep.equal(false);
  });
});

describe("#snapshotInterrupt()", function() {
  let stlinkprotocolMock;
  let probe = {
    probeId: "999990",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let myGroup = [
    {
      address: "0x12345678",
      type: apidefs.READ_TYPE_U8,
      nb: 4
    },
    {
      address: "0x12340000",
      type: apidefs.READ_TYPE_S64
    }
  ];

  let writeGroup = [];

  let elt = new Object();
  elt.name = "nbVar";
  elt.address = "0x20000000";
  elt.type = apidefs.READ_TYPE_U32;
  elt.value = "0x00"; // The nbVar value stops the target acquisition.
  writeGroup.push(elt);

  // Reset flags
  const flagGroup = [
    {
      address: this.h_flagAddStrHex,
      type: apidefs.READ_TYPE_U32,
      value: "0x00" // Reset flag
    }
  ];

  const flagGroup2 = [
    {
      address: "",
      type: apidefs.READ_TYPE_U32,
      value: "0x00" // Reset flag
    }
  ];

  let header = [
    { name: "startMark[0]", address: "", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerFieldSize", address: "", type: apidefs.READ_TYPE_S8, nb: 1 },
    { name: "headerVersion", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferStartAddr", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferEndAddr", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVarMax", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "bufferFormat", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "timestamp_base_unit", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointer", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "writePointerCopy", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointer", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "readPointerCopy", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "flags", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "subSamplingRate", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "nbVar", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].accessType", address: "", type: apidefs.READ_TYPE_U32, nb: 1 },
    { name: "g_varList[0].address", address: "", type: apidefs.READ_TYPE_U32, nb: 1 }
  ];

  let headerResponse = [];
  headerResponse.push(
    [{ x: "startMark[0]", y: "0x008f25a7" }],
    [{ x: "headerFieldSize", y: "0x00000004" }],
    [{ x: "headerVersion", y: "0x00000608" }],
    [{ x: "bufferStartAddr", y: "0x20000058" }],
    [{ x: "bufferEndAddr", y: "0x200008ef" }],
    [{ x: "nbVarMax", y: "0x0000000a" }],
    [{ x: "bufferFormat", y: "0x00000001" }],
    [{ x: "timestamp_base_unit", y: "0x000040a0" }],
    [{ x: "writePointer", y: "0x20000058" }],
    [{ x: "writePointerCopy", y: "0x20000058" }],
    [{ x: "readPointer", y: "0x200008b8" }],
    [{ x: "readPointerCopy", y: "0x200008b8" }],
    [{ x: "flags", y: "0x00000002" }],
    [{ x: "subSamplingRate", y: "0x00000001" }],
    [{ x: "nbVar", y: "0x00001006" }],
    [{ x: "g_varList[0].accessType", y: "" }],
    [{ x: "g_varList[0].address", y: "" }]
  );

  let expectedReadObject = { result: 1, data: headerResponse };

  beforeEach(function() {
    logger.info("#snapshotInterrupt()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.restore();
  });

  it("should work if the snapshotInterrupt works", function() {
    logger.info("it should work if the snapshotInterrupt works()");

    const snap = new Snapshot(probe.probeId);

    snap.myGroupToBeRead = myGroup;
    snap.h_flags = 2; // !!! use to clear timer

    snap.h_nbVarAddrStrHex = "0x20000000";

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, writeGroup)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, flagGroup)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, flagGroup)
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("readMemoryGroup")
      .withArgs(probe.probeId, header)
      .thrice()
      .returns(expectedReadObject);

    stlinkprotocolMock
      .expects("writeMemoryGroup")
      .withArgs(probe.probeId, flagGroup2)
      .returns(expectedReadObject);

    snap.snapshotInterrupt(probe.probeId);

    expect(snap.synchroFinished).to.deep.equal(false);
    expect(snap.synchroStopDone).to.deep.equal(true);
  });
});

describe("#startTrace()", function() {
  let stlinkprotocolMock;
  let probe = {
    probeId: "999999",
    probeName: "ST-Link v2-1A (...56789)",
    property: {
      JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
      SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
    }
  };

  let myGroup = [
    {
      address: "0x12345678",
      type: apidefs.READ_TYPE_U8,
      nb: 4
    },
    {
      address: "0x12340000",
      type: apidefs.READ_TYPE_S64
    }
  ];

  let expectedReadObject = { result: 1 };

  beforeEach(function() {
    logger.info("#initSnapshot()");
    stlinkprotocolMock = sinon.mock(stLinkProto);
  });

  this.afterEach(function() {
    stlinkprotocolMock.restore();
  });

  it("should work if the startTrace works", function() {
    logger.info("it should work if the startTrace works()");
    const snap = new Snapshot(probe.probeId);

    stlinkprotocolMock.expects("writeMemoryGroup").returns(expectedReadObject);

    snap.VarsToBeRead = myGroup;

    snap.startTrace(probe.probeId);

    expect(snap.traceIsStarted).to.deep.equal(true);
  });
});
