/* eslint-disable max-len */
/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
const expect = require("chai").expect;
const sinon = require("sinon");
const STLogger = require("@stm32/stm32cubemonitor-logger");
const logger = STLogger.getLogger("stlinkprotocol");

/**
 * Tested module.
 */
const stlinkprotocol = require("../src/stlinkprotocol");

/**
 * Module(s) mocked with sinonjs.
 */
const stLinkDrv = require("../src/STLinkUSBDriver").stLinkDrv;

/**
 * Module(s) mocked without sinonjs.
 */
const STLink_GetDeviceInfo2Mocked = require("./STLinkUSBDriverMock").STLink_GetDeviceInfo2Mocked;
const STLink_SendCommandSetOptions = require("./STLinkUSBDriverMock").STLink_SendCommandSetOptions;
const STLink_SendCommandTcpMocked = require("./STLinkUSBDriverMock").STLink_SendCommandTcpMocked;

/**
 * Provided definitions required for the assert.
 */
const stLinkDefs = require("../src/STLinkUSBDriver").defs;
const apidefs = require("../src/stlinkprotocol").apidefs;
const TDeviceInfo2 = require("../src/STLinkUSBDriver").TDeviceInfo2;

const TCP_MODE = "tcp";
const accessPoint = 0;

/**
 * Tests Scenarios.
 */

describe("#reenumerate() in tcp mode", function() {
  let stlinkDrvMock;
  beforeEach(function() {
    logger.info("#reenumerate()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should work if the STLink driver Reenumerate works in tcp mode", function() {
    logger.info("it should work if the STLink driver Reenumerate works() in tcp mode");
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false, "/0", "-a")
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .returns(3);

    expect(stlinkprotocol.cli.reenumerate(TCP_MODE)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the STLink driver Reenumerate fails in tcp mode", function() {
    logger.info("it should fail if the STLink driver Reenumerate fails in tcp mode");
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false, "/0", "-a")
      .returns(stLinkDefs.DRV_BAD_PARAMETER);

    expect(stlinkprotocol.cli.reenumerate(TCP_MODE)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#getNbDevices() in tcp mode", function() {
  beforeEach(function() {
    logger.info("#getNbDevices()");
  });

  it("should return the number of devices in tcp mode", function() {
    let stlinkDrvMock = sinon.mock(stLinkDrv);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .returns(3);

    stlinkprotocol.cli.reenumerate(TCP_MODE);
    expect(stlinkprotocol.cli.getNbDevices()).to.equal(3);

    stlinkDrvMock.verify();
  });
});

describe("#openDevice() in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#openDevice()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false, "/0", "-a")
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .returns(3);

    stlinkprotocol.cli.reenumerate(TCP_MODE);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should open the device", function() {
    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 0, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);

    expect(stlinkprotocol.cli.openDeviceTcp(2)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the STLink OpenDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 0, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_DEVICE_NOT_SUPPORTED);

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);

    expect(stlinkprotocol.cli.openDeviceTcp(2)).to.equal(apidefs.STLINK_DEVICE_NOT_SUPPORTED);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.openDeviceTcp(3)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#closeDevice() in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#closeDevice()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .returns(3);

    stlinkprotocol.cli.reenumerate(TCP_MODE);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should close the device", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 0, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_CloseDeviceTcp").returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    expect(stlinkprotocol.cli.closeDeviceTcp(2)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the STLink driver GetDeviceInfo2 fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_TIMEOUT);

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);

    expect(stlinkprotocol.cli.closeDeviceTcp(2)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);

    expect(stlinkprotocol.cli.closeDeviceTcp(2)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the STLink driver CloseDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 0, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_CloseDeviceTcp").returns(stLinkDefs.DRV_TCP_CLOSE_ERROR);

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    expect(stlinkprotocol.cli.closeDeviceTcp(2)).to.equal(apidefs.STLINK_DRIVER_ERR);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.closeDeviceTcp(3)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#probeDiscovery() in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#probeDiscovery()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should discover probes (with supported SWD andJTAG versions)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    // const TCP_3 = "/0";
    // const DEFAULT_TCP_SERVER_OPTION = "a";
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = function(interface, index, deviceInfo, deviceInfoLength) {
      for (let i = 0; i < deviceInfoLength; i++) {
        deviceInfo[i] = 0x30 + i;
      }
      // Set the end of the EnumUniqueId
      deviceInfo[10] = 0x00;
      // set the Vendor Id
      deviceInfo[36] = 0x83;
      deviceInfo[37] = 0x4;
      // set the Product Id
      deviceInfo[38] = 0x4a;
      deviceInfo[39] = 0x37;
      return stLinkDefs.DRV_OK;
    };

    let expectedResult = [
      {
        probeId: "456789",
        probeName: "ST-Link v2-1A (...56789)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "456789",
        probeName: "ST-Link v2-1A (...56789)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "456789",
        probeName: "ST-Link v2-1A (...56789)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      }
    ];

    expect(stlinkprotocol.probeDiscovery(TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should discover probes with short uniq Id (with supported SWIM and JTAG versions)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    let expectedResult = [
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      }
    ];

    expect(stlinkprotocol.probeDiscovery(TCP_MODE)).to.deep.equal(expectedResult);
  });

  /* eslint-disable-next-line max-len */
  it("should discover probes with short uniq Id (with supported SWD and JTAG versions and at least one V3 SLINK version)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;
    STLink_SendCommandSetOptions("STLINK_V3_VERSION");

    let expectedResult = [
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["24000 kHz", "8000 kHz"],
          SWD: ["3300 kHz", "1000 kHz", "200 kHz"]
        },
        version: {
          JTAGVersion: 4,
          STLinkVersion: 3
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["24000 kHz", "8000 kHz"],
          SWD: ["3300 kHz", "1000 kHz", "200 kHz"]
        },
        version: {
          JTAGVersion: 4,
          STLinkVersion: 3
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["24000 kHz", "8000 kHz"],
          SWD: ["3300 kHz", "1000 kHz", "200 kHz"]
        },
        version: {
          JTAGVersion: 4,
          STLinkVersion: 3
        }
      }
    ];

    expect(stlinkprotocol.probeDiscovery(TCP_MODE)).to.deep.equal(expectedResult);
  });
});

describe("#connectToTarget() in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#connectToTarget()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should connect to target with the SWD protocol", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(TCP_MODE);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461,
      MCUName: "STM32L496xx/4A6xx"
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, TCP_MODE)).to.deep.equal(
      expectedResult
    );
  });
});

describe("#closeProbes() in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#closeProbes() in tcp mode");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should close all probes (opened in JTAG mode)", function() {
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(TCP_MODE);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, TCP_MODE);

    STLink_SendCommandSetOptions("SET_JTAG_MODE");

    stlinkprotocol.closeProbes(1);
  });
});

describe("#JTAG_ReadMemory32Bit in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#JTAG_ReadMemory32Bit");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .returns(4);
    stlinkprotocol.cli.reenumerate(TCP_MODE);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should read one memory of 32 bits", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4]),
      size: 4
    };

    const res = stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 4, TCP_MODE);
    expect(res).to.deep.equal(expectedResult);
  });

  it("should read one memory of 192 bits (6 x 32 bits)", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]),
      size: 24
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 24, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is null", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 0, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is greater than 6144", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 6148, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is not a multiple of 4", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 6, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory address is not a multiple of 4", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345679, 4, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };
    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 4, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(4, 0x12345678, 4, TCP_MODE)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#JTAG_ReadMemory16Bit in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#JTAG_ReadMemory16Bit");
    stlinkDrvMock = sinon.mock(stLinkDrv);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .returns(3);
    stlinkprotocol.cli.reenumerate(TCP_MODE);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should read one memory of 16 bits", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2]),
      size: 2
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 2, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should read one memory of 64 bits (4 x 16 bits)", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4]),
      size: 4
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 4, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is null", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 0, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is greater than 6144", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 6148, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is not a multiple of 2", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 5, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(4, 0x12345678, 2, TCP_MODE)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#JTAG_ReadMemory8Bit in tcp mode", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#JTAG_ReadMemory8Bit");
    stlinkDrvMock = sinon.mock(stLinkDrv);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .returns(3);
    stlinkprotocol.cli.reenumerate(TCP_MODE);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should read one memory of 8 bits", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1]),
      size: 1
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 1, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should read one memory of 32 bits (4 x 8 bits)", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.cli.getDeviceInfo(2, TCP_MODE);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4]),
      size: 4
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345679, 4, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is null", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 0, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is greater than 64", function() {
    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, 858927408, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDeviceTcp(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 65, TCP_MODE)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(4, 0x12345678, 1, TCP_MODE)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#readMemoryGroup() in tcp mode", function() {
  let typeConnect = TCP_MODE;
  let stlinkDrvMock;
  let myClock;

  beforeEach(function() {
    logger.info(true, "#readMemoryGroup()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
    const myNow = new Date(1);
    myClock = sinon.useFakeTimers(myNow.getTime());
  });

  afterEach(function() {
    stlinkDrvMock.verify();
    myClock.restore();
  });

  it("should read a group", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        [
          {
            x: 1,
            y: 0x01
          },
          {
            x: 1,
            y: 0x02
          },
          {
            x: 1,
            y: 0x03
          },
          {
            x: 1,
            y: 0x04
          }
        ],
        [
          {
            x: 1,
            y: 0x0807060504030201
          }
        ],
        [
          {
            x: 1,
            y: 0x04030201
          },
          {
            x: 1,
            y: 0x08070605
          },
          {
            x: 1,
            y: 0x0c0b0a09
          }
        ],
        [
          {
            x: 1,
            y: 0x01
          }
        ],
        [
          {
            x: 1,
            y: 0x0102010403020104
          }
        ],
        [
          {
            x: 1,
            y: 0x0201acab
          }
        ],
        [
          {
            x: 1,
            y: 0xacab
          }
        ],
        [
          {
            x: 1,
            y: -21333
          }
        ],
        [
          {
            x: 1,
            y: -6.504395154765953e-229
          }
        ],
        [
          {
            x: 1,
            y: -2.1499718795889888e-31
          }
        ]
      ]
    };

    let myGroup = [
      {
        address: "0x12345678",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_S64
      },
      {
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      },
      {
        address: "0x12345000",
        type: apidefs.READ_TYPE_S8
      },
      {
        address: "0x123400FF",
        type: apidefs.READ_TYPE_U64
      },
      {
        address: "0x123400AA",
        type: apidefs.READ_TYPE_U32
      },
      {
        address: "0x123400AA",
        type: apidefs.READ_TYPE_U16
      },
      {
        address: "0x123400AA",
        type: apidefs.READ_TYPE_S16
      },
      {
        address: "0x12340088",
        type: apidefs.READ_TYPE_DOUBLE
      },
      {
        address: "0x12340088",
        type: apidefs.READ_TYPE_FLOAT
      }
    ];

    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should get BAD PARAMETER error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let myGroupNoAdress = [
      {
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];
    let myGroupNoType = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];
    let myGroupNoNb = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: "n°3"
      }
    ];

    expect(stlinkprotocol.readMemoryGroup("", myGroupNoAdress, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567", myGroupNoAdress, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567", myGroupNoType, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567", myGroupNoNb, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567")).to.deep.equal({ result: apidefs.STLINK_BAD_PARAMETER });
  });

  it("should get TIME OUT error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let myGroup = [
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "0x12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");
    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_TIMEOUT
    });
  });

  it("should get STLINK_NO_DEVICE error", function() {
    logger.info("it should get STLINK_NO_DEVICE error");
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let myGroup = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);
    expect(stlinkprotocol.readMemoryGroup("4568", myGroup, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_NO_DEVICE
    });
  });

  it("should get STLINK_DEVICE_NOT_OPENED error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let myGroup = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(2)
      .returns(stLinkDefs.DRV_NO_DEVICE);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    });
  });
});

describe("#writeMemoryGroup() in tcp mode", function() {
  let stlinkDrvMock;
  let myClock;
  let typeConnect = TCP_MODE;

  beforeEach(function() {
    logger.info(true, "#writeMemoryGroup()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
    const myNow = new Date(1);
    myClock = sinon.useFakeTimers(myNow.getTime());
  });

  afterEach(function() {
    stlinkDrvMock.verify();
    myClock.restore();
  });

  it("should write decimal and hexadecimal values in several variables of types READ_TYPE_U8 and READ_TYPE_S8", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345678",
        type: apidefs.READ_TYPE_U8,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S8,
        value: "0x09F"
      },
      {
        name: "Variable_Dec_1",
        address: "0x12340010",
        type: apidefs.READ_TYPE_S8,
        value: "-128"
      },
      {
        name: "Variable_Dec_2",
        address: "0x12340010",
        type: apidefs.READ_TYPE_U8,
        value: "127"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should write decimal and hexadecimal values in several variables of types READ_TYPE_U16 and READ_TYPE_S16", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "Variable_Dec_1",
        address: "0x00000002",
        type: apidefs.READ_TYPE_S16,
        value: "32767"
      },
      {
        name: "Variable_Dec_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S16,
        value: "-32768"
      },
      {
        name: "Variable_Dec_3",
        address: "0x00000002",
        type: apidefs.READ_TYPE_U16,
        value: "0"
      },
      {
        name: "Variable_Dec_4",
        address: "0x12340002",
        type: apidefs.READ_TYPE_U16,
        value: "65535"
      },
      {
        name: "Variable_Hex_1",
        address: "0x00000002",
        type: apidefs.READ_TYPE_S16,
        value: "0x0"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S16,
        value: "0xFFFF"
      },
      {
        name: "Variable_Hex_3",
        address: "0x00000002",
        type: apidefs.READ_TYPE_U16,
        value: "0x0"
      },
      {
        name: "Variable_Hex_4",
        address: "0x12340002",
        type: apidefs.READ_TYPE_U16,
        value: "0xFFFF"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should write decimal and hexadecimal values in several variables of types READ_TYPE_U32 and READ_TYPE_S32", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        value: "0x04030201"
      },
      {
        name: "Variable_Hex_2",
        address: "0x2345000",
        type: apidefs.READ_TYPE_U32,
        value: "0x1"
      },
      {
        name: "Variable_Dec_1",
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        value: "-2147483648"
      },
      {
        name: "Variable_Dec_2",
        address: "0x2345000",
        type: apidefs.READ_TYPE_U32,
        value: "4294967295"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });
});

describe("#blinkLed in tcp mode", function() {
  let stlinkDrvMock;
  let typeConnect = TCP_MODE;

  beforeEach(function() {
    logger.info("#blinkLed");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should blink the Led", function() {
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDeviceTcp(2);

    expect(stlinkprotocol.cli.JTAG_BlinkLed(2, typeConnect)).to.equal(apidefs.STLINK_OK);

    stlinkprotocol.cli.closeDeviceTcp(2);
  });

  it("should fail if the device is not opened", function() {
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    expect(stlinkprotocol.cli.JTAG_BlinkLed(2, typeConnect)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    stlinkDrvMock
      .expects("STLink_ReenumerateTcp")
      .withArgs(stLinkDefs.DRV_STLINK_TCP, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_STLINK_TCP)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDeviceTcp")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDeviceTcp")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommandTcp = STLink_SendCommandTcpMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDeviceTcp(2);

    STLink_SendCommandSetOptions("ERROR");

    expect(stlinkprotocol.cli.JTAG_BlinkLed(2, typeConnect)).to.equal(apidefs.STLINK_DRIVER_ERR);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_BlinkLed(3, typeConnect)).to.deep.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});
