/*!
 * module electron
 * Copyright(c) 2019 STMicroelectronics
 *
 * ****************************************************
 * The code is based on node-red project reference file.
 * Copyright JS Foundation and other contributors, http://js.foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var http = require('http');
var semver = require('semver');
var https = require('https');
var util = require("util");
var express = require("express");
var crypto = require("crypto");
try { bcrypt = require('bcrypt'); }
catch (e) { bcrypt = require('bcryptjs'); }
// var nopt = require("nopt");
var path = require("path");
var fs = require("fs-extra");
var RED = require("node-red");
const os = require('os');
const when = require('when');
const { productName, version, copyright } = require("./package.json");
var server;
var app = express();

const defaultDir = path.join(__dirname, 'default');
const flowsFileName = "flows.json";
const ressourceVersionFileName ="ressourcesVersion.json"
const settingsFileName = "settings.js";
const scriptFileName = "updateDesign.js"
const ressourcesDirectoryName = "ressources";
const libDirectoryName = "lib";
const flowsDirectoryName = "flows";
const themesDirectoryName = "themes";
const stThemesDirectoryName = "STMicrolectronics";

const stThemeFileName = "ST-light.js";

const package = fs.readJSONSync(path.join(__dirname, 'package.json'));
const userDir = process.platform === 'darwin' ? path.join(os.homedir(), 'Library', package.name) : path.join(os.homedir(), '.STMicroelectronics', package.name);

fs.ensureDirSync(userDir);
const settingsFile = path.join(userDir, settingsFileName);
const userFlowsFile = path.join(userDir, flowsFileName);

const ressourcesDirectoryPath = path.join(userDir, ressourcesDirectoryName);
const ressourcesVersionFile = path.join(ressourcesDirectoryPath, ressourceVersionFileName);
const flowsDirectoryPath = path.join(userDir, libDirectoryName, flowsDirectoryName);
const themesDirectoryPath = path.join(userDir, libDirectoryName, themesDirectoryName);
const stThemeDirectoryPath = path.join(themesDirectoryPath, stThemesDirectoryName);

const basicFlowFileName = productName + "_BasicFlow.json";
const libBasicflowsFile = path.join(flowsDirectoryPath, basicFlowFileName);
const advancedFlowFileName = productName + "_AdvancedFlow.json";
const libAdvancedflowsFile = path.join(flowsDirectoryPath, advancedFlowFileName);


// Copy the flow to used by default if none already installed  
if (!fs.existsSync(userFlowsFile)) {
    const defaultFlowjson = fs.readJSONSync(path.join(defaultDir, basicFlowFileName));
    fs.writeJSONSync(userFlowsFile, defaultFlowjson)
}
// Copy the settings file to used by default if none already installed  
if (!fs.existsSync(settingsFile)) {
    // fs.copyFileSync(path.join(defaultDir, settingsFileName), settingsFile);
    const settingsobj = fs.readFileSync(path.join(defaultDir, settingsFileName));
    fs.writeFileSync(settingsFile, settingsobj);
}

// Copy the ressource directory to used by default if none already existing or version of CM have been upgraded.
if (!fs.existsSync(ressourcesDirectoryPath) || !fs.existsSync(ressourcesVersionFile) || semver.gt(version,require(ressourcesVersionFile).ressourcesVersion)) {
    fs.ensureDirSync(ressourcesDirectoryPath);
    fs.copySync(path.join(defaultDir, ressourcesDirectoryName), ressourcesDirectoryPath);
}
// Create the lib folder and copy the reference flows in library if files do not exist
if (!fs.existsSync(flowsDirectoryPath)) {
    fs.ensureDirSync(flowsDirectoryPath);
}
if (!fs.existsSync(libBasicflowsFile)) {
    const basicFlowjson = fs.readJSONSync(path.join(defaultDir, basicFlowFileName));
    fs.writeJSONSync(libBasicflowsFile, basicFlowjson);
}
if (!fs.existsSync(libAdvancedflowsFile)) {
    const advancedFlowjson = fs.readJSONSync(path.join(defaultDir, advancedFlowFileName));
    fs.writeJSONSync(libAdvancedflowsFile, advancedFlowjson);
}


// Copy the default theme in library at each start of STM32CubeMonitor
if (!fs.existsSync(stThemeDirectoryPath)) {
    fs.ensureDirSync(stThemeDirectoryPath);
}
fs.copySync(path.join(defaultDir, stThemeFileName), path.join(stThemeDirectoryPath, stThemeFileName));


try {
    var settings = require(settingsFile);
    settings.settingsFile = settingsFile;
} catch (err) {
    console.log("Error loading settings file: " + settingsFile)
    if (err.code == 'MODULE_NOT_FOUND') {
        if (err.toString().indexOf(settingsFile) === -1) {
            console.log(err.toString());
        }
    } else {
        console.log(err);
    }
    process.exit();
}

// Copy the script file at each start of STM32CubeMonitor and append about information at the end of the file
let scriptFile = fs.readFileSync(path.join(defaultDir, scriptFileName),{encoding:'utf8'});
const uiPath=(settings.ui || {}).path;
if (uiPath) {
scriptFile = scriptFile.replace('href="/ui"', 'href="/'+uiPath+'"');
}
const about = String.raw`
let about = {
    productName:"${productName}",
    version:"${version}",
    copyright:"${copyright}"
};`;
fs.writeFileSync(path.join(userDir, scriptFileName), scriptFile+about);


// Add our STM32CubeMonitor script to customize design view 
// Check first is script key is already present or needs to be created
if (!settings.editorTheme) {
    Object.assign(settings, { editorTheme: { page: { scripts: path.join(userDir, scriptFileName) } } });
} else if (!settings.editorTheme.page) {
    Object.assign(settings.editorTheme, { page: { scripts: path.join(userDir, scriptFileName) } });
} else if (!settings.editorTheme.page.scripts) {
    Object.assign(settings.editorTheme.page, { scripts: path.join(userDir, scriptFileName) });
} else {
    //script key already exist in settings.js file so need to add an entry to this key
    // script key could be a single value or an array
    if (settings.editorTheme.page.scripts instanceof Array) {
        settings.editorTheme.page.scripts.push(path.join(userDir, scriptFileName))
    } else {
        settings.editorTheme.page.scripts = Array.of(settings.editorTheme.page.scripts, path.join(userDir, scriptFileName));
    }
}

if (settings.https) {
    server = https.createServer(settings.https, function (req, res) { app(req, res); });
} else {
    server = http.createServer(function (req, res) { app(req, res); });
}
server.setMaxListeners(0);

function formatRoot(root) {
    if (root[0] != "/") {
        root = "/" + root;
    }
    if (root.slice(-1) != "/") {
        root = root + "/";
    }
    return root;
}

if (settings.httpRoot === false) {
    settings.httpAdminRoot = false;
    settings.httpNodeRoot = false;
} else {
    settings.httpRoot = settings.httpRoot || "/";
    settings.disableEditor = settings.disableEditor || false;
}

if (settings.httpAdminRoot !== false) {
    settings.httpAdminRoot = formatRoot(settings.httpAdminRoot || settings.httpRoot || "/");
    settings.httpAdminAuth = settings.httpAdminAuth || settings.httpAuth;
} else {
    settings.disableEditor = true;
}

if (settings.httpNodeRoot !== false) {
    settings.httpNodeRoot = formatRoot(settings.httpNodeRoot || settings.httpRoot || "/");
    settings.httpNodeAuth = settings.httpNodeAuth || settings.httpAuth;
}

// TODO
if (settings.uiPort === undefined) {
    settings.uiPort = 1880;
}

settings.uiHost = settings.uiHost || "0.0.0.0";

if (userFlowsFile) {
    settings.flowFile = userFlowsFile;
}
settings.userDir = userDir;

try {
    RED.init(server, settings);
} catch (err) {
    if (err.code == "unsupported_version") {
        console.log("Unsupported version of node.js:", process.version);
        console.log("Node-RED requires node.js v4 or later");
    } else if (err.code == "not_built") {
        console.log("Node-RED has not been built. See README.md for details");
    } else {
        console.log("Failed to start server:");
        if (err.stack) {
            console.log(err.stack);
        } else {
            console.log(err);
        }
    }
    process.exit(1);
}

function basicAuthMiddleware(user, pass) {
    var basicAuth = require('basic-auth');
    var checkPassword;
    var localCachedPassword;
    if (pass.length == "32") {
        // Assume its a legacy md5 password
        checkPassword = function (p) {
            return crypto.createHash('md5').update(p, 'utf8').digest('hex') === pass;
        }
    } else {
        checkPassword = function (p) {
            return bcrypt.compareSync(p, pass);
        }
    }

    var checkPasswordAndCache = function (p) {
        // For BasicAuth routes we know the password cannot change without
        // a restart of Node-RED. This means we can cache the provided crypted
        // version to save recalculating each time.
        if (localCachedPassword === p) {
            return true;
        }
        var result = checkPassword(p);
        if (result) {
            localCachedPassword = p;
        }
        return result;
    }

    return function (req, res, next) {
        if (req.method === 'OPTIONS') {
            return next();
        }
        var requestUser = basicAuth(req);
        if (!requestUser || requestUser.name !== user || !checkPasswordAndCache(requestUser.pass)) {
            res.set('WWW-Authenticate', 'Basic realm=Authorization Required');
            return res.sendStatus(401);
        }
        next();
    }
}

if (settings.httpAdminRoot !== false && settings.httpAdminAuth) {
    RED.log.warn(RED.log._("server.httpadminauth-deprecated"));
    app.use(settings.httpAdminRoot, basicAuthMiddleware(settings.httpAdminAuth.user, settings.httpAdminAuth.pass));
}

if (settings.httpAdminRoot !== false) {
    app.use(settings.httpAdminRoot, RED.httpAdmin);
}
if (settings.httpNodeRoot !== false && settings.httpNodeAuth) {
    app.use(settings.httpNodeRoot, basicAuthMiddleware(settings.httpNodeAuth.user, settings.httpNodeAuth.pass));
}
if (settings.httpNodeRoot !== false) {
    app.use(settings.httpNodeRoot, RED.httpNode);
}
if (settings.httpStatic) {
    settings.httpStaticAuth = settings.httpStaticAuth || settings.httpAuth;
    if (settings.httpStaticAuth) {
        app.use("/", basicAuthMiddleware(settings.httpStaticAuth.user, settings.httpStaticAuth.pass));
    }
    app.use("/", express.static(settings.httpStatic));
}

function getListenPath() {
    var port = settings.serverPort;
    if (port === undefined) {
        port = settings.uiPort;
    }

    var listenPath = 'http' + (settings.https ? 's' : '') + '://' +
        (settings.uiHost == '::' ? 'localhost' : (settings.uiHost == '0.0.0.0' ? '127.0.0.1' : settings.uiHost)) +
        ':' + port;
    if (settings.httpAdminRoot !== false) {
        listenPath += settings.httpAdminRoot;
    } else if (settings.httpStatic) {
        listenPath += "/";
    }
    return listenPath;
}


process.on('uncaughtException', function (err) {
    util.log('[red] Uncaught Exception:');
    if (err.stack) {
        util.log(err.stack);
    } else {
        util.log(err);
    }
    process.exit(1);
});

process.on('SIGINT', function () {
    RED.stop().then(function () {
        process.exit();
    });
});

function start() {
    return RED.start().then(function () {

        return new when.promise((resolve, reject) => {

            server.once('error', function (err) {
                if (err.errno === "EADDRINUSE") {
                    RED.log.error(RED.log._("server.unable-to-listen", { listenpath: getListenPath() }));
                    RED.log.error(RED.log._("server.port-in-use"));
                } else {
                    RED.log.error(RED.log._("server.uncaught-exception"));
                    if (err.stack) {
                        RED.log.error(err.stack);
                    } else {
                        RED.log.error(err);
                    }
                }
                reject(err);
            });

            server.listen(settings.uiPort, settings.uiHost, function () {
                if (settings.httpAdminRoot === false) {
                    RED.log.info(RED.log._("server.admin-ui-disabled"));
                }
                settings.serverPort = server.address().port;
                process.title = package.name || 'node-red';
                RED.log.info(RED.log._("server.now-running", { listenpath: getListenPath() }));
                resolve();
            });
        });
    }).otherwise(function (err) {
        RED.log.error(RED.log._("server.failed-to-start"));
        if (err.stack) {
            RED.log.error(err.stack);
        } else {
            RED.log.error(err);
        }
        reject(err);
    });

};

exports.start = start;
exports.settings = settings;
