/*!
 * module st-link-protocol
 * Copyright(c) 2019 STMicroelectronics
 */
/* eslint-disable max-len */
const expect = require("chai").expect;
const sinon = require("sinon");
const STLogger = require("@stm32/stm32cubemonitor-logger");
const logger = STLogger.getLogger("stlinkprotocol");

/**
 * Tested module.
 */
const stlinkprotocol = require("../src/stlinkprotocol");

/**
 * Module(s) mocked with sinonjs.
 */
const stLinkDrv = require("../src/STLinkUSBDriver").stLinkDrv;

/**
 * Module(s) mocked without sinonjs.
 */
const STLink_GetDeviceInfo2Mocked = require("./STLinkUSBDriverMock").STLink_GetDeviceInfo2Mocked;
const STLink_SendCommandSetOptions = require("./STLinkUSBDriverMock").STLink_SendCommandSetOptions;
const STLink_SendCommandMocked = require("./STLinkUSBDriverMock").STLink_SendCommandMocked;

/**
 * Provided definitions required for the assert.
 */
const stLinkDefs = require("../src/STLinkUSBDriver").defs;
const TDeviceInfo2 = require("../src/STLinkUSBDriver").TDeviceInfo2;
const apidefs = require("../src/stlinkprotocol").apidefs;

const typeConnect = "p2p";

const accessPoint = 0;

/**
 * Tests Scenarios.
 */
describe("#reenumerate()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#reenumerate()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should work if the STLink driver Reenumerate works", function() {
    logger.info("it should work if the STLink driver Reenumerate works()");
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);

    expect(stlinkprotocol.cli.reenumerate(typeConnect)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the STLink driver Reenumerate fails", function() {
    logger.info("it should fail if the STLink driver Reenumerate fails");
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_BAD_PARAMETER);

    stlinkDrvMock.expects("STLink_GetNbDevices").never();

    expect(stlinkprotocol.cli.reenumerate(typeConnect)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#getNbDevices()", function() {
  beforeEach(function() {
    logger.info("#getNbDevices()");
  });

  it("should return the number of devices", function() {
    let stlinkDrvMock = sinon.mock(stLinkDrv);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);

    stlinkprotocol.cli.reenumerate(typeConnect);
    expect(stlinkprotocol.cli.getNbDevices()).to.equal(3);

    stlinkDrvMock.verify();
  });
});

describe("#errorString()", function() {
  beforeEach(function() {
    logger.info("#errorString()");
  });

  it("should return one dedicated string if the error code is BAD_PARAMETER", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_BAD_PARAMETER)).to.equal("Bad parameter (7)");
  });
  it("should return one dedicated string if the error code is DEVICE_NOT_OPENED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_DEVICE_NOT_OPENED)).to.equal("Device not opened (5)");
  });
  it("should return one dedicated string if the error is WRONG_DEVICE_MODE", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_WRONG_DEVICE_MODE)).to.equal("Device in wrong mode (4)");
  });
  it("should return one dedicated string if the error is DEVICE_NOT_SUPPORTED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_DEVICE_NOT_SUPPORTED)).to.equal("Device not supported (8)");
  });
  it("should return one dedicated string if the error is NO_DEVICE", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_NO_DEVICE)).to.equal("Device not found (6)");
  });

  it("should return one dedicated string if the error is ERR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_ERR)).to.equal("Error (2)");
  });
  it("should return one dedicated string if the error is DRIVER_ERR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_DRIVER_ERR)).to.equal("Driver error (3)");
  });
  it("should return one dedicated string if the error is TIMEOUT", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_TIMEOUT)).to.equal("Timeout (9)");
  });

  it("should return one dedicated string if the error is JTAG_PROTOCOL_NOT_SUPPORTED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_PROTOCOL_NOT_SUPPORTED)).to.equal(
      "JTAG protocol not supported (f0)"
    );
  });

  it("should return one dedicated string if the error is SWD_PROTOCOL_NOT_SUPPORTED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_SWD_PROTOCOL_NOT_SUPPORTED)).to.equal(
      "SWD protocol not supported (f1)"
    );
  });

  it("should return one dedicated string if the error is JTAG_FREQUENCY_NOT_SUPPORTED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_FREQUENCY_NOT_SUPPORTED)).to.equal(
      "JTAG frequency not supported (f2)"
    );
  });

  it("should return one dedicated string if the error is SWD_FREQUENCY_NOT_SUPPORTED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_SWD_FREQUENCY_NOT_SUPPORTED)).to.equal(
      "SWD frequency not supported (f3)"
    );
  });

  it("should return one dedicated string if the error is JTAG_READ_ERROR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_READ_ERROR)).to.equal("JTAG/SWD read error (f4)");
  });

  it("should return one dedicated string if the error is JTAG_ERROR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_ERROR)).to.equal("JTAG/SWD error (f5)");
  });

  it("should return one dedicated string if the error is JTAG_UNKNOWN_JTAG_CHAIN", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN)).to.equal("Unknown JTAG chain (f6)");
  });

  it("should return one dedicated string if the error is JTAG_NO_DEVICE_CONNECTED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_NO_DEVICE_CONNECTED)).to.equal(
      "JTAG/SWD no device connected (f7)"
    );
  });

  it("should return one dedicated string if the error is JTAG_WRITE_ERROR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_WRITE_ERROR)).to.equal("JTAG/SWD write error (f8)");
  });

  it("should return one dedicated string if the error is JTAG_ALREADY_OPENED_IN_OTHER_MODE", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE)).to.equal(
      "JTAG/SWD already opened in other mode (f9)"
    );
  });

  it("should return one dedicated string if the error is SWD_ERROR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_SWD_ERROR)).to.equal("SWD protocol implementation error (fa)");
  });

  it("should return one dedicated string if the error is AP_ERROR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_AP_ERROR)).to.equal("Access Point error (fb)");
  });

  it("should return one dedicated string if the error is SWV_NOT_AVAILABLE", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_SWV_NOT_AVAILABLE)).to.equal("SWV not available (fc)");
  });

  it("should return one dedicated string if the error is JTAG_TCP_ERROR", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_TCP_ERROR)).to.equal("TCP error (fd)");
  });

  it("should return one dedicated string if the error is JTAG_16BIT_ACCESS_NOT_SUPPORTED", function() {
    expect(stlinkprotocol.cli.errorString(apidefs.STLINK_JTAG_16BIT_ACCESS_NOT_SUPPORTED)).to.equal(
      "16-bit access mode not supported (fe)"
    );
  });

  it("should return one default string for the others error code", function() {
    expect(stlinkprotocol.cli.errorString(-1)).to.equal(" (-1)");
  });
});

describe("#getDeviceInfo()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#getDeviceInfo()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);

    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should return non-official device info", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    let expectedResult = {
      Cookie: 0,
      DevNumber: 2,
      StLinkUsbId: "0x00000000",
      EnumUniqueId: "",
      VendorId: "0000",
      ProductId: "0000",
      DeviceUsed: false
    };

    expect(stlinkprotocol.cli.getDeviceInfo(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should return official device info", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    let expectedResult = {
      Cookie: 0x33323130,
      DevNumber: 2,
      StLinkUsbId: "0x33323130",
      EnumUniqueId: "4567",
      VendorId: "ST-Microelectronics",
      ProductId: "ST-Link v2-1A",
      DeviceUsed: true
    };

    expect(stlinkprotocol.cli.getDeviceInfo(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should return null if the STLink driver GetDeviceInfo2 fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_NO_DEVICE);

    expect(stlinkprotocol.cli.getDeviceInfo(2, typeConnect)).to.equal(null);
  });

  it("should return null if the device index is wrong", function() {
    expect(stlinkprotocol.cli.getDeviceInfo(3, typeConnect)).to.equal(null);
  });
});

describe("#openDevice()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#openDevice()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);

    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should open the device", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    expect(stlinkprotocol.cli.openDevice(2)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the STLink OpenDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_DEVICE_NOT_SUPPORTED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    expect(stlinkprotocol.cli.openDevice(2)).to.equal(apidefs.STLINK_DEVICE_NOT_SUPPORTED);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.openDevice(3)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#closeDevice()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#closeDevice()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);

    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should close the device", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_CloseDevice").returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.closeDevice(2)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the STLink driver GetDeviceInfo2 fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_TIMEOUT);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    expect(stlinkprotocol.cli.closeDevice(2)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    expect(stlinkprotocol.cli.closeDevice(2)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the STLink driver CloseDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_CloseDevice").returns(stLinkDefs.DRV_ERR);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.closeDevice(2)).to.equal(apidefs.STLINK_DEVICE_NOT_RESPONDING);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.closeDevice(3)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#getCurrentMode()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#getCurrentMode()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should get the current mode (which is DFU mode)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_SendCommand").returns(stLinkDefs.DRV_OK);

    let expectedResult = {
      result: stLinkDefs.DRV_OK,
      mode: 0,
      DFUVersion: 0
    };

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.getCurrentMode(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should get the current mode (which is not DFU mode)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: stLinkDefs.DRV_OK,
      mode: 0x02,
      DFUVersion: 0
    };

    expect(stlinkprotocol.cli.getCurrentMode(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver GetDeviceInfo2 fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_BAD_PARAMETER);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };

    expect(stlinkprotocol.cli.getCurrentMode(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };

    expect(stlinkprotocol.cli.getCurrentMode(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_SendCommand").returns(stLinkDefs.DRV_ERR);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_RESPONDING
    };

    expect(stlinkprotocol.cli.getCurrentMode(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };
    expect(stlinkprotocol.cli.getCurrentMode(3, typeConnect)).to.deep.equal(expectedResult);
  });
});

describe("#getVersion()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#getVersion()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should get the version", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_SendCommand").returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: stLinkDefs.DRV_OK,
      STLinkVersion: 0,
      JTAGVersion: 0,
      SWIMVersion: 0
    };

    expect(stlinkprotocol.cli.getVersion(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver GetDeviceInfo2 fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };

    expect(stlinkprotocol.cli.getVersion(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };

    expect(stlinkprotocol.cli.getVersion(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_SendCommand").returns(stLinkDefs.DRV_FAILED);

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2, typeConnect);

    expect(stlinkprotocol.cli.getVersion(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };
    expect(stlinkprotocol.cli.getVersion(3, typeConnect)).to.deep.equal(expectedResult);
  });
});

describe("#modeString()", function() {
  beforeEach(function() {
    logger.info("#modeString()");
  });
  it("should return DFU string if the mode is DFU", function() {
    // 0x00 = STLINK_DEV_DFU_MODE
    expect(stlinkprotocol.cli.modeString(0x00)).to.equal("DFU");
  });
  it("should return MASS/IDLE string if the mode is MASS", function() {
    // 0x01 = STLINK_DEV_MASS_MODE
    expect(stlinkprotocol.cli.modeString(0x01)).to.equal("MASS/IDLE");
  });
  it("should return JTAG string if the mode is JTAG", function() {
    // 0x02 = STLINK_DEV_JTAG_MODE
    expect(stlinkprotocol.cli.modeString(0x02)).to.equal("JTAG");
  });
  it("should return SWIM string if the mode is SWIM", function() {
    // 0x03 = STLINK_DEV_SWIM_MODE
    expect(stlinkprotocol.cli.modeString(0x03)).to.equal("SWIM");
  });
  it("should return BTLD string if the mode is BTLD", function() {
    // 0x04 = STLINK_DEV_BTLD_MODE
    expect(stlinkprotocol.cli.modeString(0x04)).to.equal("BTLD");
  });
  it("should return empty string if the mode is unknown", function() {
    expect(stlinkprotocol.cli.modeString(-1)).to.equal("");
  });
});

describe("#exitDFUMode()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#exitDFUMode()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should exit the DFU mode", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .twice()
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    expect(stlinkprotocol.cli.exitDFUMode(2, typeConnect)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    expect(stlinkprotocol.cli.exitDFUMode(2, typeConnect)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the first call of the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_SendCommand").returns(stLinkDefs.DRV_FAILED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    expect(stlinkprotocol.cli.exitDFUMode(2, typeConnect)).to.equal(apidefs.STLINK_WRONG_DEVICE_MODE);
  });

  it("should fail if the second call of the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .onCall(0)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .onCall(1)
      .returns(stLinkDefs.DRV_FAILED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    expect(stlinkprotocol.cli.exitDFUMode(2, typeConnect)).to.equal(apidefs.STLINK_DRIVER_ERR);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.exitDFUMode(3, typeConnect)).to.deep.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#enterJTAG", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#enterJTAG");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should enter the JTAG mode", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.enterJTAG(2, typeConnect)).to.equal(apidefs.STLINK_OK);

    stlinkprotocol.cli.closeDevice(2);
  });

  it("should fail if the device is undefined", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_CloseDevice").returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.reenumerate(typeConnect);
    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.enterJTAG(2, typeConnect)).to.equal(apidefs.STLINK_JTAG_PROTOCOL_NOT_SUPPORTED);

    stlinkprotocol.cli.closeDevice(2);
  });

  it("should fail if the device is not opened", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    expect(stlinkprotocol.cli.enterJTAG(2, typeConnect)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("ENTER_SWD_OR_JTAG_RETURNS_ERROR");

    expect(stlinkprotocol.cli.enterJTAG(2, typeConnect)).to.equal(apidefs.STLINK_DRIVER_ERR);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.enterJTAG(3, typeConnect)).to.deep.equal(apidefs.STLINK_BAD_PARAMETER);
  });

  stlinkprotocol.cli.closeDevice(2);
});

describe("#enterSWD", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#enterSWD");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should enter the SWD mode", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.enterSWD(2, typeConnect)).to.equal(apidefs.STLINK_OK);

    stlinkprotocol.cli.closeDevice(2);
  });

  it("should fail if the device is undefined", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.reenumerate(typeConnect);
    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.enterSWD(2, typeConnect)).to.equal(apidefs.STLINK_SWD_PROTOCOL_NOT_SUPPORTED);

    stlinkprotocol.cli.closeDevice(2);
  });

  it("should fail if the device is not opened", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    expect(stlinkprotocol.cli.enterSWD(2, typeConnect)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("ENTER_SWD_OR_JTAG_RETURNS_ERROR");

    expect(stlinkprotocol.cli.enterSWD(2, typeConnect)).to.equal(apidefs.STLINK_DRIVER_ERR);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.enterSWD(3, typeConnect)).to.deep.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#exitJTAG", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#exitJTAG");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should exit the JTAG mode", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_SendCommand").returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.exitJTAG(2, typeConnect)).to.equal(apidefs.STLINK_OK);
  });

  it("should fail if the device is undefined", function() {
    expect(stlinkprotocol.cli.exitJTAG(2)).to.equal(apidefs.STLINK_NO_DEVICE);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    expect(stlinkprotocol.cli.exitJTAG(2, typeConnect)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .twice()
      .returns(stLinkDefs.DRV_FAILED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    expect(stlinkprotocol.cli.exitJTAG(2, typeConnect)).to.equal(apidefs.STLINK_DRIVER_ERR);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.exitJTAG(3, typeConnect)).to.deep.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#JTAG_ReadIdCodes", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#JTAG_ReadIdCodes");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);
    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should read Id codes (set in target at STM32_DBGMCU_IDCODE_ADDR address", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("STM32_MCU_CODE_IS_SET_IN_STM32_DBGMCU_IDCODE_ADDR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should read Id codes (set in target at STM32_DBGMCU_ADDR address", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("STM32_MCU_CODE_IS_SET_IN_STM32_DBGMCU_ADDR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should read Id codes (set in target at STM32H_IDCODE_ADDR address", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("STM32_MCU_CODE_IS_SET_IN_STM32H_IDCODE_ADDR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should read Id codes (set in target at STM32W_IDCODE_ADDR address", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("STM32_MCU_CODE_IS_SET_IN_STM32W_IDCODE_ADDR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should read Id codes (but id codes are not set in target", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("STM32_MCU_CODE_IS_NOT_SET_IN_TARGET");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is undefined", function() {
    let expectedResult = {
      result: apidefs.STLINK_NO_DEVICE
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the reading of Id codes set in target at STM32_DBGMCU_IDCODE_ADDR address returns one error", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_STM32_MCU_CODE_AT_STM32_DBGMCU_IDCODE_ADDR_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the reading of Id codes set in target at STM32_DBGMCU_ADDR address returns one error", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_STM32_MCU_CODE_AT_STM32_DBGMCU_ADDR_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the reading of Id codes set in target at STM32H_IDCODE_ADDR address returns one error", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_STM32_MCU_CODE_AT_STM32H_IDCODE_ADDR_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the reading of Id codes set in target at STM32W_IDCODE_ADDR address returns one error", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_STM32_MCU_CODE_AT_STM32W_IDCODE_ADDR_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .twice()
      .returns(stLinkDefs.DRV_FAILED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error UNKNOWN_JTAG_CHAIN when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_UNKNOWN_JTAG_CHAIN");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error NO_DEVICE_CONNECTED when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_NO_DEVICE_CONNECTED");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_NO_DEVICE_CONNECTED
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error WRITE_ERROR when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_WRITE_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_WRITE_ERROR
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error ALREADY_OPENED_IN_OTHER_MODE when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_ALREADY_OPENED_IN_OTHER_MODE");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ALREADY_OPENED_IN_OTHER_MODE
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error SWD_AP_WAIT when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_SWD_AP_WAIT");

    let expectedResult = {
      result: apidefs.STLINK_SWD_ERROR
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error SWD_AP_STICKY_ERROR when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_SWD_AP_STICKY_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_AP_ERROR
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error SWV_NOT_AVAILABLE when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_SWV_NOT_AVAILABLE");

    let expectedResult = {
      result: apidefs.STLINK_SWV_NOT_AVAILABLE
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets the JTAG error TCPID_NOT_FOUND when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_TCPID_NOT_FOUND");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_TCP_ERROR
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets an unknown JTAG error when reading the ID code", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_SendCommand").returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR // STLINK_JTAG_ERROR
    };
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails when reading STM32_DBGMCU_IDCODE_ADDR address", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0
    };

    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_ReadIdCodes(3, typeConnect)).to.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});

describe("#JTAG_ReadMemory32Bit", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#JTAG_ReadMemory32Bit");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);
    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should read one memory of 32 bits", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4]),
      size: 4
    };

    const res = stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 4, typeConnect);
    expect(res).to.deep.equal(expectedResult);
  });

  it("should read one memory of 192 bits (6 x 32 bits)", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]),
      size: 24
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 24, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is null", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 0, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is greater than 6144", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 6148, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is not a multiple of 4", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 6, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory address is not a multiple of 4", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345679, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };
    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .twice()
      .returns(stLinkDefs.DRV_FAILED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };
    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the last Read/Write status cannot be read", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("READ_LAST_READ_WRITE_STATUS_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the last Read/Write status is erroneous", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("LAST_READ_WRITE_STATUS_IS_ERRONEOUS");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_ReadMemory32Bit(3, 0x12345678, 4, typeConnect)).to.equal(
      apidefs.STLINK_BAD_PARAMETER
    );
  });
});

describe("#JTAG_ReadMemory16Bit", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#JTAG_ReadMemory16Bit");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should read one memory of 16 bits", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2]),
      size: 2
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 2, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should read one memory of 64 bits (4 x 16 bits)", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4]),
      size: 4
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is null", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 0, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is greater than 6144", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 6148, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is not a multiple of 2", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 5, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory address is not a multiple of 2", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345679, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };
    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .twice()
      .returns(stLinkDefs.DRV_FAILED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };
    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the last Read/Write status is erroneous", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("LAST_READ_WRITE_STATUS_IS_ERRONEOUS");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_ReadMemory16Bit(3, 0x12345678, 4, typeConnect)).to.equal(
      apidefs.STLINK_BAD_PARAMETER
    );
  });
});

describe("#JTAG_ReadMemory8Bit", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#JTAG_ReadMemory8Bit");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .returns(3);
    stlinkprotocol.cli.reenumerate(typeConnect);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should read one memory of 8 bits", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1]),
      size: 1
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 1, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should read one memory of 32 bits (4 x 8 bits)", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: Buffer.from([1, 2, 3, 4]),
      size: 4
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345679, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is null", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 0, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the memory size is greater than 64", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkprotocol.cli.getDeviceInfo(2);
    stlinkprotocol.cli.openDevice(2);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 65, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };
    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, deviceInfo2.ref(), deviceInfo2.ref().length)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .twice()
      .returns(stLinkDefs.DRV_FAILED);

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);
    stlinkprotocol.cli.getCurrentMode(2, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };
    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 4, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the last Read/Write status is erroneous", function() {
    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, 2, stLinkDefs.DRV_OPEN_STLINK_SHARED_ACCESS)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.cli.getDeviceInfo(2, typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("LAST_READ_WRITE_STATUS_IS_ERRONEOUS");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR
    };

    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(2, 0x12345678, 1, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_ReadMemory8Bit(3, 0x12345678, 4, typeConnect)).to.equal(
      apidefs.STLINK_BAD_PARAMETER
    );
  });
});

describe("#probeDiscovery()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#probeDiscovery()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should discover probes (with supported SWD and JTAG versions)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = function(interface, index, deviceInfo, deviceInfoLength) {
      for (let i = 0; i < deviceInfoLength; i++) {
        deviceInfo[i] = 0x30 + i;
      }
      // Set the end of the EnumUniqueId
      deviceInfo[10] = 0x00;
      // set the Vendor Id
      deviceInfo[36] = 0x83;
      deviceInfo[37] = 0x4;
      // set the Product Id
      deviceInfo[38] = 0x4a;
      deviceInfo[39] = 0x37;
      return stLinkDefs.DRV_OK;
    };

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    let expectedResult = [
      {
        probeId: "456789",
        probeName: "ST-Link v2-1A (...56789)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "456789",
        probeName: "ST-Link v2-1A (...56789)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "456789",
        probeName: "ST-Link v2-1A (...56789)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      }
    ];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should discover probes with short uniq Id (with supported SWD and JTAG versions)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    let expectedResult = [
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["18 MHz", "9 MHz", "4.5 MHz", "2.25 MHz", "1.12 MHz - Default", "560 kHz", "280 kHz", "140 kHz"],
          SWD: ["4.6 MHz", "1.8 MHz - Default", "950 kHz", "400 kHz", "150 kHz"]
        },
        version: {
          JTAGVersion: 48,
          STLinkVersion: 0
        }
      }
    ];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should discover probes with short uniq Id (with supported SWD and JTAG versions and at least one V3 SLINK version)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;
    STLink_SendCommandSetOptions("STLINK_V3_VERSION");

    let expectedResult = [
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["24000 kHz", "8000 kHz"],
          SWD: ["3300 kHz", "1000 kHz", "200 kHz"]
        },
        version: {
          JTAGVersion: 4,
          STLinkVersion: 3
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["24000 kHz", "8000 kHz"],
          SWD: ["3300 kHz", "1000 kHz", "200 kHz"]
        },
        version: {
          JTAGVersion: 4,
          STLinkVersion: 3
        }
      },
      {
        probeId: "4567",
        probeName: "ST-Link v2-1A (...4567)",
        property: {
          JTAG: ["24000 kHz", "8000 kHz"],
          SWD: ["3300 kHz", "1000 kHz", "200 kHz"]
        },
        version: {
          JTAGVersion: 4,
          STLinkVersion: 3
        }
      }
    ];
    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should discover probes (with NOT supported SWIM and JTAG versions)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .exactly(9)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should discover STLINKV3 probes (with NOT supported SWIM and JTAG versions)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;
    STLink_SendCommandSetOptions("STLINK_V3_NO_JTAG");

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should discover no probe", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(0);

    stlinkDrvMock.expects("STLink_GetDeviceInfo2").never();

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver Reenumerate fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_NO_DEVICE);

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver GetDeviceInfo2 fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .returns(stLinkDefs.DRV_TIMEOUT)
      .thrice();

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver OpenDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.expects("STLink_OpenDevice").returns(stLinkDefs.DRV_TIMEOUT);

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_GetDeviceInfo2")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .once()
      .returns(stLinkDefs.DRV_BAD_PARAMETER);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .once()
      .returns(stLinkDefs.DRV_OK);

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);

    stlinkprotocol.cli.closeDevice(0);
  });

  it("should fail if the STLink driver SendCommand fails when exiting the DFU Mode", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = function(interface, index, deviceInfo, deviceInfoLength) {
      for (let i = 0; i < deviceInfoLength; i++) {
        deviceInfo[i] = 0x30 + i;
      }
      // Set the end of the EnumUniqueId
      deviceInfo[10] = 0x00;
      // set the Vendor Id
      deviceInfo[36] = 0x83;
      deviceInfo[37] = 0x4;
      // set the Product Id
      deviceInfo[38] = 0x4a;
      deviceInfo[39] = 0x37;
      return stLinkDefs.DRV_OK;
    };

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    STLink_SendCommandSetOptions("EXIT_DFU_MODE_RETURNS_ERROR");

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);

    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should fail if the STLink driver SendCommand fails when getting the firmware version", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = function(interface, index, deviceInfo, deviceInfoLength) {
      for (let i = 0; i < deviceInfoLength; i++) {
        deviceInfo[i] = 0x30 + i;
      }
      // Set the end of the EnumUniqueId
      deviceInfo[10] = 0x00;
      // set the Vendor Id
      deviceInfo[36] = 0x83;
      deviceInfo[37] = 0x4;
      // set the Product Id
      deviceInfo[38] = 0x4a;
      deviceInfo[39] = 0x37;
      return stLinkDefs.DRV_OK;
    };

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    STLink_SendCommandSetOptions("GET_VERSION_RETURNS_ERROR");

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand fails when getting the STLINK V3 firmware version", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = function(interface, index, deviceInfo, deviceInfoLength) {
      for (let i = 0; i < deviceInfoLength; i++) {
        deviceInfo[i] = 0x30 + i;
      }
      // Set the end of the EnumUniqueId
      deviceInfo[10] = 0x00;
      // set the Vendor Id
      deviceInfo[36] = 0x83;
      deviceInfo[37] = 0x4;
      // set the Product Id
      deviceInfo[38] = 0x4a;
      deviceInfo[39] = 0x37;
      return stLinkDefs.DRV_OK;
    };

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    STLink_SendCommandSetOptions("GET_EXT_VERSION_RETURNS_ERROR");

    let expectedResult = [];

    expect(stlinkprotocol.probeDiscovery(typeConnect)).to.deep.equal(expectedResult);
  });
});

describe("#connectToTarget()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#connectToTarget()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should connect to target with the SWD protocol", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461,
      MCUName: "STM32L496xx/4A6xx"
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );

    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should connect to target with the SWD protocol with the rediscovery mode", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .twice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .twice()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(7) // 3x first probediscovery, 3x second probediscovery, 1x for connect to target
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(7)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461,
      MCUName: "STM32L496xx/4A6xx"
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", true, typeConnect)).to.deep.equal(
      expectedResult
    );

    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should connect to target with the SWD protocol with default frequency", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461,
      MCUName: "STM32L496xx/4A6xx"
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", undefined, false, typeConnect)).to.deep.equal(expectedResult);
    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should connect to target with the SWD protocol and the target has an unknown MCU ID ", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("UNKNOWN_MCU_ID");

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x561,
      MCUName: "Unknown MCU"
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should connect to target with the JTAG protocol", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461,
      MCUName: "STM32L496xx/4A6xx"
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should connect to target with the JTAG protocol with default frequency", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      MCUID: 0x461,
      MCUName: "STM32L496xx/4A6xx"
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", undefined, false, typeConnect)).to.deep.equal(expectedResult);
    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should fail with the SWIM protocol (SW DEVELOPMENT TO BE DONE)", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_ERR
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWIM", "HIGH - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should fail if the probe identifier is undefined", function() {
    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.connectToTarget()).to.deep.equal(expectedResult);
  });

  it("should fail if the protocol is not SWD, JTAG or SWIM", function() {
    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.connectToTarget("4567")).to.deep.equal(expectedResult);
  });

  it("should fail if the probe identifier is erroneous", function() {
    let expectedResult = {
      result: apidefs.STLINK_NO_DEVICE
    };

    expect(stlinkprotocol.connectToTarget("\u0004\u0005\u0006", "SWD", undefined, false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the SWD frequency is erroneous", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: 0xf3 // STLINK_SWD_FREQUENCY_NOT_SUPPORTED
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "90000 MHz", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the JTAG frequency is erroneous", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: 0xf2 // STLINK_JTAG_FREQUENCY_NOT_SUPPORTED
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "12345 kHz", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the probe does not support the SWD protocol", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    STLink_SendCommandSetOptions("UNSUPPORTED_FIRMWARE_VERSION");
    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: 0xf1 // STLINK_SWD_PROTOCOL_NOT_SUPPORTED
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "4.6 MHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the probe does not support the JTAG protocol", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    STLink_SendCommandSetOptions("UNSUPPORTED_FIRMWARE_VERSION");
    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: 0xf0 // STLINK_JTAG_PROTOCOL_NOT_SUPPORTED
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver OpenDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let stlinkOpenDeviceCounter = 0;

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_SendCommand")
      .exactly(9)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkDrvMock.object.STLink_OpenDevice = function() {
      stlinkOpenDeviceCounter++;
      // console.log("stlinkOpenDeviceCounter = " + stlinkOpenDeviceCounter);
      if (stlinkOpenDeviceCounter > 3) {
        // returns ERR when the function STLink_OpenDevice is called in connectToProbe function
        return stLinkDefs.DRV_ERR;
      } else {
        // returns OK when the function STLink_OpenDevice is called in probediscovery function
        return stLinkDefs.DRV_OK;
      }
    };

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets a wrong status when setting the SWD frequency", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("SET_FREQUENCY_FAILS");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN // value set in fake STLink_SendCommand (different of STLINK_JTAG_OK)
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the STLink driver SendCommand returns one error when setting the SWD frequency", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("SET_FREQUENCY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the STLink driver SendCommand sets a wrong status when entering SWD", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("ENTER_SWD_OR_JTAG_FAILS");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the STLink driver SendCommand returns one error when getting the current mode (with the SWD protocol)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("GET_CURRENT_MODE_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the STLink driver SendCommand sets a wrong dfu mode when getting the current mode (with the SWD protocol)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("GET_CURRENT_MODE_FAILS");

    let expectedResult = {
      result: apidefs.STLINK_WRONG_DEVICE_MODE
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the STLink driver SendCommand sets a wrong status when reading Id Code (with the SWD protocol)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_UNKNOWN_JTAG_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR
    };

    expect(stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the STLink driver SendCommand sets a wrong status when setting the JTAG frequency", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("SET_FREQUENCY_FAILS");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_UNKNOWN_JTAG_CHAIN
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand returns one error when setting the JTAG frequency", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("SET_FREQUENCY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets a wrong status when entering JTAG", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("ENTER_SWD_OR_JTAG_FAILS");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand returns one error when getting the current mode (with the JTAG protocol)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("GET_CURRENT_MODE_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_DRIVER_ERR
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets a wrong dfu mode when getting the current mode (with the JTAG protocol)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("GET_CURRENT_MODE_FAILS");

    let expectedResult = {
      result: apidefs.STLINK_WRONG_DEVICE_MODE
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the STLink driver SendCommand sets a wrong status when reading Id Code (with the JTAG protocol)", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("READ_ID_CODES_SETS_UNKNOWN_JTAG_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_JTAG_ERROR
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the target ID code is erroneous", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    STLink_SendCommandSetOptions("ID_CODES_ARE_ERRONEOUS");

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_SUPPORTED
    };

    expect(stlinkprotocol.connectToTarget("4567", "JTAG", "560 kHz", false, typeConnect)).to.deep.equal(expectedResult);
  });
});

describe("#closeProbes()", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#closeProbes()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should close all probes (opened in JTAG mode)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("SET_JTAG_MODE");

    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should close all probes (opened in SWIM mode)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("SET_SWIM_MODE");

    stlinkprotocol.closeProbes("", "p2p");
  });

  it("should close one opened probe (opened in JTAG mode)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("SET_JTAG_MODE");

    stlinkprotocol.closeProbes("4567", "p2p");
  });

  it("should close one opened probe (opened in SWIM mode)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("SET_SWIM_MODE");

    stlinkprotocol.closeProbes("4567", "p2p");
  });

  it("should not close any probe if the probe identifier is erroneous", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("SET_JTAG_MODE");

    stlinkprotocol.closeProbes("1234", "p2p");
  });
});

describe("#readMemoryTyped()", function() {
  let stlinkDrvMock;
  let myClock;

  beforeEach(function() {
    logger.info("#readMemoryTyped()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
    const myNow = new Date(1);
    myClock = sinon.useFakeTimers(myNow.getTime());
  });

  afterEach(function() {
    stlinkDrvMock.verify();
    myClock.restore();
  });

  it("should read four words of signed 64 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        { x: 1, y: 0x0807060504030201 },
        { x: 1, y: 0x100f0e0d0c0b0a09 },
        { x: 1, y: 0x1817161514131211 },
        { x: 1, y: 0x201f1e1d1c1b1a19 }
      ],
      elemSize: 8
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_S64, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of unsigned 64 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        { x: 1, y: 0x0807060504030201 },
        { x: 1, y: 0x100f0e0d0c0b0a09 },
        { x: 1, y: 0x1817161514131211 },
        { x: 1, y: 0x201f1e1d1c1b1a19 }
      ],
      elemSize: 8
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_U64, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of Double", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        { x: 1, y: 5.447603722011605e-270 },
        { x: 1, y: 2.500364306227096e-231 },
        { x: 1, y: 1.2650169649295773e-192 },
        { x: 1, y: 5.802147495458688e-154 }
      ],
      elemSize: 8
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_DOUBLE, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });
  it("should read four words of unsigned 32 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        {
          x: 1,
          y: 0x04030201
        },
        {
          x: 1,
          y: 0x08070605
        },
        {
          x: 1,
          y: 0x0c0b0a09
        },
        {
          x: 1,
          y: 0x100f0e0d
        }
      ],
      elemSize: 4
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_U32, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of signed 32 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        { x: 1, y: 0x04030201 },
        { x: 1, y: 0x08070605 },
        { x: 1, y: 0x0c0b0a09 },
        { x: 1, y: 0x100f0e0d }
      ],
      elemSize: 4
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_S32, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of 4 float", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        { x: 1, y: 1.539989614439558e-36 },
        { x: 1, y: 4.063216068939723e-34 },
        { x: 1, y: 1.0711190335780294e-31 },
        { x: 1, y: 2.8212601689791736e-29 }
      ],
      elemSize: 4
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_FLOAT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of unsigned 16 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        {
          x: 1,
          y: 0x0201
        },
        {
          x: 1,
          y: 0x0403
        },
        {
          x: 1,
          y: 0x0605
        },
        {
          x: 1,
          y: 0x0807
        }
      ],
      elemSize: 2
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_U16, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of signed 16 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        { x: 1, y: 0x0201 },
        { x: 1, y: 0x0403 },
        { x: 1, y: 0x0605 },
        { x: 1, y: 0x0807 }
      ],
      elemSize: 2
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345676", 4, apidefs.READ_TYPE_S16, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of unsigned 8 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        {
          x: 1,
          y: 0x01
        },
        {
          x: 1,
          y: 0x02
        },
        {
          x: 1,
          y: 0x03
        },
        {
          x: 1,
          y: 0x04
        }
      ],
      elemSize: 1
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_U8, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read four words of signed 8 bits", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        { x: 1, y: 0x01 },
        { x: 1, y: 0x02 },
        { x: 1, y: 0x03 },
        { x: 1, y: 0x04 }
      ],
      elemSize: 1
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 4, apidefs.READ_TYPE_S8, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_BAD_PARAMETER when read one array of 13 bytes in undefined AUTO mode (memory address not aligned on 32-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.readMemoryTyped("4567", 0x12345675, 13)).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_BAD_PARAMETER if the probe identifier is undefined", function() {
    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.readMemoryTyped()).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_NO_DEVICE if the probe identifier is erroneous", function() {
    let expectedResult = {
      result: apidefs.STLINK_NO_DEVICE
    };

    expect(
      stlinkprotocol.readMemoryTyped(
        "\u0004\u0005\u0006",
        "0x12345678",
        8,
        apidefs.READ_TYPE_U32,
        accessPoint,
        typeConnect
      )
    ).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_BAD_PARAMETER if the read mode is erroneous", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.readMemoryTyped("4567", "0x12345678", 8, -1, accessPoint, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail STLINK_DEVICE_NOT_OPENED if the STLink driver OpenDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let stlinkOpenDeviceCounter = 0;

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkDrvMock.object.STLink_OpenDevice = function() {
      stlinkOpenDeviceCounter++;
      // console.log("stlinkOpenDeviceCounter = " + stlinkOpenDeviceCounter);
      if (stlinkOpenDeviceCounter > 3) {
        // returns ERR when the function STLink_OpenDevice is called in connectToProbe function
        return stLinkDefs.DRV_ERR;
      } else {
        // returns OK when the function STLink_OpenDevice is called in probediscovery function
        return stLinkDefs.DRV_OK;
      }
    };

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "0x12345678", 8, apidefs.READ_TYPE_U32, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_TIMEOUT to read one memory of 32 bits if the STLink SendCommand returns one error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "12345678", 8, apidefs.READ_TYPE_U32, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_TIMEOUT to read one memory of 16 bits if the STLink SendCommand returns one error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "0x12345678", 8, apidefs.READ_TYPE_U16, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_TIMEOUT to read one memory of 8 bits if the STLink SendCommand returns one error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "0x12345678", 8, apidefs.READ_TYPE_U8, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_BAD_PARAMETER to read one array of 3 U16 if the STLink SendCommand returns one error (memory address not aligned on 16-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", 0x12345675, 3, apidefs.READ_TYPE_U16, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail STLINK_TIMEOUT to read one array of 8 U16 if the STLink SendCommand returns one error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryTyped("4567", "0x12345678", 8, apidefs.READ_TYPE_U16, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });
});

describe("#readMemoryMode()", function() {
  let stlinkDrvMock;
  let myClock;

  beforeEach(function() {
    logger.info("#readMemoryMode()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
    const myNow = new Date(1);
    myClock = sinon.useFakeTimers(myNow.getTime());
  });

  afterEach(function() {
    stlinkDrvMock.verify();
    myClock.restore();
  });

  it("should read 8 bytes in mode MODE_32_BIT", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [1, 2, 3, 4, 5, 6, 7, 8],
      time: 1
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "12345678", 8, apidefs.READ_MODE_32_BIT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read 8 bytes in mode MODE_16_BIT", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [1, 2, 3, 4, 5, 6, 7, 8],
      time: 1
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, apidefs.READ_MODE_16_BIT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read 8 bytes in mode MODE_8_BIT", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [1, 2, 3, 4, 5, 6, 7, 8],
      time: 1
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, apidefs.READ_MODE_8_BIT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read 13 bytes in undefined mode (memory address not aligned on 32-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [1, 2, 3, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2],
      time: 1
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345675", 13, apidefs.READ_MODE_AUTO, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read 2 bytes in undefined mode (memory address not aligned on 32-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [1, 2],
      time: 1
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345675", 2, apidefs.READ_MODE_AUTO, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should read one array of 12 bytes in AUTO mode (memory address aligned on 32-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
      time: 1
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 12, apidefs.READ_MODE_AUTO, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail if the probe identifier is undefined", function() {
    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.readMemoryMode()).to.deep.equal(expectedResult);
  });

  it("should fail if the probe identifier is erroneous", function() {
    let expectedResult = {
      result: apidefs.STLINK_NO_DEVICE
    };

    expect(
      stlinkprotocol.readMemoryMode(
        "\u0004\u0005\u0006",
        "0x12345678",
        8,
        apidefs.READ_MODE_32_BIT,
        accessPoint,
        typeConnect
      )
    ).to.deep.equal(expectedResult);
  });

  it("should fail if the read mode is erroneous", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, -1, accessPoint, typeConnect)).to.deep.equal(
      expectedResult
    );
  });

  it("should fail if the STLink driver OpenDevice fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let stlinkOpenDeviceCounter = 0;

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;

    stlinkDrvMock.object.STLink_OpenDevice = function() {
      stlinkOpenDeviceCounter++;
      // console.log("stlinkOpenDeviceCounter = " + stlinkOpenDeviceCounter);
      if (stlinkOpenDeviceCounter > 3) {
        // returns ERR when the function STLink_OpenDevice is called in connectToProbe function
        return stLinkDefs.DRV_ERR;
      } else {
        // returns OK when the function STLink_OpenDevice is called in probediscovery function
        return stLinkDefs.DRV_OK;
      }
    };

    stlinkprotocol.probeDiscovery(typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, apidefs.READ_MODE_32_BIT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail to read one memory of 32 bits if the STLink SendCommand returns one error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, apidefs.READ_MODE_32_BIT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail to read one memory of 16 bits if the STLink SendCommand returns one error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, apidefs.READ_MODE_16_BIT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail to read one memory of 8 bits if the STLink SendCommand returns one error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, apidefs.READ_MODE_8_BIT, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail to read one array of bytes if the STLink SendCommand returns one error (memory address not aligned on 32-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", 0x12345675, 3, apidefs.READ_MODE_AUTO, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail to read one array of bytes if the STLink SendCommand returns one error (memory address aligned on 32-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 8, apidefs.READ_MODE_AUTO, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });

  it("should fail to read one array of bytes if the second call of STLink SendCommand returns one error (memory address aligned on 32-bit boundary)", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("SECOND_READ_MEMORY_RETURNS_ERROR");

    let expectedResult = {
      result: apidefs.STLINK_TIMEOUT
    };

    expect(
      stlinkprotocol.readMemoryMode("4567", "0x12345678", 10, apidefs.READ_MODE_AUTO, accessPoint, typeConnect)
    ).to.deep.equal(expectedResult);
  });
});

describe("#readMemoryGroup()", function() {
  let stlinkDrvMock;
  let myClock;

  beforeEach(function() {
    logger.info(true, "#readMemoryGroup()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
    const myNow = new Date(1);
    myClock = sinon.useFakeTimers(myNow.getTime());
  });

  afterEach(function() {
    stlinkDrvMock.verify();
    myClock.restore();
  });

  it("should read a group", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        [
          {
            x: 1,
            y: 0x01
          },
          {
            x: 1,
            y: 0x02
          },
          {
            x: 1,
            y: 0x03
          },
          {
            x: 1,
            y: 0x04
          }
        ],
        [
          {
            x: 1,
            y: 0x0807060504030201
          }
        ],
        [
          {
            x: 1,
            y: 0x04030201
          },
          {
            x: 1,
            y: 0x08070605
          },
          {
            x: 1,
            y: 0x0c0b0a09
          }
        ],
        [
          {
            x: 1,
            y: 0x01
          }
        ],
        [
          {
            x: 1,
            y: 0x0102010403020104
          }
        ],
        [
          {
            x: 1,
            y: 0x0201acab
          }
        ],
        [
          {
            x: 1,
            y: 0xacab
          }
        ],
        [
          {
            x: 1,
            y: -21333
          }
        ],
        [
          {
            x: 1,
            y: -6.504395154765953e-229
          }
        ],
        [
          {
            x: 1,
            y: -2.1499718795889888e-31
          }
        ]
      ]
    };

    let myGroup = [
      {
        address: "0x12345678",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_S64
      },
      {
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      },
      {
        address: "0x12345000",
        type: apidefs.READ_TYPE_S8
      },
      {
        address: "0x123400FF",
        type: apidefs.READ_TYPE_U64
      },
      {
        address: "0x123400AA",
        type: apidefs.READ_TYPE_U32
      },
      {
        address: "0x123400AA",
        type: apidefs.READ_TYPE_U16
      },
      {
        address: "0x123400AA",
        type: apidefs.READ_TYPE_S16
      },
      {
        address: "0x12340088",
        type: apidefs.READ_TYPE_DOUBLE
      },
      {
        address: "0x12340088",
        type: apidefs.READ_TYPE_FLOAT
      }
    ];

    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should get BAD PARAMETER error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let myGroupNoAdress = [
      {
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];
    let myGroupNoType = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];
    let myGroupNoNb = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: "n°3"
      }
    ];

    expect(stlinkprotocol.readMemoryGroup("", myGroupNoAdress, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567", myGroupNoAdress, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567", myGroupNoType, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567", myGroupNoNb, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
    expect(stlinkprotocol.readMemoryGroup("4567")).to.deep.equal({ result: apidefs.STLINK_BAD_PARAMETER });
  });

  it("should get TIME OUT error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let myGroup = [
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "0x12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    STLink_SendCommandSetOptions("READ_MEMORY_RETURNS_ERROR");
    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_TIMEOUT
    });
  });

  it("should get STLINK_NO_DEVICE error", function() {
    logger.info("it should get STLINK_NO_DEVICE error");
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let myGroup = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);
    expect(stlinkprotocol.readMemoryGroup("4568", myGroup, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_NO_DEVICE
    });
  });

  it("should get STLINK_DEVICE_NOT_OPENED error", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);
    let myGroup = [
      {
        address: "12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "12340000",
        type: apidefs.READ_TYPE_S64,
        nb: 2
      },
      {
        address: "12345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(2)
      .returns(stLinkDefs.DRV_NO_DEVICE);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_DEVICE_NOT_OPENED
    });
  });

  it("STLINK_BAD_PARAMETER error - address is not a string", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let myGroup = [
      {
        address: 0x12345678,
        type: apidefs.READ_TYPE_U8,
        nb: 4
      },
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_S64
      },
      {
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    let expectedResult = {
      result: apidefs.STLINK_BAD_PARAMETER
    };

    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("STLINK_BAD_PARAMETER error - The number is not correctly defined", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let myGroup = [
      {
        address: "0x12345678",
        type: apidefs.READ_TYPE_U8,
        nb: "hello"
      },
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_S64
      },
      {
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        nb: 3
      }
    ];

    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal({
      result: apidefs.STLINK_BAD_PARAMETER
    });
  });

  it("varStartAddress and varEndAddress under the range of frameAddress", function() {
    //read is made in 3 steps one u16 mode, one u16 then one u8
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let myGroup = [
      {
        address: "0x12340000",
        type: apidefs.READ_TYPE_U8,
        nb: 10
      },
      {
        address: "0x12340002",
        type: apidefs.READ_TYPE_U8,
        nb: 3
      },
      {
        address: "0x12340010",
        type: apidefs.READ_TYPE_U8,
        nb: 3
      }
    ];

    let expectedResult = {
      result: apidefs.STLINK_OK,
      data: [
        [
          {
            x: 1,
            y: 1
          },
          {
            x: 1,
            y: 2
          },
          {
            x: 1,
            y: 3
          },
          {
            x: 1,
            y: 4
          },
          {
            x: 1,
            y: 5
          },
          {
            x: 1,
            y: 6
          },
          {
            x: 1,
            y: 7
          },
          {
            x: 1,
            y: 8
          },
          {
            x: 1,
            y: 9
          },
          {
            x: 1,
            y: 10
          }
        ],
        [
          {
            x: 1,
            y: 3
          },
          {
            x: 1,
            y: 4
          },
          {
            x: 1,
            y: 5
          }
        ],
        [
          {
            x: 1,
            y: 1
          },
          {
            x: 1,
            y: 2
          },
          {
            x: 1,
            y: 1
          }
        ]
      ]
    };
    expect(stlinkprotocol.readMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });
});

describe("#writeMemoryGroup()", function() {
  let stlinkDrvMock;
  let myClock;

  beforeEach(function() {
    logger.info(true, "#writeMemoryGroup()");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
    const myNow = new Date(1);
    myClock = sinon.useFakeTimers(myNow.getTime());
  });

  afterEach(function() {
    stlinkDrvMock.verify();
    myClock.restore();
  });

  it("should write decimal and hexadecimal values in several variables of types READ_TYPE_U8 and READ_TYPE_S8", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345678",
        type: apidefs.READ_TYPE_U8,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S8,
        value: "0x09F"
      },
      {
        name: "Variable_Dec_1",
        address: "0x12340010",
        type: apidefs.READ_TYPE_S8,
        value: "-128"
      },
      {
        name: "Variable_Dec_2",
        address: "0x12340010",
        type: apidefs.READ_TYPE_U8,
        value: "127"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should write decimal and hexadecimal values in several variables of types READ_TYPE_U32 and READ_TYPE_S32", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        value: "0x04030201"
      },
      {
        name: "Variable_Hex_2",
        address: "0x2345000",
        type: apidefs.READ_TYPE_U32,
        value: "0x1"
      },
      {
        name: "Variable_Dec_1",
        address: "0x2345000",
        type: apidefs.READ_TYPE_S32,
        value: "-2147483648"
      },
      {
        name: "Variable_Dec_2",
        address: "0x2345000",
        type: apidefs.READ_TYPE_U32,
        value: "4294967295"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should write decimal and hexadecimal values in several variables of types READ_TYPE_U16 and READ_TYPE_S16", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "Variable_Dec_1",
        address: "0x00000002",
        type: apidefs.READ_TYPE_S16,
        value: "32767"
      },
      {
        name: "Variable_Dec_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S16,
        value: "-32768"
      },
      {
        name: "Variable_Dec_3",
        address: "0x00000002",
        type: apidefs.READ_TYPE_U16,
        value: "0"
      },
      {
        name: "Variable_Dec_4",
        address: "0x12340002",
        type: apidefs.READ_TYPE_U16,
        value: "65535"
      },
      {
        name: "Variable_Hex_1",
        address: "0x00000002",
        type: apidefs.READ_TYPE_S16,
        value: "0x0"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S16,
        value: "0xFFFF"
      },
      {
        name: "Variable_Hex_3",
        address: "0x00000002",
        type: apidefs.READ_TYPE_U16,
        value: "0x0"
      },
      {
        name: "Variable_Hex_4",
        address: "0x12340002",
        type: apidefs.READ_TYPE_U16,
        value: "0xFFFF"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should write decimal and hexadecimal values in several variables of type DOUBLE", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "Variable_Dec_1",
        address: "0x2345000",
        type: apidefs.READ_TYPE_DOUBLE,
        value: "-6.504395154765953e-229"
      },
      {
        name: "Variable_Dec_2",
        address: "0x2345000",
        type: apidefs.READ_TYPE_DOUBLE,
        value: "-2147483648"
      },
      {
        name: "Variable_Dec_3",
        address: "0x2345000",
        type: apidefs.READ_TYPE_DOUBLE,
        value: "1.7e-308"
      },
      {
        name: "Variable_Hex_1",
        address: "0x2345000",
        type: apidefs.READ_TYPE_DOUBLE,
        value: "0x0"
      },
      {
        name: "Variable_Hex_2",
        address: "0x2345000",
        type: apidefs.READ_TYPE_DOUBLE,
        value: "0x1234567890123456"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should write decimal and hexadecimal values in several variables of type FLOAT", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: stLinkDefs.DRV_OK };

    let myGroup = [
      {
        name: "myFirstVariable",
        address: "0x00000004",
        type: apidefs.READ_TYPE_FLOAT,
        value: "0xffffffffffffffffffffffffffffffffff"
      },
      {
        name: "mySecondVariable",
        address: "0x12340004",
        type: apidefs.READ_TYPE_FLOAT,
        value: "-2147483648"
      },
      {
        name: "myThirdVariable",
        address: "0x12340010",
        type: apidefs.READ_TYPE_FLOAT,
        value: "2147483647"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should write no value", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_OK };

    let myGroup = [];

    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the probeId is undefined", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_BAD_PARAMETER };

    let myGroup = [];

    expect(stlinkprotocol.writeMemoryGroup("", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the probeId is erroneous", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_NO_DEVICE };

    let myGroup = [];

    expect(stlinkprotocol.writeMemoryGroup("1234", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the probe is not opened", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DEVICE_NOT_OPENED };

    let myGroup = [];

    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the group is undefined", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_BAD_PARAMETER };

    let myGroup = undefined;
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail if the variable type is erroneous", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_BAD_PARAMETER };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345678",
        type: 20,
        value: "0xFF"
      }
    ];

    STLink_SendCommandSetOptions("");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect)).to.deep.equal(expectedResult);
  });

  it("should fail to write one value in one variable of type READ_TYPE_U8 if JTAG_WriteMemory8Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345679",
        type: apidefs.READ_TYPE_U8,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_U8,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should fail to write one value in one variable of type READ_TYPE_S8 if JTAG_WriteMemory8Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345679",
        type: apidefs.READ_TYPE_S8,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S8,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should fail to write one value in one variable of type READ_TYPE_U16 if JTAG_WriteMemory8Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345679",
        type: apidefs.READ_TYPE_U16,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_U16,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should fail to write one value in one variable of type READ_TYPE_S16 if JTAG_WriteMemory8Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345679",
        type: apidefs.READ_TYPE_S16,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S16,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should fail to write one value in one variable of type READ_TYPE_U32 if JTAG_WriteMemory32Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345678",
        type: apidefs.READ_TYPE_U32,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_U32,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should fail to write one value in one variable of type READ_TYPE_S32 if JTAG_WriteMemory32Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345678",
        type: apidefs.READ_TYPE_S32,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_S32,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should fail to write one value in one variable of type READ_TYPE_FLOAT if JTAG_WriteMemory32Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345678",
        type: apidefs.READ_TYPE_FLOAT,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_FLOAT,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });

  it("should fail to write one value in one variable of type READ_TYPE_DOUBLE if JTAG_WriteMemory32Bit fails", function() {
    let deviceInfo2 = new TDeviceInfo2();
    deviceInfo2.ref().fill(0);

    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.connectToTarget("4567", "SWD", "1.8 MHz - Default", false, typeConnect);

    let expectedResult = { result: apidefs.STLINK_DRIVER_ERR };

    let myGroup = [
      {
        name: "Variable_Hex_1",
        address: "0x12345678",
        type: apidefs.READ_TYPE_DOUBLE,
        value: "0xFF"
      },
      {
        name: "Variable_Hex_2",
        address: "0x12340002",
        type: apidefs.READ_TYPE_DOUBLE,
        value: "0x09F"
      }
    ];

    STLink_SendCommandSetOptions("Error");
    expect(stlinkprotocol.writeMemoryGroup("4567", myGroup, accessPoint, typeConnect).result).to.deep.equal(
      expectedResult.result
    );
  });
});

describe("#blinkLed", function() {
  let stlinkDrvMock;

  beforeEach(function() {
    logger.info("#blinkLed");
    stlinkDrvMock = sinon.mock(stLinkDrv);
    STLink_SendCommandSetOptions("");
  });

  afterEach(function() {
    stlinkDrvMock.verify();
  });

  it("should blink the Led", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDevice(2);

    expect(stlinkprotocol.cli.JTAG_BlinkLed(2, typeConnect)).to.equal(apidefs.STLINK_OK);

    stlinkprotocol.cli.closeDevice(2);
  });

  it("should fail if the device is not opened", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .thrice()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(3)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);

    expect(stlinkprotocol.cli.JTAG_BlinkLed(2, typeConnect)).to.equal(apidefs.STLINK_DEVICE_NOT_OPENED);
  });

  it("should fail if the STLink driver SendCommand fails", function() {
    stlinkDrvMock
      .expects("STLink_Reenumerate")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF, false)
      .once()
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_GetNbDevices")
      .withArgs(stLinkDefs.DRV_DEFAULT_IF)
      .once()
      .returns(3);

    stlinkDrvMock
      .expects("STLink_OpenDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock
      .expects("STLink_CloseDevice")
      .exactly(4)
      .returns(stLinkDefs.DRV_OK);

    stlinkDrvMock.object.STLink_GetDeviceInfo2 = STLink_GetDeviceInfo2Mocked;
    stlinkDrvMock.object.STLink_SendCommand = STLink_SendCommandMocked;

    stlinkprotocol.probeDiscovery(typeConnect);
    stlinkprotocol.cli.openDevice(2);

    STLink_SendCommandSetOptions("ERROR");

    expect(stlinkprotocol.cli.JTAG_BlinkLed(2, typeConnect)).to.equal(apidefs.STLINK_DRIVER_ERR);
    stlinkprotocol.cli.closeDevice(2);
  });

  it("should fail if the device index is wrong", function() {
    expect(stlinkprotocol.cli.JTAG_BlinkLed(3, typeConnect)).to.deep.equal(apidefs.STLINK_BAD_PARAMETER);
  });
});
